#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

#define CHILDREN	(_POSIX_CHILD_MAX -2)	/* Number of children to fork()	*/

static	void interrupt(int);
static	void children(int);
static	void wait_up(void);

static	int		tty_fildes, loop_fildes;
static	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
static	pid_t	pid;


int main(void)
{
	setup(E_INTERFACE);
	funct(E_INTERFACE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(1, BASE)
		GTI_NOT_PROVIDED_RESULT(2, BASE)
		GTI_NOT_PROVIDED_RESULT(3, BASE)
		GTI_NOT_PROVIDED_RESULT(4, BASE)
		GTI_NOT_PROVIDED_RESULT(5, BASE)
		GTI_NOT_PROVIDED_RESULT(6, BASE)
		GTI_NOT_PROVIDED_RESULT(7, BASE)
		GTI_NOT_PROVIDED_RESULT(8, BASE)
		GTI_NOT_PROVIDED_RESULT(9, BASE)
		done();
	}

	continue_on(ERROR);

	assertion(1, BASE);
	start
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

							/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			children(CHILDREN);		/* Fork children, pause for signal.		*/
			SLEEP(WAIT_FOR_TIME); 	/* Allow for child setup.				*/
			interrupt(TRUE);		/* Send interrupt to self and children. */
			wait_up();				/* Check for interrupted children.		*/

			done();
		}
		wait_for(pid, 4 * MAX_TIME_TO_RUN);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis


			/* New session cannot acquire another's controlling terminal. */
	assertion(2, EXTENDED_TESTED);
	start
		FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

							/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			interrupt(TRUE);	/* Ensure controlling terminal established. */

			prep("Catch SIGINT, expected from child sending INTR to this");
			prep("process's controlling terminal.");
			start
				SIGNAL(SIGINT, SIG_IGN);
			finis

			FORK
				testing("Attempt to acquire an existing controlling terminal");
				testing("for a new session should fail.");
				start
					Zsetsid();		/* Establish a new session. */
					SET_ALARM(WAIT_FOR_TIME)
						tty_fildes = open_ctty(PORT_TTY, O_RDWR | O_NONBLOCK,
													PERM_CTTY);
					CLEAR_ALARM
				finis

				if(tty_fildes > 0)
					interrupt(FALSE);	 /* No controlling terminal.	*/
			KROF
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

				/* Non-session leader cannot acquire controlling terminal.	*/
	assertion(3, BASE);
	start
		FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

							/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			interrupt(TRUE);	/* Ensure controlling terminal established. */
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)

		FORK
			prep("Open PORT_TTY and PORT_LOOP.");
			start
				tty_prep(&tty_fildes, tty_termios_p);
				loop_prep(&loop_fildes, loop_termios_p);

							/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			interrupt(FALSE);	/* Controlling terminal is not established. */
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

			/* O_NOCTTY on open() prevents acquiring a controlling terminal. */
	assertion(4, BASE);
	start
		FORK
			prep("Establish session leader process then call open() on a");
			prep("terminal device with O_NOCTTY flag.");
			start
				Zsetsid();
				SET_ALARM(WAIT_FOR_TIME)
					tty_fildes = Zopen(PORT_TTY, O_RDWR|O_NONBLOCK|O_NOCTTY);
				CLEAR_ALARM
				loop_prep(&loop_fildes, loop_termios_p);
										/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			finis

			SLEEP(WAIT_FOR_TIME); 	/* PORT_LOOP knows PORT_TTY is active. */
			interrupt(FALSE);
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

			/*	Foreground process group of controlling terminal is
				set to process group of the session leader.					*/
	assertion(5, BASE);
	start
		FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);
			finis
			
			testing("Foreground process group of controlling terminal is");
			testing("set to process group of the session leader.");
			start
				if(Zgetpgrp() != Ztcgetpgrp(tty_fildes))
					error("%s%ld\n%s%ld."
						"Controlling terminal's process group is ",
						(long) tcgetpgrp(tty_fildes),
						"Process goup of session leader is ",
						(long)getpgrp());
			finis
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

			/*	Process relinquishes controlling terminal when it
				creates a new session with setsid().					*/
	assertion(6, BASE);
	start
		FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

										/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			finis

			prep("Catch SIGINT, expected from child sending INTR to this");
			prep("process's controlling terminal.");
			start
				SIGNAL(SIGINT, SIG_IGN);
			finis

			FORK
				interrupt(TRUE);
				Zsetsid();
				interrupt(FALSE);
			KROF
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis


	assertion(7, BASE);
	start
		FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

							/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			interrupt(TRUE);	/* Ensure controlling terminal established. */

			prep("Catch SIGINT, expected from child sending INTR to this");
			prep("process's controlling terminal.");
			start
				SIGNAL(SIGINT, SIG_IGN);
			finis

			FORK
				prep("Close controlling terminal.");
				start
					SIGNAL(SIGHUP, SIG_IGN);
					Zclose(tty_fildes);
					SLEEP(GTI_LINE_CLOSE_DELAY)
				finis

				testing("Child process can re-acquire controlling terminal.");
				start
					SET_ALARM(WAIT_FOR_TIME)
						tty_fildes = open_ctty(PORT_TTY, O_RDWR | O_NONBLOCK,
												PERM_CTTY);
					CLEAR_ALARM
				finis

				if(tty_fildes > 0)
					interrupt(TRUE);	/* Re-acquired controlling terminal. */
			KROF
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	assertion(8, BASE);
	start
		FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

							/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			interrupt(TRUE);	/* Ensure controlling terminal established. */
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)

		FORK
			prep("Since the controlling process terminated, the prior");
			prep("controlling terminal can be acquired by new session leader.");
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

							/* Enable SIGNALS */
				Ztcgetattr(loop_fildes, loop_termios_p);
				loop_termios_p->c_lflag |= ISIG;
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis

			interrupt(TRUE);	/* Ensure controlling terminal established. */
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	assertion(9, BASE);
	start
		char	inbuf[_POSIX_MAX_INPUT];

		if(A_D_TO_P_WHEN_C_P_T) {
			ENABLE_PSYNC
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Establish PORT_TTY as control terminal and PORT_LOOP.");
				start
					termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
									loop_termios_p);
					SIGNAL(SIGHUP, SIG_IGN);
				finis

				prep("Fork a child process, then terminate.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {
					SEND_PSYNC
								/* Allow controlling process to terminate. */
					SLEEP(WAIT_FOR_TIME);

					SET_ALARM(WAIT_FOR_TIME)
						/*	From the description of modem disconnect in
							POSIX.1-1990 (page 135, lines 274-277):
								If the [EIO] condition specified in 6.4.1.4
								that applies when the implementation supports
								job control also exists, it is unspecified
								whether the EOF condition or the [EIO] is
								returned.
						*/

						testing("Subsequent read from child process returns");
						testing("zero (EOF condition) or -1 ([EIO] condition.");
						start
							ssize_t ret_val = 0;
							
							errno = 0;
							ret_val = read(tty_fildes, inbuf, sizeof(inbuf));

							if(ret_val == 0) {	/*	EOF condition	*/
								if(errno != 0)
									error("%s\n%s%s",
										"Expected: errno value of 0.",
										"Actual: errno is ",
										err_lookup(errno) );
							}
							else {				/*	[EIO] condition	*/
								if(ret_val == (ssize_t) -1) {
									if(errno != EIO)
										error("%s\n%s%s",
											"Expected: errno set to [EIO].",
											"Actual: errno is ",
											err_lookup(errno) );
								}
								else
									error("Value from read() was %ld.\n",
											(long) ret_val);
							}
						finis

						testing("Subsequent call to write() from child");
						testing("process returns -1.");
						start
							expecting(-1);
							expecting(EIO);
							Zwrite(tty_fildes, "Try", 3);
						finis
					CLEAR_ALARM

					done();
				}
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM

				done();
			}
			WAIT_FOR_PID(pid);
			SLEEP(2 * WAIT_FOR_TIME);		/* Allow for test to complete. */
			SLEEP(GTI_LINE_CLOSE_DELAY)
			DISABLE_PSYNC
		}
		else
			not_implemented("access to session denied");
	finis

	done();
}


		/* Fork children, pause for signal. */
static	void children(int number)
{
	int i;

	prep("Fork children for termination by SIGINT.");
	start
		for(i = 0; i < number; ++i) {
			SIGNAL(SIGINT, SIG_DFL);

			if(Zfork() == (pid_t)0) {
				exit_on(ERROR);
				SET_TIMEOUT
					pause();
				CLEAR_ALARM

				done();
			}
		}
	finis
}

		/* Send SIGINT to self and children. */
static	void interrupt(int toggle)
{
	if(toggle == TRUE)
		testing("INTR sent to PORT_TTY generates SIGINT.");
	else
		testing("INTR sent to PORT_TTY does not generate SIGINT.");
	start
		SIGNAL(SIGINT, sig_int);
		sig_rcvd = 0;
		SET_ALARM(WAIT_FOR_TIME)
			Ztcgetattr(tty_fildes, tty_termios_p);	/* Initialize TTY struct. */
			WRITE_TTY_SCC(loop_fildes, tty_termios_p->c_cc[VINTR]);
		IGNORE_ALARM
		sleep(WAIT_FOR_TIME);

		if(toggle == TRUE && sig_rcvd == 0)
			error("Process did not receive SIGINT.");

		if(toggle == FALSE && sig_rcvd != 0)
			error("Process received SIGINT.");
	finis
}

			/* Verify child exited because of SIGINT signal. */
static	void wait_up(void)
{
	int stat_val;

	testing("Child exit due to SIGINT signal.");
	start
		SLEEP(WAIT_FOR_TIME);

		while((pid = wait(&stat_val)) > (pid_t)0) {
			if(WTERMSIG(stat_val) != SIGINT)
				error("%s%ld\n%s%d\n%s%d\n",
						"Unexpected child return ",
						(long) pid,
						"Expected ",
						SIGINT,
						"Received ",
						WTERMSIG(stat_val));
		}
	finis
}
