#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

static	char outbuf[]= "Hutsut ralston on the rillarah\n";
#define MSG_LEN (sizeof(outbuf)-1)


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	pid_t pid;
	sigset_t	set;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(14, BASE)
		GTI_NOT_PROVIDED_RESULT(15, BASE)
		GTI_NOT_PROVIDED_RESULT(16, BASE)
		GTI_NOT_PROVIDED_RESULT(17, BASE)
		GTI_NOT_PROVIDED_RESULT(18, BASE)
		done();
	}

	continue_on(ERROR);

	assertion(14, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
		finis

		FORK
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis
	
			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTOU, sig_int);
				SIGUNBLOCK(set, SIGTTOU);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child of background process receives SIGTTOU.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd == 0)
						error("SIGTTOU not received.");
				finis

				done();
			}

			testing("A background process, not ignoring or blocking SIGTTOU,");
			testing("attempting to write to its controlling terminal causes");
			testing("its process group to be sent a SIGTTOU signal.");
			start
				ssize_t	nb_wr;
				char	inbuf[MSG_LEN];
				
				SET_TIMEOUT
					nb_wr = write(tty_fildes, outbuf, MSG_LEN);
				CLEAR_ALARM
				if(sig_rcvd == 0)
					error("SIGTTOU not received.");
				
							/*	Test data reported as written was sent. */
				SET_TIMEOUT
					tcdrain(tty_fildes);	/* Possible interrupt by SIGTTOU. */
					tcdrain(tty_fildes);	/* If no data written, this call can
											   be interrupted by SIGTTOU. */
					dt_secs(TIME_ACQUIRE_XMITTED_DATA);
					if(nb_wr == (ssize_t)-1)
						expecting(0);
					else
						expecting(nb_wr);
					nb_wr = Zread(loop_fildes, inbuf, MSG_LEN);
				CLEAR_ALARM
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF
	finis


	assertion(15, BASE);
	start
		FORK
			prep("Clear TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTOU, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child background process does not receive SIGTTOU.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd != 0)
						error("SIGTTOU was received.");
				finis

				done();
			}

			testing("When TOSTOP is set a background process attempting");
			testing("to write to its controlling terminal succeeds and");
			testing("its process group is not sent a SIGTTOU signal.");
			start
				SET_TIMEOUT
					expecting(SUCCESS);
					Zwrite(tty_fildes, outbuf, MSG_LEN);
				CLEAR_ALARM
				if(sig_rcvd != 0)
					error("SIGTTOU was received.");
			finis

			WAIT_FOR_PID(pid);
			done();
		KROF

								/* Repeat for orphaned background process. */
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Clear TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTOU, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Fork a child process.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {
					testing("Child of orphaned background process does not");
					testing("receive SIGTTOU.");
					start
						sleep(2*WAIT_FOR_TIME);
						if(sig_rcvd != 0)
							error("SIGTTOU was received.");
					finis
	
					done();
				}

				testing("When TOSTOP is set, an orphaned background process");
				testing("attempting to write to its controlling terminal");
				testing("succeeds and its process group is not sent a");
				testing("SIGTTOU signal.");
				start
					SLEEP(WAIT_FOR_TIME);	/* Ensure process group orphaned. */
					SET_ALARM(WAIT_FOR_TIME)
						expecting(SUCCESS);
						Zwrite(tty_fildes, outbuf, MSG_LEN);
					CLEAR_ALARM
					if(sig_rcvd != 0)
						error("SIGTTOU was received.");
				finis

				WAIT_FOR_PID(pid);
				done();
			}
			done();					/* Exit process group leader. */
		}
		WAIT_FOR_PID(pid);			/* Establish orphaned process group. */
						/* Allow time for orphaned processses to complete.	*/
		SLEEP(3*WAIT_FOR_TIME);
	finis

	assertion(16, BASE);
	start
		FORK
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTOU, sig_int);
				sig_rcvd = 0;
			finis

			testing("A background process, ignoring SIGTTOU,");
			testing("attempting to write to its controlling");
			testing("terminal succeeds.");
			start
				SIGNAL(SIGTTOU, SIG_IGN);
				SET_TIMEOUT
					expecting(SUCCESS);
					Zwrite(tty_fildes, outbuf, MSG_LEN);
				CLEAR_ALARM
			finis
		KROF
	finis


	assertion(17, BASE);
	start
		FORK
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTOU, sig_int);
				sig_rcvd = 0;
			finis

			testing("A background process, blocking SIGTTOU,");
			testing("attempting to write to its controlling");
			testing("terminal succeeds.");
			start
				SIGBLOCK(set, SIGTTOU);
				SET_TIMEOUT
					expecting(SUCCESS);
					Zwrite(tty_fildes, outbuf, MSG_LEN);
				CLEAR_ALARM
			finis
		KROF
	finis


	assertion(18, BASE);
	start
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Set TOSTOP flag on PORT_TTY.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				tty_termios_p->c_lflag |= TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis
	
			prep("Establish a background process group.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTOU, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Fork a child process.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {
					testing("Child of orphaned background process group");
					testing("receives no SIGTTOU.");
					start
						sleep(2*WAIT_FOR_TIME);
						if(sig_rcvd != 0)
							error("SIGTTOU was received.");
					finis

					done();
				}

				testing("An orphaned background process neither ignoring");
				testing("nor blocking SIGTTOU and is");
				testing("attempting to write to its controlling");
				testing("terminal returns errno set to [EIO] and no SIGTTOU");
				testing("signal is sent to any process in process group.");
				start
					SLEEP(WAIT_FOR_TIME);	/* Ensure process group orphaned. */
					SIGUNBLOCK(set, SIGTTOU);
					SET_ALARM(WAIT_FOR_TIME)
						expecting((ssize_t)-1);
						expecting(EIO);
						Zwrite(tty_fildes, outbuf, MSG_LEN);
					CLEAR_ALARM

					if(sig_rcvd != 0)
						error("SIGTTOU was received.");
				finis

				WAIT_FOR_PID(pid);
				done();
			}
			done();					/* Exit process group leader. */
		}
		WAIT_FOR_PID(pid);			/* Establish orphaned process group. */
						/* Allow time for orphaned processses to complete.	*/
		SLEEP(3*WAIT_FOR_TIME);
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
