#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

static	char String[]		= "Quote the raven: Nevermore!\n";
static	char Substring[]	= "Quote the raven:";
#define Strlen		(sizeof(String)-1)
#define SubStrlen	(sizeof(Substring)-1)


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	pid_t pid;
	char inbuf[_POSIX_MAX_INPUT];
	int flags;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(19, BASE)
		GTI_NOT_PROVIDED_RESULT(20, BASE)
		GTI_NOT_PROVIDED_RESULT(21, BASE)
		GTI_NOT_PROVIDED_RESULT(22, BASE)
		done();
	}

	continue_on(ERROR);
								/* O_NONBLOCK clear, read() blocked */
	assertion(19, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			exit_on(ERROR);
			prep("Clear the O_NONBLOCK flag for PORT_TTY.");
			start
				flags = Zfcntl(tty_fildes, F_GETFL, 0);
				flags &= ~O_NONBLOCK;
				Zfcntl(tty_fildes, F_SETFL, flags);
			finis

			testing("When O_NONBLOCK is clear, a read() blocks");
			testing("until data is available.");
			start
				SET_TIMEOUT
					expecting(Strlen);
					Zread(tty_fildes, inbuf, Strlen);
				CLEAR_ALARM
			finis

			done();
		}
		prep("Write to terminal device after a delay.");
		start
			SLEEP(2 * WAIT_FOR_TIME);
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
			CLEAR_ALARM
		finis
		WAIT_FOR_PID(pid);


								/* O_NONBLOCK clear, read() blocked */
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			exit_on(ERROR);
			testing("Send signal to parent process.");
			start
				sv_flush();
				SLEEP(WAIT_FOR_TIME);
				Zkill(getppid(), SIGINT);
			finis

			done();
		}

		testing("When O_NONBLOCK is clear, a read() blocks");
		testing("until a signal is received.");
		start
			SIGNAL(SIGINT, sig_int);
			SET_TIMEOUT
				expecting(EINTR);
				Zread(tty_fildes, inbuf, Strlen);
			CLEAR_ALARM
		finis

		WAIT_FOR_PID(pid);
	finis

					/* Enough data */
	assertion(20, BASE);
	start
		prep("Clear PORT queues and set the O_NONBLOCK flag for PORT_TTY.");
		start
			Ztcflush(tty_fildes, TCIOFLUSH);
			Ztcflush(loop_fildes, TCIOFLUSH);

			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			Zfcntl(tty_fildes, F_SETFL, flags | O_NONBLOCK);
		finis

		prep("Write out data.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM
		finis

		testing("When O_NONBLOCK is set and enough data is available");
		testing("to satisfy the entire request, a read() shall complete");
		testing("successfully, without blocking, having read all the");
		testing("requested data and returns the number of bytes read.");
		testing("Canonical mode.");
		start
			SET_TIMEOUT
				expecting(Strlen);
  				Zread(tty_fildes, inbuf, Strlen);
			CLEAR_ALARM
			expecting(0);
			Zstrncmp(String, inbuf, Strlen);
		finis

						/* Repeat test for non-canonical mode */
		prep("Set O_NONBLOCK flag and non-canonical mode for PORT_TTY.");
		start
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			Zfcntl(tty_fildes, F_SETFL, flags | O_NONBLOCK);

			Ztcgetattr(tty_fildes, tty_termios_p);
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VTIME] = 0;
			tty_termios_p->c_cc[VMIN] = 1;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
		finis

		prep("Write out data.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM
		finis

		testing("O_NONBLOCK is set and enough data is available");
		testing("to satisfy the entire request, a read() shall complete");
		testing("successfully, without blocking, having read all the");
		testing("requested data and returns the number of bytes read.");
		testing("Non-canonical mode.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, inbuf, Strlen);
			CLEAR_ALARM
			expecting(0);
			Zstrncmp(String,inbuf,Strlen);
		finis
	finis

					/* Not enough data */
	assertion(21, BASE);
	start
		prep("Clear PORT queues and set the O_NONBLOCK flag for PORT_TTY.");
		start
			Ztcflush(tty_fildes, TCIOFLUSH);
			Ztcflush(loop_fildes, TCIOFLUSH);
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			Zfcntl(tty_fildes, F_SETFL, flags | O_NONBLOCK);
		finis

		prep("Write out data.");
		start
			SET_TIMEOUT
				expecting(SubStrlen);
				Zwrite(loop_fildes, String, SubStrlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM
		finis

		testing("O_NONBLOCK is set and not enough data is available");
		testing("to satisfy the entire request, a read() completes");
		testing("successfully, without blocking, having read all the");
		testing("data available and returns the number of bytes read.");
		testing("Canonical mode.");
		start
			SET_TIMEOUT
				expecting(SubStrlen);
				Zread(tty_fildes, inbuf, Strlen);
			CLEAR_ALARM
			expecting(0);
			Zstrncmp(String, inbuf, SubStrlen);
		finis

						/* Repeat test for non-canonical mode */
		prep("Set O_NONBLOCK flag and non-canonical mode for PORT_TTY.");
		start
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			Zfcntl(tty_fildes, F_SETFL, flags | O_NONBLOCK);

			Ztcgetattr(tty_fildes, tty_termios_p);
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VTIME] = 0;
			tty_termios_p->c_cc[VMIN] = 1;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
		finis

		prep("Write out data.");
		start
			SET_TIMEOUT
				expecting(SubStrlen);
				Zwrite(loop_fildes, String, SubStrlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM
		finis

		testing("O_NONBLOCK is set and not enough data is available");
		testing("to satisfy the entire request, a read() completes");
		testing("successfully, without blocking, having read all the");
		testing("data available and returns the number of bytes read.");
		testing("Non-canonical mode.");
		start
			SET_TIMEOUT
				expecting(SubStrlen);
				Zread(tty_fildes, inbuf, Strlen);
			CLEAR_ALARM
			expecting(0);
			Zstrncmp(String, inbuf, SubStrlen);
		finis
	finis

								/* No data */
	assertion(22, BASE);
	start
		prep("Clear PORT queues and set the O_NONBLOCK flag for PORT_TTY.");
		start
			Ztcflush(tty_fildes, TCIOFLUSH);
			Ztcflush(loop_fildes, TCIOFLUSH);
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			Zfcntl(tty_fildes, F_SETFL, flags | O_NONBLOCK);
		finis

		testing("O_NONBLOCK set and no data available,");
		testing("a read request completes without blocking");
		testing("and returns -1, with \"errno\" set to [EAGAIN].");
		testing("Canonical mode.");
		start
			SET_TIMEOUT
				expecting(EAGAIN);
				expecting((ssize_t)-1);
				Zread(tty_fildes, inbuf, sizeof(inbuf));
			CLEAR_ALARM
		finis

						/* Repeat test for non-canonical mode */
		prep("Set O_NONBLOCK flag and non-canonical mode for PORT_TTY.");
		start
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			Zfcntl(tty_fildes, F_SETFL, flags | O_NONBLOCK);

			Ztcgetattr(tty_fildes, tty_termios_p);
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VTIME] = 0;
			tty_termios_p->c_cc[VMIN] = 1;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
		finis

		testing("O_NONBLOCK set and no data available,");
		testing("a read request completes without blocking");
		testing("and returns -1, with \"errno\" set to [EAGAIN].");
		testing("Non-canonical mode.");
		start
			SET_TIMEOUT
				expecting(EAGAIN);
				expecting((ssize_t)-1);
				Zread(tty_fildes, inbuf, sizeof(inbuf));
			CLEAR_ALARM
		finis
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
