#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

static	char String[] = "With a capital T that rhymes with P that stands for!";
#define Strlen	(sizeof(String)-1)

unsigned char NL	= '\n';
#define	DELAY_TIME	3		/* Delay in seconds before terminating line. */

int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	pid_t	pid;
	char	inbuf[_POSIX_MAX_INPUT];

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	exit_on(ERROR);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(23, BASE)
		done();
	}
								/* Line processing */
	assertion(23, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Call open() on PORT_TTY without O_NONBLOCK flag.");
		start
			loop_prep(&loop_fildes, loop_termios_p);
			Zsetsid();
			SET_TIMEOUT
				tty_fildes = open_ctty(PORT_TTY, O_RDWR, PERM_CTTY);
				tty_init(tty_fildes, tty_termios_p);
			CLEAR_ALARM
			sleep(1);	/*	Allow time for connection to be established. */
		finis

		ENABLE_PSYNC
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Write data but delay before writing NL.");
			start
				sv_flush();
				SET_TIMEOUT
					expecting(Strlen);
					Zwrite(loop_fildes, String, Strlen);
					Ztcdrain(loop_fildes);
				CLEAR_ALARM

				SEND_PSYNC
				WAIT_PSYNCA
				SLEEP(DELAY_TIME)

				SET_TIMEOUT
					expecting(sizeof(char));
					Zwrite(loop_fildes, (char *)&NL, sizeof(char));
					Ztcdrain(loop_fildes);
				CLEAR_ALARM
			finis

			done();
		}

		testing("Call to read() does not return until NL is typed.");
		start
			SET_TIMEOUT
				WAIT_PSYNC
				SEND_PSYNCA
			CLEAR_ALARM

			SET_ALARM(DELAY_TIME +1)
				expecting(Strlen+sizeof(char));
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(unused_time > 1)
				error("%s\n%s%d\n%s%u\n",
						"Call to read() completed before NL became available.",
						"Expected time to complete to be >= to ",
						DELAY_TIME,
						"Actual time to complete was ",
						DELAY_TIME +1 -unused_time);

			if(strncmp(inbuf,String,Strlen) != 0)
				error("NL test failed. Read string != written string.");
		finis

		WAIT_FOR_PID(pid);

		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Write data but delay before writing EOL.");
			start
				sv_flush();
				SET_TIMEOUT
					expecting(Strlen);
					Zwrite(loop_fildes, String, Strlen);
					Ztcdrain(loop_fildes);
				CLEAR_ALARM

				SEND_PSYNC
				WAIT_PSYNCA
				SLEEP(DELAY_TIME)

				SET_TIMEOUT
					WRITE_TTY_SCC(loop_fildes, tty_termios_p->c_cc[VEOL]);
					Ztcdrain(loop_fildes);
				CLEAR_ALARM
			finis

			done();
		}

		testing("Call to read() does not return until EOL is typed.");
		start
			SET_TIMEOUT
				WAIT_PSYNC
				SEND_PSYNCA
			CLEAR_ALARM

			SET_ALARM(DELAY_TIME +1)
				expecting(Strlen +sizeof(cc_t));
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(unused_time > 1)
				error("%s\n%s%d\n%s%u\n",
						"Call to read() completed before EOL became available.",
						"Expected time to complete to be >= to ",
						DELAY_TIME,
						"Actual time to complete was ",
						DELAY_TIME +1 -unused_time);

			if(strncmp(inbuf, String, Strlen) != 0)
				error("EOL test failed. Read string != written string.");
		finis

		WAIT_FOR_PID(pid);

		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Write data but delay before writing EOF.");
			start
				sv_flush();
				SET_TIMEOUT
					expecting(Strlen);
					Zwrite(loop_fildes, String, Strlen);
					Ztcdrain(loop_fildes);
				CLEAR_ALARM

				SEND_PSYNC
				WAIT_PSYNCA
				SLEEP(DELAY_TIME)

				SET_TIMEOUT
					WRITE_TTY_SCC(loop_fildes, tty_termios_p->c_cc[VEOF]);
					Ztcdrain(loop_fildes);
				CLEAR_ALARM
			finis

			done();
		}

		testing("Call to read() does not return until EOF is typed.");
		start
			SET_TIMEOUT
				WAIT_PSYNC
				SEND_PSYNCA
			CLEAR_ALARM

			SET_ALARM(DELAY_TIME +1)
				expecting(Strlen);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(unused_time > 1)
				error("%s\n%s%d\n%s%u\n",
						"Call to read() completed before EOF became available.",
						"Expected time to complete to be >= to ",
						DELAY_TIME,
						"Actual time to complete was ",
						DELAY_TIME +1 -unused_time);

			if(strncmp(inbuf,String,Strlen) != 0)
				error("EOF test failed. Read string != written string.");
		finis

		WAIT_FOR_PID(pid);

		prep("Fork a child.");
		start
			SIGNAL(SIGINT, sig_int);
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Delay, then send signal to parent process.");
			start
				sv_flush();
				SET_TIMEOUT
					SEND_PSYNC
					WAIT_PSYNCA
				CLEAR_ALARM
				SLEEP(DELAY_TIME)
				Zkill(getppid(), SIGINT);
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		testing("Call to read() does not return until signal is received.");
		start
			SEND_PSYNCA
			SET_ALARM(DELAY_TIME +1)
				expecting(-1);
				expecting(EINTR);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(unused_time > 1)
				error("%s\n%s%d\n%s%u\n",
						"Call to read() completed signal was received.",
						"Expected time to complete to be >= to ",
						DELAY_TIME,
						"Actual time to complete was ",
						DELAY_TIME +1 -unused_time);
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC

	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
