#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


char String[] = "Yes Sir, that's my baby. No Sir, I don't mean maybe.";
#define Strlen	(sizeof(String)-1)

#define	DELAY_TIME	(WAIT_FOR_TIME +1)


int main(void)
{
	const cc_t READ_MIN	= 12;
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	inbuf[_POSIX_MAX_INPUT];
	pid_t	pid;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

						/*  Noncanonical Mode Input Processing */
						/* MIN > 0, TIME > 0 */

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(29, BASE)
		GTI_NOT_PROVIDED_RESULT(30, BASE)
		GTI_NOT_PROVIDED_RESULT(31, BASE)
		GTI_NOT_PROVIDED_RESULT(32, BASE)
		GTI_NOT_PROVIDED_RESULT(33, BASE)
		GTI_NOT_PROVIDED_RESULT(34, BASE)
		done();
	}

						/* Inter-byte time .1 seconds */
	assertion(29, BASE);
	start
		prep("Establish PORT_TTY as the control terminal, and PORT_LOOP.");
		prep("Setup PORT_TTY in non-canonical mode.");
		prep("Set on PORT_TTY MIN>0 and TIME=2.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			tty_termios_p->c_lflag &= ~(ICANON | ECHO);
			tty_termios_p->c_cc[VMIN] = READ_MIN;
			tty_termios_p->c_cc[VTIME] = 2;		/* .2 second inter-byte timer */
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Delay a long time, send byte, delay .1 second,");
			testing("send byte, delay .3 seconds.");
			start
				SLEEP(DELAY_TIME)
					expecting(sizeof(char));
					Zwrite(loop_fildes, "X", sizeof(char));
					Ztcdrain(loop_fildes);

					dt_secs(1);				/* Delay 1/10 of a second. */

					expecting(sizeof(char));
					Zwrite(loop_fildes, "Y", sizeof(char));
					Ztcdrain(loop_fildes);

					dt_secs(3);				/* Delay 3/10 of a second. */

					expecting(sizeof(char));
					Zwrite(loop_fildes, "Z", sizeof(char));
					Ztcdrain(loop_fildes);
			finis

			done();
		}

		testing("TIME acts as an inter-byte timer of .1 seconds granularity");
		testing("and is activated after the first byte is received.");
		start
			SET_ALARM(DELAY_TIME +4)
				expecting(2);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis
		WAIT_FOR_PID(pid);
	finis

								/* Reset inter-byte timer after each byte */
#define	INTER_BYTE_TIME	10		/* 1 second */
#define	FUDGE_FACTOR	2		/* .2 seconds */
	assertion(30, BASE);
	start
		prep("Set PORT_TTY parameters and clear queues of PORTs.");
		start
			tty_termios_p->c_cc[VTIME] = INTER_BYTE_TIME;
			tty_termios_p->c_cc[VMIN] = READ_MIN;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			Ztcflush(loop_fildes, TCIOFLUSH);
		finis

		ENABLE_PSYNC
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Write out test data in timed intervals.");
			start
				int	i,j;

				sv_flush();
				SIGNAL(SIGINT, SIG_DFL)
				SEND_PSYNC
				for(i=5; i<16; i++) {
					for(j=0; j<2; j++) {
						dt_secs(i);		/* Delay approximately i/10 seconds. */
						expecting(sizeof(char));
						Zwrite(loop_fildes, String, sizeof(char));
						Ztcdrain(loop_fildes);
					}
				}
			finis
				
			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		testing("TIME acts as an inter-byte timer of .1 second granularity");
		testing("and is reset after each byte is received.");
		start
			ssize_t nbyte;

			SET_TIMEOUT
				nbyte = Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(	(nbyte/2 +5 < INTER_BYTE_TIME -FUDGE_FACTOR) ||
				(nbyte/2 +5 > INTER_BYTE_TIME +FUDGE_FACTOR) )
				error("%s%d/10 seconds.\n%s%d/10 seconds.",
						"Expected interbyte timer to expire in ",
						INTER_BYTE_TIME,
						"Actually interbyte timer expired in ",
						nbyte/2 +5);
			else
				if(	(nbyte/2 +5 < INTER_BYTE_TIME -1) ||
					(nbyte/2 +5 > INTER_BYTE_TIME +1) )
														/* Close enough */
					message("%s%d/10 seconds.\n%s%d/10 seconds.",
							"Expected interbyte timer to expire in ",
							INTER_BYTE_TIME,
							"Actually interbyte timer expired in ",
							nbyte/2 +5);
			Zkill(pid, SIGINT);					/* Terminate child. */
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis


								/* read() exits on MIN */
	assertion(31, BASE);
	start
		prep("Set PORT_TTY parameters and clear queues of PORTs.");
		start
			tty_termios_p->c_cc[VMIN] = READ_MIN;
			tty_termios_p->c_cc[VTIME] = 10;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			Ztcflush(loop_fildes, TCIOFLUSH);
			dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			Ztcflush(tty_fildes, TCIOFLUSH);
		finis

		prep("Write out MIN bytes.");
		start
			SET_TIMEOUT
				expecting(READ_MIN);
				Zwrite(loop_fildes, String, (size_t)READ_MIN);
				Ztcdrain(loop_fildes);
			CLEAR_ALARM
		finis

		testing("When the total input queue of MIN bytes are received");
		testing("before the inter-byte timer expires, read() is satisfied.");
		start
			SET_TIMEOUT
				expecting(READ_MIN);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis
	finis


							/* inter-byte timer terminates read() */
	assertion(32, BASE);
	start
		prep("Set PORT_TTY parameters and clear queues of PORTs.");
		start
			tty_termios_p->c_cc[VMIN] = 2*Strlen;
			tty_termios_p->c_cc[VTIME] = 10;				/* 1 second */
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			Ztcflush(loop_fildes, TCIOFLUSH);
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);

				SLEEP(2)

				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
			CLEAR_ALARM

			done();
		}

		testing("When the inter-byte timer expires before MIN bytes");
		testing("are received, then the bytes received to that point are");
		testing("returned from the read().");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, inbuf, Strlen*2);
			CLEAR_ALARM
		finis
		WAIT_FOR_PID(pid);
	finis


							/* read() blocks till data received */
	assertion(33, BASE);
	start
		prep("Set PORT_TTY parameters and clear queues of PORTs.");
		start
			tty_termios_p->c_cc[VMIN] = READ_MIN;
			tty_termios_p->c_cc[VTIME] = (DELAY_TIME -1)*10;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			Ztcflush(loop_fildes, TCIOFLUSH);
			dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			Ztcflush(tty_fildes, TCIOFLUSH);
		finis

		ENABLE_PSYNC
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Delay, write data, delay, then write again.");
			start
				SET_TIMEOUT
					WAIT_PSYNC
					SLEEP(DELAY_TIME);

					expecting(sizeof(char));
					Zwrite(loop_fildes, "A", sizeof(char));
					Ztcdrain(loop_fildes);

					SLEEP(DELAY_TIME)

					expecting(sizeof(char));
					Zwrite(loop_fildes, "B", sizeof(char));
					Ztcdrain(loop_fildes);
				CLEAR_ALARM
			finis
					
			done();
		}
	
		testing("Call to read() blocks until MIN bytes are received.");
		testing("Then interbyte timer is activated.");
		start
			SET_TIMEOUT
				SEND_PSYNC
								/* Expect a single byte, before interbyte
								   timer expires. */
				expecting(sizeof(char));
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

						/* Time to read data is:
								DELAY_TIME +(DELAY_TIME -1) +transmit
						*/
			if(TIMEOUT -unused_time < DELAY_TIME +DELAY_TIME -1)
				error("%s\n%s%d\n%s%u\n",
						"Time required to receive data was too short.",
						"Expected minimum delay of ",
						DELAY_TIME +DELAY_TIME -1,
						"Actual delay was ",
						TIMEOUT - unused_time);
		finis
		WAIT_FOR_PID(pid);

							/* read() blocks till signal received */
		prep("Clear queues of PORTs.");
		start
			Ztcflush(loop_fildes, TCIOFLUSH);
			dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			Ztcflush(tty_fildes, TCIOFLUSH);
		finis

		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Delay, then send signal to parent process.");
			start
				sv_flush();
				WAIT_PSYNC
				SLEEP(DELAY_TIME)
				Zkill(getppid(), SIGINT);
			finis

			done();
		}

		testing("Call to read() does not return until signal is received.");
		start
			SIGNAL(SIGINT, sig_int);
			SEND_PSYNC
			SET_ALARM(DELAY_TIME +1)
				expecting(EINTR);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(unused_time > 1)
				error("%s\n%s%d\n%s%u\n",
						"Call to read() completed signal was received.",
						"Expected time to complete to be >= to ",
						DELAY_TIME,
						"Actual time to complete was ",
						DELAY_TIME +1 -unused_time);
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis


						/* Data in queue at start of read() */
	assertion(34, BASE);
	start
		prep("Set PORT_TTY parameters and clear queues of PORTs.");
		start
			tty_termios_p->c_cc[VMIN] = READ_MIN;
			tty_termios_p->c_cc[VTIME] = 10;		/* 1 sec timer */
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			Ztcflush(loop_fildes, TCIOFLUSH);
		finis

		ENABLE_PSYNC
		prep("Write out test data.");
		start
			expecting(sizeof(char));
			Zwrite(loop_fildes, String, sizeof(char));
			Ztcdrain(loop_fildes);
				/*	Ensure data arrives at input queue before continuing.	*/
			dt_secs(TIME_ACQUIRE_XMITTED_DATA);
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Delay then write data.");
			start
				sv_flush();
				SEND_PSYNC
				SLEEP(2);
				SET_TIMEOUT
					expecting(sizeof(char));
					Zwrite(loop_fildes, String, sizeof(char));
					Ztcdrain(loop_fildes);
				CLEAR_ALARM
			finis
				
			done();
		}

		testing("Timer is started as if data received on call to read().");
		start
			SET_TIMEOUT
				WAIT_PSYNC
				expecting(sizeof(char));
				Zread(tty_fildes, inbuf, sizeof(char)*2);
			CLEAR_ALARM
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
