#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

#define	DELAY_TIME	3


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char inbuf[_POSIX_MAX_INPUT];
	pid_t pid;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	continue_on(ERROR);

						/*  Noncanonical Mode Input Processing */
						/* MIN > 0, TIME = 0 */
	assertion(35, BASE);
	start
		GTI_NOT_PROVIDED_EXIT

		ENABLE_PSYNC
			prep("Establish PORT_TTY as the control terminal, and PORT_LOOP.");
			prep("Setup PORT_TTY in non-canonical mode.");
			prep("Set on PORT_TTY MIN>0 and TIME=0.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);
	
									/* Set for non-canonical mode. */
				tty_termios_p->c_lflag &= ~(ICANON | ECHO);
				tty_termios_p->c_cc[VMIN] = 2;
				tty_termios_p->c_cc[VTIME] = 0;
				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			finis
	
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				testing("Write out single character, delay, write another.");
				start
					sv_flush();
					SET_TIMEOUT
						expecting(sizeof(char));
						Zwrite(loop_fildes, "A", sizeof(char));
						Ztcdrain(loop_fildes);

						WAIT_PSYNC
						SLEEP(DELAY_TIME);

						expecting(sizeof(char));
						Zwrite(loop_fildes, "B", sizeof(char));
						Ztcdrain(loop_fildes);
					CLEAR_ALARM
				finis
					
				done();
			}
	
			testing("Call to read() blocks until MIN bytes are received.");
			start
				SEND_PSYNC
				SET_TIMEOUT
					expecting(sizeof(char)*2);
					Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
				CLEAR_ALARM
	
				if(TIMEOUT -unused_time < DELAY_TIME)
					error("%s\n%s%d\n%s%u\n",
							"Delay to receive data was too short.",
							"Expected minimum delay of ",
							DELAY_TIME,
							"Actual delay was ",
							TIMEOUT - unused_time);
			finis
			WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
