#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

char String[] = "Yes Sir, that's my baby. No Sir, I don't mean maybe.";
#define Strlen	(sizeof(String)-1)

int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char inbuf[_POSIX_MAX_INPUT];
	pid_t pid;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(36, BASE)
		GTI_NOT_PROVIDED_RESULT(37, BASE)
		GTI_NOT_PROVIDED_RESULT(38, BASE)
		done();
	}

	continue_on(ERROR);

						/*  Noncanonical Mode Input Processing */
						/* MIN = 0, TIME > 0 */
	assertion(36, BASE);
	start
		prep("Establish PORT_TTY as the control terminal, and PORT_LOOP.");
		prep("Setup PORT_TTY in non-canonical mode.");
		prep("Set on PORT_TTY MIN=0 and TIME>0.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

								/* Set for non-canonical mode. */
			tty_termios_p->c_lflag &= ~(ICANON | ECHO);
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 30;		/* 3 sec read timer */
			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
		finis

		prep("Place data in input queue.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
			CLEAR_ALARM
		finis

		testing("When any data is available to satisfy the read() before the");
		testing("read timer expires, the read() is satisfied.");
		start
			SET_ALARM(2);		/* Alarm time set lower than read timer. */
				expecting(SUCCESS);
				Zread(tty_fildes, inbuf, Strlen);
			CLEAR_ALARM
		finis
	finis

	assertion(37, BASE);
	start
		prep("Clear PORT queues.");
		start
			Ztcflush(loop_fildes, TCIOFLUSH);
			dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			Ztcflush(tty_fildes, TCIOFLUSH);
			tty_termios_p->c_cc[VTIME] = 10;		/* 1 sec read timer */
			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
		finis

		testing("When the input queue is empty, zero bytes shall be");
		testing("returned from the read() when the read timer expires.");
		start
			SET_TIMEOUT
				expecting(0);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM
	
			if(TIMEOUT -unused_time < 1)
				error("%s\n%s\n%s%u\n",
						"Time to complete test was too short.",
						"Expected minimum duration of 1 second",
						"Actual duration was ",
						TIMEOUT - unused_time);
		finis
	finis


						/* Data in queue at start of read(). */
	assertion(38, BASE);
	start
		prep("Clear PORT queues.");
		start
			Ztcflush(tty_fildes, TCIOFLUSH);
			Ztcflush(loop_fildes, TCIOFLUSH);
		finis
	finis

	ENABLE_PSYNC
		prep("Write out test data.");
		start
			expecting(sizeof(char));
			Zwrite(loop_fildes, String, sizeof(char));
			Ztcdrain(loop_fildes);
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Delay then write data.");
			start
				sv_flush();
				SEND_PSYNC
				SLEEP(2);
				SET_TIMEOUT
					expecting(sizeof(char));
					Zwrite(loop_fildes, String, sizeof(char));
					Ztcdrain(loop_fildes);
				CLEAR_ALARM
			finis
				
			done();
		}

		testing("Timer is started as if data received immediately after the");
		testing("call to read().");
		start
			SET_TIMEOUT
				WAIT_PSYNC
				expecting(sizeof(char));
				Zread(tty_fildes, inbuf, sizeof(char)*2);
			CLEAR_ALARM
		finis

		WAIT_FOR_PID(pid);
	DISABLE_PSYNC

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
