#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	String[_POSIX_MAX_INPUT];
	int		flags;
	ssize_t	ret;
	long	n_bytes = 0;

	setup(E_INTERFACE);
	funct(E_INTERFACE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(42, BASE)
		GTI_NOT_PROVIDED_RESULT(43, BASE)
		GTI_NOT_PROVIDED_RESULT(44, BASE)
		GTI_NOT_PROVIDED_RESULT(45, BASE)
		done();
	}

	assertion(45, EXTENDED);

	exit_on(ERROR);		/* Tests #43 and #44 require #42 to be successful. */

	assertion(42, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and open PORT_LOOP.");
		prep("Suspend communication from PORT_TTY.");
		start
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
			Ztcflow(tty_fildes, TCOOFF);
		finis

		prep("Set the O_NONBLOCK flag for PORT_TTY.");
		start
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			flags |= O_NONBLOCK;
			Zfcntl(tty_fildes, F_SETFL, flags);
		finis

		testing("Call to write() when no data can be accepted returns a");
		testing("value of -1 and sets errno to [EAGAIN].");
		start
			if(PCTS_GTI_BUFFERS_OUTPUT) {
				SET_TIMEOUT
						/*	Determine how many bytes can be written (n_bytes)
							before implementation stops buffering data.
						*/
					while(TRUE) {
						ret = write(tty_fildes, String, _POSIX_MAX_INPUT);
								/* Note: Value returned is either >0 or -1.
										 Write to terminal device "shall write
										 what it can and return the number of
										 bytes written" (POSIX.1, page 120,
										 lines 258-261).
								*/
						if(ret != (ssize_t)(_POSIX_MAX_INPUT))
							break;
						n_bytes += ret;
					}
					if(ret != -1)
						n_bytes += ret;
				CLEAR_ALARM
			}

			SET_ALARM(WAIT_FOR_TIME)
				expecting(-1);
				expecting(EAGAIN);
				Zwrite(tty_fildes, String, sizeof(char));
			CLEAR_ALARM
		finis
	finis
							/*	Continue with conditions established in #42. 
								This test requires #42 to be PASS before
								this test is to be run.
							*/
	assertion(43, BASE);
	start
		prep("Clear the O_NONBLOCK flag for PORT_TTY.");
		start
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			flags &= ~O_NONBLOCK;
			Zfcntl(tty_fildes, F_SETFL, flags);
		finis

		testing("When output is blocked, a call to write() blocks until a");
		testing("signal is received.");
		start
			SET_ALARM(WAIT_FOR_TIME)
				write(tty_fildes, String, _POSIX_MAX_INPUT);
			EXPECT_ALARM

			Ztcflow(tty_fildes, TCOON);		/*	Restart suspended output. */
		finis

		cleanup("Clear queues of ports.");
		start
			Ztcflush(tty_fildes, TCIOFLUSH);			/* Flush output */
			dt_secs(1);
			Ztcflush(loop_fildes, TCIOFLUSH);			/* input queues.*/

			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			flags |= O_NONBLOCK;
			Zfcntl(tty_fildes, F_SETFL, flags);
		finis


							/*	Output blocked until data is accepted.
								Tested only for implementations that
								do not support PCTS_GTI_BUFFERS_OUTPUT
								and those that provide a limited amount of
								buffering.
							*/
		prep("Set IXON flag for PORT_TTY.");
		start
			tty_termios_p->c_iflag |= IXON;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
		finis

		prep("Set IXON, send a STOP character to suspend output, attain");
		prep("state when no further output can be written.");
		start
			SET_TIMEOUT
				WRITE_TTY_SCC(loop_fildes, tty_termios_p->c_cc[VSTOP]);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM

			if(PCTS_GTI_BUFFERS_OUTPUT) {
				SET_TIMEOUT
					while(TRUE) {
						if(write(tty_fildes, String, _POSIX_MAX_INPUT) !=
										(ssize_t) _POSIX_MAX_INPUT)
							break;		/*	Output buffer full. */
						
					}
				CLEAR_ALARM
			}
		finis

		prep("Clear the O_NONBLOCK flag for PORT_TTY.");
		start
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			flags &= ~O_NONBLOCK;
			Zfcntl(tty_fildes, F_SETFL, flags);
		finis

		testing("When output is blocked, a call to write() blocks until");
		testing("the data can be accepted.");
		start
			SET_ALARM(WAIT_FOR_TIME)
				write(tty_fildes, String, _POSIX_MAX_INPUT);
			EXPECT_ALARM

			SET_TIMEOUT
				Ztcflush(tty_fildes, TCIOFLUSH);			/* Flush output */
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);			/*      and     */
				Ztcflush(loop_fildes, TCIOFLUSH);			/* input queues.*/
										/*	Send IXON to restart output. */
				WRITE_TTY_SCC(loop_fildes, tty_termios_p->c_cc[VSTART]);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);

				expecting(_POSIX_MAX_INPUT);
				Zwrite(tty_fildes, String, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis

		cleanup("Clear queues of ports.");
		start
			Ztcflush(tty_fildes, TCIOFLUSH);			/* Flush output */
			dt_secs(TIME_ACQUIRE_XMITTED_DATA);			/*      and     */
			Ztcflush(loop_fildes, TCIOFLUSH);			/* input queues.*/

			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			flags |= O_NONBLOCK;
			Zfcntl(tty_fildes, F_SETFL, flags);
		finis

	finis

#define	LOOP_LIMIT	100
	assertion(44, EXTENDED_TESTED);
	start
		char	*outbuf;
		ssize_t	bytes_wr, bytes_rd, i, total_wr;
		long	max_write;
		int		loop_cnt;

		break_on(ERROR);
		prep("Set the O_NONBLOCK flag for PORT_TTY.");
		prep("Suspend communication from PORT_TTY.");
		start
			flags = Zfcntl(tty_fildes, F_GETFL, 0);
			flags |= O_NONBLOCK;
			Zfcntl(tty_fildes, F_SETFL, flags);

									/* Set for non-canonical mode. */
			loop_termios_p->c_lflag &= ~(ICANON | ECHO);
			loop_termios_p->c_cc[VMIN] = 0;
			loop_termios_p->c_cc[VTIME] = 10;
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			Ztcflush(tty_fildes, TCIOFLUSH);
			Ztcflush(loop_fildes, TCIOFLUSH);
		finis

		testing("Call to write(), when some data can be accepted, returns");
		testing("the number of bytes written.");
		start
			max_write = n_bytes +PCTS_MAX_INPUT;
			for(total_wr=0, loop_cnt=0; loop_cnt<LOOP_LIMIT;
										loop_cnt++, total_wr += max_write) {
				message("Attempt to write %ld bytes, for a total of %ld bytes.",
											(long) max_write,
											(long) (total_wr +max_write));
				SET_TIMEOUT
					outbuf = (char *)Zmalloc(max_write);
					memset(outbuf, (int)'A', max_write);
                    bytes_wr = write(tty_fildes, outbuf, (size_t)max_write);
					if((bytes_wr == -1) && (errno == EAGAIN)) {
						message("write() of %ld bytes returned -1, "
								"with errno of [EAGAIN].\n",
								(long) max_write );
						bytes_wr = write(tty_fildes, outbuf, (size_t) 1);
						if((bytes_wr == -1) && (errno == EAGAIN))
							message("Subsequent write() of one byte returned "
									"-1, with errno of [EAGAIN].\n");
						else
							error("Subsequent write() of one byte returned -1, "
									"with errno of [%s].\n",
									err_lookup(errno) );
					}
					else {
						if(bytes_wr == -1)
							error("Call to write() returned -1, "
								"with errno of [%s]\n ",
								err_lookup(errno));
					}
				CLEAR_ALARM

				if(bytes_wr == 0)
					error("Bytes written returned a count of zero.");
				if(bytes_wr != (ssize_t)max_write)
					break;
			}

			if(loop_cnt==LOOP_LIMIT)
				error("Unable to fill up input buffer of terminal."); 
			if(bytes_wr > 0)
				total_wr += bytes_wr;
		finis

		testing("Read (%ld) reported bytes written.", (long) total_wr);
		start
			SET_TIMEOUT
				for(i=0; i<total_wr; i += bytes_rd) {
					bytes_rd = Zread(loop_fildes, outbuf, (size_t)max_write);
					/*
					message("Last read() read %ld bytes.\nTotal is %ld bytes.",
								(long)bytes_rd, (long)(i +bytes_rd));
					*/
					if(bytes_rd == (ssize_t)0) {
						error("Wrote %ld bytes but only read %ld bytes.",
								(long)total_wr, (long)(i +bytes_rd));
						break;
					}
				}
			CLEAR_ALARM
		finis
	finis


	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
