#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

char String[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ012345!";
const int Strlen = sizeof(String)-1;

const int FIVE		= '\037';		/*									*/
const int SIX		= '\077';		/*									*/
const int SEVEN		= '\177';		/*									*/
const int EIGHT		= '\377';		/*									*/

void csize(tcflag_t, char *, int);

static	int		tty_fildes, loop_fildes;
static	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
static	char	inbuf[_POSIX_MAX_INPUT];


int main(void)
{
	setup(E_SETTABLE);
	funct(E_SETTABLE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(39, BASE)
		GTI_NOT_PROVIDED_RESULT(40, BASE)
		GTI_NOT_PROVIDED_RESULT(41, BASE)
		GTI_NOT_PROVIDED_RESULT(42, BASE)
		done();
	}

									/* CS5 */
	assertion(39, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Set PORT_TTY and PORT_LOOP to non-canonical, MIN=TIME=0.");
		prep("Clear ISTRIP flags.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			loop_termios_p->c_iflag &= ~ISTRIP;
			Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);

			tty_termios_p->c_lflag &= ~ICANON;
					/* Return number of bytes requested or that available. */
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
									/* Disable special processing of input. */
			tty_termios_p->c_iflag &= ~(ISTRIP|IGNBRK|BRKINT|ICRNL|INLCR);
			tty_termios_p->c_lflag &= ~ISIG;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
		finis

		csize(CS5, "CS5", FIVE);
	finis

	assertion(40, BASE);
	start
		csize(CS6, "CS6", SIX);
	finis

	assertion(41, BASE);
	start
		csize(CS7, "CS7", SEVEN);
	finis

	assertion(42, BASE);
	start
		csize(CS8, "CS8", EIGHT);
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}


void csize(tcflag_t mask, char *name, int bits)
		/*	mask	CSIZE value,
			name	CSIZE name
			bits	char size bit mask
		*/
{
	unsigned char	eight_bits;
	pid_t	pid;

	FORK
		prep("Does implementation support %s ?", name);
		start
			loop_termios_p->c_cflag &= ~CSIZE;
			loop_termios_p->c_cflag |= mask;
			errno = 0;
			tcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			if(errno && (errno != EINVAL))
				error("%s\n%s%s\n",
						"tcsetattr() - failed unexpectedly",
						"Returned with errno set to ",
						err_lookup(errno) );
			Ztcgetattr(loop_fildes, loop_termios_p);
		finis
	
		if((loop_termios_p->c_cflag & CSIZE) == mask) {
			prep("Set CSIZE mask bits to %s on PORT_TTY.", name);
			start
				tty_termios_p->c_cflag &= ~CSIZE;
				tty_termios_p->c_cflag |= mask;
				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			finis
	
			testing("Attempt write and read data of character size %s.", name);
			start
				int	i;

				SET_TIMEOUT
					eight_bits = '\377';
					expecting(sizeof(char));
					Zwrite(loop_fildes, (char *)&eight_bits, sizeof(char));
					expecting(Strlen);
					Zwrite(loop_fildes, String, Strlen);
					Ztcdrain(loop_fildes);
					dt_secs(TIME_ACQUIRE_XMITTED_DATA);
				CLEAR_ALARM
	
				SET_TIMEOUT
					expecting(sizeof(char));
					Zread(tty_fildes, inbuf, sizeof(char));
					expecting(Strlen);
					Zread(tty_fildes, inbuf+1, _POSIX_MAX_INPUT);
				CLEAR_ALARM
	
							/*	Compare transmitted with received data
								based on the character size transmitted.
							*/
				if((eight_bits & bits) != (unsigned char) inbuf[0])
						error("%s%x(Hex)\n%s%x(Hex)\n",
							"Expected: ",
							eight_bits & bits,
							"Received: ",
							(unsigned char) inbuf[0]);

				for(i=1; i<Strlen +1; i++) {
					if((String[i -1] & bits) != inbuf[i])
						error("Character \"%c\"%s.\n%s%x(Hex)\n%s%x(Hex)\n",
							String[i-1],
							" improperly processed.",
							"Expected: ",
							String[i -1] & bits,
							"Received: ",
							inbuf[i]);
				}
			finis
		}
		else
			not_implemented(name);
	KROF
}
