#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

char String[] = "Test THIS";
#define Strlen	(sizeof(String)-1)


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	inbuf[_POSIX_MAX_INPUT];
	int 	i;

	setup(E_SETTABLE);
	funct(E_SETTABLE);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(49, BASE)
		GTI_NOT_PROVIDED_RESULT(50, BASE)
		GTI_NOT_PROVIDED_RESULT(51, BASE)
		GTI_NOT_PROVIDED_RESULT(52, BASE)
		done();
	}

	assertion(49, EXTENDED);

									/* PARENB and PARODD set */
	assertion(50, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Set PORT_TTY to non-canonical, MIN=TIME=0.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			tty_termios_p->c_cflag |= (PARENB | PARODD);
			tty_termios_p->c_iflag |= INPCK;
			tty_termios_p->c_iflag &= ~(IGNPAR | PARMRK);
					/* Return number of bytes requested or that available. */
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);

			loop_termios_p->c_cflag |= (PARENB | PARODD);
			Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
		finis

		testing("When PARENB and INPCK are set, then input parity");
		testing("checking is enabled and characters received with");
		testing("correct parity are as sent and do not generate an error.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM

			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(strncmp(String, inbuf, Strlen))
				error("%s\n%s\"%*s\"\n%s\"%*s\"",
						"Data sent not the same as data received.",
						"Sent: ",
						Strlen,
						String,
						"Received: ",
						Strlen,
						inbuf);
		finis

		testing("When PARENB and INPCK are set, then input parity");
		testing("checking is enabled and characters received with");
		testing("incorrect parity generate an error.");
		start
			tty_termios_p->c_cflag &= ~PARODD;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);

			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM

			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			for(i=0; i<Strlen; i++)
				if(inbuf[i] != '\0')
					error("%s %d\n%s\n%s%o",
							"Input character",
							i,
							"Expected: \\0",
							"Received: \\",
							(int)inbuf[i]);
		finis
	finis

	assertion(51, EXTENDED);
									/* PARENB set and PARODD clear */
	assertion(52, BASE);
	start
		prep("Clear PARODD on PORTs.");
		start
			tty_termios_p->c_cflag &= ~PARODD;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			loop_termios_p->c_cflag &= ~PARODD;
			Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
		finis

		testing("When PARENB and INPCK are set, then input parity");
		testing("checking is enabled and characters received with");
		testing("correct parity are as sent and do not generate an error.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM

			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			if(strncmp(String, inbuf, Strlen))
				error("%s\n%s\"%*s\"\n%s\"%*s\"",
						"Data sent not the same as data received.",
						"Sent: ",
						Strlen,
						String,
						"Received: ",
						Strlen,
						inbuf);
		finis

		testing("When PARENB and INPCK are set, then input parity");
		testing("checking is enabled and characters received with");
		testing("incorrect parity generate an error.");
		start
			tty_termios_p->c_cflag |= PARODD;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);

			SET_TIMEOUT
				expecting(Strlen);
				Zwrite(loop_fildes, String, Strlen);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM

			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM

			for(i=0; i<Strlen; i++)
				if(inbuf[i] != '\0')
					error("%s %d\n%s\n%s%o",
							"Input character",
							i,
							"Expected: \\0",
							"Received: \\",
							(int)inbuf[i]);
		finis
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
