#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


int	main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	inbuf[_POSIX_MAX_INPUT];
	char	outbuf[_POSIX_MAX_INPUT];
	struct	tms	tms_buf;
	clock_t	t_beg, t_end;
	size_t	nb_xmit, n_read;
	clock_t	t_xmit;
	long	clk_tcks;

	setup(E_TCDRAIN);
	funct(E_TCDRAIN);

	assertion(4, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Clear ICANON on PORT_LOOP, set speed to lowest rate supported.");
		start
			GTI_NOT_PROVIDED_EXIT

			exit_on(ERROR);

			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());

			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			memset(outbuf, 'A', _POSIX_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
			nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
			if(nb_xmit > _POSIX_MAX_INPUT)
				nb_xmit = _POSIX_MAX_INPUT;
		finis

		testing("Call to tcdrain() waits until all output written to the");
		testing("terminal has been transmitted before returning.");
		start
			break_on(ERROR);

			clk_tcks = Zsysconf(_SC_CLK_TCK);
			t_xmit = clk_tcks*nb_xmit*10/baud_value(baud_min());

					/*	Record the starting time before the write() is
						called.  In case write() blocks, the time to drain
						the data is monitored properly.
					*/
			t_beg = Ztimes(&tms_buf);
			if(PCTS_GTI_BUFFERS_OUTPUT) {
				SET_TIMEOUT
					expecting(nb_xmit);
					Zwrite(loop_fildes, outbuf, nb_xmit);	
				CLEAR_ALARM
			}
			else
				t_xmit = 0;

			expecting(0);
			Ztcdrain(loop_fildes);
			t_end = Ztimes(&tms_buf);
			Ztcflush(loop_fildes, TCOFLUSH);

			if(PCTS_GTI_BUFFERS_OUTPUT) {
				if(t_end -t_beg < t_xmit -t_xmit/10)
					error("%s%lu clock-ticks.\n%s%lu clock-ticks.",
							"Expected tcdrain() to take ",
							(unsigned long)t_xmit,
							"Actually time taken was ",
							(unsigned long) (t_end -t_beg) );
			}

			SET_TIMEOUT
										/* Ensure no data lost. */
				for(n_read=0; n_read != nb_xmit;)
					n_read += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
