#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	outbuf[_POSIX_MAX_INPUT];
	struct	tms	tms_buf;
	clock_t	t_beg, t_end;
	size_t	nb_xmit;
	long	clk_tcks;
	pid_t	pid;
	int		i;
	sigset_t	set;

	setup(E_TCDRAIN);
	funct(E_TCDRAIN);

	exit_on(ERROR);

	SIGNAL(SIGHUP, SIG_IGN);

	assertion(5, BASE);
	start
		GTI_NOT_PROVIDED_EXIT

		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());

			memset(outbuf, 'A', _POSIX_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
			nb_xmit = (baud_value(baud_min())/(sizeof(char)*8 +2))*TIMEOUT/2;
			if(nb_xmit > _POSIX_MAX_INPUT)
				nb_xmit = _POSIX_MAX_INPUT;
			clk_tcks = Zsysconf(_SC_CLK_TCK);
		finis

		for(i=0; i<2; i++) {
			if(i == 0) 
				prep("Set TOSTOP flag.");
			else
				prep("Clear TOSTOP flag.");
			start
				if(i == 0) 
					tty_termios_p->c_lflag |= TOSTOP;
				else
					tty_termios_p->c_lflag &= ~TOSTOP;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);
			finis
	
			if(PCTS_GTI_BUFFERS_OUTPUT) {
				prep("Write to control terminal (PORT_TTY).");
				start
								/*	Assume processing of write() takes longer
									than processing of this test for devices
									that buffer GTI output.
								*/
					SET_TIMEOUT
						expecting(nb_xmit);
						Zwrite(tty_fildes, outbuf, nb_xmit);
					CLEAR_ALARM
				finis
			}
	
			FORK
				prep("Establish background process group with two processes.");
				start
					Zsetpgid((pid_t)0, (pid_t)0);
					SIGNAL(SIGTTOU, sig_int);
					sig_rcvd = 0;
				finis
	
				prep("Fork a child process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					testing("Child process in background process group");
					testing("received SIGTTOU signal.");
					start
						sleep(WAIT_FOR_TIME);
						if(sig_rcvd == 0)
							error("Child process did not receive SIGTTOU.");
					finis
	
					done();
				}
		
				testing("Call to tcdrain(), on the controlling terminal of");
				testing("the process, from a member of background");
				testing("process group that is not blocking nor ignoring");
				testing("the SIGTTOU signal when TOSTOP");
				if(i > 1) 
					testing("is set is unsuccessful.");
				else
					testing("is clear is unsuccessful.");
				start
					SIGUNBLOCK(set, SIGTTOU);
					SET_TIMEOUT
						t_beg = Ztimes(&tms_buf);
						expecting(FAILURE);
						Ztcdrain(tty_fildes);
						expecting(SUCCESS);
						t_end = Ztimes(&tms_buf);
					CLEAR_ALARM
				finis
	
				testing("Call to tcdrain() returned immediately.");
				start
					if(t_end -t_beg >= clk_tcks)
						error("%s\n%s%lu clock-ticks.",
								"Expected tcdrain() to complete immediately.",
								"Actually time taken was ",
								(unsigned long) (t_end -t_beg) );
				finis
	
				testing("SIGTTOU signal is sent to calling process.");
				start
					dt_secs(1);
					if(sig_rcvd == 0)
						error("Calling process did not receive SIGTTOU.");
				finis
	
				WAIT_FOR_PID(pid);
			KROF
		}
	finis
	
	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis
	
	done();
}
