#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

void	com_code(void);

static	int		tty_fildes, loop_fildes;
static	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
static	sigset_t set;

static	char	outbuf[_POSIX_MAX_INPUT];
static	size_t	nb_xmit, n_read;

int main(void)
{
	pid_t	pid;

	setup(E_TCDRAIN);
	funct(E_TCDRAIN);

	assertion(8, BASE);
	start
		GTI_NOT_PROVIDED_EXIT

		prep("Establish PORT_TTY and PORT_LOOP as non-controlling");
		prep("terminals.");
		prep("Clear ICANON on PORT_LOOP, set speed to lowest rate supported.");
		start
			exit_on(ERROR);
	
			termios_nct(&tty_fildes, tty_termios_p, &loop_fildes,
															loop_termios_p);
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());

			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			memset(outbuf, 'A', _POSIX_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
			nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
			if(nb_xmit > _POSIX_MAX_INPUT)
				nb_xmit = _POSIX_MAX_INPUT;
		finis

									/* Foreground process test. */
		ENABLE_PSYNC
		FORK
			com_code();
		KROF


									/* Background process tests. */
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Catch SIGTTOU and establish a background process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			com_code();

			prep("Block SIGTTOU in background process group.");
			start
				SIGBLOCK(set, SIGTTOU);
			finis

			com_code();

			prep("Unblock and ignore SIGTTOU in background process group.");
			start
				SIGNAL(SIGTTOU, SIG_IGN);
				SIGUNBLOCK(set, SIGTTOU);
			finis

			com_code();

			done();
		}
		wait_for(pid, 3 * TIMEOUT);

									/* Orphaned background process test. */
		FORK
			prep("Catch SIGTTOU and establish an orphaned background");
			prep("process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				SLEEP(WAIT_FOR_TIME)

				com_code();
				SEND_PSYNCA

				done();
			}
		KROF
										/* Wait for test to complete. */
		SET_TIMEOUT
			WAIT_PSYNCA
		CLEAR_ALARM
		DISABLE_PSYNC
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}

void com_code(void)
{
	char	inbuf[_POSIX_MAX_INPUT];
	struct	tms	tms_buf;
	clock_t	t_beg, t_end;
	clock_t	t_xmit;
	long	clk_tcks;

	testing("Call to tcdrain() waits until all output written to the");
	testing("terminal has been transmitted before returning.");
	start
		clk_tcks = Zsysconf(_SC_CLK_TCK);
		t_xmit = clk_tcks*nb_xmit*10/baud_value(baud_min());

					/*	Record the starting time before the write() is
						called.  In case write() blocks, the time to drain
						the data is monitored properly.
					*/
		t_beg = Ztimes(&tms_buf);
		if(PCTS_GTI_BUFFERS_OUTPUT) {
			SET_TIMEOUT
				expecting(nb_xmit);
				Zwrite(loop_fildes, outbuf, nb_xmit);	
			CLEAR_ALARM
		}
		else
			t_xmit = 0;

		expecting(0);
		Ztcdrain(loop_fildes);
		t_end = Ztimes(&tms_buf);
		Ztcflush(loop_fildes, TCOFLUSH);

		if(PCTS_GTI_BUFFERS_OUTPUT) {
			if(t_end -t_beg < t_xmit -t_xmit/10)
				error("%s%lu clock-ticks.\n%s%lu clock-ticks.",
						"Expected tcdrain() to take ",
						(unsigned long)t_xmit,
						"Actually time taken was ",
						(unsigned long) (t_end -t_beg) );
		}

		SET_TIMEOUT
			break_on(ERROR);
			for(n_read=0; n_read != nb_xmit;)
				n_read += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
		CLEAR_ALARM
	finis
}
