#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

										/* declared in sv_tools/termios.c	*/
extern	const cc_t PERASE;
extern	const cc_t PKILL;
extern	const cc_t PINTR;
extern	const cc_t PQUIT;
extern	const cc_t PSTART;
extern	const cc_t PSTOP;
extern	const cc_t PEOF;
extern	const cc_t PEOL;
extern	const cc_t PSUSP;
extern	const cc_t PMIN;
extern	const cc_t PTIME;

static	int	io_diff;		/*	Zero if IUT supports different INPUT and
								OUTPUT baud rates on GTI port.			*/
#define	NSEC	5			/*	Maximum time for data transfer.			*/
#define	FUDGE	2			/*	Slop allowed in time-to-transfer data.	*/

void	get_parameters(int, struct termios *);


int	main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	inbuf[_POSIX_MAX_INPUT];
	char	outbuf[_POSIX_MAX_INPUT];
	int		flags;
	pid_t	pid;
	int		chars_sec, char_to_xmit, time_to_xmit;

	setup(E_TCGETATTR);
	funct(E_TCGETATTR);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(4, BASE)
		GTI_NOT_PROVIDED_RESULT(5, BASE)
		GTI_NOT_PROVIDED_RESULT(6, BASE)
		done();
	}

	assertion(4, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Determine if support for different rates for input and output");
		prep("on the same port exists on this IUT.");
		start
			exit_on(ERROR);

			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
			
			continue_on(ERROR);
					/* Set input/output baud rate to maximum configured. */
			Zcfsetospeed(loop_termios_p, baud_max());
			Zcfsetispeed(tty_termios_p, baud_max());
			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);

					/*	Set input/output baud rate to minimum configured
						if different INPUT and OUTPUT rates are supported
						otherwise set to maximum configured.
					 */
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			tcsetattr(tty_fildes, TCSANOW, tty_termios_p);

					/*	Determine if split baud rates are truly supported.
						POSIX.1, 7.2.1.2:  If differing baud rates are not
						supported, the rate returned as the input baud rate
						shall be either the number zero or the output rate.
					*/
			Ztcgetattr(tty_fildes, tty_termios_p);
			if(	(cfgetispeed(tty_termios_p) != baud_max()) ||
				(cfgetospeed(tty_termios_p) != baud_min()) ) {
				io_diff++;		/* GTI does NOT support different baud rates. */
				Zcfsetispeed(loop_termios_p, baud_max());
								/*	Reset both input and output speed. */
				Zcfsetospeed(tty_termios_p, baud_max());
				Zcfsetispeed(tty_termios_p, baud_max());
				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			}

					/*	Same specific settings for PORT_TTY and PORT_LOOP. */
			loop_termios_p->c_cflag &= ~CSIZE;
			loop_termios_p->c_cflag |= CS7|CSTOPB;
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
		finis

		prep("Initialize the parameters of the termios structure.");
		start
			tty_termios_p->c_iflag |= BRKINT|ICRNL|IGNBRK|IGNCR|IGNPAR;
			tty_termios_p->c_iflag |= INLCR|INPCK|ISTRIP|IXOFF|IXON|PARMRK;

			tty_termios_p->c_oflag |= OPOST;

			tty_termios_p->c_cflag &= ~CSIZE;
			tty_termios_p->c_cflag |= CLOCAL|CREAD|CS7|CSTOPB;
			tty_termios_p->c_cflag |= HUPCL|PARENB|PARODD;

			tty_termios_p->c_lflag |= ECHO|ECHOE|ECHOK|ECHONL;
			tty_termios_p->c_lflag |= ICANON|ISIG|NOFLSH;
			tty_termios_p->c_lflag |= TOSTOP;

			tty_termios_p->c_cc[VERASE] = PERASE;
			tty_termios_p->c_cc[VINTR] = PINTR;
			tty_termios_p->c_cc[VKILL] = PKILL;
			tty_termios_p->c_cc[VQUIT] = PQUIT;
			tty_termios_p->c_cc[VSUSP] = PSUSP;

			cfsetispeed(tty_termios_p, baud_max());
			cfsetospeed(tty_termios_p, baud_max());

			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);

									/* Reset ICANON to set VMIN and VTIME. */
			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VEOF] = PEOF;
			if(VEOF != VMIN)
				tty_termios_p->c_cc[VMIN] = PMIN;
			tty_termios_p->c_cc[VEOL] = PEOL;
			if(VEOL != VTIME)
				tty_termios_p->c_cc[VTIME] = PTIME;
			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
		finis

														/* Foreground */
		testing("Call to tcgetattr() from a foreground process gets the");
		testing("parameters associated with the terminal and stores them");
		testing("in a \"termios\" structure.");
		start
			get_parameters(tty_fildes, tty_termios_p);
		finis

														/* Background */
		FORK
			testing("Call to tcgetattr() from a background process gets the");
			testing("parameters associated with the terminal and stores them");
			testing("in a \"termios\" structure.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				get_parameters(tty_fildes, tty_termios_p);
			finis
		KROF
	finis


	if(io_diff == 0) {
		assertion(6, BASE);
		start
			yes_implemented("INPUT and OUTPUT baud rates can differ");
		finis
	}
	else {
		assertion(5, BASE);
		start
			not_implemented("INPUT and OUTPUT baud rates can differ");
		finis
	}

	if(io_diff == 0)
		assertion(5, BASE);
	else
		assertion(6, BASE);
		start
			prep("Initialize PORT_TTY and PORT_LOOP.");
			start
				tty_init(tty_fildes, tty_termios_p);
				loop_init(loop_fildes, loop_termios_p);

				flags = Zfcntl(loop_fildes, F_GETFL, 0);
				flags &= ~O_NONBLOCK;
				Zfcntl(loop_fildes, F_SETFL, flags);

				tty_termios_p->c_lflag &= ~ICANON;
				tty_termios_p->c_cc[VMIN] = 0;
				tty_termios_p->c_cc[VTIME] = 0;

				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);

				memset(outbuf, 'a', _POSIX_MAX_INPUT);
				time_to_xmit =
						(_POSIX_MAX_INPUT/(baud_value(baud_max())/10)) +FUDGE;
			finis

			prep("Fork process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				exit_on(ERROR);
				testing("%d %s%d within %d seconds.",
						_POSIX_MAX_INPUT,
						"bytes are written at speed B",
						baud_value(baud_max()),
						time_to_xmit);
				start
					SET_ALARM(time_to_xmit)
						expecting(_POSIX_MAX_INPUT);
						Zwrite(loop_fildes, outbuf, _POSIX_MAX_INPUT);
						Ztcdrain(loop_fildes);
					CLEAR_ALARM
				finis
	
				done();
			}

			testing("%d %s%d within %d seconds.",
						_POSIX_MAX_INPUT,
						"bytes are read at speed B",
						baud_value(baud_max()),
						time_to_xmit);
			start
				int	cnt;
	
				SET_ALARM(time_to_xmit)
					for(cnt=0; cnt<_POSIX_MAX_INPUT;)
						cnt += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
				CLEAR_ALARM

				if(unused_time > FUDGE +1)
					error("%s\n%s%d.\n%s%d.",
							"Time to transmit to low.",
							"Expected minimum: ",
							time_to_xmit -FUDGE,
							"Actual: ",
							time_to_xmit -unused_time);
			finis

			WAIT_FOR_PID(pid);

			if(io_diff != 0) {
				testing("Input rate is zero or actual rate.");
				start
					if( ! (	(cfgetispeed(loop_termios_p) == baud_max()) ||
						(cfgetispeed(loop_termios_p) == (speed_t)0) ) )
							error("%s\n%s%d\n%s%d\n",
								"Input speed not set properly.",
								"Expected: zero or ",
								baud_max(),
								"Actual: ",
								cfgetispeed(loop_termios_p));
				finis
			}

										/* Rerun at lowest baud rate. */
			prep("Set PORT_TTY to min baud rate.");
			start
				Zcfsetospeed(loop_termios_p, baud_min());
				Zcfsetispeed(tty_termios_p, baud_min());
				if(io_diff == 0) {
					Zcfsetispeed(loop_termios_p, baud_max());
					Zcfsetospeed(tty_termios_p, baud_max());
				} else {
					Zcfsetispeed(loop_termios_p, baud_min());
					Zcfsetospeed(tty_termios_p, baud_min());
				}
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
				Ztcsetattr(loop_fildes, TCSAFLUSH, loop_termios_p);

				/* Determine number of characters to transmit in NSEC or less.*/
				chars_sec = baud_value(baud_min())/(sizeof(char)*8 +2);
				time_to_xmit = _POSIX_MAX_INPUT/chars_sec +FUDGE;
				if(time_to_xmit > NSEC) {
					time_to_xmit = NSEC +FUDGE;
					char_to_xmit = NSEC * chars_sec;
				}
				else
					char_to_xmit = _POSIX_MAX_INPUT;
			finis

			prep("Fork process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				exit_on(ERROR);
				testing("%d %s%d within %d seconds.",
						char_to_xmit,
						"bytes are written at speed B",
						baud_value(baud_min()),
						time_to_xmit);
				start
					SET_ALARM(time_to_xmit)
						expecting(char_to_xmit);
						Zwrite(loop_fildes, outbuf, char_to_xmit);
						Ztcdrain(loop_fildes);
					CLEAR_ALARM
				finis
	
				done();
			}

			testing("%d %s%d within %d seconds.",
						char_to_xmit,
						"bytes are read at speed B",
						baud_value(baud_min()),
						time_to_xmit);
			start
				int	cnt;
	
				SET_ALARM(time_to_xmit)
					for(cnt=0; cnt<char_to_xmit;)
						cnt += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
				CLEAR_ALARM

				if(unused_time > FUDGE +1)
					error("%s\n%s%d.\n%s%d.",
							"Time to transmit to low.",
							"Expected minimum: ",
							time_to_xmit -FUDGE,
							"Actual: ",
							time_to_xmit -unused_time);
			finis

			WAIT_FOR_PID(pid);

			if(io_diff != 0) {
				testing("Input rate is zero or actual rate.");
				start
					if( ! (	(cfgetispeed(tty_termios_p) == baud_min()) ||
						(cfgetispeed(tty_termios_p) == (speed_t)0) ) )
							error("%s\n%s%d\n%s%d\n",
								"Input speed not set properly.",
								"Expected: zero or ",
								baud_min(),
								"Actual: ",
								cfgetispeed(tty_termios_p));
				finis
			}
		finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}


				/* Compare to initial state  */
void	get_parameters(int tty_fildes, struct termios *tty_termios_p)
{
	int i;

	tty_termios_p->c_iflag = 0;			/* input modes */
	tty_termios_p->c_oflag = 0;			/* output modes */
	tty_termios_p->c_cflag = 0;			/* control modes */
	tty_termios_p->c_lflag = 0;			/* local modes */
	for(i = 0; i < NCCS; ++i)
		tty_termios_p->c_cc[i] = 0;		/* control chars */

	Ztcgetattr(tty_fildes, tty_termios_p);

	if(io_diff == 0) {
		if(cfgetispeed(tty_termios_p) != baud_max())
			error("Input speed was not set properly.");
	} else {
		if( ! (	(cfgetispeed(tty_termios_p) == baud_max()) ||
			(cfgetispeed(tty_termios_p) == (speed_t)0) ) )
				error("%s\n%s%d\n%s%d\n",
					"Input speed not set properly.",
					"Expected: zero or ",
					baud_max(),
					"Actual: ",
					cfgetispeed(tty_termios_p));
		if(cfgetispeed(tty_termios_p) == (speed_t)0)
			warning("Usage of zero is obsolescent.");
	}
	if(cfgetospeed(tty_termios_p) != baud_max())
		error("Output speed not was was set.");

	if(!(tty_termios_p->c_iflag & BRKINT))
		error("BRKINT flag not set.");
	if(!(tty_termios_p->c_iflag & ICRNL))
		error("ICRNL flag not set.");
	if(!(tty_termios_p->c_iflag & IGNBRK))
		error("IGNBRK flag not set.");
	if(!(tty_termios_p->c_iflag & IGNCR))
		error("IGNCR flag not set.");
	if(!(tty_termios_p->c_iflag & IGNPAR))
		error("IGNPAR flag not set.");
	if(!(tty_termios_p->c_iflag & INLCR))
		error("INLCR flag not set.");
	if(!(tty_termios_p->c_iflag & INPCK))
		error("INPCK flag not set.");
	if(!(tty_termios_p->c_iflag & ISTRIP))
		error("ISTRIP flag not set.");
	if(!(tty_termios_p->c_iflag & IXOFF))
		error("IXOFF flag not set.");
	if(!(tty_termios_p->c_iflag & IXON))
		error("IXON flag not set.");
	if(!(tty_termios_p->c_iflag & PARMRK))
		error("PARMRK flag not set.");

	if(!(tty_termios_p->c_oflag & OPOST))
		error("OPOST flag not set.");

	if(GTI_MODEM_CONTROL) {
		if(!(tty_termios_p->c_cflag & CLOCAL))
			error("CLOCAL flag not set.");
		if(!(tty_termios_p->c_cflag & HUPCL))
			error("HUPCL flag not set.");
	}

											/* CREAD is optional */
	if(!(tty_termios_p->c_cflag & CS7))
		error("CS7 flag not set.");
	if(!(tty_termios_p->c_cflag & CSTOPB))
		error("CSTOPB flag not set.");
	if(!(tty_termios_p->c_cflag & PARENB))
		error("PARENB flag not set.");
	if(!(tty_termios_p->c_cflag & PARODD))
		error("PARODD flag not set.");
	if(!(tty_termios_p->c_lflag & ECHO))
		error("ECHO flag not set.");
	if(!(tty_termios_p->c_lflag & ECHOE))
		error("ECHOE flag not set.");
	if(!(tty_termios_p->c_lflag & ECHOK))
		error("ECHOK flag not set.");
	if(!(tty_termios_p->c_lflag & ECHONL))
		error("ECHONL flag not set.");
	if(tty_termios_p->c_lflag & ICANON)
		error("ICANON flag set.");
	if(!(tty_termios_p->c_lflag & ISIG))
		error("ISIG flag not set.");
	if(!(tty_termios_p->c_lflag & NOFLSH))
		error("NOFLSH flag not set.");
	if(!(tty_termios_p->c_lflag & TOSTOP))
		error("TOSTOP flag not set.");

	if(tty_termios_p->c_cc[VINTR] != PINTR)
		error("c_cc[VINTR] is incorrect. Expected %#o, received %#o",
			  PINTR, tty_termios_p->c_cc[VINTR]);

	if(tty_termios_p->c_cc[VQUIT] != PQUIT)
		error("c_cc[VQUIT] is incorrect. Expected %#o, received %#o",
			  PQUIT, tty_termios_p->c_cc[VQUIT]);

	if(tty_termios_p->c_cc[VERASE] != PERASE)
		error("c_cc[VERASE] is incorrect. Expected %#o, received %#o",
			PERASE, tty_termios_p->c_cc[VERASE]);

	if(tty_termios_p->c_cc[VKILL] != PKILL)
		error("c_cc[VKILL] is incorrect. Expected %#o, received %#o",
			  PKILL, tty_termios_p->c_cc[VKILL]);

	if(tty_termios_p->c_cc[VEOF] != PEOF)
		error("c_cc[VEOF] is incorrect. Expected %#o, received %#o",
			  PEOF, tty_termios_p->c_cc[VEOF]);
	if(VEOF != VMIN)
		if(tty_termios_p->c_cc[VMIN] != PMIN)
			error("c_cc[VMIN] is incorrect. Expected %#o, received %#o",
				  PMIN, tty_termios_p->c_cc[VMIN]);
	if(tty_termios_p->c_cc[VEOL] != PEOL)
		error("c_cc[VEOL] is incorrect. Expected %#o, received %#o",
			  PEOL, tty_termios_p->c_cc[VEOL]);
	if(VEOL != VTIME)
		if(tty_termios_p->c_cc[VTIME] != PTIME)
			error("c_cc[VTIME] is incorrect. Expected %#o, received %#o",
				  PTIME, tty_termios_p->c_cc[VTIME]);

	if(tty_termios_p->c_cc[VSUSP] != PSUSP)
		error("c_cc[VSUSP]  is incorrect. Expected %#o, received %#o",
			  		PSUSP, tty_termios_p->c_cc[VSUSP]);
}
