#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

void	com_code(void);

static	int		tty_fildes, loop_fildes;
static	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
static	sigset_t set;

int main(void)
{
	pid_t	pid;

	setup(E_TCSENDBREAK);
	funct(E_TCSENDBREAK);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(9, BASE)
		done();
	}

	assertion(9, BASE);
	start
		prep("Establish PORT_TTY and PORT_LOOP as non-controlling");
		prep("terminals.");
		start
			exit_on(ERROR);
	
			termios_nct(&tty_fildes, tty_termios_p, &loop_fildes,
															loop_termios_p);
		finis
									/* Foreground process test. */
		ENABLE_PSYNC
		FORK
			com_code();
		KROF


									/* Background process tests. */
		FORK
			prep("Catch SIGTTOU and establish a background process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			com_code();

			prep("Block SIGTTOU in background process group.");
			start
				SIGBLOCK(set, SIGTTOU);
			finis

			com_code();

			prep("Unblock and ignore SIGTTOU in background process group.");
			start
				SIGNAL(SIGTTOU, SIG_IGN);
				SIGUNBLOCK(set, SIGTTOU);
			finis

			com_code();
		KROF

									/* Orphaned background process test. */
		FORK
			prep("Catch SIGTTOU and establish an orphaned background");
			prep("process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				SLEEP(WAIT_FOR_TIME)

				com_code();
				SEND_PSYNCA

				done();
			}
		KROF
										/* Wait for test to complete. */
		SET_TIMEOUT
			WAIT_PSYNCA
		CLEAR_ALARM
		DISABLE_PSYNC
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}

void com_code(void)
{
	char	inbuf[_POSIX_MAX_INPUT];				/* read() buffer	*/
	char	c_data = 'A';							/* output data		*/
	pid_t	pid;

	ENABLE_PSYNC
	prep("Fork a child process.");
	start
		sv_flush();
		pid = Zfork();
	finis
	
	if(pid == (pid_t)0) {
		testing("Sending of a break condition.");
		start
			SIGNAL(SIGINT, SIG_IGN);	/* If generated. */

			SET_TIMEOUT
				WAIT_PSYNC
				Ztcsendbreak(loop_fildes, 0);
				expecting(1);
				Zwrite(loop_fildes, &c_data, 1);
				Ztcdrain(loop_fildes);
				dt_secs(TIME_ACQUIRE_XMITTED_DATA);
			CLEAR_ALARM
		finis

		done();
	}

	prep("Have PORT_TTY detect breaks as a single '\\0'.");
	start
		Ztcgetattr(tty_fildes, tty_termios_p);
		tty_termios_p->c_iflag &= ~IGNBRK;
		tty_termios_p->c_iflag &= ~BRKINT;
						/* Set for non-canonical mode. */
		tty_termios_p->c_lflag &= ~(ICANON | ECHO);
		tty_termios_p->c_cc[VMIN] = 1;
		tty_termios_p->c_cc[VTIME] = 10;
		Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
	
		SIGNAL(SIGINT, sig_int);	/* Catch break if it occurs. */
	finis

	testing("Non-controlling terminal receives single '\\0' on break.");
	start
		sig_rcvd = 0;
		SEND_PSYNC
		SET_TIMEOUT
			expecting(1);
			Zread(tty_fildes, inbuf, 1);
		CLEAR_ALARM
		if(inbuf[0] != '\0')
			error("%s\n%s.\n%s'\\%o\'.",
					"Improper character received.",
					"Expected an input character of value '\\0'.",
					"Actual character received was ",
					inbuf[0]);
		if(sig_rcvd != 0)
			error("Break signal detected.");

		SET_TIMEOUT
			expecting(1);
			Zread(tty_fildes, inbuf, 1);
		CLEAR_ALARM
		if(inbuf[0] != c_data)
			error("%s\n%s'\\%o'.\n%s'\\%o'.",
					"Improper character received.",
					"Expected input character of value ",
					c_data,
					"Actual character received was ",
					inbuf[0]);
	finis

	WAIT_FOR_PID(pid);
	DISABLE_PSYNC
}
