#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

char String[] = "How now, brown cow?";
#define	Strlen	(sizeof(String) -1)
#define	MAX_DELAY_TO_DISCONNECT 3


int	main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char	inbuf[_POSIX_MAX_INPUT];
	char	outbuf[_POSIX_MAX_INPUT];
	pid_t	pid;
	struct	tms	tms_buf;
	clock_t	t_beg, t_end;
	size_t	nb_xmit, n_read;

	setup(E_TCSETATTR);
	funct(E_TCSETATTR);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(5, BASE)
		GTI_NOT_PROVIDED_RESULT(6, BASE)
		GTI_NOT_PROVIDED_RESULT(7, BASE)
		GTI_NOT_PROVIDED_RESULT(8, BASE)
		GTI_NOT_PROVIDED_RESULT(9, BASE)
		GTI_NOT_PROVIDED_RESULT(10, BASE)
		done();
	}

	assertion(5, BASE);
	start
		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Clear ICANON on PORT_LOOP, set speed to lowest rate supported.");
		start
			exit_on(ERROR);

			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());

			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			memset(outbuf, 'A', _POSIX_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
			nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
			if(nb_xmit > _POSIX_MAX_INPUT)
				nb_xmit = _POSIX_MAX_INPUT;
		finis

		prep("Write out data.");
		start
			SET_TIMEOUT
				expecting(nb_xmit);
				Zwrite(loop_fildes, outbuf, nb_xmit);	
				expecting(nb_xmit);
				Zwrite(tty_fildes, outbuf, nb_xmit);	
			CLEAR_ALARM
		finis

		testing("tcsetattr(fd, TCSANOW, termios) completed without delay.");
		start
			int	i;
			clock_t	r_val;
			clock_t	tcks_slop, tcks_sec;
			clock_t	tcks_char;		/*	Number of ticks to transmit one
										character at the minimum baud rate
										supported by the implementation.
									*/

									/*	tcks_char = (ticks/sec)/(char/sec)	*/
			tcks_sec = Zsysconf(_SC_CLK_TCK);
			tcks_char = tcks_sec / (baud_value(baud_min())/10);
			message("ticks_per_sec %ld,  baud rate %ld, ticks_per_char %ld",
						(long) tcks_sec,
						(long) baud_value(baud_min()),
						(long) tcks_char);

			break_on(ERROR);
			SET_TIMEOUT
										/*	Ensure transmission has started. */
				while(Zread(loop_fildes, inbuf, (size_t)1) == (ssize_t)0)
					;
			CLEAR_ALARM
										/*	Ensure for systems with a hugh tick
											rate that the time slop is similar.
										*/
			tcks_slop = 1;
			if(tcks_sec > (clock_t)100)
				tcks_slop = tcks_sec/100;

			for(i=0; i<nb_xmit/2; i++) {
										/*	Attempt to test while data is 
											actually being transmitted.
										*/
				t_beg = Ztimes(&tms_buf);
				expecting(0);
				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
				t_end = Ztimes(&tms_buf);
				if((t_end -t_beg) > tcks_slop)
					error("Call to tcsetattr() was delayed.");
				
							/*	Delay one character transmit time or less.	*/
				r_val = times(&tms_buf);
				while(r_val +tcks_char > times(&tms_buf))
					;
			}
		finis

		testing("Input and output were not flushed.");
		start
			break_on(ERROR);
			Ztcdrain(loop_fildes);
			Ztcdrain(tty_fildes);

			SET_TIMEOUT
				for(n_read=0; n_read != nb_xmit;)
					n_read += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);

				for(n_read=0; n_read != nb_xmit -1;)
					n_read += Zread(loop_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis
	finis

	assertion(6, EXTENDED);

	assertion(7, BASE);
	start
		clock_t	t_xmit;

		testing("A call to tcsetattr(), with the TCSADRAIN option,");
		testing("waits for output to drain before returning.");
		start
			long	clk_tcks;

			break_on(ERROR);

			Ztcflush(tty_fildes, TCIOFLUSH);
			Ztcflush(loop_fildes, TCIOFLUSH);

			clk_tcks = Zsysconf(_SC_CLK_TCK);
			t_xmit = clk_tcks*nb_xmit*10/baud_value(baud_min());

			if(PCTS_GTI_BUFFERS_OUTPUT) {
					/*	Record the starting time before the write() is
						called.  In case write() blocks, the time to drain
						the data is monitored properly.
					*/
				t_beg = Ztimes(&tms_buf);
				SET_TIMEOUT
					expecting(nb_xmit);
					Zwrite(loop_fildes, outbuf, nb_xmit);	
				CLEAR_ALARM
			}
			else
				t_xmit = 0;

			expecting(0);
			Ztcsetattr(loop_fildes, TCSADRAIN, loop_termios_p);
			Ztcflush(loop_fildes, TCOFLUSH);	/* Flush any data remaining. */
			t_end = Ztimes(&tms_buf);

			if(PCTS_GTI_BUFFERS_OUTPUT) {
				if(t_end -t_beg < t_xmit -t_xmit/10)
					error("%s%lu clock-ticks.\n%s%lu clock-ticks.",
							"Expected tcsetattr() to take ",
							(unsigned long)t_xmit,
							"Actually time taken was ",
							(unsigned long) (t_end -t_beg) );
			}

										/* Ensure no data lost. */
			SET_TIMEOUT
				for(n_read=0; n_read != nb_xmit;)
					n_read += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis
	finis

	assertion(8, BASE);
	start
		prep("Write out data then call tcsetattr() with TCSAFLUSH option.");
		start
			Ztcflush(tty_fildes, TCIOFLUSH);
			Ztcflush(loop_fildes, TCIOFLUSH);

			SET_TIMEOUT
							/* Send data to PORT_TTY input queue. */
				expecting(nb_xmit);
				Zwrite(loop_fildes, outbuf, nb_xmit);
				Ztcdrain(loop_fildes);
			CLEAR_ALARM

			SET_TIMEOUT
							/* Send data to PORT_TTY output queue. */
				expecting(nb_xmit);
				Zwrite(tty_fildes, outbuf, nb_xmit);

				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			CLEAR_ALARM
		finis

		testing("On return of tcsetattr(), input data was discarded and");
		testing("output data was transmitted.");
		start
			break_on(ERROR);
			SET_TIMEOUT
				for(n_read=0; n_read != nb_xmit;)
					n_read += Zread(loop_fildes, inbuf, _POSIX_MAX_INPUT);

				expecting(0);
				Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis
	finis

	assertion(9, EXTENDED);

												/* B0 */
	assertion(10, BASE);
	start
		if(GTI_MODEM_CONTROL) {
			ENABLE_PSYNC
							/*	Fork a process, to avoid the possibility of
								tcsetattr() being interrupted by SIGHUP, which
								is sent to the controlling process.
							*/
			prep("Clear CLOCAL for PORT_TTY.");
			start
				tty_termios_p->c_cflag &= ~CLOCAL;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				SET_TIMEOUT
					Zcfsetospeed(loop_termios_p, B0);
							/*	In case split baud rates are not supported,
								set input speed also.
							*/
					Zcfsetispeed(loop_termios_p, B0);
					SEND_PSYNC
					WAIT_PSYNCA
					Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
				CLEAR_ALARM
				done();
			}

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			testing("When the output speed in the termios structure is set");
			testing("to B0, then the modem control lines are disconnected.");
			start
				SIGNAL(SIGHUP, sig_int);
				sig_rcvd = 0;
				SEND_PSYNCA
							/*	Perform needed delay even if interupted. */
				del_secs(GTI_LINE_CLOSE_DELAY);
				if(MAX_DELAY_TO_DISCONNECT > GTI_LINE_CLOSE_DELAY)
					del_secs(MAX_DELAY_TO_DISCONNECT - GTI_LINE_CLOSE_DELAY);
				if(sig_rcvd == 0)
					error("SIGHUP not generated.");
			finis

			WAIT_FOR_PID(pid);
			DISABLE_PSYNC
		}
		else
			not_implemented("modem control");
	finis

	cleanup("Ensure terminal devices are lowered.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		close(loop_fildes);
		close(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
