#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

										/* declared in sv_tools/termios.c	*/
extern	const cc_t PERASE;
extern	const cc_t PKILL;
extern	const cc_t PINTR;
extern	const cc_t PQUIT;
extern	const cc_t PSTART;
extern	const cc_t PSTOP;
extern	const cc_t PEOF;
extern	const cc_t PEOL;
extern	const cc_t PSUSP;
extern	const cc_t PMIN;
extern	const cc_t PTIME;

static	int	clocal;			/*	Zero, CLOCAL and GTI_MODEM_CONTROL supported. */
static	int	io_diff = FALSE;/*	TRUE if IUT supports different INPUT and
								OUTPUT baud rates on GTI port. */

void	get_parameters(int, struct termios *);


int	main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;

	setup(E_TCSETATTR);
	funct(E_TCSETATTR);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(12, BASE)
		done();
	}

	assertion(12, BASE);
	start
		int	opt_ns = 0;

		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Determine if support for split baud rates is provided.");
		start
			exit_on(ERROR);

			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
			
			continue_on(ERROR);
					/*	Note:	The unsupported attributes are retained
								for each call to tcsetattr().  But, if
								the implementation is working properly
								when a single attribute is supported, then
								tcsetattr() should be successful.
					*/

					/*	Output baud rates already set to maximum configured.
						Attempt to set input baud rate to minimum configured
						to determine if split baud rates are supported.
					*/
			Ztcgetattr(loop_fildes, loop_termios_p);
			Zcfsetispeed(loop_termios_p, baud_min());
			tcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			Ztcgetattr(loop_fildes, loop_termios_p);
			if( (cfgetispeed(loop_termios_p) != (speed_t)0)  &&
				(cfgetispeed(loop_termios_p) != cfgetospeed(loop_termios_p)) ){
						/*	Set output baud rate also to minimum configured. */
				Zcfsetospeed(tty_termios_p, baud_min());
				io_diff = TRUE;
			}
			else {
				opt_ns++;
				io_diff = FALSE;
			}

			tty_termios_p->c_cflag &= ~CLOCAL;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1) {
				opt_ns++;
				clocal++;
			}
			else {
									/*	CLOCAL clear normally implies support
										for GTI_MODEM_CONTROL supported.
									*/
				if(!GTI_MODEM_CONTROL)
					warning("CLOCAL flag clear supported and GTI_MODEM_CONTROL unsupported.");
			}

			tty_termios_p->c_cflag |= HUPCL;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1)
				opt_ns++;

			tty_termios_p->c_cflag |= CREAD;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1)
				opt_ns++;
			else {
				tty_termios_p->c_cflag &= ~CREAD;
				if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1)
					opt_ns++;
			}

														/* Try CSIZE values. */
			tty_termios_p->c_cflag &= ~CSIZE;
			tty_termios_p->c_cflag |= CS5;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1)
				opt_ns++;
			else {
				tty_termios_p->c_cflag &= ~CSIZE;
				tty_termios_p->c_cflag |= CS6;
				if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1)
					opt_ns++;
			}
		finis

		if(! opt_ns) {
			not_implemented("unsupported termios structure attribute");
		}

		else {
			prep("Initialize other parameters of the termios structure.");
			start
				tty_termios_p->c_iflag |= BRKINT|ICRNL|IGNBRK|IGNCR|IGNPAR;
				tty_termios_p->c_iflag |= INLCR|INPCK|ISTRIP|IXOFF|IXON|PARMRK;
	
				tty_termios_p->c_oflag |= OPOST;
	
				tty_termios_p->c_cflag |= CSTOPB;
				tty_termios_p->c_cflag |= PARENB|PARODD;
	
				tty_termios_p->c_lflag |= ECHO|ECHOE|ECHOK|ECHONL;
				tty_termios_p->c_lflag |= ICANON|ISIG|NOFLSH;
				tty_termios_p->c_lflag |= TOSTOP;
	
				tty_termios_p->c_cc[VERASE] = PERASE;
				tty_termios_p->c_cc[VINTR] = PINTR;
				tty_termios_p->c_cc[VKILL] = PKILL;
				tty_termios_p->c_cc[VQUIT] = PQUIT;
				tty_termios_p->c_cc[VSUSP] = PSUSP;
	
									/* Reset ICANON to set VMIN and VTIME. */
				tty_termios_p->c_lflag &= ~ICANON;
				tty_termios_p->c_cc[VEOF] = PEOF;
				if(VEOF != VMIN)
					tty_termios_p->c_cc[VMIN] = PMIN;
				tty_termios_p->c_cc[VEOL] = PEOL;
				if(VEOL != VTIME)
					tty_termios_p->c_cc[VTIME] = PTIME;
			finis
	
			testing("Call to tcsetattr() to change attributes when some are");
			testing("supported and others are unsupported returns zero and");
			testing("sets all supported attributes and leaves unchanged those");
			testing("that are unsupported.");
			start
				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
				get_parameters(tty_fildes, tty_termios_p);
			finis
		}
	finis
	

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}


				/* Compare to initial state  */
void	get_parameters(int tty_fildes, struct termios *tty_termios_p)
{
	int i;

	tty_termios_p->c_iflag = 0;			/* input modes */
	tty_termios_p->c_oflag = 0;			/* output modes */
	tty_termios_p->c_cflag = 0;			/* control modes */
	tty_termios_p->c_lflag = 0;			/* local modes */
	for(i = 0; i < NCCS; ++i)
		tty_termios_p->c_cc[i] = 0;		/* control chars */

	Ztcgetattr(tty_fildes, tty_termios_p);

	if(io_diff == TRUE) {
		if(cfgetospeed(tty_termios_p) != baud_min())
			error("Output speed was not set properly.");
	} else {
		if(cfgetospeed(tty_termios_p) != baud_max())
				error("%s\n%s%ld\n%s%ld\n",
					"Output speed not set properly.",
					"Expected: ",
					(long)baud_max(),
					"Actual: ",
					(long)cfgetospeed(tty_termios_p));
	}

	if(!(tty_termios_p->c_iflag & BRKINT))
		error("BRKINT flag not set.");
	if(!(tty_termios_p->c_iflag & ICRNL))
		error("ICRNL flag not set.");
	if(!(tty_termios_p->c_iflag & IGNBRK))
		error("IGNBRK flag not set.");
	if(!(tty_termios_p->c_iflag & IGNCR))
		error("IGNCR flag not set.");
	if(!(tty_termios_p->c_iflag & IGNPAR))
		error("IGNPAR flag not set.");
	if(!(tty_termios_p->c_iflag & INLCR))
		error("INLCR flag not set.");
	if(!(tty_termios_p->c_iflag & INPCK))
		error("INPCK flag not set.");
	if(!(tty_termios_p->c_iflag & ISTRIP))
		error("ISTRIP flag not set.");
	if(!(tty_termios_p->c_iflag & IXOFF))
		error("IXOFF flag not set.");
	if(!(tty_termios_p->c_iflag & IXON))
		error("IXON flag not set.");
	if(!(tty_termios_p->c_iflag & PARMRK))
		error("PARMRK flag not set.");

	if(!(tty_termios_p->c_oflag & OPOST))
		error("OPOST flag not set.");

	if(clocal) {
		if(!(tty_termios_p->c_cflag & CLOCAL))
			error("CLOCAL flag clear.");
	}
	else {
		if(tty_termios_p->c_cflag & CLOCAL)
			error("CLOCAL flag set.");
	}

											/* CREAD is optional */
	if(!(tty_termios_p->c_cflag & CSTOPB))
		error("CSTOPB flag not set.");
	if(!(tty_termios_p->c_cflag & PARENB))
		error("PARENB flag not set.");
	if(!(tty_termios_p->c_cflag & PARODD))
		error("PARODD flag not set.");
	if(!(tty_termios_p->c_lflag & ECHO))
		error("ECHO flag not set.");
	if(!(tty_termios_p->c_lflag & ECHOE))
		error("ECHOE flag not set.");
	if(!(tty_termios_p->c_lflag & ECHOK))
		error("ECHOK flag not set.");
	if(!(tty_termios_p->c_lflag & ECHONL))
		error("ECHONL flag not set.");
	if(tty_termios_p->c_lflag & ICANON)
		error("ICANON flag set.");
	if(!(tty_termios_p->c_lflag & ISIG))
		error("ISIG flag not set.");
	if(!(tty_termios_p->c_lflag & NOFLSH))
		error("NOFLSH flag not set.");
	if(!(tty_termios_p->c_lflag & TOSTOP))
		error("TOSTOP flag not set.");

	if(tty_termios_p->c_cc[VINTR] != PINTR)
		error("c_cc[VINTR] is incorrect. Expected %#o, received %#o",
			  PINTR, tty_termios_p->c_cc[VINTR]);

	if(tty_termios_p->c_cc[VQUIT] != PQUIT)
		error("c_cc[VQUIT] is incorrect. Expected %#o, received %#o",
			  PQUIT, tty_termios_p->c_cc[VQUIT]);

	if(tty_termios_p->c_cc[VERASE] != PERASE)
		error("c_cc[VERASE] is incorrect. Expected %#o, received %#o",
			PERASE, tty_termios_p->c_cc[VERASE]);

	if(tty_termios_p->c_cc[VKILL] != PKILL)
		error("c_cc[VKILL] is incorrect. Expected %#o, received %#o",
			  PKILL, tty_termios_p->c_cc[VKILL]);

	if(tty_termios_p->c_cc[VEOF] != PEOF)
		error("c_cc[VEOF] is incorrect. Expected %#o, received %#o",
			  PEOF, tty_termios_p->c_cc[VEOF]);
	if(VEOF != VMIN)
		if(tty_termios_p->c_cc[VMIN] != PMIN)
			error("c_cc[VMIN] is incorrect. Expected %#o, received %#o",
				  PMIN, tty_termios_p->c_cc[VMIN]);
	if(tty_termios_p->c_cc[VEOL] != PEOL)
		error("c_cc[VEOL] is incorrect. Expected %#o, received %#o",
			  PEOL, tty_termios_p->c_cc[VEOL]);
	if(VEOL != VTIME)
		if(tty_termios_p->c_cc[VTIME] != PTIME)
			error("c_cc[VTIME] is incorrect. Expected %#o, received %#o",
				  PTIME, tty_termios_p->c_cc[VTIME]);

	if(tty_termios_p->c_cc[VSUSP] != PSUSP)
		error("c_cc[VSUSP]  is incorrect. Expected %#o, received %#o",
			  		PSUSP, tty_termios_p->c_cc[VSUSP]);
}
