#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

void	com_code(void);

static	int		tty_fildes, loop_fildes;
static	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
static	char	inbuf[_POSIX_MAX_INPUT];
static	char	outbuf[_POSIX_MAX_INPUT];
static	sigset_t set;
static	struct	tms	tms_buf;
static	clock_t	t_beg, t_end;
static	size_t	nb_xmit, n_read;
static	clock_t	t_xmit;


int	main(void)
{
	pid_t	pid;

	setup(E_TCSETATTR);
	funct(E_TCSETATTR);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(16, BASE)
		done();
	}

	assertion(16, BASE);
	start
		prep("Establish PORT_TTY and PORT_LOOP as non-controlling terminals.");
		prep("Clear ICANON on PORT_LOOP, set speed to lowest rate supported.");
		start
			exit_on(ERROR);

			termios_nct(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

			tty_termios_p->c_lflag &= ~ICANON;
			tty_termios_p->c_cc[VMIN] = 0;
			tty_termios_p->c_cc[VTIME] = 0;
			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());

			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
			memset(outbuf, 'A', _POSIX_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
			nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
			if(nb_xmit > _POSIX_MAX_INPUT)
				nb_xmit = _POSIX_MAX_INPUT;
		finis

									/* Foreground process test. */
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			com_code();
			done();
		}
		wait_for(pid, 4 * TIMEOUT);



									/* Background process tests. */
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {

			prep("Catch SIGTTOU and establish a background process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			com_code();

			prep("Block SIGTTOU in background process group.");
			start
				SIGBLOCK(set, SIGTTOU);
			finis

			com_code();

			prep("Unblock and ignore SIGTTOU in background process group.");
			start
				SIGNAL(SIGTTOU, SIG_IGN);
				SIGUNBLOCK(set, SIGTTOU);
			finis

			com_code();

			done();
		}
		wait_for(pid, 12 * TIMEOUT);

									/* Orphaned background process test. */
		ENABLE_PSYNC
		FORK
			prep("Catch SIGTTOU and establish an orphaned background");
			prep("process group.");
			start
				SIGNAL(SIGTTOU, sig_int);
				Zsetpgid((pid_t)0, (pid_t)0);
			finis

			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				SLEEP(WAIT_FOR_TIME)

				com_code();
				SEND_PSYNC

				done();
			}
		KROF
										/* Wait for test to complete. */
		SET_ALARM(4 * TIMEOUT)
			WAIT_PSYNC
		CLEAR_ALARM
		DISABLE_PSYNC
	finis

	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}

void com_code(void)
{
	testing("Write out data then call tcsetattr() with TCSANOW option.");
	start
		SET_TIMEOUT
			expecting(nb_xmit);
			Zwrite(loop_fildes, outbuf, nb_xmit);	
			expecting(nb_xmit);
			Zwrite(tty_fildes, outbuf, nb_xmit);	
		CLEAR_ALARM

		t_beg = Ztimes(&tms_buf);
		expecting(0);
		Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
		t_end = Ztimes(&tms_buf);
	finis

	testing("Call to tcsetattr() completed without delay.");
	start
		if((t_end -t_beg) > 2 )
			error("Call to tcsetattr() was delayed.");
	finis

	testing("Input and output were not flushed.");
	start
		break_on(ERROR);
		Ztcdrain(loop_fildes);
		Ztcdrain(tty_fildes);

		SET_TIMEOUT
			for(n_read=0; n_read != nb_xmit;)
				n_read += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);

			for(n_read=0; n_read != nb_xmit;)
				n_read += Zread(loop_fildes, inbuf, _POSIX_MAX_INPUT);
		CLEAR_ALARM
	finis


	testing("Write out data then call tcsetattr() with TCSADRAIN option.");
	testing("Call to tcsetattr() waits for output to drain before returning.");
	start
		long	clk_tcks;

		break_on(ERROR);

		Ztcflush(tty_fildes, TCIOFLUSH);
		Ztcflush(loop_fildes, TCIOFLUSH);

		clk_tcks = Zsysconf(_SC_CLK_TCK);
		t_xmit = clk_tcks*nb_xmit*10/baud_value(baud_min());
		if(PCTS_GTI_BUFFERS_OUTPUT) {
			SET_TIMEOUT
					/*	Record the starting time before the write() is
						called.  In case write() blocks, the time to drain
						the data is monitored properly.
					*/
				t_beg = Ztimes(&tms_buf);
				expecting(nb_xmit);
				Zwrite(loop_fildes, outbuf, nb_xmit);	
			CLEAR_ALARM
		}
		else
			t_xmit = 0;

		expecting(0);
		Ztcsetattr(loop_fildes, TCSADRAIN, loop_termios_p);
		Ztcflush(loop_fildes, TCOFLUSH);
		t_end = Ztimes(&tms_buf);

		if(PCTS_GTI_BUFFERS_OUTPUT) {
			if(t_end -t_beg < t_xmit -t_xmit/10)
				error("%s%lu clock-ticks.\n%s%lu clock-ticks.",
						"Expected tcsetattr() to take ",
						(unsigned long)t_xmit,
						"Actually time taken was ",
						(unsigned long) (t_end -t_beg) );
		}

									/* Ensure no data lost. */
		SET_TIMEOUT
			for(n_read=0; n_read != nb_xmit;)
				n_read += Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
		CLEAR_ALARM
	finis

	testing("Write out data then call tcsetattr() with TCSAFLUSH option.");
	start
		Ztcflush(tty_fildes, TCIOFLUSH);
		Ztcflush(loop_fildes, TCIOFLUSH);

		SET_TIMEOUT
						/* Send data to PORT_TTY input queue. */
			expecting(nb_xmit);
			Zwrite(loop_fildes, outbuf, nb_xmit);
			Ztcdrain(loop_fildes);
		CLEAR_ALARM

		SET_TIMEOUT
						/* Send data to PORT_TTY output queue. */
			expecting(nb_xmit);
			Zwrite(tty_fildes, outbuf, nb_xmit);

			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			Ztcflush(tty_fildes, TCOFLUSH);
		CLEAR_ALARM
	finis

	testing("On return of tcsetattr(), input data was discarded and");
	testing("output data was transmitted.");
	start
		break_on(ERROR);
		SET_TIMEOUT
			for(n_read=0; n_read != nb_xmit;)
				n_read += Zread(loop_fildes, inbuf, _POSIX_MAX_INPUT);

			expecting(0);
			Zread(tty_fildes, inbuf, _POSIX_MAX_INPUT);
		CLEAR_ALARM
	finis
}
