#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


char String[] = "How now, brown cow?\n";
#define	Strlen	(sizeof(String) -1)

int main(void)
{
	int		notty_fildes;
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;

	setup(E_TCSETATTR);
	funct(E_TCSETATTR);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(17, BASE)
		GTI_NOT_PROVIDED_RESULT(18, BASE)
		GTI_NOT_PROVIDED_RESULT(19, BASE)
		GTI_NOT_PROVIDED_RESULT(20, BASE)
		GTI_NOT_PROVIDED_RESULT(21, BASE)
		done();
	}

	SIGNAL(SIGHUP, SIG_IGN);		/* Ignore hangups on PORTs. */

	assertion(18, BASE);
	start
		pid_t		pid;
		sigset_t	set;

		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			exit_on(ERROR);

			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

							/* Set TOSTOP before becoming background process. */
			tty_termios_p->c_lflag |= TOSTOP;
			Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
		finis

		prep("Write to PORT_LOOP.");
		start
			expecting(Strlen);
			Zwrite(loop_fildes, String, Strlen);
			Ztcdrain(loop_fildes);
		finis

		FORK
			prep("Establish background process group with two processes.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				SIGNAL(SIGTTOU, sig_int);
				sig_rcvd = 0;
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child process in background process group");
				testing("received SIGTTOU signal.");
				start
					sleep(WAIT_FOR_TIME);
					if(sig_rcvd == 0)
						error("Child process did not receive SIGTTOU.");
				finis

				done();
			}
	
			testing("When tcsetattr() is interrupted it returns [EINTR].");
			start
				SIGUNBLOCK(set, SIGTTOU);
				SET_TIMEOUT
					tty_termios_p->c_oflag |= OPOST;
					expecting(EINTR);
					Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
				CLEAR_ALARM
			finis

			testing("SIGTTOU signal is sent to calling process.");
			start
				dt_secs(1);
				if(sig_rcvd == 0)
					error("Calling process did not receive SIGTTOU.");
			finis

			testing("Terminal attribute is not changed.");
			start
				Ztcgetattr(tty_fildes, tty_termios_p);
				if((tty_termios_p->c_oflag & OPOST) == OPOST)
					error("Terminal attribute was set.");
			finis

			WAIT_FOR_PID(pid);
		KROF

		testing("PORT_TTY queue was not flushed.");
		start
			SET_TIMEOUT
				expecting(Strlen);
				Zread(tty_fildes, String, _POSIX_MAX_INPUT);
			CLEAR_ALARM
		finis
	finis


	assertion(17, BASE);
	start
		testing("Call tcsetattr() with an invalid file descriptor.");
		start
			tty_termios_p->c_oflag |= OPOST;
			expecting(-1);
			expecting(EBADF);
			Ztcsetattr(-1, TCSANOW, &tty_termios);
		finis

		testing("The termios structure is not changed.");
		start
			Ztcgetattr(tty_fildes, tty_termios_p);
			if((tty_termios_p->c_oflag & OPOST) == OPOST)
				error("The termios structure attribute OPOST was set.");
		finis
	finis


	assertion(19, BASE);
	start
		testing("Call tcsetattr() with an invalid action.");
		start
			tty_termios_p->c_oflag |= OPOST;
			expecting(-1);
			expecting(EINVAL);
			Ztcsetattr(tty_fildes, -1, &tty_termios);
		finis

		testing("The termios structure is not changed.");
		start
			Ztcgetattr(tty_fildes, &tty_termios);
			if((tty_termios_p->c_oflag & OPOST) == OPOST)
				error("The termios structure attribute OPOST was set.");
		finis
	finis


	assertion(20, BASE);
	start
		int	un_supp = 0;

		prep("When an optional attribute is not supported errno is set");
		prep("to [EINVAL].");
		start
			tty_termios_p->c_cflag |= CREAD;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
						"Call to tcgetattr() failed for attribute CREAD.",
						"Expected errno of [EINVAL].",
						"Actual errno returned was ",
						err_lookup(errno));
				un_supp++;

				Ztcgetattr(tty_fildes, tty_termios_p);
				if(tty_termios_p->c_cflag & CREAD == CREAD)
					error("Attribute CREAD set in termios structure.");
			}

			tty_termios_p->c_cflag &= ~CSIZE;
			tty_termios_p->c_cflag |= CS6;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
						"Call to tcgetattr() failed for attribute CS6.",
						"Expected errno of [EINVAL].",
						"Actual errno returned was ",
						err_lookup(errno));
				un_supp++;

				Ztcgetattr(tty_fildes, tty_termios_p);
				if(tty_termios_p->c_cflag & CSIZE == CS6)
					error("Attribute CS6 set in termios structure.");
			}

			tty_termios_p->c_cflag &= ~CSIZE;
			tty_termios_p->c_cflag |= CS5;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
						"Call to tcgetattr() failed for attribute CS5.",
						"Expected errno of [EINVAL].",
						"Actual errno returned was ",
						err_lookup(errno));
				un_supp++;

				Ztcgetattr(tty_fildes, tty_termios_p);
				if(tty_termios_p->c_cflag & CSIZE == CS5)
					error("Attribute CS5 set in termios structure.");
			}

			tty_termios_p->c_cflag &= ~CLOCAL;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
						"Call to tcgetattr() failed for attribute CLOCAL.",
						"Expected errno of [EINVAL].",
						"Actual errno returned was ",
						err_lookup(errno));
				un_supp++;

				Ztcgetattr(tty_fildes, tty_termios_p);
				if(tty_termios_p->c_cflag & CLOCAL == 0)
					error("Attribute CLOCAL clear in termios structure.");
			}

			tty_termios_p->c_cflag |= HUPCL;
			if(tcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p) == -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
						"Call to tcgetattr() failed for attribute HUPCL.",
						"Expected errno of [EINVAL].",
						"Actual errno returned was ",
						err_lookup(errno));
				un_supp++;

				Ztcgetattr(tty_fildes, tty_termios_p);
				if(tty_termios_p->c_cflag & HUPCL == HUPCL)
					error("Attribute HUPCL set in termios structure.");
			}
		finis

		if(un_supp > 0) {
			testing("Call to tcsetattr() with attribute in termios structure");
			testing("set to an unsupported value, returns a value of -1,");
			testing("sets errno to [EINVAL], and the values in the termios");
			testing("structure are left unchanged.");
			start
			finis
		}
		else
			not_implemented("unsupported termios structure attribute");
	finis


	assertion(21, BASE);
	start
		testing("Call tcsetattr() when fildes is not a tty).");
		start
			notty_fildes = temp_open("notspec", O_RDWR|O_CREAT, PERM_CTTY);
			tty_termios_p->c_oflag |= OPOST;
			expecting(-1);
			expecting(ENOTTY);
			Ztcsetattr(notty_fildes, TCSANOW, &tty_termios);
		finis

		testing("The termios structure is not changed.");
		start
			Ztcgetattr(tty_fildes, tty_termios_p);
			if((tty_termios_p->c_oflag & OPOST) == OPOST)
				error("The termios structure attribute OPOST was set.");
		finis
	finis



	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(loop_fildes);
		Zclose(tty_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
