#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.6.3_26"


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	
	setup(E_ACCESS);
	funct(E_ACCESS);

	assertion(26, BASE);
	start
		prep("Initialize a path name to file with write permission only.");
		start
			sprintf(path, "%s%s/%s", data_path(), "perms.d", "wr_owner");
		finis

		testing("Call to access() with modes of R_OK and X_OK are denied.");
		start
			expecting(-1);
			expecting(EACCES);
			Zaccess(path, R_OK);

			expecting(-1);
			expecting(EACCES);
			Zaccess(path, X_OK);
		finis

		prep("Initialize a path name to file with read permission only.");
		start
			sprintf(path, "%s%s/%s", data_path(), "perms.d", "rd_owner");
		finis

		testing("Call to access() with modes of W_OK and X_OK are denied.");
		start
			expecting(-1);
			expecting(EACCES);
			Zaccess(path, W_OK);

			expecting(-1);
			expecting(EACCES);
			Zaccess(path, X_OK);
		finis

		prep("Initialize a path name to file with execute permission only.");
		start
			sprintf(path, "%s%s/%s", data_path(), "perms.d", "ex_owner");
		finis

		testing("Call to access() with modes of R_OK and W_OK are denied.");
		start
			expecting(-1);
			expecting(EACCES);
			Zaccess(path, R_OK);

			expecting(-1);
			expecting(EACCES);
			Zaccess(path, W_OK);
		finis
	finis

	assertion(27, BASE);
	start
		prep("Initialize path name for a file where search permission");
		prep("is denied on a component of the path prefix.");
		start
			sprintf(path, "%s%s/%s", data_path(), "no_srch.d", "_tryme_");
		finis

		testing("Call to access() is denied.");
		start
			expecting(-1);
			expecting(EACCES);
			Zaccess(path, R_OK);
			expecting(-1);
			expecting(EACCES);
			Zaccess(path, W_OK);
			expecting(-1);
			expecting(EACCES);
			Zaccess(path, X_OK);
		finis
	finis

	if(tc_name_max()) {
		assertion(28, BASE);
		start
			e_NO_TRUNC(E_ACCESS);
		finis
	}
	else
		assertion(28, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(29, BASE);
		start
			e_NO_TRUNC(E_ACCESS);
		finis
	}
	else
		assertion(29, NOT_TESTABLE);

	assertion(30, BASE);
	start
		e_ENAMETOOLONG(E_ACCESS);
	finis

	assertion(31, BASE);
	start
		testing("Call to access() when path points to an empty string.");
		start
			expecting(-1);
			expecting(ENOENT);
			Zaccess("", R_OK);
		finis
	finis

	assertion(32, BASE);
	start
		testing("Call to access() when component of the path prefix");
		testing("does not exist.");
		start
			sprintf(path, "%s%s/%s", temp_path(), "_access_", "_tryme_");

			expecting(-1);
			expecting(ENOENT);
			Zaccess(path, R_OK);
		finis
	finis

	assertion(33, BASE);
	start
		testing("Call to access() for a file which does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_access_");

			expecting(-1);
			expecting(ENOENT);
			Zaccess(path, R_OK);
		finis
	finis

	assertion(34, BASE);
	start
		testing("Call to access() when pathname contains a non-directory");
		testing("component.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
			Zstrcat(path, "/dummy");

			expecting(-1);
			expecting(ENOTDIR);
			Zaccess(path, R_OK);
		finis
	finis

	assertion(36, BASE);
	start
		if(EINVAL_ACCESS) {
			prep("Setup a valid pathname.");
			start
				Zstrcpy(path, data_path());
				Zstrcat(path, "access_tests");
			finis

			testing("Invalid value specified for access mode.");
			start
				expecting(-1);
				expecting(EINVAL);
				Zaccess(path, EINVAL_ACCESS);
			finis
		}
		else
			not_implemented("[EINVAL] for access()");
	finis

	if(EINVAL_ACCESS) {
		assertion(37, EXTENDED_TESTED);
		start
			yes_implemented("supports detection of [EINVAL]");
		finis
	}
	else
		assertion(37, EXTENDED);

	done();
}
