#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Zstat.h>
#include	<Zsignal.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.6.4_25"
#define	PROT_BITS	(PROT_ALL | (mode_t)(S_ISUID | S_ISGID))

int main(void)
{
	char	path[_POSIX_PATH_MAX +1];

	setup(E_CHMOD);
	funct(E_CHMOD);

	assertion(25, BASE);
	start
		testing("Call to chmod() when search permission");
		testing("is denied on a component of the path prefix.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zstrcat(path, "/_tryme_");

			expecting(-1);
			expecting(EACCES);
			Zchmod(path, (mode_t) 0);
		finis

		prep("Set search mode on directory.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zchmod(path, PROT_ALL);
		finis

		testing("No change to file mode occurred.");
		start
			struct	stat	st_buf;

			Zstrcat(path, "/_tryme_");
			Zstat(path, &st_buf);
			if((st_buf.st_mode & PROT_BITS) != PROT_ALL)
				error("%s\n%s%lo\n%s%lo",
			  			"Call to stat() reported unexpected mode.",
			  			"Expected:	",
			  			(unsigned long) PROT_ALL,
			  			"Actual:	",
			  			(unsigned long) (st_buf.st_mode & PROT_BITS));
		finis

		cleanup("Reset data file permission changed during testing.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zchmod(path, S_IRUSR | S_IWUSR);
		finis
	finis

	if(tc_name_max()) {
		assertion(26, BASE);
		start
			e_NO_TRUNC(E_CHMOD);
		finis
	}
	else
		assertion(26, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(27, BASE);
		start
			e_NO_TRUNC(E_CHMOD);
		finis
	}
	else
		assertion(27, NOT_TESTABLE);

	assertion(28, BASE);
	start
		e_ENAMETOOLONG(E_CHMOD);
	finis

	assertion(29, BASE);
	start
		testing("A call to chmod() when a component of the path prefix");
		testing("contains a non-directory component.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
			Zstrcat(path, "/_tryme_");
		finis

		start
			expecting(-1);
			expecting(ENOTDIR);
			Zchmod(path, PROT_ALL);
		finis
	finis

	assertion(30, BASE);
	start
		testing("Call to chmod() when file named by pathname does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_chmod");

			expecting(-1);
			expecting(ENOENT);
			Zchmod(path, PROT_ALL);
		finis
	finis

	assertion(31, BASE);
	start
		testing("Call to chmod() when component of the path prefix");
		prep("does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_chmod_/_chmod_");

			expecting(-1);
			expecting(ENOENT);
			Zchmod(path, PROT_ALL);
		finis
	finis

	assertion(32, BASE);
	start
		testing("Call to chmod() when path points to an empty string.");
		start
			expecting(-1);
			expecting(ENOENT);
			Zchmod("", PROT_ALL);
		finis
	finis

	assertion(33, BASE);
	start
		testing("Call to chmod() when the effective user ID does not");
		testing("match the owner of the file and the calling process does not");
		testing("have the appropriate privileges to change the file mode.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "chmod_root");

			expecting(-1);
			expecting(EPERM);
			Zchmod(path, (mode_t) 0);
		finis

		testing("No change to file mode occurred.");
		start
			struct	stat	st_buf;

			Zstat(path, &st_buf);
			if((st_buf.st_mode & PROT_BITS) != PROT_ALL)
				error("%s\n%s%lo\n%s%lo",
			  			"Call to stat() reported unexpected mode.",
			  			"Expected:	",
			  			(unsigned long) PROT_ALL,
			  			"Actual:	",
			  			(unsigned long) (st_buf.st_mode & PROT_BITS));
		finis
	finis

	done();
}
