#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>
#include	<sv_errno.h>

/* This module has the S_ISUID mode set to ROOT. */

#define	PROT_BITS	(PROT_ALL | (mode_t)(S_ISUID | S_ISGID))


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	struct	stat st_buf, st_buf1;
	pid_t	pid;
	char	*pathm;
	time_t	s_time;

	setup(E_CHMOD);
	funct(E_CHMOD);

	if( ! PCD_READ_ONLY_FILE_SYSTEM) {
		assertion(34, BASE);
		start
			not_implemented("read-only file systems");
		finis

		done();
	}
	assertion(34, BASE);
	start
		pathm = fs_mount(MOUNT_RO, ROFS_REG_FILE);

		abstract("Ensure mounted file system gets unmounted.");
		start
			break_on(ERROR);
			prep("Establish path to implementation provided ROFS_TESTFILE");
			prep("and obtain a starting time.");
			start
				Zstrcpy(path, pathm);
				Zstrcat(path, "/");
				Zstrcat(path, ROFS_TESTFILE);

				Ztime(&s_time);			/* Obtain minimum reference time. */
				dt_secs(TRF_DELAY);		/* Delay a fixed interval.			*/
			finis
	
			preparation("Obtain mode for ROFS_TESTFILE.");
			start
				Zstat(path, &st_buf);
			finis
	
			testing("Call to chmod(), by process with appropriate privileges,");
			testing("to change a file on a read-only file system.");
			testing("Path name is \"%s\".", path);
			start
				expecting(EROFS);
				Zchmod(path, (mode_t) 0);
			finis
	
			testing("No change to file mode occurred.");
			start
				Zstat(path, &st_buf1);
				if ((st_buf.st_mode & PROT_BITS) !=
								(st_buf1.st_mode & PROT_BITS))
					error("%s\n%s%lo\n%s%lo",
			  				"Call to stat() reported unexpected mode.",
			  				"Expected:	",
			  				(unsigned long) (st_buf.st_mode & PROT_BITS),
			  				"Actual:	",
			  				(unsigned long) (st_buf1.st_mode & PROT_BITS));
			finis
	
			testing("Time related elements \"st_atime\", \"st_ctime\" and");
			testing("\"st_mtime\" of the parent directory are not changed");
			start
				trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, pathm);
				trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, pathm);
				trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, pathm);
			finis

			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				exit_on(ERROR);

				fs_umount(ROFS_REG_FILE);

				fs_mount(MOUNT_RW, ROFS_REG_FILE);

				prep("Set file owner to real user ID.");
				start
					Zchown(path, getuid(), getgid());
				finis

				fs_umount(NO_FILE);

				fs_mount(MOUNT_RO, NO_FILE);

				prep("Set effective user ID to real user ID");
				prep("and obtain a starting time.");
				start
					Zsetuid(getuid());

					Ztime(&s_time);			/* Obtain minimum reference time. */
					dt_secs(TRF_DELAY);		/* Delay a fixed interval.		*/
				finis
	
				testing("Call to chmod(), by process without appropriate");
				testing("privileges, to change file on read-only file system.");
				start
					expecting(EROFS);
					Zchmod(path, st_buf.st_mode);
				finis
	
				testing("No change to file mode occurred.");
				start
					Zstat(path, &st_buf1);
					if((st_buf.st_mode & PROT_BITS) !=
									(st_buf1.st_mode & PROT_BITS))
						error("%s\n%s%lo\n%s%lo",
			  					"Call to stat() reported unexpected mode.",
			  					"Expected:	",
			  					(unsigned long) (st_buf.st_mode & PROT_BITS),
			  					"Actual:	",
			  					(unsigned long) (st_buf1.st_mode & PROT_BITS));
				finis
	
				testing("Time related elements \"st_atime\", \"st_ctime\" and");
				testing("\"st_mtime\" of the parent directory are not changed");
				start
					trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, pathm);
					trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, pathm);
					trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, pathm);
				finis

				done();
			}
			wait_for(pid, TIME_PCTS_FS_TESTING);
		finis
	
		fs_umount(ROFS_REG_FILE);
	finis
	
	done();
}
