#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.6.5_20"

#define	S_IS_USER			1
#define	S_IS_APP			2
#define	S_ISUID_SUPPORTED	4
#define	S_ISGID_SUPPORTED	8


int main(void)
{
	struct	stat st_buf;
	char	path[_POSIX_PATH_MAX +1];

	setup(E_CHOWN);
	funct(E_CHOWN);

	assertion(20, BASE);
	start
		int			i;
		union		u {
						char	*p_beg;
						gid_t	*g_beg;
					} u;
		gid_t		*s_gids;
		long		n_grps;
		int			n_getgrp;

		
		Zstrcpy(path, data_path());
		if(pathconf(path, _PC_CHOWN_RESTRICTED) != (long) -1) {
			break_on(ERROR);

			prep("Obtain effective group ID and supplementary group IDs."); 
			start
				Zstrcpy(path, data_path());
				Zstrcat(path, "t_chown_res");
				/* Number of supplementary groups plus possibly the effective
				   group ID if not included in the count.
				*/
				n_grps = Zsysconf(_SC_NGROUPS_MAX) +1;
				/* Add +1 for adding effective group ID to array in case
				   the implementation does not provide it from getgroups()
				*/
				u.p_beg = (char *) Zmalloc((size_t)(sizeof(gid_t)*(n_grps +1)));
				n_getgrp = Zgetgroups((int)n_grps, u.g_beg);
										/* Add process group ID to list */
				s_gids = u.g_beg;
				s_gids[n_getgrp] = getegid();
			finis
		
			testing("Chown sets the group to the value of argument group");
			testing("when group is the calling processes effective group ID");
			testing("or one of its supplementary group IDs.");
			start
				continue_on(ERROR);		/* Get all errors recorded */
				for(i=0; i<n_grps; i++, s_gids++) {
					expecting(0);
					Zchown(path, UID, *s_gids);
					Zstat(path, &st_buf);
					if(st_buf.st_gid != *s_gids)
						error("%s\n\t\"%s\"\n%s%d.\n%s%lu\n%s%lu",
				    		"Call to chown() failed to set group ID for ",
				    		path,
							"on getgroups() array index of value ",
							i,
				    		"Expected:	",
				    		(unsigned long) *s_gids,
				    		"actual:	",
				    		(unsigned long) st_buf.st_gid);
					else
						testing("Tested index %d, GID \"%lu\"",
									i, (unsigned long) *s_gids);
				}
			finis
		}
		else {
			error("FIPS 151-2 required feature {_POSIX_CHOWN_RESTRICTED} is");
			error("not supported.");
		}
	finis

	assertion(21, NOT_APPLICABLE);

	if(PCTS_CHMOD_SET_IDS) {
		assertion(22, BASE);
		start
			prep("Setup path to test data file.");
			prep("Verify S_ISUID and S_ISGID bits are set on data file.");
			start
				Zstrcpy(path, data_path());
				Zstrcat(path, "suid_mode_test");

				Zstat(path, &st_buf);
				if(! (st_buf.st_mode & S_ISUID))
					error("%s\"%s\"\n%s\n%s",
						"Call to stat() reported unexpected mode for path ",
						path,
						"Expected:	S_ISUID bit set.",
						"Actual:	S_ISUID bit cleared.");
				if(! (st_buf.st_mode & S_ISGID))
					error("%s\"%s\"\n%s\n%s",
						"Call to stat() reported unexpected mode for path ",
						path,
						"Expected:	S_ISGID bit set.",
						"Actual:	S_ISGID bit cleared.");
			finis

			testing("S_ISUID and S_ISGID bits are cleared by chown().");
			start
				Zchown(path, UID, GID);
				Zstat(path, &st_buf);
				if(st_buf.st_mode & S_ISUID)
					error("%s\n%s\n%s",
				  		"Call to stat() reported unexpected mode.",
				  		"Expected:	S_ISUID bit cleared.",
				  		"Actual:	S_ISUID bit set.");
				if(st_buf.st_mode & S_ISGID)
					error("%s\n%s\n%s",
				  		"Call to stat() reported unexpected mode.",
				  		"Expected:	S_ISGID bit cleared.",
				  		"Actual:	S_ISGID bit set.");
			finis

			cleanup("Reset data file permisssion changed during testing.");
			start
				Zstrcpy(path, data_path());
				Zstrcat(path, "suid_mode_test");
				Zchmod(path, S_ISUID | S_ISGID | S_IRWXU | S_IRWXG | S_IRWXO);
			finis
		finis
	}
	else
		assertion(22, NOT_TESTABLE);

	assertion(23, BASE);
	start
		time_t	s_time;

		prep("Create a regular file type, record the time, and call chown().");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			Zchown(path, UID, gsup_gid());
		finis

		testing("Call to chown() updates the st_ctime field of the file.");
		start
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, path);
		finis
	finis

	done();
}
