#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.6.5_24"
void t_nch(uid_t uid, gid_t gid);

static	char path[_POSIX_PATH_MAX +1];


int main(void)
{
	setup(E_CHOWN);
	funct(E_CHOWN);

	assertion(24, BASE);
	start
		testing("Call to chown() when search permission is denied for a");
		testing("component of the path prefix.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zstrcat(path, "/_tryme_");

			expecting(-1);
			expecting(EACCES);
			Zchown(path, UID, GID);
		finis

		prep("Change mode of directory so search permission is permitted.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zchmod(path, PROT_ALL);
		finis

		t_nch(UID, GID);

		cleanup("Reset data file permisssion changed during testing.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zchmod(path, S_IRUSR | S_IWUSR);
		finis
	finis

	if(tc_name_max()) {
		assertion(25, BASE);
		start
			e_NO_TRUNC(E_CHOWN);
		finis
	}
	else
		assertion(25, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(26, BASE);
		start
			e_NO_TRUNC(E_CHOWN);
		finis
	}
	else
		assertion(26, NOT_TESTABLE);

	assertion(27, BASE);
	start
		e_ENAMETOOLONG(E_CHOWN);
	finis

	assertion(28, BASE);
	start
		testing("Call to chown() when path contains non-directory component.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
			Zstrcat(path, "/dummy");

			expecting(-1);
			expecting(ENOTDIR);
			Zchown(path, UID, GID);
		finis
	finis

	assertion(29, BASE);
	start
		testing("Call to chown() when file named by path does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_chown_");

			expecting(-1);
			expecting(ENOENT);
			Zchown(path, UID, GID);
		finis
	finis

	assertion(30, BASE);
	start
		testing("Call to chown() when component of path prefix does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_chown_/abc");

			expecting(-1);
			expecting(ENOENT);
			Zchown(path, UID, GID);
		finis
	finis

	assertion(31, BASE);
	start
		testing("Call to chown() when path points to an empty string.");
		start
			expecting(-1);
			expecting(ENOENT);
			Zchown("", UID, GID);
		finis
	finis

	assertion(32, BASE);
	start
		Zstrcpy(path, data_path());
		if(pathconf(path, _PC_CHOWN_RESTRICTED) != (long) -1) {
			prep("Setup a pathname where:");
			prep("    calling process does not have appropriate privileges,");
			prep("    owner of file does not match the process's");
			prep("          effective user ID,");
			prep("    and effective group ID of process matches group owner");
			prep("          of file.");
			start
				Zstrcpy(path, data_path());
				Zstrcat(path, "uid_bin");
			finis

			testing("Call to chown() using path provided.");
			start
				expecting(-1);
				expecting(EPERM);
				Zchown(path, BIN_UID, BIN_GID);
			finis

			t_nch(BIN_UID, GID);
		}
		else {
			error("FIPS 151-2 required feature {_POSIX_CHOWN_RESTRICTED} is");
			error("not supported.");
		}
	finis

	assertion(33, BASE);
	start
		Zstrcpy(path, data_path());
		if(pathconf(path, _PC_CHOWN_RESTRICTED) != (long) -1) {
			prep("Setup a pathname where:");
			prep("    calling process does not have appropriate privileges,");
			prep("    owner of file matches the process's");
			prep("          effective user ID,");
			prep("    and group owner of file does not match the process's");
			prep("    effective group ID or any supplementary group ID.");
			start
				Zstrcpy(path, data_path());
				Zstrcat(path, "gid_bin");
			finis

			testing("Call to chown() using path provided.");
			start
				expecting(-1);
				expecting(EPERM);
				Zchown(path, BIN_UID, BIN_GID);
			finis

			t_nch(UID, BIN_GID);
		}
		else {
			error("FIPS 151-2 required feature {_POSIX_CHOWN_RESTRICTED} is");
			error("not supported.");
		}
	finis

	assertion(34, BASE);
	start
		testing("Call to chown() when the argument owner does not match the");
		testing("effective user ID.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "uid_bin");		/* Owner of file is BIN. */

			expecting(-1);
			expecting(EPERM);
			Zchown(path, UID, GID);
		finis

		t_nch(BIN_UID, GID);
	finis

	done();
}


void t_nch(uid_t uid, gid_t gid)
{
	static	struct stat st_buf;

	testing("No change was made to owner or group of file.");
	start
		Zstat(path, &st_buf);
		if(st_buf.st_gid != gid)
			error("%s\n\t\"%s\"\n%s%lu\n%s%lu",
				"Call to stat() reported unexpected group ID for path",
				path,
				"Expected:	",
				(unsigned long) gid,
				"Actual:	",
				(unsigned long) st_buf.st_gid);
		if(st_buf.st_uid != uid)
			error("%s\n\t\"%s\"\n%s%lu\n%s%lu",
				"Call to stat() reported unexpected user ID for path ",
				path,
				"Expected:	",
				(unsigned long) uid,
				"Actual:	",
				(unsigned long) st_buf.st_uid);
	finis
}
