#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

/* This module has the S_ISUID mode set to ROOT. */


int main(void)
{
	time_t	s_time;
	char	path[_POSIX_PATH_MAX +1];
	char	*pathm;
	pid_t	pid;
	int		fd;				/* Ensure mounted file system gets unmounted. */

	setup(E_CREAT);
	funct(E_CREAT);

	if( ! PCD_READ_ONLY_FILE_SYSTEM) {
		assertion(52, BASE);
		start
			not_implemented("read-only file systems");
		finis
		assertion(53, BASE);
		start
			not_implemented("read-only file systems");
		finis

		done();
	}

	assertion(52, BASE);
	start
		pathm = fs_mount( MOUNT_RO, ROFS_REG_FILE );

		prep("Establish path to existing file on read-only file system");
		prep("and a starting time.");
		start
			Zstrcpy(path, pathm);
			Zstrcat(path, "/" );
			Zstrcat(path, ROFS_TESTFILE );

			Ztime(&s_time);			/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);		/* Delay a fixed interval.			*/
		finis

		testing("As appropriate privilege process, call to creat()");
		testing("when path references a file on a read-only file system");
		testing("Path name is \"%s\".", path);
		start
			expecting(-1);
			expecting(EROFS);
			fd = Zcreat(path, PROT_ALL);
			if(fd > 0)
				Zclose(fd);
		finis

		FORK
			prep("Set effective user ID to real user ID.");
			start
				Zsetuid(getuid());
			finis

			testing("As ordinary user process, call to creat()");
			testing("when path references a file on read-only file system");
			testing("Path name is \"%s\".", path);
			start
				expecting(-1);
				expecting(EROFS);
				fd = Zcreat(path, PROT_ALL);
				if(fd > 0)
					Zclose(fd);
			finis

			testing("Time related elements \"st_atime\", \"st_ctime\" and");
			testing("\"st_mtime\" of the file are not changed.");
			start
				trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);
				trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
				trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			finis
		KROF

		fs_umount(ROFS_REG_FILE);
	finis

	assertion(53, BASE);
	start
		pathm = fs_mount(MOUNT_RO, NO_FILE);

		prep("Establish path to non-existent file and obtain a");
		prep("starting time.");
		start
			Zstrcpy(path, pathm);
			Zstrcat(path, "/");
			Zstrcat(path, ROFS_TESTFILE);

			Ztime(&s_time);			/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);		/* Delay a fixed interval.			*/
		finis

		testing("As appropriate privilege process, call to creat()");
		testing("when path references a non-existent file on a read-only");
		testing("file system.");
		testing("Path name is \"%s\".", path);
		start
			expecting(-1);
			expecting(EROFS);
			fd = Zcreat(path, PROT_ALL);
			if(fd > 0)
				Zclose(fd);
		finis

		FORK
			prep("Set effective user ID to real user ID.");
			start
				Zsetuid(getuid());
			finis

			testing("As ordinary user process, call to open() when");
			testing("path references a non-existent file on a read-only");
			testing("file system.");
			start
				expecting(-1);
				expecting(EROFS);
				fd = Zcreat(path, PROT_ALL);
				if(fd > 0)
					Zclose(fd);
			finis

			testing("File is not created.");
			start
				expecting(-1);
				expecting(ENOENT);
				Zaccess(path, F_OK);
			finis

			testing("Time related elements \"st_ctime\" and \"st_mtime\"");
			testing("of the parent directory are not changed.");
			start
				trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, pathm);
				trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, pathm);
			finis
		KROF

		fs_umount(NO_FILE);
	finis

	done();
}
