#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FILENAME	"5.7.1b_26"
#define	EINVAL_FPATHCONF_NAME	(_PC_LINK_MAX +_PC_MAX_CANON +_PC_MAX_INPUT \
								+_PC_NAME_MAX +_PC_PATH_MAX +_PC_PIPE_BUF \
								+_PC_CHOWN_RESTRICTED +_PC_NO_TRUNC \
								+_PC_VDISABLE)

int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	long	r_success, r_test;
	int		fd_dir, fd_tty, fd_reg;
	pid_t	pid;


	setup(E_FPATHCONF);
	funct(E_FPATHCONF);

	assertion(26, BASE);
	start
		prep("Open a regular and a directory file type and possibly a");
		prep("terminal file type.");
		start
			exit_on(ERROR);

			SET_TIMEOUT
				Zstrcpy(path, temp_path());
				fd_dir = Zopen(path, O_RDONLY);
				Zstrcat(path, FILENAME);
				fd_reg = Zcreat(path, PROT_ALL);
				if(GTI_PROVIDED)
					fd_tty = Zopen(FILENAME_OF_TERMINAL_FILE,
												O_RDONLY | O_NONBLOCK);
			CLEAR_ALARM
		finis

		testing("Call to fpathconf() with an invalid value for name.");
		start
			expecting(-1);
			expecting(EINVAL);
			Zfpathconf(fd_dir, EINVAL_FPATHCONF_NAME);
		finis
	finis

	assertion(27, BASE);
	start
		FORK
			testing("Call to fpathconf() when fildes is not valid.");
			start
				r_success = Zfpathconf(fd_dir, _PC_NAME_MAX);
				Zclose(fd_dir);

				errno = 0;
				r_test = fpathconf(fd_dir, _PC_NAME_MAX);
				if(r_test == (long) -1) {
					if(errno != EBADF)
						error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EBADF].",
							"Actual errno value reported was ",
							err_lookup(errno) );
				}
				else
					if(r_test != r_success)
						error("%s%ld.\n%s%ld.",
								"Value returned was ",
								r_test,
								"Expected value of ",
								r_success);
			finis
		KROF
	finis

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(28, BASE);
		GTI_NOT_PROVIDED_RESULT(29, BASE);
	}
	else {
		assertion(28, BASE);
		start
			testing("Call to fpathconf(fildes, _PC_MAX_CANON) when fildes");
			testing("does not refer to a terminal file.");
			start
				r_success = Zfpathconf(fd_tty, _PC_MAX_CANON);
	
				errno = 0;
				r_test = fpathconf(fd_dir, _PC_MAX_CANON);
				if(r_test == (long) -1) {
					if(errno != EINVAL)
						error("%s\n%s\n%s%s.",
								"Value returned was -1.",
								"Expected errno value of [EINVAL].",
								"Actual errno value reported was ",
								err_lookup(errno) );
				}
				else
					if(r_test != r_success)
						error("%s%ld.\n%s%ld.",
								"Value returned was ",
								r_test,
								"Expected value of ",
								r_success);
			finis
		finis

		assertion(29, BASE);
		start
			testing("Call to fpathconf(fildes, _PC_MAX_INPUT) when fildes");
			testing("does not refer to a terminal file.");
			start
				r_success = Zfpathconf(fd_tty, _PC_MAX_INPUT);

				errno = 0;
				r_test = fpathconf(fd_dir, _PC_MAX_INPUT);
				if(r_test == (long) -1) {
					if(errno != EINVAL)
						error("%s\n%s\n%s%s.",
								"Value returned was -1.",
								"Expected errno value of [EINVAL].",
								"Actual errno value reported was ",
								err_lookup(errno) );
				}
				else
					if(r_test != r_success)
						error("%s%ld.\n%s%ld.",
								"Value returned was ",
								r_test,
								"Expected value of ",
								r_success);
			finis
		finis
	}

	assertion(30, BASE);
	start
		testing("Call to fpathconf(fildes, _PC_NAME_MAX) when fildes does");
		testing("not refer to a directory.");
		start
			r_success = Zfpathconf(fd_dir, _PC_NAME_MAX);

			if(GTI_PROVIDED) {
				errno = 0;
				r_test = fpathconf(fd_tty, _PC_NAME_MAX);
				if(r_test == (long) -1) {
					if(errno != EINVAL)
						error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
				}
			}

			errno = 0;
			r_test = fpathconf(fd_reg, _PC_NAME_MAX);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(31, BASE);
	start
		testing("Call to fpathconf(fildes, _PC_PATH_MAX) when fildes does");
		testing("not refer to a directory.");
		start
			r_success = Zfpathconf(fd_dir, _PC_PATH_MAX);

			if(GTI_PROVIDED) {
				errno = 0;
				r_test = fpathconf(fd_tty, _PC_PATH_MAX);
				if(r_test == (long) -1) {
					if(errno != EINVAL)
						error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
				}
			}

			errno = 0;
			r_test = fpathconf(fd_reg, _PC_PATH_MAX);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis


	assertion(32, BASE);
	start
		testing("Call to fpathconf(fildes, _PC_PIPE_BUF) when fildes does");
		testing("not refer to a FIFO or directory.");
		start
			r_success = Zfpathconf(fd_dir, _PC_PIPE_BUF);

			if(GTI_PROVIDED) {
				errno = 0;
				r_test = fpathconf(fd_tty, _PC_PIPE_BUF);
				if(r_test == (long) -1) {
					if(errno != EINVAL)
						error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
				}
			}

			errno = 0;
			r_test = fpathconf(fd_reg, _PC_PIPE_BUF);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(33, BASE);
	start
		testing("Call to fpathconf(fildes, _PC_NO_TRUNC) when fildes does");
		testing("not refer to a directory.");
		start
			r_success = Zfpathconf(fd_dir, _PC_NO_TRUNC);

			if(GTI_PROVIDED) {
				errno = 0;
				r_test = fpathconf(fd_tty, _PC_NO_TRUNC);
				if(r_test == (long) -1) {
					if(errno != EINVAL)
						error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
				}
			}

			errno = 0;
			r_test = fpathconf(fd_reg, _PC_NO_TRUNC);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(34, BASE);
	start
		testing("Call to fpathconf(fildes, _PC_VDISABLE) when fildes does");
		testing("not refer to a terminal file.");
		start
			errno = 0;
			r_test = fpathconf(fd_dir, _PC_VDISABLE);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else {
				if(GTI_PROVIDED) {
					r_success = Zfpathconf(fd_tty, _PC_VDISABLE);
					if(r_test != r_success)
						error("%s%ld.\n%s%ld.",
								"Value returned was ",
								r_test,
								"Expected value of ",
								r_success);
				}
				else {
					r_success = Zfpathconf(fd_reg, _PC_VDISABLE);
					if(r_test != r_success)
						error("%s%ld.\n%s%ld.",
									"Value returned was ",
									r_test,
									"Expected value of ",
									r_success);
				}
			}
		finis
	finis

	done();
}
