#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.6.2b_04"

int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	struct	stat st_buf;
	time_t	time_beg, time_end;
	int		fd;

	setup(E_FSTAT);
	funct(E_FSTAT);

	assertion(4, BASE);
	start
		prep("Establish file with known parameters for fstat() testing.");
		start
			umask(0);						/* Clear file mode creation mask. */

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Ztime(&time_beg);			/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			fd = Zcreat(path, PROT_ALL);
		finis

		testing("A call to fstat() returns a value of zero.");
		testing("Information stored by fstat().");
		start
			expecting(0);
			Zfstat(fd, &st_buf);

			dt_secs(TRF_DELAY);				/* Delay a fixed interval.		*/
			Ztime(&time_end);				/* Obtain minimum reference time. */
		finis

		testing("Call to fstat() stores in the st_mode member of the struct");
		testing("stat the proper info.");
		start
			if((st_buf.st_mode & PROT_ALL) != PROT_ALL)
				error("%s%o (octal)\n%s%o (octal)",
					"Expected: ",
					PROT_ALL,
					"Received: ",
					(int)(st_buf.st_mode & PROT_ALL));
		finis
	finis

	assertion(5, BASE);
	start
		char	path2[_POSIX_PATH_MAX +1];
		struct	stat st_buf2;

		prep("Create a second file.");
		start
			Zstrcpy(path2, temp_path());
			Zstrcat(path2, "AA");
			Zclose(Zcreat(path2, PROT_ALL));
			Zstat(path2, &st_buf2);
		finis

		testing("Call to fstat() stores in the st_ino and st_dev members of");
		testing("the struct stat the proper info.");
		start
			if(st_buf.st_ino == st_buf2.st_ino)
				error("%s%s\n%s%lu.",
					"Expected inode values of two files on",
					"same device to be different.",
					"Received the same inode value of ",
					(unsigned long) st_buf.st_ino);

			if(st_buf.st_dev != st_buf2.st_dev)
				error("%s%s\n%s%lu %lu.",
					"Expected device values of two files on",
					"same device to be the same.",
					"Received device values of ",
					(unsigned long) st_buf.st_dev,
					(unsigned long) st_buf2.st_dev);
		finis
	finis

	assertion(6, BASE);
	start
		testing("Call to fstat() stores in the st_nlink member of the struct");
		testing("stat the proper info.");
		start
			if(st_buf.st_nlink != (nlink_t) 1)
				error("%s\n%s%lu",
					"Expected: 1 ",
					"Received: ",
					(unsigned long) st_buf.st_nlink);
		finis
	finis

	assertion(7, BASE);
	start
		testing("Call to fstat() stores in the st_uid member of the struct");
		testing("stat the proper info.");
		start
			if(st_buf.st_uid != (uid_t)UID)
				error("%s%lu\n%s%lu",
					"Expected: ",
					(unsigned long) UID,
					"Received: ",
					(unsigned long) st_buf.st_uid);
		finis
	finis

	assertion(8, BASE);
	start
		testing("Call to fstat() stores in the st_gid member of the struct");
		testing("stat the proper info.");
		start
			if(st_buf.st_gid != (gid_t)GID)
				error("%s%lu\n%s%lu",
					"Expected: ",
					(unsigned long) GID,
					"Received: ",
					(unsigned long) st_buf.st_gid);
		finis
	finis

	assertion(9, BASE);
	start
		testing("Call to fstat() stores in the st_size member of the struct");
		testing("stat the proper info.");
		start
			if(st_buf.st_size != (off_t)0)
				error("%s\n%s%lu",
					"Expected: 0",
					"Received: ",
					(unsigned long) st_buf.st_size);
		finis
	finis

	assertion(10, BASE);
	start
		testing("Call to fstat() stores in the st_atime member of the struct");
		testing("stat the proper info.");
		start
			if((st_buf.st_atime <= time_beg) || (st_buf.st_atime >= time_end))
				error("%s%lu%s%lu.\n%s%lu",
					"Expected a value greater than ",
					(unsigned long) time_beg,
					" and less than ",
					(unsigned long) time_end,
					"Received: ",
					(unsigned long) st_buf.st_atime);
		finis
	finis

	assertion(11, BASE);
	start
		testing("Call to fstat() stores in the st_mtime member of the struct");
		testing("stat the proper info.");
		start
			if((st_buf.st_mtime <= time_beg) || (st_buf.st_mtime >= time_end))
				error("%s%lu%s%lu.\n%s%lu",
					"Expected a value greater than ",
					(unsigned long) time_beg,
					" and less than ",
					(unsigned long) time_end,
					"Received: ",
					(unsigned long) st_buf.st_mtime);
		finis
	finis

	assertion(12, BASE);
	start
		testing("Call to fstat() stores in the st_ctime member of the struct");
		testing("stat the proper info.");
		start
			if((st_buf.st_ctime <= time_beg) || (st_buf.st_ctime >= time_end))
				error("%s%lu%s%lu.\n%s%lu",
					"Expected a value greater than ",
					(unsigned long) time_beg,
					" and less than ",
					(unsigned long) time_end,
					"Received: ",
					(unsigned long) st_buf.st_ctime);
		finis
	finis

										/* This assertion is tested for
										   every POSIX function, in the
										   time related tests for that
										   function.
										*/
	assertion(13, BASE);
	start
		time_t	s_time;

		prep("Record current time, then create a regular file.");
		start
			unlink(path);
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			fd = Zcreat(path, PROT_ALL);
		finis

		testing("Call to fstat() updates those time related fields marked");
		testing("for update.");
		start
			trf_ck(CHK_CTIME, TRUE, s_time, fd, NO_PATH);
			trf_ck(CHK_MTIME, TRUE, s_time, fd, NO_PATH);
			trf_ck(CHK_ATIME, TRUE, s_time, fd, NO_PATH);
		finis
	finis

	done();
}
