#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>
#include	<sv_errno.h>

/* This module has the S_ISUID mode set to ROOT. */


#define	FILENAME		"5.3.4_61"
#define	FILENAME_NEW	"5.3.4_61_n"
#define	DIRNAME			"5.3.4_61.d"

static	char	path_ex[_POSIX_PATH_MAX +1], path_new[_POSIX_PATH_MAX +1];
static	char	*pathm;
void	t_exdev(void);


int main(void)
{
	pid_t	pid;
	struct	stat st_buf1, st_buf2;

	setup(E_LINK);
	funct(E_LINK);

	assertion(61, BASE);
	start
		if(PCD_LINK_TO_DIRECTORY == TRUE) {
			preparation("Establish paths for linking to a directory.");
			start
				Zstrcpy(path_ex, temp_path());
				Zstrcat(path_ex, DIRNAME);
				Zmkdir(path_ex, PROT_ALL);
	
				Zstrcpy(path_new, temp_path());
				Zstrcat(path_new, FILENAME_NEW);
			finis
	
			FORK
				prep("Set effective user ID to real user ID.");
				start
					Zsetuid(getuid());
				finis

				testing("Ordinary user process linking to a directory.");
				testing("Implementation supports links to directories.");
				start
					Zstat(path_ex, &st_buf1);
					expecting(-1);
					expecting(EPERM);
					Zlink(path_ex, path_new);
									/* No link was created. */
					expecting(ENOENT);
					Zaccess(path_new, F_OK);

										/* Link count unchanged. */
					Zstat(path_ex, &st_buf2);
					if(st_buf1.st_nlink != st_buf2.st_nlink)
						error("Link count changed from %lu to %lu.",
									(unsigned long) st_buf1.st_nlink,
									(unsigned long) st_buf2.st_nlink);
				finis
			KROF

			cleanup("Remove files possibly created.");
			start
				unlink(path_new);
				Zrmdir(path_ex);
			finis
		}
		else
			not_implemented("{PCD_LINK_TO_DIRECTORY} of TRUE");
	finis

	assertion(62, BASE);
	start
		if(PCD_LINK_TO_DIRECTORY == FALSE) {
			nlink_t	cur_nlink;

			preparation("Establish paths for linking to a directory.");
			start
				Zstrcpy(path_ex, temp_path());
				Zstrcat(path_ex, DIRNAME);
				Zmkdir(path_ex, PROT_ALL);

				Zstat(path_ex, &st_buf1);
				cur_nlink = st_buf1.st_nlink;
	
				Zstrcpy(path_new, temp_path());
				Zstrcat(path_new, FILENAME_NEW);
			finis
	
			FORK
				testing("User with appropriate privileges linking to a");
				testing("directory when implementation does not support links");
				testing("to directories.");
				start
					expecting(-1);
					expecting(EPERM);
					Zlink(path_ex, path_new);
									/* No link was created. */
					expecting(ENOENT);
					Zaccess(path_new, F_OK);

									/* Link count unchanged. */
					Zstat(path_ex, &st_buf1);
					if(st_buf1.st_nlink != cur_nlink)
						error("Link count was changed from %lu to %lu.",
								(unsigned long) cur_nlink,
								(unsigned long) st_buf1.st_nlink);
				finis
			KROF

			cleanup("Remove files possibly created.");
			start
				unlink(path_new);
				Zrmdir(path_ex);
			finis
		}
		else
			not_implemented("{PCD_LINK_TO_DIRECTORY} of FALSE");
	finis

	assertion(63, NOT_APPLICABLE);
	
	assertion(64, BASE);
	start
		time_t	s_time;

		if(PCD_READ_ONLY_FILE_SYSTEM) {
			break_on(ERROR);
			pathm = fs_mount(MOUNT_RO, ROFS_REG_FILE);
	
			abstract("Ensure mounted file system gets unmounted.");
			start
				break_on(ERROR);
				prep("Establish path to implementation provided ROFS_TESTFILE");
				start
					Zstrcpy(path_new, pathm);
					Zstrcat(path_new, "/link_rofs_su");
	
					Zstrcpy(path_ex, pathm);
					Zstrcat(path_ex, "/");
					Zstrcat(path_ex, ROFS_TESTFILE);
				finis
	
				testing("Appropriate privilege process creating link to file");
				testing("on a read only file system.");
				testing("link(%s, %s)", path_ex, path_new);
				start
					expecting(EROFS);
					Zlink(path_ex, path_new);
				finis
	
				FORK
					exit_on(ERROR);
					prep("Set effective user ID to real user ID.");
					start
						Zsetuid(getuid());
					finis
	
					testing("Ordinary user process creating link() to a file");
					testing("on a read only file system.");
					start
						Ztime(&s_time);		/* Obtain minimum reference time. */
						dt_secs(TRF_DELAY);	/* Delay a fixed interval.		*/
	
						continue_on(ERROR);
						expecting(-1);
						expecting(EROFS);
						Zlink(path_ex, path_new);
	
										/* No link was created. */
						expecting(ENOENT);
						Zaccess(path_new, F_OK);
	
										/* Link count unchanged. */
						Zstat(path_ex, &st_buf1);
						if(st_buf1.st_nlink != (nlink_t)1)
							error("Link count set to %lu.",
									(unsigned long) st_buf1.st_nlink);
					finis
	
					testing("Time related elements \"st_atime\", \"st_ctime\"");
					testing("and \"st_mtime\" of parent directory are not");
					testing("changed.");
					start
						trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, pathm);
						trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, pathm);
						trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, pathm);
					finis
				KROF
			finis
	
			fs_umount(ROFS_REG_FILE);
		}
		else
			not_implemented("read-only file systems");
	finis


	if(strlen(PCTS_FS) == 0)
		assertion(65, NOT_TESTABLE);
	else {
		assertion(65, BASE);
		start
			if(PCD_LINK_FILE_SYSTEM == FALSE)
				t_exdev();
			else
				not_implemented("{PCD_LINK_FILE_SYSTEM} of FALSE");
		finis
	}

	assertion(66, NOT_APPLICABLE);

	done();
}


void t_exdev(void)
{
	pathm = fs_mount(MOUNT_RW, ROFS_REG_FILE);

	abstract("Ensure mounted file system gets unmounted.");
	start
		break_on(ERROR);
		preparation("Establish paths for linking.");
		start
			Zstrcpy(path_new, pathm);
			Zstrcat(path_new, "/link_exdev");
	
			temp_creat("exdev_ck", PROT_ALL);
			Zstrcpy(path_ex, temp_path());
			Zstrcat(path_ex, "exdev_ck");
		finis
	
		testing("Appropriate privilege process creating a link to a");
		testing("file on a different file system.");
		testing("Path name is \"%s\".", path_new);
		start
			if(PCD_LINK_FILE_SYSTEM == FALSE)
				expecting(EXDEV);
			Zlink(path_ex, path_new);
		finis
	finis
	
	(void) fs_umount(ROFS_REG_FILE);
}
