#include <pcts.h>

#include <Zunistd.h>
#include <sys/Zstat.h>
#include <Zfcntl.h>
#include <Ztermios.h>

#include <sv_env.h>
#include <sv_errno.h>
#include <pcts_timer.h>


#define	TIME_TO_DELAY	(4)
#define	TIMING_SLOP		(1)
#define	FIFONAME		"5.3.1_35_fifo"
#define	String			"DaTa"
#define	Strlen			(sizeof(String) -1)

int main(void)
{
	pid_t	pid;
	char	path[_POSIX_PATH_MAX +1];
	char	inbuf[Strlen +1];
	int		fd_rd, fd_wr;
	int		trunc;

	setup(E_OPEN);
	funct(E_OPEN);

	assertion(35, BASE);
	start
		prep("Establish FIFO test file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "FIFONAME");
			Zmkfifo(path, PROT_ALL);
		finis

		testing("Call to open() on FIFO with O_RDONLY and O_NONBLOCK set");
		testing("returns without delay.");
		start
			SET_ALARM(1);
				fd_rd = Zopen(path, O_RDONLY | O_NONBLOCK);
			CLEAR_ALARM;
			Zclose(fd_rd);
		finis
	finis

	assertion(36, BASE);
	start
	  ENABLE_PSYNC
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis
	
		if(pid == (pid_t)0) {
			prep("Child opens FIFO with O_WRONLY and O_NONBLOCK set.");
			start
				sv_flush();
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM

				SLEEP(TIME_TO_DELAY);
				Zopen(path, O_WRONLY | O_NONBLOCK);
				SLEEP(WAIT_FOR_TIME);
			finis
	
			done();
		}
	
		prep("Parent opens FIFO with O_RDONLY set and O_NONBLOCK clear.");
		start
			SEND_PSYNC
			SET_TIMEOUT
				fd_rd = Zopen(path, O_RDONLY);
			CLEAR_ALARM
		finis

		testing("Read channel blocked until the write channel was opened.");
		start
			if(unused_time > (TIMEOUT -TIME_TO_DELAY +TIMING_SLOP))
					error("The open() call did not block.");
		finis
	
		WAIT_FOR_PID(pid);
	
		cleanup("Close open FIFO file.");
		start
			Zclose(fd_rd);
		finis
	  DISABLE_PSYNC
	finis

	assertion(37, BASE);
	start
	  ENABLE_PSYNC
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Child opens FIFO with O_RDONLY and O_NONBLOCK set.");
			start
				sv_flush();
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
				SLEEP(TIME_TO_DELAY);
				Zopen(path, O_RDONLY | O_NONBLOCK);
				SLEEP(WAIT_FOR_TIME);
			finis

			done();
		}

		prep("Parent opens FIFO with O_WRONLY set and O_NONBLOCK clear.");
		start
			SEND_PSYNC
			SET_TIMEOUT
				fd_wr = Zopen(path, O_WRONLY);
			CLEAR_ALARM
		finis

		testing("Write channel blocked until the read channel was opened.");
		start
			if(unused_time > (MAX_TIME_TO_RUN -TIME_TO_DELAY +TIMING_SLOP))
					error("The open() call did not block.");
		finis

		WAIT_FOR_PID(pid);

		cleanup("Close open FIFO file.");
		start
			Zclose(fd_wr);
		finis
	  DISABLE_PSYNC
	finis

	assertion(38, EXTENDED);
	assertion(39, EXTENDED);
	assertion(40, EXTENDED);
	assertion(41, EXTENDED);

	assertion(43, BASE);
	start
		SET_ALARM(TIMEOUT)
			prep("Creat a FIFO file containing data.");
			start
				fd_rd = Zopen(path, O_RDONLY | O_NONBLOCK);
				fd_wr = Zopen(path, O_WRONLY);
				Zwrite(fd_wr, String, Strlen);
				Zclose(fd_wr);
			finis

			prep("Open write channel of FIFO file with O_TRUNC clear,");
			prep("then attempt to read FIFO.");
			start
				fd_wr = Zopen(path, O_WRONLY);
				trunc = TRUE;
				if(read(fd_rd, inbuf, Strlen) == Strlen)
					trunc = FALSE;

				Zwrite(fd_wr, String, Strlen);
				Zclose(fd_wr);
			finis

			testing("Call to open() with the O_TRUNC flag set for a FIFO file");
			testing("is the same as with the O_TRUNC flag clear.");
			start
				fd_wr = Zopen(path, O_WRONLY | O_TRUNC);
				if(trunc == FALSE) {
					expecting(Strlen);
					Zread(fd_rd, inbuf, Strlen);
				}
				else {
					expecting(EAGAIN);
					Zread(fd_rd, inbuf, Strlen);
				}
			finis

			cleanup("Close FIFO files.");
			start
				Zclose(fd_wr);
				Zclose(fd_rd);
			finis
		CLEAR_ALARM
	finis

	done();
}
