#include <pcts.h>

#include <Zunistd.h>
#include <sys/Zstat.h>
#include <Zfcntl.h>
#include <Ztermios.h>

#include <sv_env.h>
#include <sv_errno.h>
#include <pcts_timer.h>

#define	String	"DaTa\n"
#define Strlen	(sizeof(String)-1)

int main(void)
{
	int		tty_fildes, loop_fildes, fd_wr, fd_rd;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	pid_t 	pid;
	char 	inbuf[_POSIX_MAX_INPUT];
	int		trunc;

	setup(E_OPEN);
	funct(E_OPEN);

	assertion(44, BASE);
	start
		GTI_NOT_PROVIDED_EXIT

		exit_on(ERROR);

	    FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);
				tty_termios_p->c_lflag &= ~ICANON;
				tty_termios_p->c_cc[VMIN] = 0;
				tty_termios_p->c_cc[VTIME] = 0;
				Ztcsetattr(tty_fildes, TCSAFLUSH, tty_termios_p);
			finis

											/* Test on input buffer. */
			prep("Write to PORT_LOOP.");
			start
		  		SET_TIMEOUT
					expecting(Strlen);
					Zwrite(loop_fildes, String, Strlen);
					Ztcdrain(loop_fildes);
					dt_secs(TIME_ACQUIRE_XMITTED_DATA);
		  		CLEAR_ALARM
			finis

			prep("Open a second file for PORT_TTY with O_TRUNC clear");
			prep("then attempt to read PORT_TTY.");
			start
		  		SET_TIMEOUT
					fd_rd = Zopen(PORT_TTY, O_RDWR);
					trunc = TRUE;
					if(read(tty_fildes, inbuf, Strlen) == Strlen)
						trunc = FALSE;

					expecting(Strlen);
					Zwrite(loop_fildes, String, Strlen);
					Ztcdrain(loop_fildes);
					dt_secs(TIME_ACQUIRE_XMITTED_DATA);
		  		CLEAR_ALARM
			finis

			testing("Call to open() with the O_TRUNC flag set for terminal");
			testing("device file is the same as with the O_TRUNC flag clear.");
			start
		  		SET_TIMEOUT
					fd_rd = Zopen(PORT_LOOP, O_RDWR | O_TRUNC);
					if(trunc == FALSE) {
						expecting(Strlen);
						Zread(tty_fildes, inbuf, Strlen);
					}
					else {
						expecting(EAGAIN);
						Zread(tty_fildes, inbuf, Strlen);
					}
		  		CLEAR_ALARM
			finis

										/*	Test on output buffer. */
			if(!PCTS_GTI_BUFFERS_OUTPUT) {
					/*	Implementations that do not provide a buffering
						mechanism on write() will timeout when terminal
						output is suspended.

						Since the output cannot be buffered, O_TRUNC will
						never truncate the file.  Testing that the
						PCTS_GTI_BUFFERS_OUTPUT "config" flag does apply for the
						implementation is performed in the tcflow() tests.
					*/
				testing("Call to open() with the O_TRUNC flag set for");
				testing("terminal device file is the same as with the O_TRUNC");
				testing("flag clear.");
				start
					message("%s\n%s",
							"Implementation does not buffer output on GTI",
							"devices, output buffer testing is not performed.");
				finis
			}
			else {
				prep("Write to PORT_LOOP.");
				start
		  			SET_TIMEOUT
						Ztcflow(loop_fildes, TCOOFF);
						expecting(Strlen);
						Zwrite(loop_fildes, String, Strlen);
		  			CLEAR_ALARM
				finis

				prep("Open a second file for PORT_LOOP with O_TRUNC clear");
				prep("then attempt to read PORT_TTY.");
				start
		  			SET_TIMEOUT
						fd_wr = Zopen(PORT_LOOP, O_RDWR);
						Ztcflow(fd_wr, TCOON);
						Ztcdrain(fd_wr);
						dt_secs(TIME_ACQUIRE_XMITTED_DATA);
						trunc = TRUE;
						if(read(tty_fildes, inbuf, Strlen) == Strlen)
							trunc = FALSE;

						Ztcflow(loop_fildes, TCOOFF);
						expecting(Strlen);
						Zwrite(loop_fildes, String, Strlen);
		  			CLEAR_ALARM
				finis

				testing("Call to open() with the O_TRUNC flag set on terminal");
				testing("device file is the same as with O_TRUNC flag clear.");
				start
		  			SET_TIMEOUT
						fd_wr = Zopen(PORT_LOOP, O_RDWR | O_TRUNC);
						Ztcflow(fd_wr, TCOON);
						Ztcdrain(fd_wr);
						dt_secs(TIME_ACQUIRE_XMITTED_DATA);
						if(trunc == FALSE) {
							expecting(Strlen);
							Zread(tty_fildes, inbuf, Strlen);
						}
						else {
							expecting(EAGAIN);
							Zread(tty_fildes, inbuf, Strlen);
						}
		  			CLEAR_ALARM
				finis
			}
		KROF
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
