#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	String	"This is a test!!!"
#define	Strlen	(sizeof(String) -1)


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	char	inbuf[Strlen +1];
	int		fd;
	time_t	s_time;

	setup(E_OPEN);
	funct(E_OPEN);

	assertion(47, BASE);
	start
		prep("Initialize path name for a file where search permission");
		prep("is denied on a component of the path prefix.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perm.d");
			Zstrcat(path, "/open_tryme");
		finis

		testing("A call to open() when search permission is denied.");
		start
			expecting(-1);
			expecting(EACCES);
			Zopen(path, O_RDONLY);
		finis
	finis

	assertion(48, BASE);
	start
		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perms.d");
			Zstrcat(path, "/no_read");
		finis

		testing("Read permission along the path is denied.");
		start
			expecting(-1);
			expecting(EACCES);
			Zopen(path, O_RDONLY);
		finis

		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perms.d");
			Zstrcat(path, "/no_write");
		finis

		testing("Write permission along the path is denied.");
		start
			expecting(-1);
			expecting(EACCES);
			Zopen(path, O_WRONLY);
		finis
	finis

	assertion(49, BASE);
	start
		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_write.d");
			Zstrcat(path, "/new_file");
		finis

		testing("File does not exist and parent directory does not");
		testing("permit writing.");
		start
			expecting(-1);
			expecting(EACCES);
			Zopen(path, O_WRONLY | O_CREAT, PROT_ALL);
		finis

		testing("File is not created.");
		start
			expecting(-1);
			expecting(ENOENT);
			Zaccess(path, F_OK);
		finis
	finis

	assertion(50, BASE);
	start
		prep("Establish path where write permission is denied to owner.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_write");
		finis

		testing("Call to open() when O_TRUNC flag is set and write permission");
		testing("is denied on the path referenced.");
		start
			expecting(-1);
			expecting(EACCES);
			Zopen(path, O_WRONLY | O_TRUNC);
		finis

		testing("File referenced is not truncated.");
		start
			fd = Zopen(path, O_RDONLY);
			if(Zread(fd, inbuf, Strlen) == 0)
				error("File was truncated.");
		finis
	finis

	assertion(51, EXTENDED);

	assertion(52, BASE);
	start
		time_t	s_time;

		prep("Initialize path name for testing.");
		prep("Obtain a starting time.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "Open_Test");

			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis
	
		testing("Call to open() when O_CREAT and O_EXCL flags are set");
		testing("and the named file exists.");
		start
			expecting(-1);
			expecting(EEXIST);
			Zopen(path, O_WRONLY | O_CREAT | O_EXCL, PROT_ALL);
		finis

		testing("Time related elements \"st_ctime\", \"st_mtime\", and");
		testing("\"st_atime\" of the file and \"st_ctime\"");
		testing("and \"st_mtime\" of the parent directory are not changed.");
		start
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);

			Zstrcpy(path, data_path());
			strcat(path, ".");
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	assertion(53, BASE);
	start
		pid_t	pid;
		struct	stat st_buf;

		prep("Create a FIFO file and obtain a starting time.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "er_eintr");
			Zmkfifo(path, PROT_ALL);

			Zstrcpy(path, temp_path());
			strcat(path, ".");
			Zstat(path, &st_buf);	/*	Ensure created FIFO's time-related
										fields are updated.	*/
			Zstrcpy(path, temp_path());
			Zstrcat(path, "er_eintr");
			Zstat(path, &st_buf);	/*	Ensure FIFO's parent directory
										time-related fields are updated.	*/

			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis

		prep("Fork a child.");
		start
			sv_flush( );
			pid = Zfork();
		finis
	
		if(pid == (pid_t)0) {
			SIGNAL(SIGINT, sig_int);
			prep("When open() is interrupted, errno is set to [EINTR].");
			start
				expecting(-1);
				expecting(EINTR);
				Zopen(path, O_WRONLY);
			finis

			testing("Time related elements \"st_ctime\", \"st_mtime\", and");
			testing("\"st_atime\" of the file and \"st_ctime\" and");
			testing("\"st_mtime\" of the parent directory are not changed.");
			start
				trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
				trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
				trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);

				Zstrcpy(path, temp_path());
				strcat(path, ".");
				trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
				trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			finis

			done();
		}

		SLEEP(1);
		Zkill(pid, SIGINT);			/* Send signal to child. */
		WAIT_FOR_PID(pid);
	finis

	assertion(54, BASE);
	start
		struct	stat st_buf;

		prep("Initialize a path name.");
		prep("Obtain a starting time.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "open_dir");
			Zmkdir(path, PROT_ALL);

			Zstat(path, &st_buf);		/*	Set base-line times for directory */
			strcat(path, "/..");
			Zstat(path, &st_buf);			/*	and parent directory.*/

			Zstrcpy(path, temp_path());
			Zstrcat(path, "open_dir");
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis

		testing("Call to open() when named file is a directory and O_WRONLY");
		testing("flag is set.");
		start
			expecting(-1);
			expecting(EISDIR);
			Zopen(path, O_WRONLY);
		finis

		testing("Call to open() when named file is a directory and O_RDWR");
		testing("flag is set.");
		start
			expecting(-1);
			expecting(EISDIR);
			Zopen(path, O_RDWR);
		finis

		testing("Time related elements \"st_ctime\", \"st_mtime\", and");
		testing("\"st_atime\" of the file and \"st_ctime\" and");
		testing("\"st_mtime\" of the parent directory are not changed.");
		start
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);

			strcat(path, "/..");
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	assertion(55, BASE);
	start
		int		open_max, topen_max, i;

		break_on(ERROR);

		prep("Initialize a path name.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "fill_up");
		finis

		open_max = g_open_max();	/* Get value for {OPEN_MAX} */
									/* Get testable value for {OPEN_MAX} */
		topen_max = (open_max <= PCTS_OPEN_MAX) ? open_max : PCTS_OPEN_MAX;

		prep("Close all file descriptors except 0, 1 and 2.");
		start
			for(i=3; i<topen_max; i++)
				(void) close(i);
		finis

		continue_on(ERROR);			/* Close all open file descriptors. */
		if(open_max <= PCTS_OPEN_MAX) {
			prep("{OPEN_MAX} equal to {PCTS_OPEN_MAX}.");
			prep("Consume all available file descriptors.");
			start
				for(i=3; i<open_max; i++)
					fd = Zopen(path, O_RDWR | O_TRUNC | O_CREAT, PROT_ALL);
			finis

			testing("Call to open() when {OPEN_MAX} file descriptors in use");
			testing("O_TRUNC flag is set, and file exists.");
			testing("File is not truncated.");
			start
				Zwrite(fd, String, Strlen);			/* Write data to file. */
				Zlseek(fd, (off_t)0, SEEK_SET);

				expecting(-1);
				expecting(EMFILE);
				Zopen(path, O_RDWR | O_TRUNC, PROT_ALL);

				expecting(Strlen);
				Zread(fd, inbuf, Strlen);
			finis

			testing("Call to open() when {OPEN_MAX} file descriptors in use,");
			testing("O_CREAT flag is set, and file does not exist.");
			testing("File is not created.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, "emfile_tst");

				expecting(-1);
				expecting(EMFILE);
				Zopen(path, O_WRONLY | O_CREAT, PROT_ALL);

				expecting(-1);
				expecting(ENOENT);
				Zaccess(path, F_OK);
			finis
		}
		else {
			prep("Testing for {PCTS_OPEN_MAX} < {OPEN_MAX}.");
			prep("Consume PCTS_OPEN_MAX -1 file descriptors."); 
			start
				for(i=3; i<(topen_max -1); i++)
					(void) Zcreat(path, PROT_ALL);
			finis

			prep("Initialize a path name.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, "emfile_tst");
			finis
	
			testing("open() on the {PCTS_OPEN_MAX} -1 file descriptor.");
			start
				expecting(SUCCESS);
				(void) Zopen(path, O_WRONLY | O_CREAT, PROT_ALL);
			finis
		}

		cleanup("Close all file descriptors except 0, 1 and 2.");
		start
			for(i=3; i<topen_max; i++)
				(void) close(i);
		finis
	finis

	if(tc_name_max()) {
		assertion(56, BASE);
		start
			e_NO_TRUNC(E_OPEN);
		finis
	}
	else
		assertion(56, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(57, BASE);
		start
			e_NO_TRUNC(E_OPEN);
		finis
	}
	else
		assertion(57, NOT_TESTABLE);

	assertion(58, BASE);
	start
		e_ENAMETOOLONG(E_OPEN);
	finis

	assertion(59, EXTENDED);

	assertion(60, BASE);
	start
		testing("Call to open() when O_CREAT is not set and the file does");
		testing("not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_open_");

			expecting(-1);
			expecting(ENOENT);
			Zopen(path, O_WRONLY, PROT_ALL);
		finis
	finis

	assertion(61, BASE);
	start
		testing("Call to open() when path prefix does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_open_/test");

			expecting(-1);
			expecting(ENOENT);
			Zopen(path, O_WRONLY | O_CREAT, PROT_ALL);
		finis

		testing("Call to open() when path points to an empty string.");
		start
			expecting(-1);
			expecting(ENOENT);
			Zopen("", O_WRONLY | O_CREAT, PROT_ALL);
		finis
	finis

	assertion(62, EXTENDED);

	assertion(63, BASE);
	start
		testing("Call to open() when pathname contains a non-directory");
		testing("component.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "Open_Test");
			Zstrcat(path, "/dummy");

			expecting(-1);
			expecting(ENOTDIR);
			Zopen(path, O_RDONLY);
		finis
	finis

	assertion(64, BASE);
	start
		prep("Create a FIFO file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "er_enxio");
			Zmkfifo(path, PROT_ALL);
		finis

		testing("Call to open() on a FIFO file with O_NONBLOCK and O_WRONLY");
		testing("set and no process has the file opened for reading.");
		start
			expecting(-1);
			expecting(ENXIO);
			Zopen(path, O_WRONLY | O_NONBLOCK);
		finis
	finis

	done();
}
