#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define	FILENAME	"5.7.1a_42"
#define	EINVAL_PATHCONF_NAME	(_PC_LINK_MAX +_PC_MAX_CANON +_PC_MAX_INPUT \
								+_PC_NAME_MAX +_PC_PATH_MAX +_PC_PIPE_BUF \
								+_PC_CHOWN_RESTRICTED +_PC_NO_TRUNC \
								+_PC_VDISABLE)
void	f_ENAMETOOLONG(void);
void	f_NO_TRUNC(void);


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	long	r_success, r_test;

	setup(E_PATHCONF);
	funct(E_PATHCONF);

	assertion(42, BASE);
	start
		prep("Create a test file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
		finis

		testing("Call to pathconf() with an invalid value for name.");
		start
			Zstrcpy(path, temp_path());

			expecting(-1);
			expecting(EINVAL);
			Zpathconf(path, EINVAL_PATHCONF_NAME);
		finis
	finis

	assertion(43, BASE);
	start
		testing("Call to pathconf() when search permission is denied on path.");
		start
			Zstrcpy(path, data_path());
			r_success = Zpathconf(path, _PC_NAME_MAX);

			sprintf(path, "%s%s/%s", data_path(), "no_srch.d", "_tryme_.d");

			errno = 0;
			r_test = pathconf(path, _PC_NAME_MAX);
			if(r_test == (long) -1) {
				if(errno != EACCES)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EACCES].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(44, BASE);
	start
		testing("Call to pathconf(path, _PC_MAX_CANON) when path does");
		testing("not refer to a terminal file.");
		start
			Zstrcpy(path, data_path());

			errno = 0;
			r_test = pathconf(path, _PC_MAX_CANON);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else {
				r_success = Zpathconf(FILENAME_OF_TERMINAL_FILE, _PC_MAX_CANON);
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
			}
		finis
	finis

	assertion(45, BASE);
	start
		testing("Call to pathconf(path, _PC_MAX_INPUT) when path does");
		testing("not refer to a terminal file.");
		start
			Zstrcpy(path, data_path());

			errno = 0;
			r_test = pathconf(path, _PC_MAX_INPUT);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else {
				r_success = Zpathconf(FILENAME_OF_TERMINAL_FILE, _PC_MAX_INPUT);
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
			}
		finis
	finis

	assertion(46, BASE);
	start
		testing("Call to pathconf(path, _PC_NAME_MAX) when path does");
		testing("not refer to a directory.");
		start
			Zstrcpy(path, data_path());
			r_success = Zpathconf(path, _PC_NAME_MAX);

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			errno = 0;
			r_test = pathconf(path, _PC_NAME_MAX);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(47, BASE);
	start
		testing("Call to pathconf(path, _PC_PATH_MAX) when path does");
		testing("not refer to a directory.");
		start
			Zstrcpy(path, data_path());
			r_success = Zpathconf(path, _PC_PATH_MAX);

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			errno = 0;
			r_test = pathconf(path, _PC_PATH_MAX);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis


	assertion(48, BASE);
	start
		testing("Call to pathconf(path, _PC_PIPE_BUF) when path does");
		testing("not refer to a FIFO or directory.");
		start
			Zstrcpy(path, data_path());
			r_success = Zpathconf(path, _PC_PIPE_BUF);

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);			/* Path to a regular file. */
			errno = 0;
			r_test = pathconf(path, _PC_PIPE_BUF);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(49, BASE);
	start
		testing("Call to pathconf(path, _PC_NO_TRUNC) when path does");
		testing("not refer to a directory.");
		start
			Zstrcpy(path, data_path());
			r_success = Zpathconf(path, _PC_NO_TRUNC);

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);			/* Path to a regular file. */
			errno = 0;
			r_test = pathconf(path, _PC_NO_TRUNC);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(50, BASE);
	start
		testing("Call to pathconf(path, _PC_VDISABLE) when path does");
		testing("not refer to a terminal file.");
		start
			Zstrcpy(path, data_path());

			errno = 0;
			r_test = pathconf(path, _PC_VDISABLE);
			if(r_test == (long) -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [EINVAL].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else {
				r_success = Zpathconf(FILENAME_OF_TERMINAL_FILE, _PC_VDISABLE);
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
			}
		finis
	finis

	if(tc_name_max()) {
		assertion(51, BASE);
		start
			f_NO_TRUNC();
		finis
	}
	else
		assertion(51, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(52, BASE);
		start
			f_NO_TRUNC();
		finis
	}
	else
		assertion(52, NOT_TESTABLE);

	assertion(53, BASE);
	start
		f_ENAMETOOLONG();
	finis

	assertion(54, BASE);
	start
		testing("Call to pathconf() when path does not exist.");
		start
			Zstrcpy(path, temp_path());
			r_success = Zpathconf(path, _PC_NO_TRUNC);
			Zstrcat(path, "_tryme_");

			errno = 0;
			r_test = pathconf(path, _PC_NO_TRUNC);
			if(r_test == (long) -1) {
				if(errno != ENOENT)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [ENOENT].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(55, BASE);
	start
		testing("Call to pathconf() when component of the path prefix of path");
		testing("does not exist.");
		start
			Zstrcpy(path, temp_path());
			r_success = Zpathconf(path, _PC_NO_TRUNC);
			Zstrcat(path, "_tryme_/tryme");

			errno = 0;
			r_test = pathconf(path, _PC_NO_TRUNC);
			if(r_test == (long) -1) {
				if(errno != ENOENT)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [ENOENT].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(56, BASE);
	start
		testing("Call to pathconf() when path points to an empty string.");
		start
			Zstrcpy(path, data_path());
			r_success = Zpathconf(path, _PC_NO_TRUNC);

			errno = 0;
			r_test = pathconf("", _PC_NO_TRUNC);
			if(r_test == (long) -1) {
				if(errno != ENOENT)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [ENOENT].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	assertion(57, BASE);
	start
		testing("Call to pathconf() when component of the path prefix is not");
		testing("a directory");
		start
			Zstrcpy(path, temp_path());
			r_success = Zpathconf(path, _PC_NO_TRUNC);
			Zstrcat(path, FILENAME);
			Zstrcat(path, "/");
			Zstrcat(path, FILENAME);

			errno = 0;
			r_test = pathconf(path, _PC_NO_TRUNC);
			if(r_test == (long) -1) {
				if(errno != ENOTDIR)
					error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [ENOTDIR].",
							"Actual errno value reported was ",
							err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	finis

	done();
}


void f_NO_TRUNC(void)
{
	char	*lpath;
	long	r_test, r_success;
	char	path[_POSIX_PATH_MAX +1];

	Zstrcpy(path, temp_path());
	r_success = pathconf(path, _PC_NO_TRUNC);

	if(r_success != (long) -1 )  {
		lpath = g_lname();

		if(tc_name_max() == TRUE) {
			testing("Path component longer than {NAME_MAX}.");
			start
				errno = 0;
				r_test = pathconf(path, _PC_NO_TRUNC);
				if(r_test == (long) -1) {
					if(errno != ENAMETOOLONG)
						error("%s\n%s\n%s%s.",
							"Value returned was -1.",
							"Expected errno value of [ENAMETOOLONG].",
							"Actual errno value reported was ",
							err_lookup(errno) );
				}
				else
					if(r_test != r_success)
						error("%s%ld.\n%s%ld.",
								"Value returned was ",
								r_test,
								"Expected value of ",
								r_success);
			finis
		}
		else {
			testing("Path component equal to {PCTS_NAME_MAX}.");
			start
				expecting(SUCCESS);
				r_test = Zpathconf(path, _PC_NO_TRUNC);
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
			finis
		}

		cleanup("Remove buffer space allocated for long path.");
		start
			Zfree(lpath);
		finis
	}
	else
		error("{_POSIX_NO_TRUNC} not supported.");
}

void f_ENAMETOOLONG(void)
{
	char	*lpath;
	int		test_pm;
	long	r_success, r_test;
	char	path[_POSIX_PATH_MAX +1];

	Zstrcpy(path, temp_path());
	r_success = pathconf(path, _PC_NO_TRUNC);

	if(test_pm = tc_path_max()) {
		prep("Establish pathname of {PATH_MAX} +2 bytes.");
		start
			lpath = g_lpath();
		finis
	}
	else {
		prep("Establish pathname of {PCTS_PATH_MAX} +2 bytes.");
		start
			char	*ptr, *ptr_end;

					/* Allocate buffer for {PCTS_PATH_MAX} +1 + "/X" */
			lpath = (char *) Zmalloc(PCTS_PATH_MAX +3);

			Zstrcpy(lpath, temp_path());
			Zstrcat(lpath, "ENAMETOOLONG");
			if(access(lpath, F_OK))
				Zmkdir(lpath, PROT_ALL);
			ptr = lpath +Zstrlen(lpath);

			ptr_end = PCTS_PATH_MAX +lpath;
			for( ; ptr +16 < ptr_end; ptr += 16 )
        		(void) strcpy(ptr, "/../ENAMETOOLONG");

			for( ; ptr < ptr_end; ptr += 2 )
        			(void) strcpy(ptr, "/.");

							/* End of path may be "/.//X" "or /./X". */
        	(void) strcpy(ptr_end, "/X");
		finis
	}

	if(test_pm = tc_path_max()) {
		testing("For path exceeds {PATH_MAX}.");
		start
			errno = 0;
			r_test = pathconf(path, _PC_NO_TRUNC);
			if(r_test == (long) -1) {
				if(errno != ENAMETOOLONG)
					error("%s\n%s\n%s%s.",
						"Value returned was -1.",
						"Expected errno value of [ENAMETOOLONG].",
						"Actual errno value reported was ",
						err_lookup(errno) );
			}
			else
				if(r_test != r_success)
					error("%s%ld.\n%s%ld.",
							"Value returned was ",
							r_test,
							"Expected value of ",
							r_success);
		finis
	}
	else {
		testing("For path length of {PCTS_PATH_MAX}.");
		start
			expecting(SUCCESS);
			r_test = Zpathconf(path, _PC_NO_TRUNC);
			if(r_test != r_success)
				error("%s%ld.\n%s%ld.",
						"Value returned was ",
						r_test,
						"Expected value of ",
						r_success);
		finis
	}

	cleanup("Remove buffer space allocated for long path.");
	start
		Zfree(lpath);
	finis
}
