#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	DIRNAME			"5.5.3_33.d"
#define	DIRNAME_OLD		"5.5.3_33.do"
#define	DIRNAME_NEW		"5.5.3_33.dn"
#define	FILENAME_OLD	"5.5.3_33"
#define	FILENAME_NEW	"5.5.3_33_n"
#define	String			"DaTa for link testing"
#define	Strlen			(sizeof(String) -1)


int main(void)
{
	struct	stat st_buf1, st_buf2;
	char	path_new[_POSIX_PATH_MAX +1];
	char	inbuf[Strlen +1];
	int		fd;
	pid_t	pid;

	setup(E_RENAME);
	funct(E_RENAME);

	assertion(33, BASE);
	start
		prep("Create old and new files.");
		start
			Zchdir("../tmp/rename.d");
			Zclose(Zcreat(FILENAME_OLD, PROT_ALL));
			Zclose(Zcreat(FILENAME_NEW, PROT_ALL));
		finis

		testing("Call to rename(old, new) removes old pathname");
		testing("and returns a value of zero.");
		start
			expecting(0);
			Zrename(FILENAME_OLD, FILENAME_NEW);
			expecting(ENOENT);
			Zaccess(FILENAME_OLD, F_OK);
		finis
	finis

	assertion(34, BASE);
	start
		testing("When old and new both refer to links to the same existing");
		testing("file, then a call to rename(old, new) returns successfully");
		testing("without taking any other action.");
		start
			Zlink(FILENAME_NEW, FILENAME_OLD);

			expecting(SUCCESS);
			Zrename(FILENAME_OLD, FILENAME_NEW);

			Zaccess(FILENAME_OLD, F_OK);
			Zaccess(FILENAME_NEW, F_OK);
		finis

		cleanup("Remove new.");
		start
			unlink(FILENAME_NEW);
		finis
	finis

	assertion(35, EXTENDED);

	assertion(36, BASE);
	start
		testing("When the link named by the new argument exists, then a call");
		testing("to rename(old, new) removes new and renames old to new.");
		start
			Zclose(Zcreat(FILENAME_NEW, PROT_ALL));
			Zstat(FILENAME_OLD, &st_buf1);

			Zrename(FILENAME_OLD, FILENAME_NEW);
			Zstat(FILENAME_NEW, &st_buf2);
			if(st_buf1.st_ino != st_buf2.st_ino) 
				error("%s\n%s%lu.\n%s%lu.\n",
						"File inode not expected inode.",
						"Expected inode of file to be: ",
						(unsigned long) st_buf1.st_ino,
						"Actual inode of file is: ",
						(unsigned long) st_buf2.st_ino);

			Zunlink(FILENAME_NEW);					/* Cleanup */
		finis
	finis

	assertion(37, BASE);
	start
		testing("When the directory named by the new argument exists and is");
		testing("empty and old points to a directory, then a call");
		testing("to rename(old, new) removes new and renames old to new.");
		start
			Zmkdir(DIRNAME_NEW, S_IRWXU);
			Zmkdir(DIRNAME_OLD, S_IRWXU);
			Zstat(DIRNAME_OLD, &st_buf1);

			Zrename(DIRNAME_OLD, DIRNAME_NEW);
			Zstat(DIRNAME_NEW, &st_buf2);
			if(st_buf1.st_ino != st_buf2.st_ino) 
				error("%s\n%s%lu.\n%s%lu.\n",
						"File inode not expected inode.",
						"Expected inode of file to be: ",
						(unsigned long) st_buf1.st_ino,
						"Actual inode of file is: ",
						(unsigned long) st_buf2.st_ino);

			Zrmdir(DIRNAME_NEW);					/* Cleanup */
		finis
	finis

	assertion(38, EXTENDED);

	assertion(39, BASE);
	start
		prep("Establish test files.");
		start
														/* Cleanup */
			unlink(FILENAME_OLD);
			unlink(FILENAME_NEW);

			Zclose(Zcreat(FILENAME_OLD, PROT_ALL));
			fd = Zcreat(FILENAME_NEW, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd, String, Strlen);
			Zclose(fd);
		finis

		ENABLE_PSYNC
		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Open new.");
			start
				fd = Zopen(FILENAME_NEW, O_RDONLY);
				SEND_PSYNC
				SET_TIMEOUT
					WAIT_PSYNCA
				CLEAR_ALARM
			finis

			testing("Link to new has been removed but data is");
			testing("still accessible .");
			start
				Zlseek(fd, (off_t)0, SEEK_SET);
				expecting(Strlen);
				Zread(fd, inbuf, Strlen);
			finis

			done();
		}

		testing("Call to rename() removes link pointed to by new.");
		start
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
			Zrename(FILENAME_OLD, FILENAME_NEW);

			expecting(ENOENT);
			Zaccess(FILENAME_OLD, F_OK);

			SEND_PSYNCA
		finis

		WAIT_FOR_PID(pid);

		testing("Contents of file new have been removed.");
		start
			fd = Zopen(FILENAME_NEW, O_RDONLY);
			expecting(0);
			Zread(fd, inbuf, Strlen);
			Zclose(fd);

			Zunlink(FILENAME_NEW);						/* Cleanup */
		finis

		DISABLE_PSYNC
	finis

	assertion(40, BASE);
	start
		if(PCD_WRITE_PERM_TO_RENAME == FALSE) {
			testing("When write permission is denied for a directory pointed");
			testing("to by the old or new arguments, then a call to");
			testing("rename(old, new) completes successfully.");
			start
				rmdir(DIRNAME_NEW);
				rmdir(DIRNAME_OLD);
				Zmkdir(DIRNAME_NEW, S_IRUSR | S_IXUSR);
				Zmkdir(DIRNAME_OLD, S_IRUSR | S_IXUSR);

				Zrename(DIRNAME_OLD, DIRNAME_NEW);
				expecting(ENOENT);
				Zaccess(DIRNAME_OLD, F_OK);
			finis
		}
		else
			not_implemented("{PCD_WRITE_PERM_TO_RENAME} of FALSE");
	finis

	assertion(41, EXTENDED);

	assertion(43, BASE);
	start
		time_t	s_time;

		prep("Obtain a starting time, then call rename().");
		start
			Zclose(Zcreat(FILENAME_OLD, PROT_ALL));
			Zmkdir(DIRNAME, PROT_ALL);
			sprintf(path_new, "%s/%s", DIRNAME, FILENAME_NEW);

			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/

			Zrename(FILENAME_OLD, path_new);
		finis
	
		testing("Time related elements \"st_ctime\" and \"st_mtime\"");
		testing("of parent directory of each file are marked for update.");
		start
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, DIRNAME);
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, DIRNAME);

			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, "../rename.d");
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, "../rename.d");
		finis
	finis


	done();
}
