#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>
#include	<Zdirent.h>

#include	<sv_env.h>
#include	<sv_errno.h>

#define		DIRNAME	"5.5.2_19"


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	DIR		*dirp;
	struct	dirent	*dirent;

	setup(E_RMDIR);
	funct(E_RMDIR);

	assertion(19, BASE);
	start
		prep("Create a directory.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, DIRNAME);
			Zmkdir(path, PROT_ALL);
			Zaccess(path, F_OK);
		finis

		testing("A call to rmdir() removes directory and returns value zero.");
		start
			expecting(0);
			Zrmdir(path);

			expecting(ENOENT);
			Zaccess(path, F_OK);
		finis
	finis

	assertion(20, EXTENDED);

	assertion(21, BASE);
	start
		DIR	*dirp;

		prep("Create test directory and open it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, DIRNAME);
			Zmkdir(path, PROT_ALL);
	
			dirp = Zopendir(path);
		finis
	
		prep("Remove last link to directory.");
		start
			Zrmdir(path);
		finis
	
		testing("No new entry can be created in directory.");
		start
			Zstrcat(path, "/A");
			expecting(FAILURE);
			Zcreat(path, PROT_ALL);
		finis
	
		testing("Dot and dot-dot entry do not exist in directory.");
		start
			Zrewinddir(dirp);			/*	Get current state of directory. */
			while((dirent = readdir(dirp)) != NULL) {
				if(!strcmp(dirent->d_name, "."))
					error("Entry for dot detected.");
				if(!strcmp(dirent->d_name, ".."))
					error("Entry for dot-dot detected.");
			}
		finis
	finis

	assertion(22, EXTENDED_TESTED);
	start
		prep("Create a test directory and open it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, DIRNAME);
			Zmkdir(path, PROT_ALL);
	
			dirp = Zopendir(path);
		finis
	
		prep("Remove last link to directory.");
		start
			Zrmdir(path);
		finis
	
		testing("Directory still available.");
		start
			Zrewinddir(dirp);			/*	Get current state of directory. */
			errno = 0;
							/*	Next read returns end-of-directory indicator
								and errno is not changed.
							*/
			expecting((struct dirent *) 0);
			dirent = Zreaddir(dirp);
			if(errno != 0)
				error("%s\n%s\n%s%s\n",
					  "Value of errno returned incorrect.",
					  "Expected: no change from value set.",
					  "Actual:	",
					  err_lookup(errno));
		finis

		prep("Close last reference to directory stream.");
		start
			Zclosedir(dirp);
		finis

		testing("Directory no longer available.");
		start
			if(EBADF_READDIR) {
				expecting(EBADF);
				dirent = Zreaddir(dirp);
			}
			else {
				expecting(FAILURE);
				dirent = Zreaddir(dirp);
				message("Call to readdir() set errno to %s.\n",
						err_lookup(errno));
			}
		finis
	finis

	assertion(23, BASE);
	start
		time_t	s_time;

		prep("Record a starting time, then remove a directory.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, DIRNAME);
			Zmkdir(path, PROT_ALL);

			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			Zrmdir(path);
		finis
	
		testing("Time related elements \"st_ctime\" and \"st_mtime\"");
		testing("of parent directory are marked for update.");
		start
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, temp_path());
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, temp_path());
		finis
	finis

	assertion(24, EXTENDED);

	done();
}
