#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<sys/Zstat.h>

#include	<sv_env.h>

#define	T_DIR	0
#define	T_CHR	1
#define	T_BLK	2
#define	T_REG	3
#define	T_FIFO	4

void	ck_macro(int, int, char *);


int main(void)
{
	char	r_path[_POSIX_PATH_MAX +1];
	char	d_path[_POSIX_PATH_MAX +1];
	char	f_path[_POSIX_PATH_MAX +1];
	char	b_path[_POSIX_PATH_MAX +1];
	char	c_path[_POSIX_PATH_MAX +1];

	setup(E_STAT_H);
	funct(E_STAT_H);

	assertion(3, BASE);
	start
		prep("Establish regular file for testing macros.");
		start
			Zstrcpy(r_path, temp_path());
			Zstrcat(r_path, "reg_type");
			Zclose(Zcreat(r_path, PROT_ALL));
		finis

		prep("Establish directory for testing macros.");
		start
			Zstrcpy(d_path, temp_path());
			Zstrcat(d_path, "dir_type");
			Zmkdir(d_path, PROT_ALL);
		finis

		prep("Establish FIFO file for testing macros.");
		start
			Zstrcpy(f_path, temp_path());
			Zstrcat(f_path, "fifo_type");
			Zmkfifo(f_path, PROT_ALL);
		finis

		if(BLOCK_SPECIAL == 1) {
			prep("Establish path for block special file.");
			start
				Zstrcpy(b_path, temp_path());
				Zstrcat(b_path, "../../../../");
				Zstrcat(b_path, BLOCK_SPECIAL_PATH);
			finis
		}

		if(CHAR_SPECIAL == 1) {
			prep("Establish path for character special file.");
			start
				Zstrcpy(c_path, temp_path());
				Zstrcat(c_path, "../../../../");
				Zstrcat(c_path, CHAR_SPECIAL_PATH);
			finis
		}

		testing("S_ISDIR macro on directory file type returns TRUE.");
		start
			ck_macro(TRUE, T_DIR, d_path);
		finis

		testing("S_ISDIR macro on non-directory file types return FALSE.");
		start
			ck_macro(FALSE, T_DIR, r_path);
			ck_macro(FALSE, T_DIR, f_path);
			if(BLOCK_SPECIAL == 1)
				ck_macro(FALSE, T_DIR, b_path);
			if(CHAR_SPECIAL == 1)
				ck_macro(FALSE, T_DIR, c_path);
		finis
	finis

	assertion(4, BASE);
	start
		if(CHAR_SPECIAL == 1) {
			testing("S_ISCHR macro on character special file type returns");
			testing("TRUE.");
			start
				ck_macro(TRUE, T_CHR, c_path);
			finis
	
			testing("S_ISCHR macro on non-character special file types");
			testing("return FALSE.");
			start
				ck_macro(FALSE, T_CHR, d_path);
				ck_macro(FALSE, T_CHR, r_path);
				ck_macro(FALSE, T_CHR, f_path);
				if(BLOCK_SPECIAL == 1)
					ck_macro(FALSE, T_CHR, b_path);
			finis
		}
		else
			not_implemented("character special files");
	finis

	assertion(5, BASE);
	start
		if(BLOCK_SPECIAL == 1) {
			testing("S_ISBLK macro on block special file type returns TRUE.");
			start
				ck_macro(TRUE, T_BLK, b_path);
			finis

			testing("S_ISBLK macro on non-block special file types return");
			testing("FALSE.");
			start
				ck_macro(FALSE, T_BLK, d_path);
				ck_macro(FALSE, T_BLK, r_path);
				if(CHAR_SPECIAL == 1)
					ck_macro(FALSE, T_BLK, c_path);
				ck_macro(FALSE, T_BLK, f_path);
			finis
		}
		else
			not_implemented("block special files");
	finis

	assertion(6, BASE);
	start
		testing("S_ISREG macro on regular file type returns TRUE.");
		start
			ck_macro(TRUE, T_REG, r_path);
		finis

		testing("S_ISREG macro on non-regular file types return FALSE.");
		start
			ck_macro(FALSE, T_REG, d_path);
			ck_macro(FALSE, T_REG, f_path);
			if(BLOCK_SPECIAL == 1)
				ck_macro(FALSE, T_REG, b_path);
			if(CHAR_SPECIAL == 1)
				ck_macro(FALSE, T_REG, c_path);
		finis
	finis

	assertion(7, BASE);
	start
		testing("S_ISFIFO macro on FIFO file type returns TRUE.");
		start
			ck_macro(TRUE, T_FIFO, f_path);
		finis

		testing("S_ISFIFO macro on non-FIFO file types return FALSE.");
		start
			ck_macro(FALSE, T_FIFO, d_path);
			ck_macro(FALSE, T_FIFO, r_path);
			if(BLOCK_SPECIAL == 1)
				ck_macro(FALSE, T_FIFO, b_path);
			if(CHAR_SPECIAL == 1)
				ck_macro(FALSE, T_FIFO, c_path);
		finis
	finis

	done();
}
			


void	ck_macro(int exp, int t_macro, char *path)
{
	struct	stat	st_buf;
	int	ret;

	Zstat(path, &st_buf);
	switch(t_macro) {
		case T_DIR:
		if((ret = S_ISDIR(st_buf.st_mode)) && exp)
			break;
		if(ret == exp)
			break;
				error("S_ISDIR for path \"%s\" returned %d\n%s\n",
				path,
				ret,
					(exp) ? "Expected non-zero value." :
					"Expected 0");
		break;

		case T_CHR:
		if((ret = S_ISCHR(st_buf.st_mode)) && exp)
			break;
		if(ret == exp)
			break;
				error("S_ISCHR for path \"%s\" returned %d\n%s\n",
				path,
				ret,
					(exp) ? "Expected non-zero value." :
					"Expected 0");
		break;

		case T_BLK:
		if((ret = S_ISBLK(st_buf.st_mode)) && exp)
			break;
		if(ret == exp)
			break;
				error("S_ISBLK for path \"%s\" returned %d\n%s\n",
				path,
				ret,
					(exp) ? "Expected non-zero value." :
					"Expected 0");
		break;

		case T_REG:
		if((ret = S_ISREG(st_buf.st_mode)) && exp)
			break;
		if(ret == exp)
			break;
				error("S_ISREG for path \"%s\" returned %d\n%s\n",
				path,
				ret,
					(exp) ? "Expected non-zero value." :
					"Expected 0");
		break;

		case T_FIFO:
		if((ret = S_ISFIFO(st_buf.st_mode)) && exp)
			break;
		if(ret == exp)
			break;
				error("S_ISFIFO for path \"%s\" returned %d\n%s\n",
				path,
				ret,
					(exp) ? "Expected non-zero value." :
					"Expected 0");
		break;

	}
}
