#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

void ck_stat(char *, mode_t);

/*	Mode values for POSIX	*/ 
mode_t p_mode_o[] = {
	0,
	S_IXOTH,
	S_IXOTH | S_IWOTH,
	S_IXOTH | S_IROTH,
	S_IXOTH | S_IWOTH | S_IROTH,
	S_IWOTH,
	S_IWOTH | S_IROTH,
	S_IROTH,
};
#define	N_OMODES	8

mode_t p_mode_g[] = {
	0,
	S_IXGRP,
	S_IXGRP | S_IWGRP,
	S_IXGRP | S_IRGRP,
	S_IXGRP | S_IWGRP | S_IRGRP,
	S_IWGRP,
	S_IWGRP | S_IRGRP,
	S_IRGRP,
};
#define	N_GMODES	8

mode_t p_mode_u[] = {
	0,
	S_IXUSR,
	S_IXUSR | S_IWUSR,
	S_IXUSR | S_IRUSR,
	S_IXUSR | S_IWUSR | S_IRUSR,
	S_IWUSR,
	S_IWUSR | S_IRUSR,
	S_IRUSR,
};
#define	N_UMODES	8

#define	FILENAME	"5.3.3_04"


int main(void)
{
	mode_t	prev_mask;
	char	path[_POSIX_PATH_MAX +1];

	setup(E_UMASK);
	funct(E_UMASK);

	assertion(4, BASE);
	start
		prep("Set umask to an initial value.");
		start
			umask(0);
			prev_mask = 0;
		finis

		testing("Cycle through the masks.");
		start
			mode_t	p_mask;
			mode_t	*p_m_o, *p_m_g, *p_m_u;
			int	m_u, m_g, m_o;

			break_on(ERROR);
			for(m_u = 0, p_m_u = &p_mode_u[0];
					m_u < N_UMODES; m_u++, p_m_u++)
			for (m_g = 0, p_m_g = &p_mode_g[0];
					m_g < N_GMODES; m_g++, p_m_g++)
			for (m_o = 0, p_m_o = &p_mode_o[0];
					m_o < N_OMODES; m_o++, p_m_o++) {
				p_mask = *p_m_u | *p_m_g | *p_m_o;
				expecting(prev_mask);
				umask(p_mask);
				prev_mask = p_mask;
			}
		finis
	finis

	assertion(5, BASE);
	start
		prep("Set umask() to allow all file permission bits.");
		start
			umask((mode_t)0);

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
		finis

		testing("A call to open(), creat(), mkdir(), and mkfifo()");
		testing("will set all permission bits.");
		start
			Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
			ck_stat(path, PROT_ALL);
			Zunlink(path);

			Zcreat(path, PROT_ALL);
			ck_stat(path, PROT_ALL);
			Zunlink(path);

			Zmkdir(path, PROT_ALL);
			ck_stat(path, PROT_ALL);
			Zrmdir(path);

			Zmkfifo(path, PROT_ALL);
			ck_stat(path, PROT_ALL);
			Zunlink(path);
		finis

		prep("Set umask() to inhibit all file permission bits.");
		start
			umask(PROT_ALL);
		finis

		testing("A call to open(), creat(), mkdir(), and mkfifo() will");
		testing("not set any permission bits.");
		start
			Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
			ck_stat(path, (mode_t) 0);
			Zunlink(path);

			Zcreat(path, PROT_ALL);
			ck_stat(path, (mode_t) 0);
			Zunlink(path);

			Zmkdir(path, PROT_ALL);
			ck_stat(path, (mode_t) 0);
			Zrmdir(path);

			Zmkfifo(path, PROT_ALL);
			ck_stat(path, (mode_t) 0);
			Zunlink(path);
		finis
	finis

	assertion(6, BASE);
	start
		prep("Set umask() to inhibit all file permission bits and test it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);

			umask(PROT_ALL);
			Zcreat(path, PROT_ALL);
			ck_stat(path, (mode_t) 0);
			Zunlink(path);
		finis

		prep("Set umask() to allow all file permission bits and test it.");
		start
			prev_mask = umask((mode_t)0);
			open(path, O_RDWR | O_CREAT, PROT_ALL);
			ck_stat(path, PROT_ALL);
			Zunlink(path);
		finis

		testing("When cmask is the return value from a previous call to");
		testing("umask(), then a call to umask(cmask) restores the process's");
		testing("file mode creation mask to the same value as that prior to");
		testing("the previous call to umask().");
		start
			umask(prev_mask);
			Zcreat(path, PROT_ALL);
			ck_stat(path, (mode_t) 0);
			Zunlink(path);
		finis
	finis

	done();
}


/*	Report an ERROR if the mode of file specified by path
	is not identical to that specified by "prot".
*/

void ck_stat(char *path, mode_t prot)
{
	struct	stat buf;

	Zstat(path, &buf);
	if((buf.st_mode & (PROT_ALL)) != prot)
		error("%s\n%s%o\n%s%o",
				"Unexpected permission mode obtained.",
			 	"Expected: ",
			 	(unsigned long)prot,
			 	"Actual: ",
			 	(unsigned long)(buf.st_mode & PROT_ALL));
}
