#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<locks.h>

#define	 String		"What goes up must come down."
#define	 Strlen		(sizeof(String) -1)

#define 	TMP_FILE	"dup.tmp"

int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	int 	fd_org, fd_dup;
	struct	stat st_org, st_dup;
	struct	flock flock_org, flock_dup;

	setup(E_DUP);
	funct(E_DUP);

	assertion(4, BASE);
	start
		prep("Open a regular type file and set a lock.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, TMP_FILE);
			fd_org = Zopen(path, O_RDWR | O_CREAT | O_APPEND, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd_org, String, Strlen);

			SET_LOCK(F_WRLCK, fd_org, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
						&flock_org);
			Zfstat(fd_org, &st_org);
		finis

		prep("Close file descriptor 0.");
		start
			close(0);
		finis

		testing("Call to dup() returns a new file descriptor which is the");
		testing("lowest numbered available at the time of the dup() call.");
		start
			expecting(0);
			fd_dup = Zdup(fd_org);
		finis

		testing("New file description refers to the same open file");
		testing("description as the original file description.");
		start
			off_t	off_org, off_dup;
										/* Test file ofset. */
			if( (off_org = Zlseek(fd_org, (off_t)0, SEEK_CUR)) != 
						(off_dup = Zlseek(fd_dup, (off_t)0, SEEK_CUR)) )
				error("%s\n%s%lu\n%s%lu\n",
					"The file offsets differ from expected offset:",
					"Expecting:	",
					(unsigned long) off_org,
					"Actual offset for duped file descriptor:	",
					(unsigned long) off_dup);
		
			Zfstat(fd_dup, &st_dup);

										/* Test access mode. */
			if(st_org.st_mode != st_dup.st_mode)
				error("%s\n%s%lo\n%s%lo\n",
					"The file status modes differ:",
					"Expecting:	",
					(unsigned long)st_org.st_mode,
					"Actual:	",
					(unsigned long)st_dup.st_mode);

										/* Test status flags. */
			if( (Zfcntl(fd_org, F_GETFL) & O_ACCMODE) !=
								(Zfcntl(fd_dup, F_GETFL) & O_ACCMODE) )
				error("%s\n%s%o\n%s%o\n",
					"File status flags differ.",
					"Expecting:	",
					Zfcntl(fd_org, F_GETFL) & O_ACCMODE,
					"Actual:	",
					Zfcntl(fd_dup, F_GETFL) & O_ACCMODE );
		finis

		prep("Initialize lock structure.");
		start
			INIT_LOCK_STRUCT(F_WRLCK, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
												flock_dup);
		finis

		testing("File descriptors have the same locks.");
		start
			Zfcntl(fd_org, F_GETLK, &flock_org);
			Zfcntl(fd_dup, F_GETLK, &flock_dup);
			same_lock(&flock_org, &flock_dup);
		finis

		testing("The FD_CLOEXEC flag for the new file descriptor is clear.");
		start
			if((Zfcntl(fd_dup, F_GETFD) & FD_CLOEXEC) != 0)
				error("FD_CLOEXEC flag is set.");
		finis
	finis

	done();
}
