#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<locks.h>

#define	 String		"What goes up must come down."
#define	 Strlen		(sizeof(String) -1)

#define	FILENAME	"6.5.2_04"

int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	int 	fd, fd2;
	struct	stat st_org, st_dup;
	struct	flock flock_org, flock_dup;

	setup(E_FCNTL);
	funct(E_FCNTL);

	exit_on(ERROR);			

	assertion(4, BASE);
	start
		prep("Create a file to test with and set a lock on it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_RDWR | O_CREAT | O_APPEND, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd, String, Strlen);

			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
						&flock_org);
			Zfstat(fd, &st_org);
		finis

		prep("Ensure file descriptor after the prior one allocated is closed.");
		start
			close(fd +1);
		finis

		testing("Call to fcntl(fd, F_DUPFD, fd2) returns the lowest");
		testing("numbered file descriptor >= fd2.");
		start
			fd2 = Zfcntl(fd, F_DUPFD, fd +2);
			if(fd2 < (fd +2))
				error("%s\n%s%d.\n%s%d.\n",
					"A call to fcntl() returned the wrong file descriptor.",
					"Expected a file descriptor of value >= ",
					fd +2,
					"Actual file descriptor value is ",
					fd2);
		finis

		testing("New file description refers to the same open file");
		testing("description as the original file description.");
		start
			off_t	off_org, off_dupfd;
										/* Test file ofset. */
			if( (off_org = Zlseek(fd, (off_t)0, SEEK_CUR)) != 
						(off_dupfd = Zlseek(fd2, (off_t)0, SEEK_CUR)) )
				error("%s\n%s%lu\n%s%lu\n",
					"The file offsets differ from expected offset:",
					"Expecting:	",
					(unsigned long) off_org,
					"Actual offset for duplicated file descriptor:	",
					(unsigned long) off_dupfd);

			Zfstat(fd2, &st_dup);
										/* Test access mode. */
			if(st_org.st_mode != st_dup.st_mode)
				error("%s\n%s%lo\n%s%lo\n",
					"The file status modes differ:",
					"Expecting:	",
					(unsigned long)st_org.st_mode,
					"Actual:	",
					(unsigned long)st_dup.st_mode);

										/* Test status flags. */
			if( (Zfcntl(fd, F_GETFL) & O_ACCMODE) !=
								(Zfcntl(fd2, F_GETFL) & O_ACCMODE) )
				error("%s\n%s%o\n%s%o\n",
					"File status flags differ.",
					"Expecting:	",
					Zfcntl(fd, F_GETFL) & O_ACCMODE,
					"Actual:	",
					Zfcntl(fd2, F_GETFL) & O_ACCMODE );
		finis

		prep("Initialize lock structure.");
		start
			INIT_LOCK_STRUCT(F_WRLCK, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
												flock_dup);
		finis

		testing("File descriptors have the same locks.");
		start
			Zfcntl(fd, F_GETLK, &flock_org);
			Zfcntl(fd2, F_GETLK, &flock_dup);
			same_lock(&flock_org, &flock_dup);
		finis

		testing("The FD_CLOEXEC flag for the new file descriptor is clear.");
		start
			if((Zfcntl(fd2, F_GETFD) & FD_CLOEXEC) != 0)
				error("FD_CLOEXEC flag is set.");
		finis
	finis

	done();
}
