#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FILENAME	"6.5.2_08"
#define	FIFONAME	"6.5.2_08_fifo"


int main(void)
{
	int		pipefd[2];
	int		fd;
	int		fifowr, fiford;
	char	path[_POSIX_PATH_MAX +1], fifop[_POSIX_PATH_MAX +1];
	int		ret;
	pid_t	pid;

	setup(E_FCNTL);
	funct(E_FCNTL);

	exit_on(ERROR);

		/*	fcntl(fildes, F_GETFL) will return the file status flags");
			O_APPEND and O_NONBLOCK and file access modes O_RDONLY, ORDWR,
			O_WRONLY.");
		*/
	assertion(8, BASE);
	start
		prep("Create a regular file and a FIFO for testing.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));

			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

		FORK
			prep("Open temp file with O_RDONLY flag set.");
			start
				fd = Zopen(path, O_RDONLY);
			finis
		
			testing("Access mode O_RDONLY only set.");
			start
				ret = Zfcntl(fd, F_GETFL, 0);
				if( (ret & O_ACCMODE) != O_RDONLY)
					error("Expected access mode to be O_RDONLY (%d)\n%s%d\n",
						O_RDONLY,
						"Actual access mode is ",
						ret & O_ACCMODE);
			finis
		
			prep("Open a temp file for O_WRONLY and O_APPEND flags set.");
			start
				close(fd);
				fd = Zopen(path, O_WRONLY | O_APPEND);
			finis
		
			testing("Access mode O_WRONLY only set.");
			start
				ret = Zfcntl(fd, F_GETFL, 0);
				if( (ret & O_ACCMODE) != O_WRONLY)
					error("Expected access mode to be O_WRONLY (%d)\n%s%d\n",
						O_WRONLY,
						"Actual access mode is ",
						ret & O_ACCMODE);
			finis
		
			testing("O_APPEND mode is set.");
			start
				if( (ret & O_APPEND) != O_APPEND)
					error("O_APPEND mode not set.");
			finis
		
			prep("Open a temp file for O_RDWR flag set.");
			start
				close(fd);
				fd = Zopen(path, O_RDWR);
			finis
		
			testing("Access mode O_RDWR only set.");
			start
				ret = Zfcntl(fd, F_GETFL, 0);
				if( (ret & O_ACCMODE) != O_RDWR)
					error("Expected access mode to be O_RDWR (%d)\n%s%d\n",
						O_RDWR,
						"Actual access mode is ",
						ret & O_ACCMODE);
			finis
	
			prep("Open regular file with O_NONBLOCK flag set.");
			start 
				fd = Zopen(path, O_RDONLY | O_NONBLOCK);
			finis
		
			testing("O_NONBLOCK flag is set.");
			start
				if((fcntl(fd, F_GETFL, 0) & O_NONBLOCK) != O_NONBLOCK)
					error("O_NONBLOCK clear!");
			finis
		KROF
	finis

	assertion(9, BASE);
	start
		FORK
											/* Test O_APPEND cleared. */
			prep("Open regular file with O_APPEND flag set, then");
			prep("call fcntl(fd, F_SETFL, 0).");
			start 
				fd = Zopen(path, O_RDWR | O_APPEND|O_CREAT | O_TRUNC, PROT_ALL);
				ret = Zfcntl(fd, F_SETFL, 0);
			finis
		
			testing("O_APPEND flag is clear.");
			start
				char	rd_buf[10];

				Zwrite(fd, "Test", 4);
				Zlseek(fd, (off_t) 0, SEEK_SET);	/* Rewind file. */
				Zwrite(fd, "This", 4);
				Zlseek(fd, (off_t) 0, SEEK_SET);	/* Rewind file. */

				expecting(4);
				Zread(fd, rd_buf, 4);
				if(strncmp( "This", rd_buf, 4) )
						error("%s\"%s\"\n%s\"%s\"\n",
							"Read of test file returned:  ",
							rd_buf,
							"Expected:	",
							"This");
			finis

											/* Test O_NONBLOCK cleared. */
			prep("Open FIFO with O_NONBLOCK set on read channel, then");
			prep("call fcntl(rd_chan, F_SETFL, 0).");
			start 
				SET_ALARM(TIMEOUT)
					fiford = Zopen(fifop, O_RDONLY | O_NONBLOCK);
					fifowr = Zopen(fifop, O_WRONLY);
					Zfcntl(fiford, F_SETFL, 0);
				CLEAR_ALARM
			finis

			testing("Check that the O_NONBLOCK flag is clear.  A read()");
			testing("on pipe blocks waiting for data.");
			start
				char	data;

				SET_ALARM(2)
					read(fiford, &data, 1);
				EXPECT_ALARM
			finis
		KROF
	finis

	assertion(10, BASE);
	start
		FORK
			prep("Open FIFO, set read channel flag O_NONBLOCK with fcntl().");
			start 
				SET_ALARM(TIMEOUT)
					fiford = Zopen(fifop, O_RDONLY | O_NONBLOCK);
					fifowr = Zopen(fifop, O_WRONLY);
					Zclose(fiford);
					fiford = Zopen(fifop, O_RDONLY);
					Zfcntl(fiford, F_SETFL, O_NONBLOCK);
				CLEAR_ALARM
			finis

			testing("Check that the O_NONBLOCK flag is set.  A read()");
			testing("returns immediately if no data is available for reading.");
			start
				char	data;

				SET_ALARM(TIMEOUT)
					expecting(EAGAIN);
					Zread(fiford, &data, 1);
				CLEAR_ALARM
			finis
		KROF
	finis
	
	assertion(11, BASE);
	start
		FORK
			prep("Open the temp file, then set O_APPEND flag.");
			start
				fd = Zopen(path, O_RDWR | O_CREAT | O_TRUNC, PROT_ALL);
				Zfcntl(fd, F_SETFL, O_APPEND);
			finis

			testing("File opened with O_APPEND has the file offset set");
			testing("to end-of-file prior to each write.");
			start
				char	rd_buf[10];

				Zwrite(fd, "Test", 4);
				Zlseek(fd, (off_t) 0, SEEK_SET);	/* Rewind file. */
				Zwrite(fd, "This", 4);
				Zlseek(fd, (off_t) 0, SEEK_SET);	/* Rewind file. */

				expecting(8);
				Zread(fd, rd_buf, 8);
				if(strncmp("TestThis", rd_buf, 8) )
						error("%s\"%s\"\n%s\"%s\"\n",
							"Read of test file returned:  ",
							rd_buf,
							"Expected:	",
							"TestThis");
			finis
		KROF
	finis

	assertion(12, BASE);
	start
		FORK
			prep("Open a pipe and set the O_APPEND and O_NONBLOCK flags");
			prep("on the write channel.");
			start 
				Zpipe(pipefd);
				Zfcntl(pipefd[1], F_SETFL, O_APPEND | O_NONBLOCK);
			finis
	
			testing("O_APPEND flag is set");
			testing("Note: should perform a more dynamic test.");
			start
				if((Zfcntl(pipefd[1], F_GETFL, 0) & O_APPEND) != O_APPEND)
					error("O_APPEND not set!");
			finis
	
			testing("O_NONBLOCK flag is set");
			testing("Note: should perform a more dynamic test.");
			start
				if((Zfcntl(pipefd[1], F_GETFL, 0) & O_NONBLOCK) != O_NONBLOCK)
					error("O_NONBLOCK not set!");
			finis
		KROF
	finis

				/*	When the third argument arg contains the file status
					flags O_APPEND and O_NONBLOCK and any combination of
					file access mode flags O_RDONLY, O_RDWR, and O_WRONLY
					and oflag values O_CREAT, O_EXCL, O_NOCTTY, and O_TRUNC,
					then a call to fcntl(fildes, F_SETFL, arg) does not alter
					the file access mode and ignores the oflag values.
				*/
	assertion(13, BASE);
	start
		FORK
			prep("Open temp file with O_RDONLY flag set.");
			prep("Call fcntl(fildes, F_SETFL, O_APPEND | O_NONBLOCK).");
			start
				fd = Zopen(path, O_RDONLY | O_CREAT, PROT_ALL);
				Zfcntl(fd, F_SETFL, O_APPEND | O_NONBLOCK);
			finis
		
			testing("File access modes O_RDONLY, O_RDWR, O_WRONLY are not");
			testing("altered and the oflag values O_CREAT, O_EXCL, O_NOCTTY,");
			testing("and O_TRUNC are ignored.");
			start
				ret = Zfcntl(fd, F_GETFL, 0);
				if( (ret & O_ACCMODE) != O_RDONLY)
					error("Expected access mode to be O_RDONLY (%d).\n%s%d.\n",
						O_RDONLY,
						"Actual access mode is ",
						ret & O_ACCMODE);

				Zfcntl(fd, F_SETFL, O_APPEND | O_NONBLOCK | O_CREAT |
											O_EXCL | O_NOCTTY | O_TRUNC);
				ret = Zfcntl(fd, F_GETFL, 0);
				if( (ret & (O_ACCMODE)) != O_RDONLY)
					error("Expected access mode to be O_RDONLY (%d).\n%s%d.",
						O_RDONLY,
						"Actual access mode is ",
						ret & O_ACCMODE);

				if((fcntl(fd, F_GETFL, 0) & O_NONBLOCK) != O_NONBLOCK)
					error("O_NONBLOCK clear!");
				if((fcntl(fd, F_GETFL, 0) & O_APPEND) != O_APPEND)
					error("O_APPEND clear!");
			finis
		KROF
	finis

	done();
}
