#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	String		"Woolly bully, woolly bully\n"
#define	Strlen		(sizeof(String) -1)
#include	<locks.h>

#define		FILENAME	"6.5.2_25"


int main(void)
{
	int		fd;
	pid_t	pid;
	char	tmp_buf[FILE_LENGTH];
	struct	flock flock1, flock2;

	setup(E_FCNTL);
	funct(E_FCNTL);

		/*	When a shared lock has been set on a segment of a
			file, other processes are be able to set shared
			locks on that segment or a portion of it.
		*/
	assertion(25, BASE);
	start
		break_on(ERROR);
		prep("Open a temp file, then set a shared lock on entire file.");
		start
			exit_on(ERROR);
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
			SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, QUARTER_FILE, &flock1);
		finis   

		FORK
			testing("Have child set a shared lock on entire file.");
			start
				SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock2);
			finis

			testing("Have child set a shared lock on a portion of file.");
			start
				SET_LOCK(F_RDLCK, fd, SEEK_SET, QUARTER_FILE,QUARTER_FILE,
																	&flock2);
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);
	finis

		/*	A shared lock prevents any other process from
			setting an exclusive lock on any portion of the
			part of the file to which the shared lock applies.
		*/
	assertion(26, BASE);
	start
		break_on(ERROR);
		prep("Open a temp file, then set a shared lock on entire file.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
			SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, QUARTER_FILE, &flock1);
		finis   

		FORK
			testing("Have child set an exclusive lock on entire file.");
			start
				expecting(FAILURE);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock2);
			finis

			testing("Have child set an exclusive lock on portion of file.");
			start
				expecting(FAILURE);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, QUARTER_FILE,
																	&flock2);
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);
	finis

		/*	An exclusive lock prevents any other process from
			setting a shared lock or an exclusive lock on any
			portion of the protected area.
		*/
	assertion(27, BASE);
	start
		break_on(ERROR);
		prep("Create a temp file, set an exclusive lock.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
			expecting(SUCCESS);
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
		finis

		FORK
			testing("Setting a shared lock or an exclusive lock on the");
			testing("whole file fails.");
			start
				expecting(FAILURE);
				SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock1);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock1);
			finis

			testing("Setting a shared lock or an exclusive on a portion");
			testing("of the file fails.");
			start
				expecting(FAILURE);
				SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET,
													QUARTER_FILE, &flock1);
				SET_LOCK(F_RDLCK, fd, SEEK_SET, HALF_FILE + QUARTER_FILE,
													QUARTER_FILE, &flock1);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET,
													QUARTER_FILE, &flock1);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, HALF_FILE + QUARTER_FILE,
													QUARTER_FILE, &flock1);
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);
	finis


		/*	Locks can start and extend beyond the current end of a file.  */
	assertion(28, BASE);
	start
		break_on(ERROR);
		prep("Open a temp file.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
		finis

		testing("Set a shared lock that begins and ends beyond the ");
		testing("end of the file.");
		start
			SET_LOCK(F_RDLCK, fd, SEEK_SET, TOTAL_FILE + TOTAL_FILE,
														HALF_FILE, &flock1);
		finis

		testing("Set a write lock that begins before the end of file");
		testing("and extends beyond the end of file.");
		start
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, 
								TOTAL_FILE + TOTAL_FILE + TOTAL_FILE, &flock1);
		finis
	finis

		/*	Locks can be set to always extend to the end-of-file
			by setting l_len to zero.
		*/
	assertion(29, BASE);
	start
		break_on(ERROR);
		prep("Open a temp file and set a shared lock using l_len = 0.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
			SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, (off_t) 0, &flock1);
		finis

		FORK
			testing("Attempt an exclusive lock on last byte of temp file.");
			start
				expecting(FAILURE);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET,
										TOTAL_FILE_LEN -((off_t) 1), &flock1);
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);

				/*	When the fields of the structure flock have l_whence
					and l_start point to the beginning of the file.
					and l_len of zero, the entire file shall be locked.
				*/
		prep("Open a temp file, set a read lock");
		prep("with l_whence and l_start pointing to the beginning of the");
		prep("file and l_len equal to zero.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, tmp_buf, FILE_LENGTH);
			SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, (off_t) 0, &flock1);
		finis

		FORK
			testing("Attempt an exclusive lock on areas of temp file.");
			start
				expecting(FAILURE);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, TOTAL_FILE -(off_t)1, (off_t)1,
																	&flock1);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, QUARTER_FILE, (off_t) 1,&flock1);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, (off_t) 1, &flock1);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,&flock1);
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);
	finis

	done();
}
