#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<Ztermios.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FILENAME	"6.5.2_34"
#define	String		"Woolly bully, woolly bully\n"
#define	Strlen		(sizeof(String) -1)

#include	<locks.h>

void v_error(void);


int main(void)
{
	int		fd;
	pid_t	pid;
	struct	flock flock1, flock2;

	setup(E_FCNTL);
	funct(E_FCNTL);

	assertion(34, BASE);
	start
		break_on(ERROR);
		prep("Create a temp file and set an exclusive lock");
		prep("on a portion of the file.");
		start
			exit_on(ERROR);
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, String, FILE_LENGTH);
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, HALF_FILE, &flock1);
		finis

		FORK
			testing("Setting shared lock on exclusively locked segment of");
			testing("file returns -1 and sets errno to [EACCES] or [EAGAIN].");
			start
				expecting(FAILURE);
				SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, QUARTER_FILE,
																	&flock2);
				v_error();
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);
	finis

	assertion(35, BASE);
	start
		break_on(ERROR);
		prep("Create a temp file and set an exclusive lock");
		prep("on a portion of the file.");
		start
			exit_on(ERROR);
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, String, FILE_LENGTH);
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, HALF_FILE, &flock1);
		finis

		FORK
			testing("Setting exclusive lock on exclusively locked segment of");
			testing("file returns -1 and sets errno to [EACCES] or [EAGAIN].");
			start
				expecting(FAILURE);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, QUARTER_FILE,
																	&flock2);
				v_error();
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);
	finis

	assertion(36, BASE);
	start
		break_on(ERROR);
		prep("Create a temp file and set a shared lock");
		prep("on a portion of the file.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, String, FILE_LENGTH);
			SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, HALF_FILE, &flock1);
		finis

		FORK
			testing("Setting exclusive lock on shared locked segment of");
			testing("file returns -1 and sets errno to [EACCES] or [EAGAIN].");
			start
				expecting(FAILURE);
				SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET,
														QUARTER_FILE, &flock2);
				v_error();
			finis
		KROF

		CLOSE_LOCKED_FILE(fd);
	finis

	assertion(37, BASE);
	start
		break_on(ERROR);
		testing("Using a closed file descriptor.");
		start
			close(fd);
			expecting(EBADF);
			Zfcntl(fd, F_DUPFD, 0);
			expecting(EBADF);
			Zfcntl(fd, F_GETFD);
			expecting(EBADF);
			Zfcntl(fd, F_SETFD, FD_CLOEXEC);
			expecting(EBADF);
			Zfcntl(fd, F_SETFD, 0);
			expecting(EBADF);
			Zfcntl(fd, F_GETFL);
			expecting(EBADF);
			Zfcntl(fd, F_SETFL, O_NONBLOCK);
			expecting(EBADF);
			Zfcntl(fd, F_SETFL, O_APPEND);
			expecting(EBADF);
			GET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
			expecting(EBADF);
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
			expecting(EBADF);
			SET_LOCKW(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
		finis
	finis

		/*	fcntl(fildes, F_SETLK, arg) and fcntl(fildes, F_SETLKW, arg)
			of type lock shared, but file descriptor in not a valid file
			descriptor open for reading.
		*/
	assertion(38, BASE);
	start
		break_on(ERROR);
		prep("Create temp file with data, close it, then open with O_WRONLY.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			expecting(FILE_LENGTH);
			Zwrite(fd, String, FILE_LENGTH);
			Zclose(fd);
	  		fd = temp_open(FILENAME, O_WRONLY, S_IRWXU);
		finis

		testing("Set a shared lock on a portion of the file.");
		start
			expecting(-1);
			expecting(EBADF);
			SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, HALF_FILE, &flock1);
		finis

		testing("Set a shared lock (wait if blocked) on portion of the file.");
		start
			expecting(-1);
			expecting(EBADF);
			SET_LOCKW(F_RDLCK,fd, SEEK_SET, ZERO_OFFSET, HALF_FILE, &flock1);
		finis
	finis

		/*	fcntl(fildes, F_SETLK/F_SETLKW, arg) of type lock
			exclusive but file descriptor in not a valid file
			descriptor open for writing.
		*/
	assertion(39, BASE);
	start
		break_on(ERROR);
		prep("Open temp file with O_RDONLY.");
		start
	  		fd = temp_open(FILENAME, O_RDONLY, S_IRWXU);
		finis

		testing("F_SETLK, Set exclusive lock on portion of the file.");
		start
			expecting(-1);
			expecting(EBADF);
			SET_LOCK(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET, HALF_FILE, &flock1);
		finis

		testing("F_SETLKW, Set exclusive lock on a portion of the file.");
		start
			expecting(-1);
			expecting(EBADF);
			SET_LOCKW(F_WRLCK,fd, SEEK_SET, ZERO_OFFSET, HALF_FILE, &flock1);
		finis
	finis

		/*	Call to fcntl(fildes, F_SETLKW, arg) that is interrupted while
			waiting for a region returns -1 and set errno to EINTR.
		*/
	assertion(40, BASE);
	start
		prep("Create a temp file, set a shared lock on the entire file.");
		start
	  		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, String, FILE_LENGTH);
			SET_LOCK(F_RDLCK, fd, SEEK_SET, ZERO_OFFSET, TOTAL_FILE, &flock1);
		finis

		ENABLE_PSYNC
		prep("Fork a child process.");
		start
			sv_flush( );
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Have child process use F_SETLKW to wait to place");
			testing("a write lock on the file.");
			start
				SIGNAL(SIGINT, sig_int);
				SEND_PSYNC
				expecting(EINTR);
				SET_LOCKW(F_WRLCK, fd, SEEK_SET, ZERO_OFFSET,
														QUARTER_FILE, &flock2);
			finis

			done();
		}

							/*	Should be a "prep", but child's message
								precedes parent's message in journal. */
		testing("Wait till child process is ready, then interrupt via SIGINT.");
		start
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
			SLEEP(1);		/*  Ensure SET_LOCKW becomes initiated in child
								process before continuing. */
			kill(pid, SIGINT);
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC
		CLOSE_LOCKED_FILE(fd);
	finis


		/*	EINVAL is returned if the argument cmd is F_DUPFD
			and arg is negative.
		*/
	assertion(41, BASE);
	start
		break_on(ERROR);
		prep("Open a temp file.");
		start
	   		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, String, FILE_LENGTH);
		finis

		testing("Dup the file descriptor using -1 as arg.");
		start
			expecting(-1);
			expecting(EINVAL);
			Zfcntl(fd, F_DUPFD, -1);
		finis
	finis

	assertion(42, BASE);
	start
		int	open_max;

		errno = 0;				/* If value not defined, errno unchanged. */
		open_max = (int)sysconf(_SC_OPEN_MAX);

		if(open_max != -1) {
			testing("When arg >= {OPEN_MAX}, then a call to fcntl()");
			testing("returns -1 and sets errno to [EINVAL].");
			start
				expecting(-1);
				expecting(EINVAL);
				Zfcntl(fd, F_DUPFD, open_max);

				expecting(-1);
				expecting(EINVAL);
				Zfcntl(fd, F_DUPFD, open_max +1);
			finis
		}
		else {
			if(errno == 0) {
				testing("When arg = {PCTS_OPEN_MAX}, then a call to");
				testing("fcntl() does not give an [EINVAL] error.");
				start
					expecting(SUCCESS);
					Zfcntl(fd, F_DUPFD, PCTS_OPEN_MAX);
				finis
			}
			else
				error("%s\n%s[%s]\n",
					"Implementation is not specifying a value for {OPEN_MAX}.",
					"A call to sysconf(_SC_OPEN_MAX) returns -1 with errno of ",
					err_lookup(errno));

		}
	finis

	done();
}


void v_error(void)
{
	if( ! ((errno == EACCES) || (errno == EAGAIN)))
		error("%s\n%s\n%s[%s]\n",
				"Incorrect errno set.",
				"Expected: [EACCES] or [EAGAIN]",
				"Actual: ",
				err_lookup(errno));
}
