#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>
#include	<Ztermios.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FILENAME	"6.5.2_43"
#define	String		"Woolly bully, woolly bully\n"
#define	Strlen		(sizeof(String) -1)

#include	<locks.h>


int main(void)
{
	int		fd;
	pid_t	pid;
	char	path[_POSIX_PATH_MAX +1];
	struct	flock flock1;
	int		i;
	int 	tty_fildes, loop_fildes;
	struct termios tty_termios, *tty_termios_p = &tty_termios, loop_termios,
									*loop_termios_p = &loop_termios;

	setup(E_FCNTL);
	funct(E_FCNTL);

		/*	For "cmd" of F_GETLK, F_SETLK, and F_SETLKW the data "arg"
			references a position prior to the beginning of the file
			and returns -1 and sets errno to [EINVAL].
		*/
	assertion(43, BASE);
	start
		break_on(ERROR);
		prep("Create a test file.");
		start
	   		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, String, FILE_LENGTH);
		finis

		testing("fcntl(fd, F_GETLK, arg) with invalid \"l_start\".");
		start
			expecting(-1);
			expecting(EINVAL);
			GET_LOCK(F_RDLCK, fd, SEEK_SET,(off_t) -4, TOTAL_FILE, &flock1);
		finis

		testing("fcntl(fd, F_SETLK, arg) with invalid \"l_start\".");
		start
			expecting(-1);
			expecting(EINVAL);
			SET_LOCK(F_RDLCK, fd, SEEK_SET,(off_t) -4, TOTAL_FILE, &flock1);
		finis

		testing("fcntl(fd, F_SETLKW, arg) with invalid \"l_start\".");
		start
			expecting(-1);
			expecting(EINVAL);
			SET_LOCKW(F_RDLCK, fd, SEEK_SET,(off_t) -4, TOTAL_FILE, &flock1);
		finis

		CLOSE_LOCKED_FILE(fd);
	finis

		/*	For "cmd" of F_GETLK, F_SETLK, and F_SETLKW when l_type field
			of the structure pointed to be "arg" is invalid shall
			return -1 and sets errno to [EINVAL].
		*/
	assertion(44, BASE);
	start
		break_on(ERROR);
		prep("Create a temporary file.");
		start
	   		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, String, FILE_LENGTH);
		finis

		testing("fcntl(fd, F_GETLK, arg) with invalid \"l_type\".");
		start
			expecting(-1);
			expecting(EINVAL);
			GET_LOCK(F_RDLCK +F_WRLCK +F_UNLCK, fd, SEEK_SET, ZERO_OFFSET,
														TOTAL_FILE, &flock1);
		finis

		testing("fcntl(fd, F_SETLK, arg) with invalid \"l_type\".");
		start
			expecting(-1);
			expecting(EINVAL);
			SET_LOCK(F_RDLCK +F_WRLCK +F_UNLCK, fd, SEEK_SET, ZERO_OFFSET,
														TOTAL_FILE, &flock1);
		finis

		testing("fcntl(fd, F_SETLKW, arg) with invalid \"l_type\".");
		start
			expecting(-1);
			expecting(EINVAL);
			SET_LOCKW(F_RDLCK +F_WRLCK +F_UNLCK, fd, SEEK_SET, ZERO_OFFSET,
														TOTAL_FILE, &flock1);
		finis

		CLOSE_LOCKED_FILE(fd);
	finis

		/*	For "cmd" of F_GETLK, F_SETLK, and F_SETLKW when l_whence field
			of the structure pointed to be "arg" is invalid shall
			return -1 and sets errno to [EINVAL].
		*/
	assertion(45, BASE);
	start
		break_on(ERROR);
		prep("Create a temporary file.");
		start
	   		fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, String, FILE_LENGTH);
		finis

		testing("fcntl(fd, F_GETLK, arg) with invalid \"l_whence\".");
		start
			expecting(-1);
			expecting(EINVAL);
			GET_LOCK(F_RDLCK, fd, SEEK_SET +SEEK_CUR +SEEK_END, ZERO_OFFSET,
														TOTAL_FILE, &flock1);
		finis

		testing("fcntl(fd, F_SETLK, arg) with invalid \"l_whence\".");
		start
			expecting(-1);
			expecting(EINVAL);
			SET_LOCK(F_RDLCK, fd, SEEK_SET +SEEK_CUR +SEEK_END, ZERO_OFFSET,
														TOTAL_FILE, &flock1);
		finis

		testing("fcntl(fd, F_SETLKW, arg) with invalid \"l_whence\".");
		start
			expecting(-1);
			expecting(EINVAL);
			SET_LOCKW(F_RDLCK, fd, SEEK_SET +SEEK_CUR +SEEK_END, ZERO_OFFSET,
														TOTAL_FILE, &flock1);
		finis

		CLOSE_LOCKED_FILE(fd);
	finis

		/*	The argument cmd is F_GETLK/F_SETLK/F_SETLKW and filedes
			refers to a file that does not support locking.
		*/
	assertion(46, BASE);
	start
		if(PCD_NO_LOCK_FILE_TYPE == TRUE) {
		  if(!GTI_NOT_PROVIDED) {
			if(GTI_LOCKS == 0) {
				FORK
					exit_on(ERROR);
					prep("Open character special files.");
					start
						termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
										loop_termios_p);
		
						SET_ALARM(TIMEOUT)
							Zwrite(loop_fildes, String, Strlen);
							Ztcdrain(loop_fildes);
						CLEAR_ALARM
					finis
	
					testing("Got a lock on this file.");
					start
						expecting(-1);
						expecting(EINVAL);
						GET_LOCK(F_WRLCK, tty_fildes, SEEK_SET, ZERO_OFFSET,
											TOTAL_FILE, &flock1);
					finis

					testing("Set a lock on this file.");
					start
						expecting(-1);
						expecting(EINVAL);
						SET_LOCK(F_WRLCK, tty_fildes, SEEK_SET, ZERO_OFFSET,
											TOTAL_FILE, &flock1);
					finis

					testing("Set a lock (wait if blocked) on this file.");
					start
						expecting(-1);
						expecting(EINVAL);
						SET_LOCKW(F_WRLCK, tty_fildes, SEEK_SET, ZERO_OFFSET,
											TOTAL_FILE, &flock1);
					finis

					cleanup("Close terminal device files.");
					start
						SIGNAL(SIGHUP, SIG_IGN);
						Zclose(tty_fildes);
						Zclose(loop_fildes);
						SLEEP(GTI_LINE_CLOSE_DELAY)
					finis
				KROF
		  	}
		  	else
				not_implemented("GTI devices support locks");
		  }
		  else
			not_implemented("GTI devices");
		}
		else
			not_implemented("{PCD_NO_LOCK_FILE_TYPE} of TRUE");
	finis

	assertion(47, NOT_APPLICABLE);

		/*	[EMFILE] is returned if the argument cmd is F_DUPFD
			and {OPEN_MAX} file descriptors are currently in use
			by the process.
		*/
	assertion(48, BASE);
	start
		int		open_max, topen_max;

		break_on(ERROR);

		open_max = g_open_max();	/* Get value for {OPEN_MAX} */
									/* Get testable value for {OPEN_MAX} */
		topen_max = (open_max <= PCTS_OPEN_MAX) ? open_max : PCTS_OPEN_MAX;

		prep("Close all (%d) file descriptors except 0, 1 and 2.", topen_max);
		start
			sv_flush();
			for(i=3; i<topen_max; i++)
				close(i);
		finis

		prep("Initialize a path name.");
		start
			sv_flush();
			Zstrcpy(path, temp_path());
			Zstrcat(path, "fill_up");
		finis

		if(open_max <= PCTS_OPEN_MAX) {
			prep("{OPEN_MAX} equal to {PCTS_OPEN_MAX}.");
			prep("Consume all available file descriptors.");
			start
				sv_flush();
				for(i=3; i<open_max; i++)
					(void) Zcreat(path, PROT_ALL);
			finis

			testing("F_DUPFD when {OPEN_MAX} file descriptors are in use.");
			start
				expecting(-1);
				expecting(EMFILE);
				Zfcntl(fd, F_DUPFD, 0);
			finis
		}
		else {
			prep("Testing for {PCTS_OPEN_MAX} < {OPEN_MAX}.");
			prep("Consume PCTS_OPEN_MAX -1 file descriptors."); 
			start
				sv_flush();
				for(i=3; i<(topen_max -1); i++)
					(void) Zcreat(path, PROT_ALL);
			finis

			testing("fcntl(fildes, F_DUPFD, arg) when {PCTS_OPEN_MAX} -1 file");
			testing("descriptors are in use.");
			start
				expecting(SUCCESS);
				Zfcntl(fd, F_DUPFD, 0);
			finis
		}

		cleanup("Close all file descriptors except 0, 1 and 2.");
		start
			for(i=3; i<topen_max; i++)
				(void) close(i);
		finis
	finis

	assertion(49, BASE);
	start
		int		open_max;

		break_on(ERROR);
		prep("Open testfile.");
		start
			fd = Zcreat(path, PROT_ALL);
		finis

		open_max = g_open_max();		/* Get value of {OPEN_MAX}	*/

		prep("Acquire the greatest numbered file descriptors less than");
		prep("OPEN_MAX");
		start
			expecting(SUCCESS);
			Zfcntl(fd, F_DUPFD, open_max -1);
		finis

		testing("fcntl(fildes, F_DUPFD, fildes2) when no file descriptors");
		testing("above fildes (%d) are available.", open_max);
		start
			expecting(-1);
			expecting(EMFILE);
			Zfcntl(fd, F_DUPFD, open_max -1);
		finis

		cleanup("close all descriptors except 0, 1, and 2.");
		start
			(void) close(fd);
			(void) close(open_max -1);
		finis
	finis

	assertion(50, EXTENDED);

	done();
}
