#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	String	"Rolling down the river."
#define	Strlen	(sizeof(String) -1)
#define	POS_BEG	((off_t)0)			/* Position to start of file */
#define	POS_END	((off_t) Strlen)	/* Position to end of file */

#define	FILENAME	"6.5.3_09"
#define	FIFONAME	"6.5.3_09_fifo"


int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	char	inbuf[Strlen];
	int		fd;
	off_t	ret;
	pid_t	pid;

	setup(E_LSEEK);
	funct(E_LSEEK);

	exit_on(ERROR);

	assertion(9, BASE);
	start
		prep("Create a test file and write to it.");
		start
			strcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zcreat(path, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd, String, Strlen);
			Zclose(fd);
		finis

		FORK
			testing("Call to lseek() when the file descriptor argument is not");
			testing("a valid file descriptor returns a value of (off_t)-1");
			testing("and sets errno to EBADF.");
			start
				expecting((off_t)-1);
				expecting(EBADF);
				ret = Zlseek(fd, POS_BEG, SEEK_SET);
			finis
		KROF
	finis

	assertion(10, BASE);
	start
		FORK
			testing("Call to lseek() when the \"whence\" argument is not");
			testing("valid returns a value of (off_t)-1 and sets errno");
			testing("to [EINVAL], and does not alter the value of the");
			testing("file pointer.");
			start
	#ifdef	SIGSYS
											/* In case a SIGSYS is generated. */
				SIGNAL(SIGSYS, sig_int);
				sig_rcvd = 0;
	#endif
				fd = Zopen(path, O_RDWR);
				Zlseek(fd, POS_BEG, SEEK_CUR);
				expecting((off_t)-1);
				expecting(EINVAL);
				Zlseek(fd, POS_END, SEEK_SET +SEEK_CUR +SEEK_END);
	
				expecting(SUCCESS);
				ret = Zlseek(fd, POS_BEG, SEEK_CUR);
				if(ret != POS_BEG)
					error("%s\n%s%lu.\n%s%lu.\n",
						"File pointer value altered.",
						"Expected ",
						(unsigned long) POS_BEG,
						"Actual ",
						(unsigned long) ret);
			finis

#ifdef	SIGSYS
										/*	In case a SIGSYS is generated.
											Report UNRESOLVED, need PCD info.
										*/
			if(sig_rcvd)
				unresolved("Signal SIGSYS generated by lseek() for invalid whence.", "6.5.3.4");
#endif
		KROF
	finis

	assertion(11, BASE);
	start
		FORK
			testing("Call to lseek() when the resulting file offset would be");
			testing("invalid returns a value of (off_t)-1 and sets errno");
			testing("to [EINVAL], and does not alter the value of the");
			testing("file pointer.");
			start
				fd = Zopen(path, O_RDWR);
				Zlseek(fd, POS_BEG, SEEK_CUR);
				expecting((off_t)-1);
				expecting(EINVAL);
				Zlseek(fd, (off_t)(-(Strlen +1)), SEEK_CUR);
	
				expecting(SUCCESS);
				ret = Zlseek(fd, POS_BEG, SEEK_CUR);
				if(ret != POS_BEG)
					error("%s\n%s%lu.\n%s%lu.\n",
						"File pointer value altered.",
						"Expected ",
						(unsigned long) POS_BEG,
						"Actual ",
						(unsigned long) ret);
			finis
		KROF
	finis

	assertion(12, BASE);
	start
		int 	fildes[2];

		FORK
			prep("Create a pipe and write to it.");
			start
				Zpipe(fildes);
				expecting(Strlen);
				Zwrite(fildes[1], String, Strlen);
			finis

			testing("Call to lseek() when the fildes argument is associated");
			testing("with a pipe returns a value of (off_t)-1 and sets errno");
			testing("to [ESPIPE].");
			start
				expecting((off_t)-1);
				expecting(ESPIPE);
				Zlseek(fildes[0], POS_END, SEEK_CUR);
				expecting((off_t)-1);
				expecting(ESPIPE);
				Zlseek(fildes[1], POS_END, SEEK_CUR);
			finis

			testing("File pointer is not altered.");
			start
				expecting(Strlen);
				Zread(fildes[0], inbuf, Strlen);
			finis
		KROF
	finis

	assertion(13, BASE);
	start
		char	fifop[_POSIX_PATH_MAX +1];
		int 	fiford;						/* FIFO read fildes		*/
		int 	fifowr;						/* FIFO write fildes	*/

		FORK
			prep("Create a FIFO and write to it.");
			start
				strcpy(fifop, temp_path());
				Zstrcat(fifop, FIFONAME);
				Zmkfifo(fifop, PROT_ALL);
				fiford = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifowr =  Zopen(fifop, O_WRONLY);
				expecting(Strlen);
				Zwrite(fifowr, String, Strlen);
			finis
	
			testing("Call to lseek() when the fildes argument is associated");
			testing("with a FIFO returns a value of (off_t)-1 and sets errno");
			testing("to [ESPIPE].");
			start
				expecting((off_t)-1);
				expecting(ESPIPE);
				Zlseek(fiford, POS_END, SEEK_CUR);
				expecting((off_t)-1);
				expecting(ESPIPE);
				Zlseek(fifowr, POS_END, SEEK_CUR);
			finis

			testing("File pointer is not altered.");
			start
				expecting(Strlen);
				Zread(fiford, inbuf, Strlen);
			finis
		KROF
	finis

	done();
}
