#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME "6.4.1_04"
#define FIFONAME "6.4.1_04.fifo"


int main(void)
{
	char	path[_POSIX_PATH_MAX +1], fifop[_POSIX_PATH_MAX +1];
	int		pipfd[2];
	int 	fd;
	char	*inbuf, *outbuf;
	int 	fifo_rd, fifo_wr;
	int		i;
	int 	tty_fildes, loop_fildes;
	struct termios tty_termios, *tty_termios_p = &tty_termios;
	struct termios loop_termios, *loop_termios_p = &loop_termios;
	size_t		pipe_buf_sz, fifo_buf_sz, max_buf_sz, max_input;
	int	nb;


	setup(E_READ);
	funct(E_READ);

	assertion(4, BASE);
	start
		break_on(ERROR);
		prep("Establish path to FIFO and create it");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

		max_buf_sz = (size_t) g_pipe_buf(fifop, &pipe_buf_sz, &fifo_buf_sz);


		prep("Initialize input and output data buffers and create a");
		prep("regular file type.");
		start
			inbuf = (char *) Zmalloc(max_buf_sz);
			outbuf = (char *) Zmalloc(max_buf_sz);
			memset(inbuf, '\0', max_buf_sz);

			for(i = 0; i < max_buf_sz; ++i)
				outbuf[i] = i % CHAR_MAX;

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zcreat(path, PROT_ALL);
			expecting((int)pipe_buf_sz);
			Zwrite(fd, outbuf, pipe_buf_sz);
			Zclose(fd);
		finis

		testing("Regular file type:");
		testing("Call to read() reads the specified bytes of data");
		testing("from referenced file descriptor into the buffer referenced");
		testing("and returns the number of bytes read.");
		start
			break_on(ERROR);

			SET_ALARM(TIMEOUT)
				for(i = 0; i <= max_buf_sz; i += max_buf_sz/4) {
					fd = Zopen(path, O_WRONLY | O_TRUNC);
					expecting(i);
					Zwrite(fd, outbuf, (size_t)i);
					Zclose(fd);

					fd =  Zopen(path, O_RDONLY);
					expecting(i);
					Zread(fd, inbuf, pipe_buf_sz);

					if(strncmp(inbuf, outbuf, pipe_buf_sz))
						error("Write/read buffers do not contain same data.");

					Zclose(fd);
				}
			CLEAR_ALARM
		finis

		testing("Pipe file type:");
		testing("Write, read and verify various combinations of nbyte");
		testing("bytes to pipe.");
		start
			break_on(ERROR);

			SET_ALARM(TIMEOUT)
								/* write 1, pipe_buf_sz/4, pipe_buf_sz/2 */
				for(i = 1;
						i <= pipe_buf_sz;
						i +=(int)((i==1) ? (pipe_buf_sz/4) -1 : pipe_buf_sz/4)){
					Zpipe(pipfd);

					expecting(i);
					Zwrite(pipfd[1], outbuf, (unsigned)i);
					Zread(pipfd[0], inbuf, (unsigned)i);

					if(strncmp(inbuf, outbuf, i))
						error("Write/read buffers do not contain same data.");

					Zclose(pipfd[1]);
					Zclose(pipfd[0]);
				}
			CLEAR_ALARM
		finis

		testing("FIFO file type:");
		testing("Write, read and verify various combinations of nbyte");
		testing("bytes to FIFO.");
		start
			break_on(ERROR);

			SET_ALARM(TIMEOUT)
								/* write 1, fifo_buf_sz/4, fifo_buf_sz/2 */
				for(i = 1;
						i <= fifo_buf_sz;
						i +=(int)((i==1) ? (fifo_buf_sz/4) -1 : fifo_buf_sz/4)){
					fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
					fifo_wr =  Zopen(fifop, O_WRONLY);

					expecting(i);
					Zwrite(fifo_wr, outbuf, (unsigned)i);
					Zread(fifo_rd, inbuf, (unsigned)i);

					if(strncmp(inbuf, outbuf, i))
						error("Write/read buffers do not contain same data.");

					Zclose(fifo_wr);
					Zclose(fifo_rd);
				}
			CLEAR_ALARM
		finis

										/* GTI device */
		if( ! GTI_NOT_PROVIDED) {
			prep("Initialize terminal files.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);
			finis
	
			max_input = g_max_input(PORT_TTY);
									/* Limit testing to size of VMIN */
			if(max_input > (unsigned)((cc_t)-1))
				max_input = (unsigned)((cc_t)-1);
	
			prep("Allocate I/O buffers.");
			start
				inbuf = (char *) Zmalloc(max_input);
				outbuf = (char *) Zmalloc(max_input);
			finis
	
			nb = max_input/16;
			for(i = 1; i <= max_input; i += (i==1) ? 15 : nb) {
				prep("Initialize input and output data buffers.");
				start
					memset(inbuf, 'Z', max_input);
					memset(outbuf, 'D', max_input);
				finis
	
				testing("Character special file type:");
				testing("Write and verify %d bytes", i);
				testing("to character special file.");
				start
					int	cnt;

					SET_ALARM(TIMEOUT)
						loop_termios_p->c_cc[VMIN] = i;
						loop_termios_p->c_cc[VTIME] = 10;
						Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
	
						expecting(i);
						Zwrite(tty_fildes, outbuf, (size_t)i);
						Ztcdrain(tty_fildes);
  						for(cnt=0; cnt != i;)
  							cnt += Zread(loop_fildes, inbuf +cnt, max_input);
					CLEAR_ALARM
	
					if(memcmp(inbuf, outbuf, i))
						error("Write/read buffers do not contain same data.");
				finis
			}
	
			cleanup("Close terminal device file.");
			start
				SIGNAL(SIGHUP, SIG_IGN);
				Zclose(tty_fildes);
				Zclose(loop_fildes);
				SLEEP(GTI_LINE_CLOSE_DELAY)
			finis
		}
	finis
	
	done();
}
