#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME "6.4.1_05"
#define FIFONAME "6.4.1_05_fifo"

char	String[] = "Hello, world!";
char	String1[] = "Hello,";
char	String2[] = " world!";

#define Strlen	(sizeof(String)-1)
#define Strlen1	(sizeof(String1)-1)
#define Strlen2	(sizeof(String2)-1)

int main(void)
{
	pid_t	pid;
	char	path[_POSIX_PATH_MAX +1];
	char	fifop[_POSIX_PATH_MAX +1];
	int		pipfd[2];
	int 	fd;
	int		flags;
	char	*inbuf, *outbuf;
	int 	fifo_rd, fifo_wr;
	int		i;
	size_t	pipe_buf_sz, fifo_buf_sz, max_buf_sz;

	setup(E_READ);
	funct(E_READ);

	exit_on(ERROR);

	assertion(5, BASE);
	start
		prep("Establish a FIFO to be used for later testing.");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

		max_buf_sz = (size_t) g_pipe_buf(fifop, &pipe_buf_sz, &fifo_buf_sz);

		prep("Allocate and initialize input and output buffers.");
		start
			inbuf = (char *) Zmalloc(max_buf_sz);
			outbuf = (char *) Zmalloc(max_buf_sz);
			memset(inbuf, '\0', max_buf_sz);
			for(i = 0; i < max_buf_sz; ++i)
				outbuf[i] = i % CHAR_MAX;
		finis

		prep("Create a regular file to use for testing.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zcreat(path, PROT_ALL);
			expecting((int)pipe_buf_sz);
			Zwrite(fd, outbuf, pipe_buf_sz);
			Zclose(fd);
		finis

		FORK
			testing("A read of 0 bytes does not move file offset pointer.");
			start
				fd = Zopen(path, O_RDONLY, PROT_ALL);
				expecting(0);
				Zread(fd, inbuf, 0);
				expecting((off_t)0);
				Zlseek(fd, (off_t)0, SEEK_CUR);
	
				expecting((off_t)5);
				Zlseek(fd, (off_t)5, SEEK_CUR);
				expecting(0);
				Zread(fd, inbuf, 0);
				expecting((off_t)5);
				Zlseek(fd, (off_t)0, SEEK_CUR);
				Zclose(fd);
			finis
		KROF
	finis

	assertion(6, BASE);
	start
		FORK
			prep("Open test file.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, FILENAME);
				fd = Zopen(path, O_RDONLY);
			finis

			testing("On a regular file, read() starts at position in the");
			testing("file given by the file offset associated with fildes.");
			start
				off_t	offset;

				for(offset = 1; offset < pipe_buf_sz; offset += pipe_buf_sz/8) {
					expecting(offset);
					Zlseek(fd, offset, SEEK_SET);
					*inbuf =(char)0;
					expecting(1);
					Zread(fd, inbuf, 1);
					if(*inbuf != outbuf[(int)offset])
						error("Improper read position.");
				}
			finis
		KROF
	finis

	assertion(7, BASE);
	start
		FORK
			prep("Open test file.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, FILENAME);
				fd = Zopen(path, O_RDONLY);
			finis

			testing("Before read() returns, the file offset is");
			testing("incremented by the number of bytes actually read.");
			start
				off_t	before, after;

				before = Zlseek(fd, 0, SEEK_CUR);
				Zread(fd, inbuf, pipe_buf_sz);

				expecting(before + pipe_buf_sz);
				after = Zlseek(fd, 0, SEEK_CUR);
			finis
		KROF
	finis

	assertion(8, BASE);
	start
		FORK
								/* FIFO files are not capable of seeking. */
			prep("Open FIFO file and write to it.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
				SET_ALARM(TIMEOUT)
					Zwrite(fifo_wr, String1, Strlen1);
				CLEAR_ALARM
			finis

			testing("Attempt to seek to beginning.");
			start
				expecting(ESPIPE);
				Zlseek(fifo_wr,(off_t)0, SEEK_SET);
			finis

			testing("Write to FIFO.");
			start
				SET_ALARM(TIMEOUT)
					Zwrite(fifo_wr, String2, Strlen2);
				CLEAR_ALARM
			finis

			testing("Read from FIFO.");
			start
				memset(inbuf,'\0',pipe_buf_sz);

				SET_ALARM(TIMEOUT)
					expecting(Strlen);
					Zread(fifo_rd, inbuf, Strlen);
				CLEAR_ALARM

				if(strncmp(inbuf, String, Strlen))
					error("Non-seeking write -\nExpected: %s\nReceived: %s",
							String, inbuf);
			finis
		KROF
	finis

	assertion(9, BASE);
	start
		FORK
			prep("Open test file.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, FILENAME);
				fd = Zopen(path, O_RDONLY);
			finis

			testing("Call to read() when more data is requested than is");
			testing("available returns the number of bytes available.");
			start
				off_t	offset;

				for(offset = (off_t)1; offset < (off_t)pipe_buf_sz;
								offset += (off_t)(pipe_buf_sz/8)) {
					Zlseek(fd, -offset, SEEK_END);

					expecting((int)offset);
					Zread(fd, inbuf, pipe_buf_sz + 1);
				}
			finis
		KROF
	finis

	assertion(10, BASE);
	start
		FORK
			prep("Open FIFO file.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
			finis

			testing("Write to FIFO.");
			start
				SET_ALARM(TIMEOUT)
					Zwrite(fifo_wr, String1, Strlen1);
				CLEAR_ALARM
			finis

			testing("Call to read() on a FIFO when more data is requested");
			testing("than is available returns the number of bytes available.");
			start
				SET_ALARM(TIMEOUT)
					expecting((int)Strlen1);
					Zread(fifo_rd, inbuf, Strlen1 +10);
				CLEAR_ALARM
			finis

			prep("Establish a pipe.");
			start
				Zpipe(pipfd);
				flags = Zfcntl(pipfd[0], F_GETFL);
				Zfcntl(pipfd[0], F_SETFL, flags | O_NONBLOCK);
			finis

			testing("Write to pipe.");
			start
				SET_ALARM(TIMEOUT)
					Zwrite(pipfd[1], String1, Strlen1);
				CLEAR_ALARM
			finis

			testing("Call to read() on a pipe when more data is requested");
			testing("than is available returns the number of bytes available.");
			start
				SET_ALARM(TIMEOUT)
					expecting((int)Strlen1);
					Zread(pipfd[0], inbuf, Strlen1 +10);
				CLEAR_ALARM
			finis
		KROF
	finis

	assertion(18, BASE);
	start
		FORK
			prep("Open test file.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, FILENAME);
				fd = Zopen(path, O_RDONLY);
			finis

			testing("A read of a regular file, when the file offset is at or");
			testing("beyond the end-of-file shall return zero.");
			start
				Zlseek(fd, (off_t)0, SEEK_END);
				expecting(0);
				Zread(fd, inbuf, pipe_buf_sz + 1);

				Zlseek(fd, (off_t)5, SEEK_END);
				expecting(0);
				Zread(fd, inbuf, pipe_buf_sz + 1);
			finis
		KROF
	finis

	done();
}
