#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME "6.4.1_19"
#define FIFONAME "6.4.1_19_fifo"
#define	DELAY	2


int main(void)
{
	pid_t	pid;
	char	path[_POSIX_PATH_MAX +1];
	char	fifop[_POSIX_PATH_MAX +1];
	int 	fifo_rd, fifo_wr, pipfd[2], fd;
	char	*inbuf, *outbuf;
	int		i;
	size_t	pipe_buf_sz, fifo_buf_sz, max_buf_sz;

	setup(E_READ);
	funct(E_READ);

	exit_on(ERROR);

	assertion(19, BASE);
	start
		FORK
			prep("Create a pipe and close the write channel.");
			start
				inbuf = (char *) Zmalloc(1);

				Zpipe(pipfd);
				Zclose(pipfd[1]);
			finis

			testing("A read from an empty pipe where no process has the pipe");
			testing("open for writing shall return zero");
			start
				expecting(0);
				Zread(pipfd[0], inbuf, 1);
			finis

			prep("Create FIFO and open for reading only.");
			start
				Zstrcpy(fifop, temp_path());
				Zstrcat(fifop, FIFONAME);
				Zmkfifo(fifop, PROT_ALL);
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
			finis

			testing("A read from an empty FIFO where no process has the FIFO");
			testing("open for writing shall return zero.");
			start
				SET_ALARM(TIMEOUT)
					expecting(0);
					Zread(fifo_rd, inbuf, 1);
				CLEAR_ALARM
			finis
		KROF
	finis

	assertion(20, BASE);
	start
		size_t	nbytes;

		prep("Establish path for FIFO.");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
		finis
	
		max_buf_sz = g_pipe_buf(fifop, &pipe_buf_sz, &fifo_buf_sz);
	
		prep("Allocate and initialize I/O buffers.");
		start
			inbuf = (char *) Zmalloc(max_buf_sz);
			outbuf = (char *) Zmalloc(max_buf_sz);
			memset(inbuf, '\0', max_buf_sz);
	
			for(i = 0; i < max_buf_sz; ++i)
				outbuf[i] = i % CHAR_MAX;
		finis
	
				/*	A read from an empty FIFO where some process has the FIFO
					open for writing and O_NONBLOCK is clear, blocks until
					some data is written or until the FIFO is closed by all
					processes that had the FIFO opened for writing.
				*/
		FORK
			ENABLE_PSYNC

			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Write to FIFO.");
				start
					fifo_wr = Zopen(fifop, O_WRONLY);
					sv_flush();
					SET_TIMEOUT
						WAIT_PSYNC
					CLEAR_ALARM
					SLEEP(DELAY);
					Zwrite(fifo_wr, outbuf, fifo_buf_sz);
					SET_TIMEOUT
						WAIT_PSYNC
					CLEAR_ALARM
					SLEEP(DELAY);
					Zclose(fifo_wr);
				finis

				done();
			}
	
			prep("Open a FIFO for reading.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY);
			finis
	
			testing("Block until data is available.");
			start
				SET_ALARM(DELAY +1)
					SEND_PSYNC
					expecting((int)fifo_buf_sz);
					nbytes = (unsigned) Zread(fifo_rd, inbuf, fifo_buf_sz);
				CLEAR_ALARM
	
				if(unused_time >1)
					error("read() - did not block until data was written");
	
				if(memcmp(inbuf,outbuf,nbytes))
					error("read() - did not read proper data.");
			finis
	
			testing("A read() returns when the FIFO is");
			testing("closed by all writing processes.");
			start
				SET_ALARM(DELAY +1)
					SEND_PSYNC
					read(fifo_rd, inbuf, fifo_buf_sz);
				CLEAR_ALARM
	
				if(unused_time > 1)
					error("read() - didn't block.");
			finis
			WAIT_FOR_PID(pid);
	
				/*	A read from an empty pipe where some process has the pipe
					open for writing and O_NONBLOCK is clear, blocks until
					some data is written or until the pipe is closed by all
					processes that had the pipe opened for writing.
				*/
			prep("Allocate a pipe.");
			start
				Zpipe(pipfd);
			finis
	
			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid ==(pid_t)0) {
				prep("Write to pipe.");
				start
					sv_flush();
					Zclose(pipfd[0]);
					SET_TIMEOUT
						WAIT_PSYNC
					CLEAR_ALARM
					SLEEP(DELAY);
					Zwrite(pipfd[1], outbuf, pipe_buf_sz);
					SET_TIMEOUT
						WAIT_PSYNC
					CLEAR_ALARM
					SLEEP(DELAY);
					Zclose(pipfd[1]);
				finis

				done();
			}
	
			testing("Block until data is available.");
			start
				SET_ALARM(DELAY +1)
					SEND_PSYNC
					nbytes = (unsigned) Zread(pipfd[0], inbuf, pipe_buf_sz);
				CLEAR_ALARM
	
				if(unused_time > 1)
					error("read() - did not block.");
	
				if(memcmp(inbuf,outbuf,nbytes))
					error("read() - did not read proper data.");
			finis
	
			testing("A read() returns when the pipe is");
			testing("closed by all writing processes.");
			start
				Zclose(pipfd[1]);
				SET_ALARM(DELAY +1)
					SEND_PSYNC
					read(pipfd[0], inbuf, pipe_buf_sz);
				CLEAR_ALARM
	
				if(unused_time > 1)
					error("read() - did not block.");
			finis
			WAIT_FOR_PID(pid);

			DISABLE_PSYNC
		KROF
	finis

	assertion(21, EXTENDED);
	assertion(22, EXTENDED);

	assertion(23, BASE);
	start
		prep("Establish path to test file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
		finis
	
		FORK
			prep("Create a file for testing.");
			start
				char	c;
	
				fd = Zopen(path, (O_RDWR | O_CREAT | O_TRUNC), PROT_ALL);
	
								/* Put out some bytes of known data. */
				for(c = (char)1; c <= (char)25; c++) {
					expecting(1);
					Zwrite(fd, &c, 1);
				}
	
								/* Seek out past the end-of-file. */
				expecting((off_t)75);
				Zlseek(fd, (off_t)50, SEEK_END);
	
								/* Fill in a little bit more. */
				while(c <= (char)50) {
					expecting(1);
					Zwrite(fd, &c, 1);
					c++;
				}
	
								/* We now have: 1-25, 50 byte hole, 26-50. */
				expecting((off_t)0);
				Zlseek(fd,(off_t)0, SEEK_SET);
			finis
	
				/*	read() from a regular file, prior to the end-of-file that
					has not had data written to that part of the file shall
					return bytes with the value zero.
				*/
			testing("The \"hole\" in this file reads as zero values.");
			start
				expecting(100);
				Zread(fd, inbuf, 100);
	
				for(i = 0; i < 25; i++)
					if(inbuf[i] != (char)((i + 1) & 0xff))
						error("First part of read incorrect.");
	
				for(i = 25; i < 75; i++)
					if(inbuf[i] != (char)0)
						error("File \"hole\" is not of value zero.");
	
				for(i = 75; i < 100; i++)
					if(inbuf[i] != (char)((i - 49) & 0xff))
						error("Second part of read incorrect.");
			finis
	
			testing("Use lseek() to seek to start of \"hole\", reread data.");
			testing("The \"hole\" in this file reads as zero values.");
			start
				expecting((off_t)25);
				Zlseek(fd,(off_t)25, SEEK_SET);
				expecting(75);
				Zread(fd, inbuf, 75);
				for(i = 0; i < 50; i++)
					if(inbuf[i] != (char)0)
						error("File \"hole\" is not of value zero.");

				for(i = 50; i < 75; i++)
					if(inbuf[i] != (char)((i - 24) & 0xff))
						error("Second part of read incorrect.");
			finis

			testing("Use fseek() to seek to start of \"hole\", reread data.");
			testing("The \"hole\" in this file must read as zero value.");
			start
				FILE	*strm;

				strm = Zfopen(path, "r");
				Zfseek(strm, (long)25, SEEK_SET);
				expecting(75);
				Zread(fileno(strm), inbuf, 75);
				for(i = 0; i < 50; i++)
					if(inbuf[i] != (char)0)
						error("File \"hole\" is not of value zero.");

				for(i = 50; i < 75; i++)
					if(inbuf[i] != (char)((i - 24) & 0xff))
						error("Second part of read incorrect.");
			finis
		KROF
	finis

	done();
}
