#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME "6.4.1_27"
#define FIFONAME "6.4.1_27_fifo"

int main(void)
{
	pid_t	pid;
	int		pipfd[2];
	char	buf_in[10];

	setup(E_READ);
	funct(E_READ);

	exit_on(ERROR);

	assertion(27, BASE);
	start
		int		flags;
		int 	fifo_rd, fifo_wr;
		char	fifop[_POSIX_PATH_MAX +1];

		FORK
			prep("Establish a pipe and set the O_NONBLOCK flag.");
			start
				Zpipe(pipfd);
				flags = Zfcntl(pipfd[0], F_GETFL);
				Zfcntl(pipfd[0], F_SETFL, flags | O_NONBLOCK);
			finis
	
				/*	[EAGAIN], read() from an empty pipe where some process
					has the pipe open for writing and O_NONBLOCK
					is set, returns -1 and sets errno to [EAGAIN].
				*/
			testing("read() from an empty pipe with O_NONBLOCK set.");
			start
				expecting(EAGAIN);
				Zread(pipfd[0], buf_in, 1);
			finis
	
			prep("Open FIFO for read and write.");
			prep("Set O_NONBLOCK on read channel.");
			start
				Zstrcpy(fifop, temp_path());
				Zstrcat(fifop, FIFONAME);
				Zmkfifo(fifop, PROT_ALL);
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
			finis
	
				/*	[EAGAIN], read() from an empty FIFO where some process
					has the FIFO open for writing and no delay O_NONBLOCK
					is set, shall return -1 and set errno to [EAGAIN].
				*/
			testing("read() from an empty FIFO with O_NONBLOCK set.");
			start
				expecting(-1);
				expecting(EAGAIN);
				Zread(fifo_rd, buf_in, 1);
			finis
		KROF
	finis
	
	assertion(28, EXTENDED);

	assertion(29, BASE);
	start
		testing("[EBADF] returned on read() with invalid file descriptor.");
		start
			expecting(-1);
			expecting(EBADF);
			Zread(-1, buf_in, 1);
		finis
	finis

	assertion(30, BASE);
	start
		int 	fd;
		char	path[_POSIX_PATH_MAX +1];

		FORK
			testing("[EBADF] returned on read() when file descriptor is not");
			testing("open for reading");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, FILENAME);
				fd = Zcreat(path, PROT_ALL);
				Zclose(fd);
				fd =  Zopen(path, O_WRONLY);
				expecting(-1);
				expecting(EBADF);
				Zread(fd, buf_in, 1);
			finis
		KROF
	finis
		
	assertion(31, BASE);
	start
		FORK
			prep("Establish a pipe for testing.");
			start
				Zpipe(pipfd);
			finis

			testing("read() interrupted by a signal before it reads any data,");
			testing("returns -1 with errno set to [EINTR].");
			start
				SIGNAL(SIGALRM, sig_int);
				alarm(2);
				expecting(-1);
				expecting(EINTR);
				Zread(pipfd[0], buf_in, 1);
			finis
		KROF
	finis
	
	assertion(32, BASE);
	start
		sigset_t	set;

		FORK
			prep("Establish this process as a background process group.");
			start
				Zsetpgid((pid_t)0, getpid());
			finis
	
			testing("Attempt to read() from controlling terminal, when");
			testing("process is ignoring the SIGTTIN signal, returns");
			testing("a value of -1 and sets errno to [EIO].");
			start
				SIGNAL(SIGTTIN, SIG_IGN);
				expecting(-1);
				expecting(EIO);
				Zread(STDIN_FILENO, buf_in, 1);
			finis
	
			testing("Attempt to read() from controlling terminal, when");
			testing("process is blocking the SIGTTIN signal, returns");
			testing("a value of -1 and sets errno to [EIO].");
			start
				SIGBLOCK(set, SIGTTIN);
				expecting(-1);
				expecting(EIO);
				Zread(STDIN_FILENO, buf_in, 1);
			finis
		KROF
	finis
	
	assertion(33, BASE);
	start
		FORK
			FORK
				prep("Establish this process as background process group.");
				prep("Exit on a SIGHUP signal.");
				start
					Zsetpgid((pid_t)0, (pid_t)0);
					SIGNAL(SIGHUP, SIG_DFL);
				finis

				FORK
					prep("Ignore the SIGHUP signal.");
					start
						SIGNAL(SIGHUP, SIG_IGN);
					finis
	
					prep("Kill the grandchild (process group leader) to");
					prep("create an orphaned process group.");
					start
						Zkill(getppid(), SIGHUP);
						sleep(WAIT_FOR_TIME);
					finis
	
					testing("Attempt a read() from controlling terminal");
					testing("sets errno to [EIO].");
					start
						expecting(EIO);
						Zread(STDIN_FILENO, buf_in, 1);
					finis
				KROF
			KROF
		KROF
		SLEEP(2 * WAIT_FOR_TIME);		/* Ensure test has completed. */
	finis
	
	done();
}
