#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME "6.4.2_4"
#define FIFONAME "6.4.2_4_fifo"


int main(void)
{
	pid_t	pid;
	char	path[_POSIX_PATH_MAX +1], fifop[_POSIX_PATH_MAX +1];
	int		pipfd[2];
	char	inbuf[_POSIX_PIPE_BUF], outbuf[_POSIX_PIPE_BUF];
	int 	fifo_rd, fifo_wr;
	size_t	i;
	int 	tty_fildes, loop_fildes;
	struct termios tty_termios, *tty_termios_p = &tty_termios, loop_termios,
										*loop_termios_p = &loop_termios;

	setup(E_WRITE);
	funct(E_WRITE);

	exit_on(ERROR);

	assertion(4, BASE);
	start
		prep("Establish path to FIFO and create it");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

		prep("Initialize input and output data buffers and create a");
		prep("regular file type.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
		finis

				/*	Regular file type */
		FORK
			testing("Write and verify various combinations of nbyte bytes");
			testing("to regular file.");
			start
				int	fd_read, fd_write;

				for(i = 0; i < _POSIX_PIPE_BUF; ++i)
					outbuf[i] = (i+1) % CHAR_MAX;

				SET_ALARM(TIMEOUT)
					for(i = 0; i <= _POSIX_PIPE_BUF; i += 128) {
						fd_write = Zopen(path, O_WRONLY | O_TRUNC);
						fd_read =  Zopen(path, O_RDONLY);

						expecting(i);
						Zwrite(fd_write, outbuf, i);

						expecting(i);
						Zread(fd_read, inbuf, i);

						if(memcmp(inbuf, outbuf, i))
						error("Write/read error.");
					}
				CLEAR_ALARM
			finis
		KROF

										/* PIPE */
		FORK
			testing("Write and verify various combinations of nbyte bytes");
			testing("to pipe.");
			start
				for(i = 0; i < _POSIX_PIPE_BUF; ++i)
					outbuf[i] = (i+2) % CHAR_MAX;

				Zpipe(pipfd);
	
				SET_ALARM(TIMEOUT)
					for(i = 1; i <= _POSIX_PIPE_BUF; i +=(i==1)?127:128) {
						expecting(i);
						Zwrite(pipfd[1], outbuf, i);
	
						expecting(i);
						Zread(pipfd[0], inbuf, i);
	
						if(memcmp(inbuf, outbuf,i))
							error("Write/read error.");
					}
				CLEAR_ALARM
			finis
	
											/* FIFO */
			testing("Write and verify various combinations of nbyte bytes");
			testing("to FIFO.");
			start
				for(i = 0; i < _POSIX_PIPE_BUF; ++i)
					outbuf[i] = (i+3) % CHAR_MAX;

				SET_ALARM(TIMEOUT)
					for(i = 1; i <= _POSIX_PIPE_BUF; i +=(i==1)?127:128) {
						fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
						fifo_wr =  Zopen(fifop, O_WRONLY);
	
						expecting(i);
						Zwrite(fifo_wr, outbuf, i);
						expecting(i);
						Zread(fifo_rd, inbuf, i);
	
						if(memcmp(inbuf, outbuf, i))
							error("Write/read error.");
					}
				CLEAR_ALARM
			finis
		KROF

								/* Terminal device type file */
		if( ! GTI_NOT_PROVIDED) {
			prep("Initialize terminal files.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
										loop_termios_p);

				tty_termios_p->c_lflag &= ~(ICANON | ECHO);
				tty_termios_p->c_cc[VMIN] = 0;
				tty_termios_p->c_cc[VTIME] = 0;
				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);

				memset(outbuf, 'Z', _POSIX_MAX_INPUT);
			finis

			for(i = 1; i <= _POSIX_MAX_INPUT; i += 64) {
				testing("Write and verify %d bytes", i);
				testing("to terminal device file.");
				start
					int	cnt;

					SET_TIMEOUT
						expecting(i);
						Zwrite(loop_fildes, outbuf, i);
						Ztcdrain(loop_fildes);
  						for(cnt=0; cnt != i;)
  							cnt += Zread(tty_fildes, inbuf +cnt,
															_POSIX_MAX_INPUT);
					CLEAR_ALARM

					if(memcmp(inbuf, outbuf, i))
						error("Write/read error.");
				finis
			}
		}
	finis

	done();
}
