#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FIFONAME "6.4.2_20_fifo"

char	String[] = "Hello, world!";
char	String1[] = "Hello,";
char	String2[] = " world!";
#define Strlen	(sizeof(String)-1)
#define Strlen1	(sizeof(String1)-1)
#define Strlen2	(sizeof(String2)-1)


int main(void)
{
	char	fifop[_POSIX_PATH_MAX +1];
	int		pipfd[2];
	char	*buf_in, *outbuf;
	int 	fifo_rd, fifo_wr;
	int		i;
	size_t	pipe_buf_sz, fifo_buf_sz, max_buf_sz, pipe_max, fifo_max;
	ssize_t	nbytes;
	int		flags;
	pid_t	pid;

	setup(E_WRITE);
	funct(E_WRITE);

	exit_on(ERROR);

	assertion(20, EXTENDED);

							/*	When the file's type is a pipe or FIFO, then
								a call to write() appends to the end of the
								pipe or FIFO.
							*/
	assertion(21, BASE);
	start
		prep("Establish path to FIFO and create it.");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

							/* Maximum bytes written in pipe and FIFO */
		max_buf_sz=g_max_pipe(&pipe_buf_sz, &fifo_buf_sz, &pipe_max, &fifo_max);

		prep("Allocate and initialize I/O buffers.");
		start
			buf_in = (char *) Zmalloc(max_buf_sz);
			outbuf = (char *) Zmalloc(max_buf_sz);
			memset(buf_in, '\0', max_buf_sz);

			for(i = 0; i < max_buf_sz; ++i)
				outbuf[i] = i % CHAR_MAX;
		finis

		FORK
			prep("Create a pipe.");
			start
				Zpipe(pipfd);
			finis
	
			testing("A call to write() on a pipe appends the data to the");
			testing("end of the pipe.");
			start
				SET_ALARM(TIMEOUT)
					expecting(Strlen1);
					Zwrite(pipfd[1], String1, Strlen1);
					expecting(Strlen2);
					Zwrite(pipfd[1], String2, Strlen2);
	
					expecting(Strlen);
					Zread(pipfd[0], buf_in, Strlen);
				CLEAR_ALARM
	
				if(strncmp(buf_in, String, Strlen))
					error("Data not appended to end of FIFO.");
			finis
	
			testing("A call to write() on a FIFO appends the data to the");
			testing("end of the FIFO.");
			start
				SET_ALARM(TIMEOUT)
					fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
					fifo_wr = Zopen(fifop, O_WRONLY);
	
					expecting(Strlen1);
					Zwrite(fifo_wr, String1, Strlen1);
					expecting(Strlen2);
					Zwrite(fifo_wr, String2, Strlen2);
	
					expecting(Strlen);
					Zread(fifo_rd, buf_in, Strlen);
				CLEAR_ALARM
	
				if(strncmp(buf_in, String, Strlen))
					error("Data not appended to end of FIFO.");
			finis
		KROF
	finis


		/*	Write() to a pipe or FIFO of {PIPE_BUF} bytes(or less),
			is not interleaved with data from other processes
			doing writes on the same pipe.
		*/
	assertion(22, EXTENDED_TESTED);
	start
		int child;
		char String[3];


		FORK
			prep("Create pipe.");
			start
				Zpipe(pipfd);
			finis

			for(child = 0; child < 3; ++child) {
				prep("Fork a child.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {
					prep("Write to pipe simultaneously with other processes.");
					start
						SET_ALARM(TIMEOUT)
							sprintf(String, "%d", child);
							memset(outbuf, String[0], pipe_buf_sz);
							expecting(pipe_buf_sz/3);
							nbytes = Zwrite(pipfd[1], outbuf, pipe_buf_sz/3);
						CLEAR_ALARM
					finis

					done();
				}
			}
	
			for(child = 0; child < 3; ++child)
				Zwait(0);
	
			testing("Data should not be interleaved.");
			start
				expecting((int)((pipe_buf_sz/3)*3));
				nbytes = Zread(pipfd[0], buf_in, pipe_buf_sz);
	
				/*	Locate first occurence of character and check for
					pipe_buf_sz/3 contiguous characters.
				*/
				for(child = 0; child < 3; ++child) {
					sprintf(String, "%d", child);
					if(((unsigned)strspn(strchr(buf_in, String[0]), String))
															< pipe_buf_sz/3)
						error("Data was interleaved.");
				}
			finis
	
			/*	Write() to a FIFO of {PIPE_BUF} bytes shall not be interleaved
				with data from other processes doing writes on the same FIFO.
			*/
			prep("Open FIFO file.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
			finis

			for(child = 0; child < 3; ++child) {
				prep("Fork a child.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {
					prep("Each child writes to pipe asynchronously.");
					start
						SET_ALARM(TIMEOUT)
							sprintf(String, "%d", child);
							memset(outbuf, String[0], fifo_buf_sz);
							nbytes = Zwrite(fifo_wr, outbuf, fifo_buf_sz/3);
						CLEAR_ALARM
	
						if(nbytes < (size_t)(fifo_buf_sz/3))
							error("%s\n%s%d\n%s%d\n",
								"Write to FIFO failed.",
								"Expected ",
								(int)(fifo_buf_sz/3),
								"Received ",
								(int)nbytes);
					finis
					done();
				}
			}
	
			for (child = 0; child < 3; ++child)
				Zwait(0);
	
						/* Read data and check for contiguous strings. */
			testing("Data should not be interleaved.");
			start
				expecting((int)((fifo_buf_sz/3)*3));
				nbytes = read(fifo_rd, buf_in, fifo_buf_sz);
				/*	Locate first occurence of character and check for
					fifo_buf_sz/3 contiguous characters.
				*/
				for(child = 0; child < 3; ++child) {
					sprintf(String, "%d", child);
					if(((unsigned)strspn(strchr(buf_in, String[0]), String))
															< fifo_buf_sz/3)
						error ("Data was interleaved.");
				}
			finis
		KROF
	finis

	assertion(23, EXTENDED);
	assertion(24, EXTENDED);
	
				/*	When the file's type is a pipe or a FIFO, the O_NONBLOCK
					flag is set, nbyte is less than or equal to {PIPE_BUF},
					and space exists in the pipe or FIFO for nbyte bytes of
					data, then a call to write(fildes, buf, nbyte) succeeds
					completely and returns nbyte.
				*/
	assertion(25, BASE);
	start
		FORK
														/* pipe test */
			prep("Create a pipe and set O_NONBLOCK flag on write channel.");
			start
				Zpipe(pipfd);
				flags = Zfcntl(pipfd[1], F_GETFL);
				Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
			finis

			prep("Fill up pipe's write channel.");
			start
				size_t	n_bytes;
				size_t	nb_wr = 0;

				SET_TIMEOUT
					for(n_bytes = pipe_buf_sz; n_bytes; n_bytes /= 2)
						while(pipe_max -nb_wr >= n_bytes) {
							expecting(n_bytes);
							nb_wr += Zwrite(pipfd[1], outbuf, n_bytes);
						}
				CLEAR_ALARM
			finis
	
			testing("A write() to a pipe when the pipe can hold an additional");
			testing("%lu bytes succeeds.", (unsigned long)(.8 * pipe_buf_sz));
			start
				SET_ALARM(TIMEOUT)
						/*	Read some data so that additional data
							can be written.  On some systems, the data
							read must be such that the remaining data
							is below the "low water mark".
						*/
					expecting((int)(pipe_buf_sz * .8));
					Zread(pipfd[0], buf_in, pipe_buf_sz * .8);
					expecting((int)(pipe_buf_sz * .8));
					Zwrite(pipfd[1], outbuf, pipe_buf_sz * .8);
				CLEAR_ALARM
			finis
	
														/* FIFO test */
			prep("Opening FIFO and set O_NONBLOCK flag.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY | O_NONBLOCK);
			finis

			prep("Fill up FIFO's write channel.");
			start
				size_t	n_bytes;
				size_t	nb_wr = 0;

				SET_TIMEOUT
					for(n_bytes = fifo_buf_sz; n_bytes; n_bytes /= 2)
						while(fifo_max -nb_wr >= n_bytes) {
							expecting(n_bytes);
							nb_wr += Zwrite(fifo_wr, outbuf, n_bytes);
						}
				CLEAR_ALARM
			finis

			testing("A write() to a FIFO when the FIFO can hold an additional");
			testing("%lu bytes succeeds.", (unsigned long)(.8 * pipe_buf_sz));
			start
				SET_ALARM(TIMEOUT)
					expecting((int)(pipe_buf_sz * .8));
					Zread(fifo_rd, buf_in, pipe_buf_sz * .8);
					expecting((int)(pipe_buf_sz * .8));
					Zwrite(fifo_wr, outbuf, pipe_buf_sz * .8);
				CLEAR_ALARM
			finis
		KROF
	finis
	
	done();
}
