#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>
#include	<locks.h>

#define FILENAME "8.1.49_03"
char	String[] = "Hello, world!";
#define	Strlen	(sizeof(String) -1)

static	void t_abort(int);

int	main(void)
{
	setup(E_ABORT);
	funct(E_ABORT);

	assertion(3, BASE);
	start
	  prep("Set SIGABRT to SIG_DFL.");
	  start
		SIGNAL(SIGABRT, SIG_DFL);
	  finis

	  t_abort(10);
	finis

	assertion(4, BASE);
	start
		sigset_t	set;
		int			i;

		for(i=0; i<3; i++) {
			if(i==0) {
				SIGNAL(SIGABRT, SIG_IGN);
			} else {
				if(i==1) {
					SIGBLOCK(set, SIGABRT);
				} else {
					SIGUNBLOCK(set, SIGABRT);
					SIGNAL(SIGABRT, sig_int);
				}
			}
	
	  		t_abort(i);
		}
	finis

	done();
}

static	void t_abort(int t_indx)
{
	char	path[_POSIX_PATH_MAX +1];
	int 	fd;
	struct	flock flock;
	short	lock_typ;
	int		i;
	int		stat_val;
	pid_t	pid;
	time_t	s_time;

	ENABLE_PSYNC
		prep("Create a test file.");
		start

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd, String, Strlen);
		finis

		for(lock_typ=F_RDLCK, i=0; i<2; lock_typ=F_WRLCK, i++) {
			prep("Initialize lock structure, open file, then fork() a child.");
			start
				INIT_LOCK_STRUCT(lock_typ, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
												flock);
				sv_flush ();
				pid = Zfork();
			finis
		
			if(pid == (pid_t)0) {
				prep("Have child set a lock, then delay a while.");
				start
					Zfcntl(fd, F_SETLK, &flock);
					sv_flush();
		
					SEND_PSYNC
					SET_TIMEOUT
						WAIT_PSYNCA
					CLEAR_ALARM
				finis
		
				abort();
				error("Call to abort() continued processing.");
				done();
			}
		
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
			prep("Parent ensures lock is currently set on test file.");
			start
				flock.l_type = F_WRLCK;
				Zfcntl(fd, F_GETLK, &flock);
				if(flock.l_pid != pid)
					error("Child does not have lock as expected.");
				SEND_PSYNCA
			finis

			WAIT_FOR_PID(pid);
		
			if(t_indx < 3)
				testing("SIGABRT signal is %s.",
								(t_indx==0) ? "being ignored" :
								(t_indx==1) ? "being blocked" :
								"caught and returns");
			testing("Parent tests lock removed when child calls abort().");
			start
				Zfcntl(fd, F_GETLK, &flock);
				if(flock.l_type != F_UNLCK)
					error("Lock being held by aborted child process.");
			finis
		}
		Zclose(fd);

		prep("Create a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis
		
		if(pid == (pid_t)0) {
			prep("Establish test file for testing time-related fields when");
			prep("abort() is performed.");
			start
				sv_flush();
				Ztime(&s_time);				/* Obtain minimum reference time. */
				expecting(sizeof(time_t));
				Zwrite(p_syn[1], (char *)&s_time, sizeof(time_t));
				dt_secs(TRF_DELAY);			/* Delay a fixed interval. */
				fd = Zopen(path, O_RDWR, PROT_ALL);
				expecting(Strlen);
				Zwrite(fd, String, Strlen);
			finis

			abort();
			error("Call to abort() continued processing.");
			done();
		}

		prep("Obtain test file's minimum reference time from child process.");
		start
			expecting(sizeof(time_t));
			Zread(p_syn[0], (char *)&s_time, sizeof(time_t));
		finis

		if(t_indx < 3)
			testing("SIGABRT signal is %s.",
								(t_indx==0) ? "being ignored" :
								(t_indx==1) ? "being blocked" :
								"caught and returns");
		testing("Call to abort() terminates child process with SIGABRT and");
		testing("updates all time related fields marked for update.");
		start
			waitpid(pid, &stat_val, 0);
			if( ! WIFEXITED(stat_val)) {
				if(WIFSIGNALED(stat_val)) {
					if(WTERMSIG(stat_val) != SIGABRT)
					  error("%s\n%s%d",
						"Expected: child process to be terminated by SIGABRT. ",
						"Actual: value returned from macro WTERMSIG is ",
						WTERMSIG(stat_val) );
				}
				else
					error("WIFSIGNALED reports no signal terminated child.");
			}
			else
				error("WIFEXITED terminated normally.");

			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, path);
		finis
	DISABLE_PSYNC
}
