extern	char **environ;

#define	SECS_PER_MINUTE		60
#define	SECS_PER_HOUR		(SECS_PER_MINUTE * 60)
#define	SECS_PER_DAY		(SECS_PER_HOUR * 24)
#define	DAYS_PER_YEAR		365
#define	SECS_PER_YEAR		(SECS_PER_DAY * DAYS_PER_YEAR)

#define	JLEAP_DAY_MINUS		57
#define	JLEAP_DAY_PLUS		60
#define	LEAP_DAY_MINUS		56
#define	LEAP_DAY_PLUS		59

#define	UCT_DST_COR		(-1)
#define	UCT_DST_COR_HRS	(-hours)
#define	STday	((leap && (jst_day > LEAP_DAY_PLUS))?jst_day:jst_day-1)
#define	ENDday	((leap && (jend_day > LEAP_DAY_PLUS))?jend_day:jend_day-1)

static	char	str_exp[26], *utc_time, *loc_time;
static	struct	tm tm_exp, *tm_utc, *tm_loc;


/* Compute "26 character string" obtained from a value of argument time. */
void	exp_time(time_t timer)
{
	time_t	c_timer;
	int		year, c_day, day;
	int		hrs, mins, secs;
	int		i, leap;
	char	*mth_yr[] = {
							"Jan", "Feb", "Mar", "Apr", "May", "Jun",
							"Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
						 };
	char	*day_wk[] =	 {
							"Thu", "Fri", "Sat", "Sun", "Mon", "Tue", "Wed",
						 };
	int		daym[2][12] = {
							{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}, 
							{31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}, 
						 };

	hrs = mins = secs = 0;
	c_timer = timer;
							/* Compute year */
	for(i=0, c_day=0, year=1970; c_timer>=SECS_PER_YEAR; i++, year++) {
		if((i%4) == 2) {
			if(c_timer >= (SECS_PER_YEAR +SECS_PER_DAY)) {
				c_timer -= SECS_PER_DAY;
				c_day++;
			}
			else
				break;
		}
		c_timer -= SECS_PER_YEAR;
		c_day += DAYS_PER_YEAR;
	}

							/* Compute day of week. */
	c_day += c_timer/SECS_PER_DAY;

							/* Compute month of year and day of month. */
	leap = ((year%4 == 0) && (year%100) != 0) || (year%400 == 0);
	day = (int)(c_timer/SECS_PER_DAY +1);
	for(i=0; day > daym[leap][i]; i++)
		day -= daym[leap][i];

	tm_exp.tm_yday = (int)(c_timer/SECS_PER_DAY);
	c_timer = c_timer%SECS_PER_DAY;
	hrs = (int)(c_timer/SECS_PER_HOUR);
	c_timer = c_timer%SECS_PER_HOUR;
	mins = (int) (c_timer/SECS_PER_MINUTE);
	secs = (int) (c_timer%SECS_PER_MINUTE);

	sprintf(str_exp, "%s %s %2.2d %2.2d:%2.2d:%2.2d %4d\n",
						day_wk[c_day%7],
						mth_yr[i],
						day,
						hrs,
						mins,
						secs,
						year);
								/* Pad for date, if needed, is ' '. */
	if(str_exp[8] == '0')
		str_exp[8] = ' ';

	tm_exp.tm_sec = secs;
	tm_exp.tm_min = mins;
	tm_exp.tm_hour = hrs;
	tm_exp.tm_mday = day;
	tm_exp.tm_mon = i;
	tm_exp.tm_year = year -1900;
	tm_exp.tm_wday = (((c_day%7)+4)%7);
}
	

/* Compute the POSIX.1 timer value in seconds.  */
time_t	c_timer(int year, int yday, int hours, int mins, int secs)
{
	return((time_t) (secs +mins*SECS_PER_MINUTE +hours*SECS_PER_HOUR
				 			+yday*SECS_PER_DAY +(year -1970)*SECS_PER_YEAR
				 			+((year -1969)/4)*SECS_PER_DAY));
}

/* Get Coordinated Universal Time (UTC).*/
#define	UTC_TIME \
	exp_time(timer); \
	utc_time = asctime(gmtime(&timer)); \
	if(strncmp(str_exp, utc_time, 24)) \
		error("%s\n%s%24.24s\n%s%24.24s", \
					"Invalid string returned by call to asctime(gmtime()).", \
					"Expected: ", \
					str_exp, \
					"Actual:   ", \
					utc_time);

#define	UTC_TIME_TM \
	exp_time(timer); \
	tm_utc = gmtime(&timer); \
	if( (tm_utc->tm_sec != tm_exp.tm_sec)   || \
		(tm_utc->tm_min != tm_exp.tm_min)   || \
		(tm_utc->tm_hour != tm_exp.tm_hour) || \
		(tm_utc->tm_mday != tm_exp.tm_mday) || \
		(tm_utc->tm_mon != tm_exp.tm_mon)   || \
		(tm_utc->tm_year != tm_exp.tm_year) || \
		(tm_utc->tm_wday != tm_exp.tm_wday) || \
		(tm_utc->tm_yday != tm_exp.tm_yday) \
	  ) \
		error("%s\n%s\n%s%2.2d  %2.2d  %2.2d   %2.2d   %2.2d  %4.4d %1.1d    %3.3d\n%s%2.2d  %2.2d  %2.2d   %2.2d   %2.2d  %4.4d %1.1d    %3.3d", \
				"Invalid value returned by call to gmtime().", \
				"          sec min hour mday mon year wday yday", \
				"Expected: ", \
				tm_exp.tm_sec, tm_exp.tm_min, tm_exp.tm_hour, \
				tm_exp.tm_mday, tm_exp.tm_mon, tm_exp.tm_year, \
				tm_exp.tm_wday, tm_exp.tm_yday, \
				"Actual:   ", \
				tm_utc->tm_sec, tm_utc->tm_min, tm_utc->tm_hour, \
				tm_utc->tm_mday, tm_utc->tm_mon, tm_utc->tm_year, \
				tm_utc->tm_wday, tm_utc->tm_yday);


/* Get local time as altered by TZ	 */
#define	LOC_TIME(diff) \
	exp_time(timer +diff); \
	loc_time = ctime(&timer); \
message("%s%24.24s\n", "Expecting ctime() return of: ", str_exp); \
	if(strncmp(str_exp, loc_time, 24)) \
		error("%s\n%s%24.24s\n%s%24.24s", \
					"Invalid string returned by call to ctime().", \
					"Expected: ", \
					str_exp, \
					"Actual:   ", \
					loc_time);

#define	LOC_TIME_TM(diff) \
	exp_time(timer +diff); \
	tm_loc = localtime(&timer); \
	if( (tm_loc->tm_sec != tm_exp.tm_sec)   || \
		(tm_loc->tm_min != tm_exp.tm_min)   || \
		(tm_loc->tm_hour != tm_exp.tm_hour) || \
		(tm_loc->tm_mday != tm_exp.tm_mday) || \
		(tm_loc->tm_mon != tm_exp.tm_mon)   || \
		(tm_loc->tm_year != tm_exp.tm_year) || \
		(tm_loc->tm_wday != tm_exp.tm_wday) || \
		(tm_loc->tm_yday != tm_exp.tm_yday) \
	  ) \
		error("%s\n%s\n%s%2.2d  %2.2d  %2.2d   %2.2d   %2.2d  %4.4d %1.1d    %3.3d\n%s%2.2d  %2.2d  %2.2d   %2.2d   %2.2d  %4.4d %1.1d    %3.3d", \
				"Invalid value returned by call to localtime().", \
				"          sec min hour mday mon year wday yday", \
				"Expected: ", \
				tm_exp.tm_sec, tm_exp.tm_min, tm_exp.tm_hour, \
				tm_exp.tm_mday, tm_exp.tm_mon, tm_exp.tm_year, \
				tm_exp.tm_wday, tm_exp.tm_yday, \
				"Actual:   ", \
				tm_loc->tm_sec, tm_loc->tm_min, tm_loc->tm_hour, \
				tm_loc->tm_mday, tm_loc->tm_mon, tm_loc->tm_year, \
				tm_loc->tm_wday, tm_loc->tm_yday);

time_t	ret_loc, ret_uct0;
char	tz_env_uct0[100];
									/*	For mktime() testing. accept ambiguous
										times on specific dates. 	*/
#define	ACCEPT_BOTH_STD_OR_DST	-SECS_PER_HOUR
#define	ACCEPT_BOTH_DST_OR_STD	SECS_PER_HOUR
#define	ACCEPT_ONLY_STD_OR_DST	0	/*	Time must be correct.		*/
									/*	When set, accept STD or DST time.	*/
int		ambig_flag = ACCEPT_ONLY_STD_OR_DST;
			/*	diff = TZ versus UCT time difference.	*/
#define	LOC_VS_CUT0_ENV(diff) \
	exp_time(loc_timer); \
message("%s%24.24s\n", "Testing local time of: ", str_exp); \
	tm_exp.tm_isdst = -1; \
	environ[0] = (char *)0; \
	put_env(tz_env); \
	ret_loc = mktime(&tm_exp); \
	if(ret_loc == (time_t)-1) \
		error("Call to mktime() failed (returned -1).\n"); \
	sprintf(tz_env_uct0, "TZ=UCT0"); \
	environ[0] = (char *)0; \
	put_env(tz_env_uct0); \
	tm_exp.tm_isdst = -1; \
	ret_uct0 = mktime(&tm_exp); \
	if(ret_uct0 +diff != ret_loc) { \
		if(ambig_flag == ACCEPT_ONLY_STD_OR_DST) \
			error("%s\n%s%ld\n%s%ld", \
				"Incorrect time returned by call to mktime().", \
				"Expected: ", \
				(long) (ret_uct0 +(diff)), \
				"Actual:   ", \
				(long) ret_loc); \
		else \
			if(ret_uct0 != ret_loc) \
				error("%s\nExpected: %ld  Or: %ld\n%s%ld", \
					"Incorrect time returned by call to mktime().", \
					(long) (ret_uct0 +(diff)), \
					(long) ret_uct0, \
					"Actual:   ", \
					(long) ret_loc); \
	} \
	environ[0] = (char *)0; \
	put_env(tz_env);

struct	mwd {
					int	month_beg;
					int	week_beg;
					int	day_beg;
					int	start_day;		/*	Day at which DST begins.	*/
					int	month_end;
					int	week_end;
					int	day_end;
					int	end_day;		/*	Day at which DST ends.	*/
				 } mwd[] =	{
						/*	Test data for year 1992.	*/
						/* From:  Wed, Jan 1  To: Fri, Feb 28 */
					{1, 1, 3, 0, 2, 5, 5, 58},
						/* From:  Sun, Jan 5  To: Sun, Mar 1 */
					{1, 1, 0, 4, 3, 1, 0, 60},
						/* From:  Sun, Apr 5  To: Sun, Jul 5 */
					{4, 1, 0, 95, 7, 1, 0, 186},
						/* From:  Mon, Apr 6  To: Mon, Jul 6 */
					{4, 1, 1, 96, 7, 1, 1, 187},
				};
#define	MWD_ENTRIES	(sizeof(mwd)/sizeof(struct mwd))
