#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>
#include	<locks.h>

#define FILENAME "8.1.50_07"
char	String[] = "Hello, world!";
#define	Strlen	(sizeof(String) -1)
#define	Status	0xFF


int	main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	int 	fd;
	struct	flock flock;
	pid_t	pid;
	int		stat_val;

	setup(E_EXIT);
	funct(E_EXIT);

	assertion(7, BASE);
	start
		FORK
			prep("Have child exit.");
			start
			finis

			done_e();

			error("Process continued after exit() was called.");
		KROF

		testing("Exit terminated the process.");
		start
			break_on(ERROR);
			expecting(FAILURE);
			Zkill(pid, SIGKILL);
		finis
	finis

				/*	File descriptors in calling process are closed.	*/
	assertion(8, BASE);
	start
		int	pipfd[2];

		prep("Create a regular file for testing.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_CREAT | O_RDWR, PROT_ALL);

			flock.l_type = F_WRLCK;
			flock.l_whence = SEEK_SET;
			flock.l_start = (off_t)0;
			flock.l_len = (off_t)0;
		finis

		prep(" Create a pipe and fork a child process.");
		start
			Zpipe(pipfd);
			sv_flush();
			pid = Zfork();
		finis

	   	if(pid == (pid_t)0) {
			prep("Lock regular file.");
			start
				Zfcntl(fd, F_SETLK, &flock);
			finis

			prep("Have child process delay then exit.");
			start
				SLEEP(3 * WAIT_FOR_TIME);
			finis

			done_e();
		}

		SLEEP(WAIT_FOR_TIME);

		prep("Test regular file for lock denial.");
		start
			Zfcntl(fd, F_GETLK, &flock);

			if(flock.l_type != F_WRLCK)
				error("Regular file accesible for locking.");
		finis

		prep("Have parent close pipe for reading, write to child");
		prep("then wait till child exits.");
		start
			Zclose(pipfd[0]);

			SET_TIMEOUT
				Zwrite(pipfd[1], String, Strlen);
				Zwaitpid(pid, &stat_val, 0);
			CLEAR_ALARM
		finis

		testing("File descriptors of child have been closed on child's exit.");
		start
			SIGNAL(SIGPIPE, SIG_IGN);		/* Ignore expected SIGPIPE */
			expecting(EPIPE);		/*  No process has pipe open for reading. */
			expecting(FAILURE);
			Zwrite(pipfd[1], String, Strlen);

		finis

		testing("Test regular file for lock access.");
		start
			Zfcntl(fd, F_GETLK, &flock);

			if(flock.l_type != F_UNLCK)
				error("Child's regular file descriptor was not closed.");
		finis

		cleanup("Close and unlink open files.");
		start
			Zclose(pipfd[1]);
			Zclose(fd);
			Zunlink(path);
		finis
	finis

	assertion(9, EXTENDED);

	assertion(10, BASE);
	start
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			sv_set_exit_val(Status);
			done_e();
		}

    	testing("Low order 8 bits of status are made available to parent");
		testing("process, that is waiting, when child exits.");
		start
			Zwaitpid(pid, &stat_val, 0);
			if( (WEXITSTATUS(stat_val) & 0xff) != Status)
	    		error("%s\n%s%x.\n%s%x.\n",
					"Unexpected status value returned from waitpid().",
					"Expected 0x",
					Status,
					"Received 0x",
					WEXITSTATUS(stat_val) & 0xff );
		finis
	finis

	assertion(11, BASE);
	start
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			sv_set_exit_val(Status);
			done_e();
		}

		prep("Parent delays to ensure child has exited.");
		start
			SLEEP(WAIT_FOR_TIME +1);
		finis

		testing("Exit status code is saved for return to the parent process");
		testing("on a susequent wait() call.");
		start
			Zwait(&stat_val);
			if((WEXITSTATUS(stat_val) & 0xff) != Status)
				error("%s\n%s%x.\n%s%x.\n",
				"Unexpected status value returned from wait.",
				"Expected 0x",
				Status,
				"Received 0x",
				WEXITSTATUS(stat_val) & 0xff);
		finis

		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			sv_set_exit_val(Status);
			done_e();
		}

		prep("Parent delays to ensure child has exited.");
		start
			SLEEP(WAIT_FOR_TIME +1);
		finis

		testing("Exit status code is saved for return to the parent process");
		testing("on a susequent waitpid() call.");
		start
			Zwaitpid(pid, &stat_val, 0);

			if((WEXITSTATUS(stat_val) & 0xff) != Status)
				error("%s\n%s%x.\n%s%x.\n",
				"Unexpected status value returned from waitpid.",
				"Expected 0x",
				Status,
				"Received 0x",
				WEXITSTATUS(stat_val) & 0xff);
		finis
	finis

	assertion(12, BASE);
	start
		ENABLE_PSYNC
		prep("Fork a process for testing with.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				pid_t	i_pid;

				prep("Have child process delay to ensure parent process");
				prep("has exited.");
				start
					i_pid = getppid();
					SEND_PSYNC
					SLEEP(WAIT_FOR_TIME);
				finis

				testing("Exit of parent process does not terminate child");
				testing("process and child process is assigned a new parent");
				testing("process ID.");
				start
					if(i_pid == getppid())
						error("New parent process ID was not assigned.");
					SEND_PSYNC
				finis

				done_e();
			}

			prep("Parent process calls exit().");
			start
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
				sv_set_exit_val(4);
			finis

			done_e();
		}
		WAIT_FOR_PID(pid);
		SET_TIMEOUT
						/* Wait for grandchild to be ready to exit. */
			WAIT_PSYNC
		CLEAR_ALARM
		SLEEP(WAIT_FOR_TIME)	/* Allow time for test to complete. */

		DISABLE_PSYNC
	finis
	
	assertion(13, BASE);
	start
		FORK
			prep("Fork a grandchild.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Have child delay, then call exit().");
				start
					sv_flush();
					SLEEP(1);
					sv_set_exit_val(6);
				finis

				done_e();
			}

			prep("Have parent catch SIGCHLD signal");
			start
				sig_rcvd = 0;
				SIGNAL(SIGCHLD, sig_int);
			finis

			sleep(2);		/* SIGHUP delivered during this interval. */

			testing("SIGCHLD signal received by parent process.");
			start
				if( ! sig_rcvd)
					error("SIGCHLD not generated when child exited.");
			finis

			WAIT_FOR_PID(pid);
		KROF
	finis

	assertion(14, NOT_APPLICABLE);

	done();
}
