#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>


int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	pid_t	pid;

	setup(E_EXIT);
	funct(E_EXIT);

	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(15, BASE)
		GTI_NOT_PROVIDED_RESULT(16, BASE)
		done();
	}

	assertion(15, BASE);
	start
		ENABLE_PSYNC
		prep("Fork a process to test with.");
		start
			sv_flush();
			pid = Zfork();
		finis
	
		if(pid == (pid_t)0) {
			prep("Establish controlling terminal process.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {			/* Child process */
				pid_t	c_ppid, gc_ppid, ggc_ppid;
	
				prep("Have child processes catch SIGHUP.");
				start
					SIGNAL(SIGHUP, sig_int);
					sig_rcvd = 0;
				finis
	
				prep("Fork a grandchild process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {		/* Grandchild process */
					prep("Fork a great grandchild process.");
					start
						sv_flush();
						pid = Zfork();
					finis
	
					if(pid == (pid_t)0) {	/* great grandchild */
						testing("Great grandchild received SIGHUP signal.");
						start
							ggc_ppid = getppid();
							SEND_PSYNC
							sleep(3 * WAIT_FOR_TIME);  /* Interrupt by SIGHUP */
							if(sig_rcvd == 0)
								error("SIGHUP not received.");
						finis
	
						done();
					}
	
					/* Grand child process */
					testing("Grandchild process received a SIGHUP signal.");
					start
						gc_ppid = getppid();
						sleep(3 * WAIT_FOR_TIME);	/* Interrupted by SIGHUP */
						if(sig_rcvd == 0)
							error("SIGHUP not received.");
					finis
	
					WAIT_FOR_PID(pid);
	
					done();
				}
	
				/* Child process */
				testing("Child process received a SIGHUP signal.");
				start
					c_ppid = getppid();
 								/* Allow ample time for parent to exit. */
					sleep(3 * WAIT_FOR_TIME);	/* Interrupted by SIGHUP */
					if(sig_rcvd == 0)
						error("SIGHUP not received.");
				finis
	
				testing("Child's parent process ID has changed.");
				start
					if(c_ppid == getppid()) {
								/*	Allow for SIGHUP to be processed before
									the implementation has changed the ppid
									of the children.
								*/
						SLEEP(WAIT_FOR_TIME);
						if(c_ppid == getppid())
							error("Child's ppid still is %ld.", (long)c_ppid);
					}
				finis
	
				WAIT_FOR_PID(pid);
	
				done();
			}
	
			prep("Parent calls exit().");
			start
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
				SLEEP(WAIT_FOR_TIME);	/* All child processes now sleeping. */
			finis

			done_e();					/* calls exit() */
		}
		WAIT_FOR_PID(pid);
		DISABLE_PSYNC

		SLEEP(WAIT_FOR_TIME);		/* Sleep, ensure completion of test. */
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	assertion(16, BASE);
	start
		pid_t	ct_pgrp;

		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

				SIGNAL(SIGHUP, SIG_IGN);
				ct_pgrp = Ztcgetpgrp(tty_fildes);
			finis

			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("When the controlling parent process calls exit(),");
				testing("the controlling terminal is disassociated from the");
				testing("session.");
				start
					SLEEP(WAIT_FOR_TIME);	/* Allow time for parent to exit. */
					if(tcgetpgrp(tty_fildes) == ct_pgrp)
						error("Controlling terminal not disassociated from session.");
				finis

				done();
			}

			done_e();					/* Controlling process calls exit(). */
		}
		WAIT_FOR_PID(pid);
		SLEEP(2 * WAIT_FOR_TIME)
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
