#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>
#include	<locks.h>

#define FILENAME "8.1.8_04"
#define FIFONAME "8.1.8_04_fifo"
char	String[] = "Hello, world!";
#define	Strlen	(sizeof(String) -1)


int	main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	int 	fd;
	FILE	*sfd;
	struct	flock flock;
	char	inbuf[Strlen +1];

	setup(E_FCLOSE);
	funct(E_FCLOSE);

	assertion(4,BASE);
	start
		exit_on(ERROR);			/* If these tests fail nothing will work. */
		prep("Create a test file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
			sfd = Zfdopen(fd, "r+");
		finis

		testing("fclose() of test file's stream pointer returns the value 0.");
		start
			expecting(0);
			Zfclose(sfd);
		finis

		testing("File descriptor is available for subsequent open.");
		start
			expecting(fd);
			Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
		finis

		cleanup("Close used file descriptor.");
		start
			Zclose(fd);
		finis
	finis


	assertion(5, BASE);
	start
		short	lock_typ;
		int		i;
		pid_t	pid;

		ENABLE_PSYNC
		for(lock_typ=F_RDLCK, i=0; i<2; lock_typ=F_WRLCK, i++) {
			prep("Initialize lock structure, open file, then fork() a child.");
			start
				INIT_LOCK_STRUCT(lock_typ, SEEK_SET, ZERO_OFFSET, TOTAL_FILE,
												flock);
				fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
				sfd = Zfdopen(fd, "r+");
				sv_flush ();
				pid = Zfork();
			finis
		
			if(pid == (pid_t)0) {
				prep("Have child set a lock, then delay a while.");
				start
					Zfcntl(fd, F_SETLK, &flock);
					sv_flush();
		
					SEND_PSYNC
					SET_TIMEOUT
						WAIT_PSYNCA
					CLEAR_ALARM
				finis
		
				prep("Close stream pointer, then delay while parent checks");
				prep("file's locks.");
				start
					Zfclose(sfd);
					sv_flush();
		
					SEND_PSYNC
					SET_TIMEOUT
						WAIT_PSYNCA
					CLEAR_ALARM
				finis

				done();
			}
		
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
			prep("Parent ensures lock is currently set on test file.");
			start
				flock.l_type = F_WRLCK;
				Zfcntl(fd, F_GETLK, &flock);
				if(flock.l_pid != pid)
					error("Child does not have lock as expected.");
				sv_flush();
				SEND_PSYNCA
			finis
		
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
			testing("Parent test lock removed when stream pointer is closed.");
			start
				Zfcntl(fd, F_GETLK, &flock);
				if(flock.l_type != F_UNLCK)
					error("A lock is still set on the closed file.");
				SEND_PSYNCA
			finis
		
			WAIT_FOR_PID(pid);

			cleanup("Close used stream and underlying file descriptor.");
			start
				Zfclose(sfd);
			finis
		}
		DISABLE_PSYNC
	finis
		
	assertion(6, EXTENDED);

	assertion(7, BASE);
	start
		int		fd_rd, fd_wr;
		FILE	*sfd_rd, *sfd_wr;

		break_on(ERROR);
		prep("Establish FIFO and write data to it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FIFONAME);
			Zmkfifo(path, PROT_ALL);

			fd_rd = Zopen(path, O_RDONLY | O_NONBLOCK);
			fd_wr = Zopen(path, O_WRONLY);

			expecting(Strlen);
			Zwrite(fd_wr, String, Strlen);
		finis

		prep("Close FIFO stream pointers.");
		start
			sfd_rd = Zfdopen(fd_rd, "r");
			sfd_wr = Zfdopen(fd_wr, "w");
			Zfclose(sfd_rd);
			Zfclose(sfd_wr);
		finis

		SET_ALARM(MAX_TIME_TO_RUN)
			prep("Open FIFO for reading and writing.");
			start
				fd_rd = Zopen(path, O_RDONLY | O_NONBLOCK);
				fd_wr = Zopen(path, O_WRONLY);
			finis

			testing("When all file descriptors associated with a FIFO");
			testing("have been closed, data remaining in the pipe is");
			testing("discarded.");
			start
				expecting(EAGAIN);
				Zread(fd_rd, inbuf, 1); /* Try reading single byte. */
			finis

			cleanup("Remove FIFO file.");
			start
				close(fd_rd);
				close(fd_wr);
				Zunlink(path);
			finis
		CLEAR_ALARM
	finis

	assertion(8, EXTENDED);
	assertion(9, EXTENDED);
	assertion(10, EXTENDED);

							/* Note: assertions 11 & 12 are complementary. */
	assertion(11, BASE);
	start
		time_t	s_time, cur_time;
		struct	stat st_buf;

		prep("Establish test file for testing time-related fields when an");
		prep("fclose() is performed.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_RDWR | O_CREAT | O_TRUNC, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd, String, Strlen);
			Zclose(fd);

			sfd = Zfopen(path, "r");
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval. */
			expecting(Strlen);
			Zfread(inbuf, sizeof(char), Strlen, sfd);
			Zfclose(sfd);
			Ztime(&cur_time);
			dt_secs(TRF_DELAY);
			Zstat(path, &st_buf);
		finis

		testing("When a file is closed, via fclose() by the last process");
		testing("that had it open, all time-related fields still marked");
		testing("for update are updated.");
		start
					/* Avoid calling ctime() twice in single function call. */
			char	s1[26];
											/* st_atime updated */
			if(st_buf.st_atime <= s_time) {
				strcpy(s1, ctime(&s_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_atime\" in error?.",
					"Expected minimum time is ",
					s1,
					"Actual time recorded is ",
					ctime(&(st_buf.st_atime)) +4);
			}
	
			if(st_buf.st_atime > cur_time) {
				strcpy(s1, ctime(&cur_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_atime\" larger than current time.",
					"Current time is ",
					s1,
					"Time-related field \"st_atime\" is ",
					ctime(&(st_buf.st_atime)) +4);
			}
		finis
	finis

	assertion(12, BASE);
	start
		time_t	s_time, cur_time;
		struct	stat st_buf;

		prep("Establish test file for testing time-related fields when an");
		prep("fclose() is performed.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval. */
			fd = Zopen(path, O_RDWR | O_CREAT | O_TRUNC, PROT_ALL);
			sfd = Zfdopen(fd, "r+");
			Zfputs(String, sfd);
			Zfclose(sfd);
			Ztime(&cur_time);
			dt_secs(TRF_DELAY);
			Zstat(path, &st_buf);
		finis

		testing("Call to fclose() updated all time related fields marked");
		testing("for update.");
		start
					/* Avoid calling ctime() twice in single function call. */
			char	s1[26];
											/* st_ctime updated */
			if(st_buf.st_ctime <= s_time) {
				strcpy(s1, ctime(&s_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_ctime\" in error?.",
					"Expected minimum time is ",
					s1,
					"Actual time recorded is ",
					ctime(&(st_buf.st_ctime)) +4);
			}
	
			if(st_buf.st_ctime > cur_time) {
				strcpy(s1, ctime(&cur_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_ctime\" larger than current time.",
					"Current time is ",
					s1,
					"Time-related field \"st_ctime\" is ",
					ctime(&(st_buf.st_ctime)) +4);
			}
											/* st_mtime updated */
			if(st_buf.st_mtime <= s_time) {
				strcpy(s1, ctime(&s_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_mtime\" in error?.",
					"Expected minimum time is ",
					s1,
					"Actual time recorded is ",
					ctime(&(st_buf.st_mtime)) +4);
			}
	
			if(st_buf.st_mtime > cur_time) {
				strcpy(s1, ctime(&cur_time) +4);
				error("%s\n%s%20.20s.\n%s%20.20s.\n",
					"Time-related field \"st_mtime\" larger than current time.",
					"Current time is ",
					s1,
					"Time-related field \"st_mtime\" is ",
					ctime(&(st_buf.st_mtime)) +4);
			}
		finis
	finis

	done();
}
