#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>


#define	String		"Test"
#define	Strlen		(sizeof(String) -1)
#define	FILENAME	"8.2.2_04"

void o_read_tst(int);
void o_creat(void);
void o_clean(void);


static	int		fd, fd2;
static	FILE	*sfd;
static	char	path[_POSIX_PATH_MAX +1];

int	main(void)
{
	char	inbuf[Strlen +1];
	struct	stat st_buf1, st_buf2;

	setup(E_FDOPEN);
	funct(E_FDOPEN);

	assertion(4, BASE);
	start
		prep("Establish path to test file.");
		start
			strcpy(path, temp_path());
			strcat(path, FILENAME);
		finis

		o_creat();

		prep("Open test file with mode of \"r\".");
		start
			fd = Zopen(path, O_RDONLY);
			sfd = Zfdopen(fd, "r");
		finis

		testing("Test file can be read.");
		testing("Test file cannot be written to.");
		start
			o_read_tst(Strlen);
			rewind(sfd);

			expecting(EBADF);
			Zfwrite(String, 1, Strlen, sfd);
		finis

		o_clean();
	finis

	assertion(5, BASE);
	start
		o_creat();

		prep("Open test file with mode of \"w\".");
		start
			stat(path, &st_buf1);
			fd = Zopen(path, O_WRONLY);
			sfd = Zfdopen(fd, "w");
			fd2 = fileno(sfd);
		finis

		testing("File length remains unchanged by fdopen().");
		start
			fstat(fd2, &st_buf2);
			if(st_buf1.st_size != st_buf2.st_size)
				error("%s\n%s%lu\n%s%lu\n",
				"File length was changed after fdopen().",
				"Size of file before fdopen() is: ",
				(unsigned long) st_buf1.st_size,
				"Size of file after fdopen() is: ",
				(unsigned long) st_buf2.st_size);
		finis

		testing("Test file can be written to.");
		testing("Test file cannot be read.");
		start
			Zfwrite(String, 1, Strlen, sfd);
			rewind(sfd);

			expecting(EBADF);
			Zfread(inbuf, sizeof(*inbuf), Strlen, sfd);
		finis

		o_clean();
	finis

	assertion(6, BASE);
	start
		o_creat(); 

		prep("Open test file with mode of \"a\".");
		start
			fd = Zopen(path, O_WRONLY | O_APPEND );
			sfd = Zfdopen(fd, "a");
		finis

		testing("Test file can be written to.");
		testing("Test file cannot be read.");
		start
			Zfwrite(String, 1, Strlen, sfd);
			rewind(sfd);
			Zfwrite(String, 1, Strlen, sfd);
			rewind(sfd);

			expecting(EBADF);
			Zfread(inbuf, 1, Strlen, sfd);
			clearerr(sfd);	/*	Clear any error indicators that may be set. */
			Zfclose(sfd);
		finis

		prep("Open test file with mode of \"r\".");
		start
			fd = Zopen(path, O_RDONLY );
			sfd = Zfdopen(fd, "r");
		finis

		testing("All prior writes were at the end-of-file.");
		start
			o_read_tst(Strlen *3); 
		finis

		o_clean();
	finis

	assertion(7, BASE);
	start
		o_creat();

		prep("Open test file with mode of \"r+\".");
		start
			fd = Zopen(path, O_RDWR);
			sfd = Zfdopen(fd, "r+");
		finis

		testing("Stream returned by fdopen() can be both read and written.");
		start
			o_read_tst(Strlen);
			Zfseek(sfd, 0L, SEEK_CUR);	/*	Input to Output. */

			expecting(Strlen);
			Zfwrite(String, 1, Strlen, sfd);
			rewind(sfd);

			o_read_tst(2 * Strlen);
		finis

		o_clean();
	finis

	assertion(8, BASE);
	start
		o_creat();

		prep("Open test file with mode of \"w+\".");
		start
			stat(path, &st_buf1);
			fd = Zopen(path, O_RDWR);
			sfd = Zfdopen(fd, "w+");
			fd2 = fileno(sfd);
		finis

		testing("File length remains unchanged by fdopen().");
		start
			fstat(fd2, &st_buf2);
			if(st_buf1.st_size != st_buf2.st_size)
				error("%s\n%s%lu\n%s%lu\n",
				"File length was changed after fdopen().",
				"Size of file before fdopen() is: ",
				(unsigned long) st_buf1.st_size,
				"Size of file after fdopen() is: ",
				(unsigned long) st_buf2.st_size);
		finis

		testing("Stream returned by fdopen() can be both read and written.");
		start
			o_read_tst(Strlen);
			Zfseek(sfd, 0L, SEEK_CUR);	/*	Input to Output. */

			expecting(Strlen);
			Zfwrite(String, 1, Strlen, sfd);
			rewind(sfd);

			o_read_tst(2 * Strlen);
		finis

		o_clean();
	finis

	assertion(9, BASE);
	start
		o_creat();

		prep("Open test file with mode of \"a+\".");
		start
			fd = Zopen(path, O_RDWR | O_APPEND);
			sfd = Zfdopen(fd, "a+");
		finis

		testing("Test file can be both written and read at end-of-file.");
		start
			expecting(Strlen);
			Zfwrite(String, 1, Strlen, sfd);
			rewind(sfd);
			o_read_tst(Strlen);

			rewind(sfd);
			expecting(Strlen);
			Zfwrite(String, 1, Strlen, sfd);
			rewind(sfd);
			expecting(Strlen);
			Zfwrite(String, 1, Strlen, sfd);
			Zfclose(sfd);
		finis

		testing("All prior writes were at the end-of-file.");
		start
			sfd = Zfopen(path, "r" );
			o_read_tst(Strlen *4); 
		finis

		o_clean();
	finis

	assertion(10, BASE);
	start
		long int	pos, spos;

		o_creat();

		prep("Open test file and write to it.");
		start
			fd = Zopen(path, O_RDWR);
			Zwrite(fd, String, Strlen);
		finis

		testing("File position of file descriptor is associated with");
		testing("file position of stream.");
		start
			pos = (long int) lseek(fd, (off_t)0, SEEK_CUR);

			sfd = Zfdopen(fd, "w+");
			spos = ftell(sfd);

			if(spos != pos)
				error("%s\n%s%ld\n%s%ld\n",
					"File position of stream and file descriptor are not equal",
					"Stream file offset is: ",
					spos,
					"File descriptor offset is: ",
					pos);
		finis

		o_clean();
	finis

	assertion(11, BASE);
	start
		o_creat();

		prep("Open and read entire test file.");
		start
			fd = Zopen(path, O_RDWR);
			expecting(Strlen);
			Zread(fd, inbuf, Strlen);
		finis

		testing("Call to fdopen() creates a stream with the error");
		testing("indicator and end-of-file indicator clear.");
		start
			sfd = Zfdopen(fd, "w+");
			if(feof(sfd) != 0)
				error("Call to feof() returned non-zero.");
			if(ferror(sfd) != 0)
				error("Call to ferror() returned non-zero.");
		finis

		o_clean();
	finis

	assertion(12, BASE);
	start
		testing("This assertion is tested in assertions #4-#10.");
		start
		finis
	finis

	assertion(13, EXTENDED);

	done();
}

/*
 *	Create a data file and place a specific set of data in it.
*/

void o_creat(void)
{
	prep("Create a test file containing data.");
	start
		fd = Zcreat(path, PROT_ALL);
		Zwrite(fd, String, Strlen);
		Zclose(fd);
	finis
}


/*	Close and unlink test file */
void o_clean()
{
	cleanup("Close and unlink test file.");
	start
		Zclose(fd);
		Zunlink(path);
	finis
}


/*	Read data test */
void o_read_tst(int cnt)
{
	char	rd_buf[Strlen];
	int		i;

	for(i=Strlen -1; i<cnt; i+=Strlen) {
		expecting(Strlen);
		Zfread(rd_buf, 1, Strlen, sfd);
	}
}
