#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME "8.1.11_05"
#define FIFONAME "8.1.11_05_fifo"
#define	BUF_SIZE 512


int main(void)
{
	char	path[_POSIX_PATH_MAX +1], fifop[_POSIX_PATH_MAX +1];
	char	buffer[BUF_SIZE];
	int		pipfd[2];
	char	*outbuf;
	int 	fifo_rd, fifo_wr;
	size_t	pipe_buf_sz, fifo_buf_sz, max_buf_sz, pipe_max, fifo_max;
	pid_t	pid;
	int 	fd;
	FILE	*sfd;
	int		i;
	long	nbytes;
	int		flags;
	long	nb_wr, nb_to_wr;

	setup(E_FFLUSH);
	funct(E_FFLUSH);

	exit_on(ERROR);

								/*	These assertion tests contain code similar
									to those of fclose(). */
	assertion(5, BASE);
	start
		prep("Establish path to FIFO and create it");
		start
			Zstrcpy(fifop, temp_path());
			Zstrcat(fifop, FIFONAME);
			Zmkfifo(fifop, PROT_ALL);
		finis

							/* Maximum bytes written in pipe and FIFO */
		max_buf_sz=g_max_pipe(&pipe_buf_sz, &fifo_buf_sz, &pipe_max, &fifo_max);

		prep("Initialize input and output data buffers and create a");
		prep("regular file type.");
		start
			outbuf = (char *) Zmalloc(max_buf_sz);

			for(i = 0; i < max_buf_sz; ++i)
				outbuf[i] = i % CHAR_MAX;

			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			Zclose(Zcreat(path, PROT_ALL));
		finis

		FORK
			prep("Create a pipe with O_NONBLOCK flag set.");
			start
				Zpipe(pipfd);
				flags = Zfcntl( pipfd[1], F_GETFL);
				Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
			finis

			prep("Fill up pipe.");
			start
						/*	Must eventually write one byte at a time to ensure
							that if data can be placed in pipe it will be
							transferred to the pipe for all POSIX.1
							implementations.
						*/
				SET_ALARM(5 * TIMEOUT)
					for(nb_to_wr=pipe_buf_sz; nb_to_wr; nb_to_wr /= 2)
						while((nb_wr = write(pipfd[1], outbuf, nb_to_wr)) != -1)
							nbytes += nb_wr;
				CLEAR_ALARM
			finis

			prep("Set up a buffered stream which writes to the filled pipe.");
			start
				sfd = Zfdopen(pipfd[1], "w");
				setvbuf(sfd, buffer, _IOFBF, _POSIX_PIPE_BUF);
				expecting(_POSIX_PIPE_BUF/2);
				Zfwrite(outbuf, 1, _POSIX_PIPE_BUF/2, sfd);
			finis

			testing("An fflush() on the pipe when insufficient capacity");
			testing("exists to accept the buffered data.");
			start
				SET_ALARM(TIMEOUT)
					expecting(EOF);
					expecting(EAGAIN);
					Zfflush(sfd);
				CLEAR_ALARM
			finis
	
			prep("Open FIFO and set O_NONBLOCK flag.");
			start
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY | O_NONBLOCK);
			finis

			prep("Fill up FIFO.");
			start
						/*	Must eventually write one byte at a time to ensure
							that if data can be placed in FIFO it will be
							transferred to the FIFO for all POSIX.1
							implementations.
						*/
				SET_ALARM(5 * TIMEOUT)
					for(nb_to_wr=fifo_buf_sz; nb_to_wr; nb_to_wr /= 2)
						while((nb_wr = write(fifo_wr, outbuf, nb_to_wr)) != -1)
							nbytes += nb_wr;
				CLEAR_ALARM
			finis

			prep("Set up a buffered stream which writes to the filled FIFO.");
			start
				sfd = Zfdopen(fifo_wr, "w");
				setvbuf(sfd, buffer, _IOFBF, _POSIX_PIPE_BUF);
				expecting(_POSIX_PIPE_BUF/2);
				Zfwrite(outbuf, 1, _POSIX_PIPE_BUF/2, sfd);
			finis

			testing("An fflush() on the FIFO when insufficient capacity");
			testing("exists to accept the buffered data.");
			start
				SET_ALARM(TIMEOUT)
					expecting(EOF);
					expecting(EAGAIN);
					Zfflush(sfd);
				CLEAR_ALARM
			finis
		KROF
	finis

	assertion(6, BASE);
	start
		FORK
							/*	POSIX.1-1992 specifies the underlying functions
								for fflush() as write() and lseek().  Setup the
								test case to ensure buffer does exist in the
								stream when fflush() is called on a file
								descriptor that is closed.
							*/
			prep("Setup a stream buffer containing some data.");
			start
				fd = Zopen(path, O_RDWR | O_CREAT, PROT_ALL);
				sfd = Zfdopen(fd, "r+");
				setvbuf(sfd, buffer, _IOFBF, _POSIX_PIPE_BUF);
				Zfputc((int)'D', sfd);
			finis

			testing("When the stream pointer argument addresses a file");
			testing("descriptor that is closed, a call to fflush() returns");
			testing("a value of [EOF] and sets errno to [EBADF].");
			start
				Zclose(fd);
				expecting(EOF);
				expecting(EBADF);
				Zfflush(sfd);
			finis
		KROF
	finis


							/*	When fflush() is terminated due to receipt
								of a signal and no data was transferred, then
								fflush() returns EOF and sets errno to [EINTR].
							*/
	assertion(7, BASE);
	start
		FORK
			prep("Create a pipe with O_NONBLOCK flag clear and fill it up.");
			start
				Zpipe(pipfd);
						/*	Must eventually write one byte at a time to ensure
							that if data can be placed in pipe it will be
							transferred to the pipe for all POSIX.1
							implementations.
						*/
				SET_ALARM(5 * TIMEOUT)
					flags = Zfcntl( pipfd[1], F_GETFL);
					Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
					for(nb_to_wr=pipe_buf_sz; nb_to_wr; nb_to_wr /= 2)
						while((nb_wr = write(pipfd[1], outbuf, nb_to_wr)) != -1)
							nbytes += nb_wr;
					Zfcntl(pipfd[1], F_SETFL, flags & ~O_NONBLOCK);
				CLEAR_ALARM
			finis

			prep("Set up a buffered stream which writes to the filled pipe.");
			start
				sfd = Zfdopen(pipfd[1], "w");
				setvbuf(sfd, buffer, _IOFBF, _POSIX_PIPE_BUF);
				expecting(_POSIX_PIPE_BUF/2);
				Zfwrite(outbuf, 1, _POSIX_PIPE_BUF/2, sfd);
			finis

			testing("When fflush() is terminated by a signal,");
			testing("fflush() returns EOF and sets errno to [EINTR].");
			start
				SIGNAL(SIGALRM, sig_int);
				alarm(2);
					expecting(EOF);
					expecting(EINTR);
					Zfflush(sfd);
				alarm(0);
			finis
		KROF
	finis


	assertion(8, EXTENDED);
	assertion(9, EXTENDED);

	assertion(10, BASE);
	start
		FORK
											/* pipe test */
			prep("Catch signal, create pipe, buffer data, and close read");
			prep("descriptor.");
			start
				SIGNAL(SIGPIPE, sig_int);
				Zpipe(pipfd);
				sfd = Zfdopen(pipfd[1], "w");
				Zsetvbuf(sfd, buffer, _IOFBF, BUF_SIZE);
				Zfwrite("S", sizeof(char), 1, sfd);
				Zclose(pipfd[0]);
			finis
	
			testing("Call to fflush(), when attempting to write to a pipe");
			testing("that is not open for reading by any process.");
			start
				sig_rcvd = 0;
	
				SET_ALARM(TIMEOUT)
					errno = 0;
					expecting(EOF);
					expecting(EPIPE);
					Zfflush(sfd);
				CLEAR_ALARM
	
				if(sig_rcvd == 0)
					error("Did not receive signal SIGPIPE.");
			finis
	
											/* FIFO test */
			prep("Catch signal, create FIFO, buffer data, and close read");
			prep("descriptor.");
			start
				SIGNAL(SIGPIPE, sig_int);
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
				sfd = Zfdopen(fifo_wr, "w");
				Zsetvbuf(sfd, buffer, _IOFBF, BUF_SIZE);
				Zfwrite("S", sizeof(char), 1, sfd);
				Zclose(fifo_rd);
			finis
	
			testing("Call to fflush(), when attempting to write to a FIFO");
			testing("that is not open for reading by any process.");
			start
				sig_rcvd = 0;
	
				SET_ALARM(TIMEOUT)
					errno = 0;
					expecting(EOF);
					expecting(EPIPE);
					Zfflush(sfd);
				CLEAR_ALARM
	
				if(sig_rcvd == 0)
					error("Did not receive signal SIGPIPE.");
			finis
		KROF

		cleanup("Remove FIFO file.");
		start
			Zunlink(fifop);
		finis
	finis

	done();
}
