#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define	FILENAME	"8.1.14_20"
#define	String	"Some data"
#define	Strlen	(sizeof(String) -1)


int	main(void)
{
	char	path[_POSIX_PATH_MAX +1];

	setup(E_FOPEN);
	funct(E_FOPEN);

	assertion(20, BASE);
	start
		prep("Initialize path name for a file where search permission");
		prep("is denied on a component of the path prefix.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perm.d");
			Zstrcat(path, "/fopen_tryme");
		finis

		testing("Expecting [EACCES] result.");
		start
			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "r");
		finis
	finis

	assertion(21, BASE);
	start
		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perms.d");
			Zstrcat(path, "/no_read");
		finis

		testing("Read permission along the path is denied.");
		start
			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "r");
		finis

		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_perms.d");
			Zstrcat(path, "/no_write");
		finis

		testing("Write permission along the path is denied.");
		start
			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "w");

			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "w+");

			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "wb");

			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "wb+");

			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "w+b");
		finis
	finis


	assertion(22, BASE);
	start
		prep("Initialize a path name.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_write.d");
			Zstrcat(path, "/new_file");
		finis

		testing("When file does not exist and parent directory does not");
		testing("permit writing [EACCES] expected.");
		start
			expecting(NULL);
			expecting(EACCES);
			Zfopen(path, "w");
		finis

		testing("File is not created.");
		start
			expecting(FAILURE);
			Zaccess(path, F_OK);
		finis
	finis

	assertion(23, BASE);
	start
		pid_t	pid;
		time_t	s_time;
		struct	stat st_buf;

		prep("Create a FIFO file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "er_eintr");
			Zmkfifo(path, PROT_ALL);

			Zstat(path, &st_buf);			/*	Set base-line times for FIFO */
			Zstrcpy(path, temp_path());
			Zstat(path, &st_buf);			/*	and parent directory.		*/

			Zstrcpy(path, temp_path());
			Zstrcat(path, "er_eintr");

			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis


		ENABLE_PSYNC
		prep("Fork a child.");
		start
			sv_flush( );
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {		/* Child */
			exit_on(ERROR);
			SIGNAL(SIGINT, sig_int);

			testing("On fopen() interrupted, errno set to EINTR.");
			start
				SEND_PSYNC
				expecting(NULL);
				expecting(EINTR);
				Zfopen(path, "w");
			finis

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		prep("Send a signal to the child.");
		start
			SLEEP(WAIT_FOR_TIME);
			Zkill(pid, SIGINT);
		finis

		WAIT_FOR_PID(pid);
		DISABLE_PSYNC

		testing("Time related fields \"st_ctime\", \"st_mtime\", and");
		testing("\"st_atime\" of the file and \"st_ctime\", \"st_mtime\" of");
		testing("of the parent directory are not updated.");
		start
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);

			Zstrcpy(path, temp_path());
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	assertion(24, BASE);
	start
		time_t	s_time;
		struct	stat st_buf;

		prep("Initialize a path name.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "open_dir");
			Zmkdir(path, PROT_ALL);
			Zstat(path, &st_buf);	/*	Set base-line time for created file. */
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis

		testing("Named file is a directory and type is set for writing.");
		start
			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "w");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "a");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "r+");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "w+");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "a+");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "wb");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "ab");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "r+b");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "rb+");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "w+b");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "wb+");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "a+b");

			expecting(NULL);
			expecting(EISDIR);
			Zfopen(path, "ab+");
		finis

		testing("Time related fields \"st_ctime\", \"st_mtime\", and");
		testing("\"st_atime\" of the file are not updated.");
		start
			trf_ck(CHK_CTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, FALSE, s_time, NO_FILDES, path);
			trf_ck(CHK_ATIME, FALSE, s_time, NO_FILDES, path);
		finis
	finis

	assertion(25, BASE);
	start
		int		open_max, topen_max, i;
		char	path_t[_POSIX_PATH_MAX +1];

		break_on(ERROR);

		prep("Create a temporary file with data for possible use later on.");
		start
			int	fd;

			Zstrcpy(path_t, temp_path());
			Zstrcat(path_t, FILENAME);
			fd = Zcreat(path_t, PROT_ALL);
			Zwrite(fd, String, Strlen);
			Zclose(fd);
		finis

		prep("Initialize a path name.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "fill_up");
		finis

		open_max = g_open_max();	/* Get value for {OPEN_MAX} */
									/* Get testable value for {OPEN_MAX} */
		topen_max = (open_max <= PCTS_OPEN_MAX) ? open_max : PCTS_OPEN_MAX;

		prep("Close all file descriptors except 0, 1 and 2.");
		start
			for(i=3; i<topen_max; i++)
				(void) close(i);
		finis

		if(open_max <= PCTS_OPEN_MAX) {
			prep("{OPEN_MAX} equal to {PCTS_OPEN_MAX}.");
			prep("Consume all available file descriptors.");
			start
				for(i=3; i<open_max; i++)
					(void) Zcreat(path, PROT_ALL);
			finis

			prep("Initialize a path name and create.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, "emfile_tst");
			finis
	

			testing("Call fopen() when {OPEN_MAX} file descriptors are in use");
			testing("and the file to be opened does not exist.");
			start
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "w");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "a");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "w+");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "a+");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "wb");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "ab");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "wb+");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "ab+");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "w+b");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path, "a+b");
			finis

			testing("File is not created.");
			start
				expecting(FAILURE);
				Zaccess(path, F_OK);
			finis

			testing("Call fopen() when {OPEN_MAX} file descriptors are in use");
			testing("and the file to be opened exists.");
			start
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path_t, "w");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path_t, "w+");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path_t, "wb");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path_t, "wb+");
	
				expecting(NULL);
				expecting(EMFILE);
				Zfopen(path_t, "w+b");
			finis

			testing("File is not truncated.");
			start
				struct	stat st_buf;

				stat(path_t, &st_buf);
				if(st_buf.st_size == (off_t)0)
					error("File has been truncated.");
			finis
		}
		else {
			prep("Testing for {PCTS_OPEN_MAX} < {OPEN_MAX}.");
			prep("Consume PCTS_OPEN_MAX -1 file descriptors."); 
			start
				for(i=3; i<(topen_max -1); i++)
					(void) Zcreat(path, PROT_ALL);
			finis

			prep("Initialize a path name.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, "emfile_tst");
				Zclose(Zcreat(path, PROT_ALL));
			finis
	
			testing("fopen() on the {PCTS_OPEN_MAX} -1 file descriptor.");
			start
				Zfopen(path, "r+");
			finis
		}

		cleanup("Close all file descriptors except 0, 1 and 2.");
		start
			for(i=3; i<topen_max; i++)
				close(i);
		finis
	finis

	if(tc_name_max()) {
		assertion(26, BASE);
		start
			e_NO_TRUNC(E_FOPEN);
		finis
	}
	else
		assertion(26, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(27, BASE);
		start
			e_NO_TRUNC(E_FOPEN);
		finis
	}
	else
		assertion(27, NOT_TESTABLE);

	if(tc_path_max()) {
		assertion(28, BASE);
		start
			e_ENAMETOOLONG(E_FOPEN);
		finis
	}
	else
		assertion(28, NOT_TESTABLE);

	assertion(29, EXTENDED);

	assertion(30, BASE);
	start
		prep("Setup a pathname for a file which does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_fopen_");
		finis

		testing("Named file does not exist.");
		start
			expecting(NULL);
			expecting(ENOENT);
			Zfopen(path, "r");

			expecting(NULL);
			expecting(ENOENT);
			Zfopen(path, "r+");

			expecting(NULL);
			expecting(ENOENT);
			Zfopen(path, "rb");

			expecting(NULL);
			expecting(ENOENT);
			Zfopen(path, "r+b");

			expecting(NULL);
			expecting(ENOENT);
			Zfopen(path, "rb+");
		finis
	finis

	assertion(31, BASE);
	start
		testing("Call to fopen() when a path prefix does not exist.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "_fopen_/test");

			expecting(NULL);
			expecting(ENOENT);
			Zfopen(path, "w");
		finis

		testing("Call to fopen() when path points to an empty string.");
		start
			expecting(NULL);
			expecting(ENOENT);
			Zfopen("", "w");
		finis
	finis

	assertion(32, EXTENDED);

	assertion(33, BASE);
	start
		testing("Call to fopen() when path contains a non-directory component.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "Fopen_Test");
			Zstrcat(path, "/dummy");

			expecting(NULL);
			expecting(ENOTDIR);
			Zfopen(path, "r");
		finis
	finis

	done();

}
