#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME	"8.1.21_04"
#define	String		"DaTa\n"
#define	Strlen		(sizeof(String) -1)


int	main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	FILE	*sfd;
	pid_t	pid;
	time_t	s_time;
	int		pipfd[2];
	char	*buf_out;
	int		i;
	size_t	pipe_buf_sz, fifo_buf_sz, max_buf_sz, pipe_max, fifo_max;
	size_t	n_bytes;
	int		flags;

	setup(E_FWRITE);
	funct(E_FWRITE);

								/*	This assertion test contains code similar
									to that of fputc(). */
	assertion(4, BASE);
	start
						/* Determine maximum bytes writable to pipe and FIFO. */
						/* The g_max_pipe() info is needed for assertion #5. */
		max_buf_sz=g_max_pipe(&pipe_buf_sz, &fifo_buf_sz, &pipe_max, &fifo_max);

		prep("Establish path to test file and create it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			sfd = Zfopen(path, "w");
		finis

		testing("A call to fwrite() marks for update the \"st_ctime\" and");
		testing("\"st_mtime\" time-related fields of the underlying file");
		testing("when data is caused to be written.");
		testing("Call to fwrite() used in conjunction with fflush().");
		start
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			Zfwrite(String, sizeof(char), Strlen, sfd);
			Zfflush(sfd);
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, path);
		finis

		testing("A call to fwrite() marks for update the \"st_ctime\" and");
		testing("\"st_mtime\" time-related fields of the underlying file");
		testing("when data is caused to be written.");
		testing("Call to fwrite() used in conjunction with fclose().");
		start
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			Zfwrite(String, sizeof(char), Strlen, sfd);
			Zfclose(sfd);
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, path);
		finis

		prep("Establish path to test file and create it.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			sfd = Zfopen(path, "w");
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
		finis
	
		FORK
			prep("Call to fwrite() used in conjunction with exit().");
			start
				Zfwrite(String, sizeof(char), Strlen, sfd);
			finis
			done_e();					/* done_e() calls exit(). */
		KROF

		testing("A call to fwrite() marks for update the \"st_ctime\" and");
		testing("\"st_mtime\" time-related fields of the underlying file");
		testing("when data is caused to be written.");
		start
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, path);
		finis
	
		prep("Establish minimum reference time.");
		prep("Catch SIGABRT signal.");
		start
			Ztime(&s_time);				/* Obtain minimum reference time. */
			dt_secs(TRF_DELAY);			/* Delay a fixed interval.			*/
			SIGNAL(SIGABRT, sig_int);
		finis
	
		FORK
			prep("Call to fwrite() used in conjunction with abort().");
			start
				Zfwrite(String, sizeof(char), Strlen, sfd);
			finis
			abort();
		KROF

		testing("A call to fwrite() marks for update the \"st_ctime\" and");
		testing("\"st_mtime\" time-related fields of the underlying file");
		testing("when data is caused to be written.");
		start
			trf_ck(CHK_CTIME, TRUE, s_time, NO_FILDES, path);
			trf_ck(CHK_MTIME, TRUE, s_time, NO_FILDES, path);
		finis
	finis



	if((pipe_buf_sz > PCTS_PIPE_BUF) || (fifo_buf_sz > PCTS_PIPE_BUF)) {
		assertion(5, NOT_TESTABLE);
		done();
	}

	assertion(5, BASE);
	start
		FORK
			prep("Allocate an output buffer.");
			start
				buf_out = (char *) Zmalloc(pipe_buf_sz +1);
				for(i=0; i<pipe_buf_sz +1; ++i)
					buf_out[i] = i % CHAR_MAX;
			finis
														/* pipe test */
			prep("Create a pipe and set O_NONBLOCK flag on write channel.");
			start
				Zpipe(pipfd);
				flags = Zfcntl(pipfd[1], F_GETFL);
				Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
				sfd = fdopen(pipfd[1], "w");
				Zsetvbuf(sfd, NULL, _IONBF, 0);		/* Unbuffered stream. */
			finis

			prep("Write to pipe, if needed, so that an additional write");
			prep("of {PIPE_BUF} will exceed the data that can \"currently\"");
			prep("be written to the pipe.");
			start
				SET_TIMEOUT
									/* Fill up pipe (almost). */
					for(i=pipe_buf_sz; i<pipe_max; i+=pipe_buf_sz) { 
						expecting(pipe_buf_sz);
						n_bytes = Zfwrite(buf_out, sizeof(char),
														pipe_buf_sz,sfd);
						message("fwrite() to pipe wrote additional %lu bytes.",
											(unsigned long)(n_bytes));
					}
				CLEAR_ALARM
			finis

			testing("When fwrite() writes greater than zero bytes but fewer");
			testing("than requested, then the error indicator for the stream");
			testing("is set.");
			start
				SET_ALARM(WAIT_FOR_TIME)
							/* On some systems, the pipe_max size computed
								by this PCTS is that of the "low water mark"
								for the pipe.  Therefore, additional calls
								to write() are performed, if necessary, to
								get beyond the "high water mark" for the pipe."
							*/
					while( (n_bytes = fwrite(buf_out, sizeof(char),
														pipe_buf_sz +10, sfd))
							== pipe_buf_sz +10) {
						message("fwrite() to pipe wrote additional %lu bytes.",
											(unsigned long)(n_bytes));
					}

					if(n_bytes == 0)
						error("Last fwrite(), to pipe, wrote no data.");
					else
						message("Last fwrite() to pipe wrote %lu bytes.",
									(unsigned long)(n_bytes));
					if(!ferror(sfd))
						error("Error indicator for stream is not set.");
				CLEAR_ALARM
			finis
		KROF
	finis

	done();
}
