#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FIFONAME "8.1.37_05_fifo"
#define FILENAME "8.1.37_05"


int main(void)
{
	pid_t	pid;
	int		pipfd[2];
	char	buf_in[10];
	FILE	*sfd;
	int		fd, tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;

	setup(E_SCANF);
	funct(E_SCANF);

	exit_on(ERROR);

								/*	These assertion tests contain code similar
									to those of fgetc(). */
	assertion(5, BASE);
	start
		int		flags;
		int 	fifo_rd, fifo_wr;
		char	fifop[_POSIX_PATH_MAX +1];

		FORK
			prep("Establish a pipe and set the O_NONBLOCK flag.");
			start
				Zpipe(pipfd);
				flags = Zfcntl(pipfd[0], F_GETFL);
				Zfcntl(pipfd[0], F_SETFL, flags | O_NONBLOCK);
			finis

			prep("Set underlying file descriptor of stdin to pipfd[0].");
			start
				if(STDIN_FILENO != (fd = Zdup2(pipfd[0], fileno(stdin))))
					error("%s\n%s%d\n%s%d",
							"Returned file descriptor was not STDIN_FILENO.",
							"Expected: ",
							STDIN_FILENO,
							"Actual: ",
							fd);
				Zfdopen(STDIN_FILENO, "r");
			finis

			testing("scanf() from an empty pipe with O_NONBLOCK set.");
			start
				expecting(EOF);
				expecting(EAGAIN);
				Zscanf("%s", buf_in);
			finis
	
			prep("Open FIFO for read and write.");
			prep("Set O_NONBLOCK on read channel.");
			start
				Zstrcpy(fifop, temp_path());
				Zstrcat(fifop, FIFONAME);
				Zmkfifo(fifop, PROT_ALL);
				fifo_rd = Zopen(fifop, O_RDONLY | O_NONBLOCK);
				fifo_wr = Zopen(fifop, O_WRONLY);
			finis
	
			prep("Set underlying file descriptor stdin to FIFO read channel.");
			start
				if(STDIN_FILENO != (fd = Zdup2(fifo_rd, fileno(stdin))))
					error("%s\n%s%d\n%s%d",
							"Returned file descriptor was not STDIN_FILENO.",
							"Expected: ",
							STDIN_FILENO,
							"Actual: ",
							fd);
				Zfdopen(STDIN_FILENO, "r");
			finis

			testing("scanf() from an empty FIFO with O_NONBLOCK set.");
			start
				expecting(EOF);
				expecting(EAGAIN);
				Zscanf("%s", buf_in);
			finis

			cleanup("Remove FIFO file.");
			start
				Zunlink(fifop);
			finis
		KROF
	finis
	
	assertion(6, BASE);
	start
		char	path[_POSIX_PATH_MAX +1];

		FORK
			testing("Stream pointer argument addresses a file descriptor that");
			testing("is not open for reading.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, FILENAME);

				SET_ALARM(WAIT_FOR_TIME)
					fd = Zcreat(path, PROT_ALL);
					if(STDIN_FILENO != (fd = Zdup2(fd, fileno(stdin))))
						error("%s\n%s%d\n%s%d",
								"Returned file descriptor not STDIN_FILENO.",
								"Expected: ",
								STDIN_FILENO,
								"Actual: ",
								fd);

					expecting(EOF);
					expecting(EBADF);
					Zscanf("%s", buf_in);
				CLEAR_ALARM
			finis
		KROF
	finis

	assertion(7, BASE);
	start
		FORK
			prep("Establish a pipe for testing.");
			start
				Zpipe(pipfd);
				if(STDIN_FILENO != (fd = Zdup2(pipfd[0], fileno(stdin))))
					error("%s\n%s%d\n%s%d",
							"Returned file descriptor was not STDIN_FILENO.",
							"Expected: ",
							STDIN_FILENO,
							"Actual: ",
							fd);
				Zfdopen(pipfd[0], "r");
			finis

			testing("scanf() interrupted by signal before reading any data.");
			start
				SIGNAL(SIGALRM, sig_int);
				alarm(2);
				expecting(EOF);
				expecting(EINTR);
				Zscanf("%s", buf_in);
				alarm(0);
			finis
		KROF
	finis
	
	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(8, BASE);
		GTI_NOT_PROVIDED_RESULT(9, BASE);
		done();
	}

	assertion(8, BASE);
	start
		sigset_t	set;

		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		start
			exit_on(ERROR);
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);
			sfd = Zfdopen(tty_fildes, "r+");
		finis

		FORK
			prep("Establish this process as a background process group.");
			start
				Zsetpgid((pid_t)0, getpid());
				if(STDIN_FILENO != (fd = Zdup2(tty_fildes, fileno(stdin))))
					error("%s\n%s%d\n%s%d",
							"Returned file descriptor was not STDIN_FILENO.",
							"Expected: ",
							STDIN_FILENO,
							"Actual: ",
							fd);
				Zfdopen(tty_fildes, "r+");
			finis
	
			testing("Attempt to scanf() from controlling terminal, when");
			testing("process is ignoring the SIGTTIN signal, returns");
			testing("a NULL pointer and sets errno to [EIO].");
			start
				SIGNAL(SIGTTIN, SIG_IGN);
				expecting(EOF);
				expecting(EIO);
				Zscanf("%s", buf_in);
			finis
	
			testing("Attempt to scanf() from controlling terminal, when");
			testing("process is blocking the SIGTTIN signal, returns");
			testing("a NULL pointer and sets errno to [EIO].");
			start
				SIGBLOCK(set, SIGTTIN);
				expecting(EOF);
				expecting(EIO);
				Zscanf("%s", buf_in);
			finis
		KROF
	finis
	
	assertion(9, BASE);
	start
		FORK
			FORK
				prep("Establish this process as background process group.");
				prep("Exit on a SIGHUP signal.");
				start
					Zsetpgid((pid_t)0, (pid_t)0);
					SIGNAL(SIGHUP, SIG_DFL);
					if(STDIN_FILENO != (fd = Zdup2(tty_fildes, fileno(stdin))))
						error("%s\n%s%d\n%s%d",
							"Returned file descriptor was not STDIN_FILENO.",
							"Expected: ",
							STDIN_FILENO,
							"Actual: ",
							fd);
					Zfdopen(tty_fildes, "r+");
				finis

				FORK
					prep("Ignore the SIGHUP signal.");
					start
						SIGNAL(SIGHUP, SIG_IGN);
					finis
	
					prep("Kill the grandchild (process group leader) to");
					prep("create an orphaned process group.");
					start
						Zkill(getppid(), SIGHUP);
						sleep(WAIT_FOR_TIME);
					finis
	
					testing("Call to scanf() for controlling terminal.");
					start
						expecting(EOF);
						expecting(EIO);
						Zscanf("%s", buf_in);
					finis
				KROF
			KROF
		KROF
		SLEEP(2 * WAIT_FOR_TIME);		/* Ensure test completes. */
	finis
	
	cleanup("Close terminal devices.");
	start
		SIGNAL(SIGHUP, SIG_IGN);
		Zclose(tty_fildes);
		Zclose(loop_fildes);
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis
	
	done();
}
