#include	<pcts.h>

#include	<Zunistd.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>

#define FILENAME "8.1.29_05"

static	struct	stat st_buf;
void	trf_cks(int, time_t);
void vprtargs(char *, ...);


int	main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	FILE	*sfd;
	int		fd, fd_stdout;
	int		stat_info;
	pid_t	pid;
	time_t	s_time;

	setup(E_VPRINTF);
	funct(E_VPRINTF);

	exit_on(ERROR);
									/* 	Code similar to that for printf[4]. */
	assertion(5, BASE);
	start
		prep("Establish path for temporary stdout.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
		finis

		FORK
			testing("Call to vprintf() marks for update the \"st_ctime\" and");
			testing("\"st_mtime\" time-related fields of the underlying file");
			testing("when data is caused to be written.");
			testing("Call to vprintf() used in conjunction with fflush().");
			start
							/*
								Redirect stdout retaining journal file.
								Get stdout as stream descriptor for the pipe.
								Lose automatic error checking.
							*/
				fflush(stdout);				/* Flush outbut buffer. */
				fd_stdout = Zdup(STDOUT_FILENO);
				sfd = freopen(path, "w", stdout);

				Ztime(&s_time);				/* Obtain minimum reference time. */
				dt_secs(TRF_DELAY);			/* Delay a fixed interval.	*/
				vprtargs("%s %s\n", "Do", "it!");
				fflush(sfd);				/* Mark for update. */
				dt_secs(TRF_DELAY);			/* Delay a fixed interval.	*/
				fstat(STDOUT_FILENO, &st_buf);

							/* Reset stdout back to journal file */
				fclose(sfd);
				dup2(fd_stdout, STDOUT_FILENO);

				trf_cks(FALSE, s_time);
			finis
		KROF

		FORK
			testing("Call to vprintf() marks for update the \"st_ctime\" and");
			testing("\"st_mtime\" time-related fields of the underlying file");
			testing("when data is caused to be written.");
			testing("Call to vprintf() used in conjunction with fclose().");
			start
				fflush(stdout);				/* Flush outbut buffer. */
				fd_stdout = Zdup(STDOUT_FILENO);
				sfd = freopen(path, "w", stdout);

				Ztime(&s_time);				/* Obtain minimum reference time. */
				dt_secs(TRF_DELAY);			/* Delay a fixed interval.	*/
				vprtargs("%s %s\n", "Do", "it!");
				fclose(sfd);				/* Mark for update. */
				dt_secs(TRF_DELAY);			/* Delay a fixed interval.	*/
				stat(path, &st_buf);

				dup2(fd_stdout, STDOUT_FILENO);

				trf_cks(FALSE, s_time);
			finis
		KROF

		ENABLE_PSYNC
		FORK
			prep("Call to vprintf() used in conjunction with exit().");
			start
				fflush(stdout);				/* Flush outbut buffer. */

				if(fork() == 0) {
					freopen(path, "w", stdout);
					Ztime(&s_time);		/* Obtain minimum reference time. */
					write(p_syn[1], (time_t *)&s_time, sizeof(time_t));
					dt_secs(TRF_DELAY);		/* Delay a fixed interval.	*/
				vprtargs("%s %s\n", "Do", "it!");
					exit(0);				/* Update time-related fields. */
				}
				wait(&stat_info);
				dt_secs(TRF_DELAY);			/* Delay a fixed interval.	*/

				stat(path, &st_buf);
			finis
	
			testing("Call to vprintf() marks for update the \"st_ctime\" and");
			testing("\"st_mtime\" time-related fields of the underlying file");
			testing("when data is caused to be written.");
			start
				expecting(sizeof(time_t));
				Zread(p_syn[0], (time_t *)&s_time, sizeof(time_t));
				trf_cks(TRUE, s_time);
			finis
		KROF
		
		prep("Catch SIGABRT signal.");
		start
			SIGNAL(SIGABRT, sig_int);
		finis
		
		FORK
			prep("Call to vprintf() used in conjunction with abort().");
			start
				fflush(stdout);				/* Flush outbut buffer. */

				if(fork() == 0) {
					freopen(path, "w", stdout);
					Ztime(&s_time);		/* Obtain minimum reference time. */
					write(p_syn[1], (time_t *)&s_time, sizeof(time_t));
					dt_secs(TRF_DELAY);		/* Delay a fixed interval.	*/
				vprtargs("%s %s\n", "Do", "it!");
					abort();				/* Update time-related fields. */
				}
				wait(&stat_info);
				dt_secs(TRF_DELAY);			/* Delay a fixed interval.	*/

				stat(path, &st_buf);
			finis
	
			testing("Call to vprintf() marks for update the \"st_ctime\" and");
			testing("\"st_mtime\" time-related fields of the underlying file");
			testing("when data is caused to be written.");
			start
				expecting(sizeof(time_t));
				Zread(p_syn[0], (time_t *)&s_time, sizeof(time_t));
				trf_cks(TRUE, s_time);
			finis
		KROF

		DISABLE_PSYNC
	finis
	
	done();
}


void trf_cks(int updated, time_t s_time)
{
	/*
		updated == TRUE, time-related field is updated
		updated == FALSE, time-related field is marked for update
	*/

	time_t	cur_time;
	char	s1[26]; /* Avoid calling ctime() twice in single function call. */

	Ztime(&cur_time);

	strcpy(s1, ctime(&st_buf.st_mtime) +4);
	if(st_buf.st_mtime <= s_time)
		error("%s\n%s%20.20s.\n%s%20.20s.",
				"Time-related field \"st_mtime\" in error?.",
				"Expected minimum time is ",
				ctime(&s_time) +4,
				"Actual time recorded is ",
				s1);

	if(updated == TRUE) {
		if(st_buf.st_mtime >= cur_time)
			error("%s\n%s%20.20s.\n%s%20.20s.",
					"Time-related field \"st_mtime\" >= to current time.",
					"Current time is ",
					ctime(&cur_time) +4,
					"Time-related field \"st_mtime\" is ",
					s1);
	}
	else {
		if(st_buf.st_mtime > cur_time)
			error("%s\n%s%20.20s.\n%s%20.20s.",
					"Time-related field \"st_mtime\" larger than current time.",
					"Current time is ",
					ctime(&cur_time) +4,
					"Time-related field \"st_mtime\" is ",
					s1);
	}

	strcpy(s1, ctime(&st_buf.st_ctime) +4);
	if(st_buf.st_ctime <= s_time)
		error("%s\n%s%20.20s.\n%s%20.20s.",
				"Time-related field \"st_ctime\" in error?.",
				"Expected minimum time is ",
				ctime(&s_time) +4,
				"Actual time recorded is ",
				s1);

	if(updated == TRUE) {
		if(st_buf.st_ctime >= cur_time)
			error("%s\n%s%20.20s.\n%s%20.20s.",
					"Time-related field \"st_ctime\" >= to current time.",
					"Current time is ",
					ctime(&cur_time) +4,
					"Time-related field \"st_ctime\" is ",
					s1);
	}
	else {
		if(st_buf.st_ctime > cur_time)
			error("%s\n%s%20.20s.\n%s%20.20s.",
					"Time-related field \"st_ctime\" larger than current time.",
					"Current time is ",
					ctime(&cur_time) +4,
					"Time-related field \"st_ctime\" is ",
					s1);
	}
}


void vprtargs(char *format, ...)
{
	int		ret_val;
	va_list	ap;
	va_start(ap, format);

	errno = 0;
	ret_val = vprintf(format, ap);
	if(ret_val < 0)
		error("%s\n%s%s.",
				"Expected call to vprintf() to be successful.",
				"Actually vprintf() failed with errno set to ",
				err_lookup(errno));
	va_end(ap);
}
