#include	<pcts.h>

#include	<Zunistd.h>
#include	<sv_env.h>

/*	This module has S_ISGID of BIN_GID.  */

						/*	Define the absolute maximum number of
							supplementary groups an implementation
							may define for {NGROUPS_MAX}.
						*/
#define	MAX_SGA_NGROUPS_MAX	200


int	main(void)
{
	gid_t	gr_lst1[NGROUPS_MAX +MAX_SGA_NGROUPS_MAX];
	gid_t	gr_lst2[NGROUPS_MAX +MAX_SGA_NGROUPS_MAX];
	int		ngrps1, ngrps2, n_grps, i;

	ass_nu = 5;
	sv_config();					/* Required for sv_nproc() call.	  */
	sv_nproc();						/* Maintain proper journal ID sequence.*/

	prep("Ensure process's real group ID is GID");
	prep("and effective group ID is BIN_GID.");
	start
		exit_on(ERROR);

		expecting(GID);
		(void) Zgetgid();

		expecting(BIN_GID);
		(void) Zgetegid();
	finis

	prep("Set process's effective group ID to GID.");
	start
		(void) Zsetgid(GID);		/* Reference #1 */
		expecting(GID);
		(void) Zgetgid();
		expecting(GID);
		(void) Zgetegid();
	finis

	prep("Get IDs for supplementary groups inuse.");
	start
		exit_on(ERROR);
		n_grps = Zsysconf(_SC_NGROUPS_MAX);
		if(n_grps == 0) {
			un_implemented("{NGROUPS_MAX}");
		}
		ngrps1 = Zgetgroups(n_grps, gr_lst1);
		if(ngrps1 != n_grps)
			error("%s\n%s%d\n%s%d\n",
						"getgroups() returned incorrect count.",
						"Expected count of ",
						n_grps,
						"Actual count is ",
						ngrps1);
	finis

	testing("Process without appropriate privileges and only the saved");
	testing("set-group-ID matches the setgid() calling argument, then a call");
	testing("to setgid() returns zero and sets the effective group ID.");
	testing("The real group ID, saved set-group-ID, and any supplementary");
	testing("group IDs remain unchanged.");
	start
		int	egid_test = TRUE;

		expecting(0);
		(void) Zsetgid(BIN_GID);
		expecting(BIN_GID);			/* Effective group ID is set. */
		(void) Zgetegid();

		expecting(GID);				/* Real group ID remains unchanged. */
		(void) Zgetgid();

									/* Supplementary group IDs are unchanged. */
		ngrps2 = Zgetgroups(n_grps, gr_lst2);
		if(ngrps2 != n_grps)
			error("%s\n%s%d\n%s%d\n",
						"getgroups() returned incorrect count.",
						"Expected count of ",
						n_grps,
						"Actual count is ",
						ngrps2);

		for(i=0; i<n_grps; i++) {
			if(egid_test) {
				if(gr_lst2[i] == getegid()) {
						egid_test = FALSE;
						continue;
				}
			}
			if(gr_lst1[i] != gr_lst2[i])
				error("%s\n%s%ld.\n%s%ld.\n",
						"Supplementary group ID changed.",
						"Expected value of ",
						(long)gr_lst1[i],
						"Actual value was ",
						(long)gr_lst2[i]);
		}

									/*	Saved set-group-ID remains unchanged.
										See Reference #1 above, if changed
										then the setgid() call in this block
										would have failed.
									*/
	finis

	testing("Process without appropriate privileges and only the real");
	testing("group ID matches the setgid() calling argument, then a call");
	testing("to setgid() sets the effective group ID.");
	testing("The real group ID, saved set-group-ID, and any supplementary");
	testing("group IDs remain unchanged.");
	start
									/*	Recapping from above results:
											saved set-group-ID -> BIN_GID
											effective group ID -> BIN_GID
											real group ID      -> GID
									*/
		(void) Zsetgid(GID);

		expecting(GID);				/* Effective group ID is set. */
		(void) Zgetegid();

		expecting(GID);				/* Real group ID remains unchanged. */
		(void) Zgetgid();

									/* Supplementary group IDs are unchanged. */
		ngrps2 = Zgetgroups(n_grps, gr_lst2);
		if(ngrps2 != n_grps)
			error("%s\n%s%d\n%s%d\n",
						"getgroups() returned incorrect count.",
						"Expected count of ",
						n_grps,
						"Actual count is ",
						ngrps2);

		for(i=0; i<n_grps; i++) {
			if(gr_lst1[i] == getegid())
					continue;
			if(gr_lst1[i] != gr_lst2[i])
				error("%s\n%s%ld.\n%s%ld.\n",
						"Supplementary group ID changed.",
						"Expected value of ",
						(long)gr_lst1[i],
						"Actual value was ",
						(long)gr_lst2[i]);
		}

									/*	Recapping at this point:
											saved set-group-ID -> BIN_GID
											effective group ID -> GID
											real group ID      -> GID

										Saved set-group-ID remains unchanged.
									*/
		expecting(SUCCESS);
		(void) Zsetgid(BIN_GID);
	finis

	done();
}
