#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<pcts_timer.h>


		/* Accumulate some user and system CPU time. */
#define ACCUM_TIMES1 { \
	Ztimes(&beg_buf); \
	Ztimes(&end_buf); \
	while(	((accum_usr = end_buf.tms_utime -beg_buf.tms_utime) < clk_tck) || \
			((accum_sys = end_buf.tms_stime -beg_buf.tms_stime) < clk_tck) ) {
#define ACCUM_TIMES2 \
		for(i=0, dum_cnt=0; i<10; i++) { \
			dum_cnt = (dum_cnt++%3)*2;			/* Acquire user time.	*/ \
		} \
		Ztimes(&end_buf);						/* Acquire system time.	*/ \
	} \
}
#define ACCUM_TIMES ACCUM_TIMES1 ACCUM_TIMES2

int	main(void)
{
	pid_t	pid;
	time_t	time_beg, time_end;
	clock_t	ret_val1, ret_val2;
	clock_t	accum_usr, accum_sys;	/* User and system CPU accumulated time */
	struct	tms	beg_buf, end_buf, child_buf; 
	clock_t	clk_tck;
	int		dum_cnt;
	int		i;
	int		stat_val;
	int		pipfd[2];
	clock_t	t_usr_exit, t_sys_exit;

	setup(E_TIMES);
	funct(E_TIMES);

	assertion(5, BASE);
	start
		prep("Get value for {CLK_TCK}.");		/* Used throughout */
		start
			exit_on(ERROR);
			clk_tck = (clock_t) Zsysconf(_SC_CLK_TCK);
		finis

		prep("Get a reference time, then delay awhile.");
		start
			SLEEP(1);			/* Attempt to synchronize clock with delay. */
			ret_val1 =  Ztimes(&end_buf);
			SLEEP(WAIT_FOR_TIME);
		finis

		testing("Call to times() returns the elapsed real time in");
		testing("clock ticks of a second from an arbitrary point");
		testing("in the past.");
		start
			ret_val2 = Ztimes(&end_buf);

			if(ret_val2 < ret_val1 +((WAIT_FOR_TIME -1)*clk_tck))
				error("%s\n%s%lu.\n%s%lu.",
					"Invalid return from times().",
					"Expected a return value (in clock ticks) >= to ",
					(unsigned long) (ret_val1 +((WAIT_FOR_TIME -1)*clk_tck)),
					"Actual return was ",
					(unsigned long) ret_val2);
		finis
	finis

										/*	tms_utime */
	assertion(6, BASE);
	start
		testing("After some processor usage, times() records a reasonable");
		testing("value for user (tms_utime) and system (tms_stime)");
		testing("CPU time used.");
		start
			time_beg = Ztime((time_t *) 0);
			while(Ztime((time_t *) 0) == time_beg)
				;
			time_beg = Ztime((time_t *) 0);

				/* If no TIMEOUT, user and system CPU time is being updated. */
			SET_ALARM(MAX_TIME_TO_RUN)
				ACCUM_TIMES
			CLEAR_ALARM

			time_end = Ztime((time_t *) 0);

			if( (time_end -time_beg -((accum_sys +accum_usr)/clk_tck) < 0) ||
				(time_end -time_beg -((accum_sys +accum_usr)/clk_tck) > 1) )
				error("%s\n%s%lu.\n%s%lu.\n%s%lu.",
					"CPU times accumulated incorrectly.",
					"User CPU time accumulated (in clock ticks) is ",
					(unsigned long) accum_usr,
					"System CPU time accumulated (in clock ticks) is ",
					(unsigned long) accum_sys,
					"Wall clock time accumulated (in seconds) ",
					(unsigned long) (time_end -time_beg) );
		finis
	finis

										/*	tms_stime */
	assertion(7, BASE);
	start
		testing("Test for this assertion is performed in assertion #6.");
		start
		finis
	finis

										/*	tms_cutime */
	assertion(8, BASE);
	start
		long	t_usr_par, t_usr_chld;	/* Signed values. */
		long	t_sys_par, t_sys_chld;	/* Signed values. */

		break_on(ERROR);
										/* Test after using wait(). */

		ENABLE_PSYNC
					/*	Obtain the times the child uses to fork a process,
						write to a pipe, and exit. */
		prep("Fork a child process.");
		start
			Zpipe(pipfd);
			Ztimes(&beg_buf);
			pid = Zfork();
		finis

		if(pid == 0) {
			Zwrite(pipfd[1], (char *)&child_buf, (unsigned)sizeof(struct tms));
			done();
		}
		Zwait(&stat_val);
		Ztimes(&end_buf);
		t_usr_exit = end_buf.tms_cutime -beg_buf.tms_cutime;
		t_sys_exit = end_buf.tms_cstime -beg_buf.tms_cstime;
									/* Clear pipe. */
		Zread(pipfd[0], (char *)&end_buf, (unsigned)sizeof(struct tms));

		prep("Obtain current parent times for reference.");
		start
			Ztimes(&beg_buf);
		finis

		prep("Fork a child process.");
		start
			pid = Zfork();
		finis

		if(pid == 0) {
			prep("Fork a grandchild process.");
			start
				pid = Zfork();
			finis

			if(pid == 0) {
				prep("Grandchild acquires measured system and user CPU time.");
				start
					SET_ALARM(MAX_TIME_TO_RUN)
						ACCUM_TIMES
					CLEAR_ALARM
				finis
	
				done();
			}

			SET_ALARM(MAX_TIME_TO_RUN +1)
				Zwait(&stat_val);
			CLEAR_ALARM

			prep("Child acquires measured system and user CPU time.");
			start
				SET_ALARM(MAX_TIME_TO_RUN)
					ACCUM_TIMES
				CLEAR_ALARM
			finis

			prep("Pass child and grandchild times to parent.");
			start
				(void) Ztimes(&end_buf);
				Zwrite(pipfd[1], (char *)&end_buf,(unsigned)sizeof(struct tms));
			finis

			SEND_PSYNC

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		(void) Ztimes(&end_buf);

		testing("Unwaited for child processes user and system times");
		testing("not included in times() results.");
		start
			continue_on(ERROR);
			if(	((end_buf.tms_cutime -beg_buf.tms_cutime) != 0) ||
				((end_buf.tms_cstime -beg_buf.tms_cstime) != 0)	)
				error("%s\n%s%lu.\n%s%lu.",
					"Unwaited for child process CPU times included.",
					"Accumulated user CPU time of ",
					(unsigned long) (end_buf.tms_cutime -beg_buf.tms_cutime),
					"Accumulated system CPU time of ",
					(unsigned long) (end_buf.tms_cstime -beg_buf.tms_cstime));
		finis

		SET_TIMEOUT
			Zwait(&stat_val);
		CLEAR_ALARM

		(void) Ztimes(&end_buf);

		testing("After a call to wait() has been processed, times()");
		testing("returns the system and user CPU time of terminated");
		testing("child processes.");
		start
					/* Pick up times computed by child processes. */
			Zread(pipfd[0], (char *)&child_buf, (unsigned)sizeof(struct tms));

						/* Perform arithmetic with signed values. */
			t_usr_par = end_buf.tms_cutime -beg_buf.tms_cutime;
			t_usr_chld = child_buf.tms_cutime +child_buf.tms_utime +t_usr_exit;
			if(abs((int)(t_usr_par -t_usr_chld)) >2)
				error("%s\n%s%ld.\n%s%lu %lu %lu.\n%s%ld.",
					"Incorrect CPU usage time reported for child processes.",
					"Expected an additional child user CPU time of ",
					t_usr_chld,
					"Obtained from child, grandchild, and exit user times of ",
					(unsigned long) child_buf.tms_cutime,
					(unsigned long) child_buf.tms_utime,
					(unsigned long) t_usr_exit,
					"Actual additional user CPU time reported was ",
					t_usr_par);
		finis

		testing("After a call to wait() has been processed, times()");
		testing("returns the system CPU time of terminated");
		testing("child processes.");
		start
						/* Perform arithmetic with signed values. */
			t_sys_par = end_buf.tms_cstime -beg_buf.tms_cstime;
			t_sys_chld = child_buf.tms_cstime +child_buf.tms_stime +t_sys_exit;
			if(abs((int)(t_sys_par -t_sys_chld)) >2)
				error("%s\n%s%ld.\n%s%lu %lu %lu.\n%s%ld.",
					"Incorrect system CPU usage time reported for child processes.",
					"Expected an additional child system CPU time of ",
					t_sys_chld,
					"Obtained from child, grandchild, and exit times of ",
					(unsigned long) child_buf.tms_cstime,
					(unsigned long) child_buf.tms_stime,
					(unsigned long) t_sys_exit,
					"Actual additional system CPU time reported was ",
					t_sys_par);
		finis

										/* Repeat test using waitpid(). */
		prep("Obtain current parent times for reference.");
		start
			Ztimes(&beg_buf);
		finis

		prep("Fork a child process.");
		start
			pid = Zfork();
		finis

		if(pid == 0) {
			prep("Fork a grandchild process.");
			start
				pid = Zfork();
			finis

			if(pid == 0) {
				prep("Grandchild acquires measured system and user CPU time.");
				start
					SET_ALARM(MAX_TIME_TO_RUN)
						ACCUM_TIMES
					CLEAR_ALARM
				finis
	
				done();
			}

			SET_ALARM(MAX_TIME_TO_RUN +1)
				Zwaitpid(pid, &stat_val, 0);
			CLEAR_ALARM

			prep("Child acquires measured system and user CPU time.");
			start
				SET_ALARM(MAX_TIME_TO_RUN)
					ACCUM_TIMES
				CLEAR_ALARM
			finis

			prep("Pass child and grandchild times to parent.");
			start
				(void) Ztimes(&end_buf);
				Zwrite(pipfd[1], (char *)&end_buf,(unsigned)sizeof(struct tms));
			finis

			SEND_PSYNC

			done();
		}

		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		(void) Ztimes(&end_buf);

		testing("Unwaited for child processes user and system times");
		testing("not included in times() results.");
		start
			continue_on(ERROR);
			if(	((end_buf.tms_cutime -beg_buf.tms_cutime) != 0) ||
				((end_buf.tms_cstime -beg_buf.tms_cstime) != 0)	)
				error("%s\n%s%lu.\n%s%lu.",
					"Unwaited for child process CPU times included.",
					"Accumulated user CPU time of ",
					(unsigned long) (end_buf.tms_cutime -beg_buf.tms_cutime),
					"Accumulated system CPU time of ",
					(unsigned long) (end_buf.tms_cstime -beg_buf.tms_cstime));
		finis

		SET_TIMEOUT
			Zwaitpid(pid, &stat_val, 0);
		CLEAR_ALARM

		(void) Ztimes(&end_buf);

		testing("After a call to waitpid() has been processed, times()");
		testing("returns the user CPU time of terminated");
		testing("child processes.");
		start
			Zread(pipfd[0], (char *)&child_buf, (unsigned)sizeof(struct tms));

						/* Perform arithmetic with signed values. */
			t_usr_par = end_buf.tms_cutime -beg_buf.tms_cutime;
			t_usr_chld = child_buf.tms_cutime +child_buf.tms_utime +t_usr_exit;
			if(abs((int)(t_usr_par -t_usr_chld)) >2)
				error("%s\n%s%ld.\n%s%lu %lu %lu.\n%s%ld.",
					"Incorrect CPU usage time reported for child processes.",
					"Expected an additional child user CPU time of ",
					t_usr_chld,
					"Obtained from child, grandchild, and exit user times of ",
					(unsigned long) child_buf.tms_cutime,
					(unsigned long) child_buf.tms_utime,
					(unsigned long) t_usr_exit,
					"Actual additional user CPU time reported was ",
					t_usr_par);
		finis

		testing("After a call to waitpid() has been processed, times()");
		testing("returns the system CPU time of terminated");
		testing("child processes.");
		start
						/* Perform arithmetic with signed values. */
			t_sys_par = end_buf.tms_cstime -beg_buf.tms_cstime;
			t_sys_chld = child_buf.tms_cstime +child_buf.tms_stime +t_sys_exit;
			if(abs((int)(t_sys_par -t_sys_chld)) >2)
				error("%s\n%s%ld.\n%s%lu %lu %lu.\n%s%ld.",
					"Incorrect system CPU usage time reported for child processes.",
					"Expected an additional child system CPU time of ",
					t_sys_chld,
					"Obtained from child, grandchild, and exit times of ",
					(unsigned long) child_buf.tms_cstime,
					(unsigned long) child_buf.tms_stime,
					(unsigned long) t_sys_exit,
					"Actual additional system CPU time reported was ",
					t_sys_par);
		finis

		DISABLE_PSYNC

		cleanup("Close pipe.");
		start
			Zclose(pipfd[0]);
			Zclose(pipfd[1]);
		finis
	finis


										/*	tms_cstime */
	assertion(9, BASE);
	start
		testing("Test for this assertion is performed in assertion #8.");
		start
		finis
	finis

	done();
}
