#include	<pcts.h>

#include	<Zunistd.h>
#include	<Ztime.h>

#include	<sv_env.h>
#include	<pcts_timer.h>

#define TIME_TO_TEST_FOR_1	5
#define TIME_TO_TEST_FOR_2	10
#define	TIME_FOR_CHANGE		2
#define SLACK_TIME	1


int main(void)
{
	time_t	start_time, end_time;
	pid_t	pid;

	setup(E_ALARM);
	funct(E_ALARM);

	assertion(4, BASE);
	start
		prep("Start alarm clock and pause until a signal is received.");
		start
			start_time = Ztime((time_t *)0);

			SET_ALARM(TIME_TO_TEST_FOR_1)
				pause();
			EXPECT_ALARM
		finis

		testing("Alarm occurs at proper time.");
		start
			end_time = Ztime((time_t *)0);
			if((end_time - start_time) > (TIME_TO_TEST_FOR_1 + SLACK_TIME))
				error("Alarm slept longer than expected.");
			if((end_time - start_time) < (TIME_TO_TEST_FOR_1 -(time_t)1))
				error("Alarm did not sleep as long as expected.");
		finis

		ENABLE_PSYNC
	    	prep("Fork a child process.");
	    	start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Start alarm clock and hang.");
				start
					sv_flush();
					SIGNAL(SIGALRM, SIG_DFL);
					alarm(2 * WAIT_FOR_TIME);

					del_secs(1);			/* Delay a second */
					SEND_PSYNC
					while(1)	/* Keep busy */
						;
				finis

				done();
			}

			prep("Delay.");
			start
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
			finis

			testing("Test that child process was terminated by SIGALRM at");
			testing("about the time the alarm was to go off.");
			start
				int	stat_val;

				SET_ALARM((2 * WAIT_FOR_TIME) +1)
					wait(&stat_val);
					if(WIFSIGNALED(stat_val)) {
						if(WTERMSIG(stat_val) != SIGALRM)
							error("%s\n%s%d\n",
								"Expected child process terminated by SIGALRM.",
								"Value returned from macro WTERMSIG is ",
								WTERMSIG(stat_val) );
					}
					else
						error("Child not terminated by signal.");
				CLEAR_ALARM
			finis
		DISABLE_PSYNC
	finis

	assertion(5, BASE);
	start
		prep("Obtain current time.");
		start
			SLEEP(1);		/* Attempt to synch test with clock. */
			start_time = Ztime((time_t *)0);
		finis

		prep("Start alarm clock, then delay a short time, then reset");
		prep("alarm clock then pause until a signal is received.");
		start
			SIGNAL(SIGALRM, sig_int);
			alarm(TIME_TO_TEST_FOR_1);
			del_secs(TIME_FOR_CHANGE);
			alarm(TIME_TO_TEST_FOR_2);
			pause();
		finis

		testing("Alarm occurs at proper time.");
		start
			end_time = Ztime((time_t *)0);
			if((end_time - start_time) > (time_t)(TIME_TO_TEST_FOR_2 +
												TIME_FOR_CHANGE +SLACK_TIME))
				error("Alarm slept longer than expected.");
			if((end_time - start_time) <
									(time_t)(TIME_TO_TEST_FOR_2 -SLACK_TIME))
				error("Alarm did not sleep as long as expected.");
			finis
	finis

	assertion(6, BASE);
	start
		break_on(ERROR);
		SET_ALARM(TIME_TO_TEST_FOR_1)

			prep("Start alarm clock and delay awhile.");
			start
				del_secs(TIME_FOR_CHANGE);
			finis

			prep("Call alarm(0).");
			start
				alarm(0);
			finis

			testing("Previously made alarm was canceled.");
			start
				del_secs(TIME_TO_TEST_FOR_1);
			finis

			/* If an alarm goes off, an error will be reported
			   by the CLEAR_ALARM macro.
			*/

		CLEAR_ALARM
	finis

	assertion(7, BASE);
	start
		SET_ALARM(TIME_TO_TEST_FOR_1)
			prep("Start alarm clock and delay awhile.");
			start
				del_secs(TIME_FOR_CHANGE);
			finis
		CLEAR_ALARM

		testing("Value returned from alarm() call is time remaining.");
		start
			if(unused_time > (TIME_TO_TEST_FOR_1 -TIME_FOR_CHANGE +
															(2*SLACK_TIME)))
				error("%s\n%s%d\n%s%d\n",
						"Time returned is to large.",
						"Expected a value less than or equal to: ",
						TIME_TO_TEST_FOR_1 -TIME_FOR_CHANGE +(2*SLACK_TIME),
						"Actual value is: ",
						unused_time);

			if(unused_time < (TIME_TO_TEST_FOR_1 -TIME_FOR_CHANGE -
															(2*SLACK_TIME)))
				error("%s\n%s%d\n%s%d\n",
						"Time returned is to small.",
						"Expected a value greater than or equal to: ",
						TIME_TO_TEST_FOR_1 -TIME_FOR_CHANGE -(2*SLACK_TIME),
						"Actual value is: ",
						unused_time);
		finis
	finis

	assertion(8, BASE);
	start
		unsigned int ret;

		testing("When there is no previous alarm() request, the return");
		testing("value of alarm() is zero.");
		start
			ret = alarm(0);
			if(ret != 0)
				error("%s %u.",
						"Call alarm() with no previous alarm request returned",
						ret);
		finis
	finis

	assertion(9, BASE);
	start
		unsigned int ret;

		testing("When there is a previous alarm() request with less than");
		testing("one second remaining, then the return from a call to alarm()");
		testing("returns the value 1.");
		start
				/*	Attempt to get alarm() in sync with system clock. */
			signal(SIGALRM, sig_int);
			alarm(1);
			pause();
			signal(SIGALRM, SIG_DFL);
			dt_secs(3);	/*  Ensure alarm() is NOT in sync with system clock */
									
			alarm(1);
			dt_secs(4);			/* Delay .4 seconds. */
			ret = alarm(0);
			if(ret != 1)
				error("%s\n%s\n%s%u.",
						"Return vale from alarm() incorrect.",
						"Expected: 1",
						"Actual: ",
						ret);
		finis
	finis

	done();
}
