#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<pcts_timer.h>

#define	ID_LENGTH	30
#define	MODE_LENGTH	10
#define	SIG_LENGTH	10
static	char	a_pid[ID_LENGTH], a_ppid[ID_LENGTH], a_pgrp[ID_LENGTH];
static	char	a_ruid[ID_LENGTH], a_rgid[ID_LENGTH];
static	char	a_cwd[_POSIX_PATH_MAX +1];
static	char	a_umask[MODE_LENGTH];
static	char	a_blk_sig[SIG_LENGTH];
#undef	ID_LENGTH

#define	FILENAME	"./3.1.2_47x"
#define	PARAMETERS	a_indx, a_pid, a_ppid, a_pgrp, a_ruid, a_rgid, a_cwd, a_umask, a_blk_sig
#include	"exec.h"


int	main(int argc, char *argv[])
{
	char	path[_POSIX_PATH_MAX +1];

	SETUP_FOR_EXEC

	assertion(47, BASE);
	start
		FORK
			prep("Reset \"environ\".");
			start
				environ[0] = (char *)0;
			finis
	
			prep("Get process user ID.");
			start
				sprintf(a_pid, "%ld", (long)getpid());
			finis
		
			prep("Get parent process user ID.");
			start
				sprintf(a_ppid, "%ld", (long)getppid());
			finis
		
			prep("Get process group ID.");
			start
				sprintf(a_pgrp, "%ld", (long)getpgrp());
			finis
	
			prep("Get real user ID.");
			start
				expecting(UID);
				sprintf(a_ruid, "%lu", (unsigned long)Zgetuid());
			finis
	
			prep("Get real group ID.");
			start
				expecting(GID);
				sprintf(a_rgid, "%lu", (unsigned long)Zgetgid());
			finis
	
			prep("Obtain supplementary group IDs and write to a temp file."); 
			start
				int		fd;
				gid_t	*g_beg;
				int		n_grps;
	
				n_grps = Zgetgroups(0, (gid_t *)0);	/* Get number of groups */
				g_beg =(gid_t *)Zmalloc((size_t)(sizeof(gid_t) * n_grps));
				(void) Zgetgroups(n_grps, g_beg);
	
				Zstrcpy(path, temp_path());
				Zstrcat(path, "sup_grps");
				fd = Zcreat(path, PROT_ALL);
				expecting(sizeof(gid_t) * n_grps);
				write(fd, (char *)g_beg, sizeof(gid_t) * n_grps);
				expecting(SUCCESS);
				Zclose(fd);
			finis
			
			prep("Get current working directory and root directory.");
			start
				Zgetcwd(a_cwd, _POSIX_PATH_MAX);
			finis
	
			prep("Set file mode creation mask.");
			start
				(void) umask((mode_t)(S_IRWXO));
				sprintf(a_umask, "%lu", (unsigned long)(S_IRWXO));
			finis
	
			prep("Set process signal mask to block a signal.");
			start
				sigset_t	set;
	
				SIGBLOCK(set, SIGUSR1);
				sprintf(a_blk_sig, "%d", SIGUSR1);
			finis
	
			prep("Send the signal that was blocked to this process to obtain");
			prep("a pending signal.");
			start
				Zkill(getpid(), SIGUSR1);
			finis
	
			prep("Accumulate some child processing time.");
			start
				int	i;
				pid_t	t_pid;
	
				if( ! (t_pid = fork()) ) {
					for(i=0; i<10; i++)
						Zaccess(path, R_OK);
					sv_flush();
					done();
				}
				WAIT_FOR_PID(t_pid);
			finis
	
			prep("Obtain current process times and write to temp file.");
			start
				int	fd;
				struct	tms	buffer;
	
				Ztimes(&buffer);
				Zstrcpy(path, temp_path());
				Zstrcat(path, "times");
	
				fd = Zcreat(path, PROT_ALL);
				write(fd, (char *)&buffer.tms_utime, sizeof(clock_t));
				write(fd, (char *)&buffer.tms_stime, sizeof(clock_t));
				write(fd, (char *)&buffer.tms_cutime, sizeof(clock_t));
				write(fd, (char *)&buffer.tms_cstime, sizeof(clock_t));
				Zclose(fd);
			finis
					
			SET_ALARM(MAX_TIME_TO_RUN);
				EXEC_MODULE(FILENAME)
			CLEAR_ALARM;
		KROF
	finis

	assertion(20, BASE);
	start
		testing("This assertion is tested by assertion #47 and many other");
		testing("assertions that pass arguments to the routine that is");
		testing("exec-ed.");
		start
		finis
	finis
	
	done();
}
