#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_errno.h>

extern	char	**environ;
static	int		e_index;

#define	FILENAME	"./exec_p/good/ex_p_good"
#define	TEST_FILE	"reg_file"

static	void exec_tst(char *, char *, char **);
static	void el_NO_TRUNC(void);
static	void el_ENAMETOOLONG(void);
static	char *p_arg;
char	*varg_set[3];

int	main(int argc, char *argv[])
{
	char	path[_POSIX_PATH_MAX +1];
	pid_t	pid;
	long	arg_max;

	e_index = atoi(*++argv);
	setup(e_index);
	funct(e_index);

	arg_max = Zsysconf(_SC_ARG_MAX);
	if( (arg_max <= (long)PCTS_ARG_MAX) && (arg_max != (long)-1) ) {
		assertion(49, BASE);
		start
			char	*ptr;

			prep("Initialize an array of length {ARG_MAX} +2.");
			start
				p_arg = (char *) Zmalloc((size_t) arg_max +2);
				for(ptr = p_arg; ptr <(p_arg +arg_max +1); ptr++)
					*ptr = 'z';
				*ptr = '\0';

				varg_set[0] = "errno_test";
				varg_set[1] = p_arg;
				varg_set[2] = (char *)0;
			finis

			FORK
				testing("Return of [E2BIG] when called with an");
				testing("argument greater than {ARG_MAX} bytes.");
				start
					Zstrcpy(path, FILENAME);
					expecting(-1);
					expecting(E2BIG);
					exec_tst(path, p_arg, varg_set);
				finis
			KROF
		finis
	}
	else
		assertion(49, NOT_TESTABLE);

	assertion(50, BASE);
	start
		prep("Initialize path name for a path where search permission");
		prep("is denied.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_srch.d");
			Zstrcat(path, "/test");

			p_arg = (char *) Zmalloc(_POSIX_PATH_MAX +1);
			Zstrcpy(p_arg, "Hi");
			varg_set[0] = "errno_test";
			varg_set[1] = (char *)0;
		finis

		FORK
			testing("Search permission is denied on path\n\t\"%s\".", path);
			start
				expecting(-1);
				expecting(EACCES);
				exec_tst(path, p_arg, varg_set);
			finis
		KROF
	finis

	assertion(51, BASE);
	start
		prep("Initialize path name for path where execute permission");
		prep("is denied.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "no_exec");
		finis

		FORK
			testing("Execute permission is denied on path\n\t\"%s\".", path);
			start
				expecting(-1);
				expecting(EACCES);
				exec_tst(path, p_arg, varg_set);
			finis
		KROF
	finis

	assertion(52, EXTENDED);

	if(tc_path_max()) {
		assertion(53, BASE);
		start
			el_ENAMETOOLONG();
		finis
	}
	else
		assertion(53, NOT_TESTABLE);


	if(tc_name_max()) {
		assertion(54, BASE);
		start
			el_NO_TRUNC();
		finis
	}
	else
		assertion(54, NOT_TESTABLE);

	if(!tc_name_max()) {
		assertion(55, BASE);
		start
			el_NO_TRUNC();
		finis
	}
	else
		assertion(55, NOT_TESTABLE);

	assertion(56, BASE);
	start
		prep("Initialize path name for a non-existing path component.");
		start
			Zstrcpy(path, data_path());
			Zstrcat(path, "/noent/exec");
		finis

		FORK
			testing("A non-existing path component\n\t\"%s\".", path);
			start
				expecting(-1);
				expecting(ENOENT);
				exec_tst(path, p_arg, varg_set);
			finis
		KROF
	finis

	assertion(57, BASE);
	start
		prep("Initialize path name to an empty string.");
		start
			Zstrcpy(path, "");
		finis

		FORK
			testing("Path points to an empty string.");
			start
				expecting(-1);
				expecting(ENOENT);
				exec_tst(path, p_arg, varg_set);
			finis
		KROF
	finis

	assertion(58, BASE);
	start
		int	fd;

		prep("Initialize path name for a path where path component is not");
		prep("a directory.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, TEST_FILE);
			fd = Zcreat(path, PROT_ALL);
			Zwrite(fd, "ABC", 3);
			Zclose(fd);
			Zstrcat(path, "/test");
		finis

		FORK
			testing("Path \"%s\" contains a non-directory component.", path);
			start
				expecting(-1);
				expecting(ENOTDIR);
				exec_tst(path, p_arg, varg_set);
			finis
		KROF
	finis

	if((e_index == (int)E_EXECLP) || (e_index == (int)E_EXECVP))
		assertion(59, UNUSED);
	else {
		assertion(59, BASE);
		start
			prep("Initialize path name to an ascii data file.");
			start
				Zstrcpy(path, temp_path());
				Zstrcat(path, TEST_FILE);
			finis
	
			FORK
				testing("An exec of a file of path \"%s\" containing", path);
				testing("improper file format.");
				start
					expecting(ENOEXEC);
					exec_tst(path, p_arg, varg_set);
				finis
			KROF
		finis
	}

	assertion(60, EXTENDED);

	done();
}

static	void exec_tst(char *path, char *p_arg, char **varg_set)
{
	switch(e_index) {
		case E_EXECL:
				Zexecl(path, varg_set[0], p_arg, (char *) 0);
				break;
		case E_EXECLE:
				Zexecle(path, varg_set[0], p_arg, (char *) 0, environ);
				break;
		case E_EXECV:
    			Zexecv(path, varg_set);
				break;
		case E_EXECVE:
    			Zexecve(path, varg_set, environ);
				break;
		case E_EXECLP:
    			Zexeclp(path, varg_set[0], p_arg, (char *) 0);
				break;
		case E_EXECVP:
    			Zexecvp(path, varg_set);
				break;
		default:
				error("Bad indx \"%d\" detected.", e_index);
	}
}

				/*	Test assertions asociated with {NAME_MAX} for exec type
					functions.  When ex_path is TRUE, path_ex is the path
					where the long name is generated for.
				*/
static	void el_NO_TRUNC(void)
{
	char	*lpath;
	long	res_p;
	char	path[_POSIX_PATH_MAX +1];
	pid_t	pid;

	Zstrcpy(path, temp_path());
	res_p = pathconf(path, _PC_NO_TRUNC);
	if(res_p != (long) -1 )  {

		lpath = g_lname();

		if(tc_name_max() == TRUE) {
			testing("Path component longer than {NAME_MAX}.");
			start
				if((pid = fork()) == (pid_t)0) {
					expecting(-1);
					expecting(ENAMETOOLONG);
					exec_tst(lpath, p_arg, varg_set);

					done();
				}
				WAIT_FOR_PID(pid);
			finis
		}
		else {
			testing("Path component equal to {PCTS_NAME_MAX}.");
			start
				if((pid = fork()) == (pid_t)0) {
					expecting(SUCCESS);
					exec_tst(lpath, p_arg, varg_set);

					done();
				}
				WAIT_FOR_PID(pid);
			finis
		}

		cleanup("Remove buffer space allocated for long path.");
		start
			Zfree(lpath);
		finis
	}
	else
		error("{_POSIX_NO_TRUNC} not supported.");
}


				/*	Test [ENAMETOOLONG] for exec type functions.  */
static	void el_ENAMETOOLONG(void)
{
	char	*lpath;
	int		test_pm;
	pid_t	pid;

	if(test_pm = tc_path_max()) {
		prep("Establish pathname of {PATH_MAX} +2 bytes.");
		start
			lpath = g_lpath();
		finis
	}
	else {
		prep("Establish pathname of {PCTS_PATH_MAX} +2 bytes.");
		start
			char	*ptr, *ptr_end;

					/* Allocate buffer for {PCTS_PATH_MAX} +1 + "/X" */
			lpath = (char *) Zmalloc(PCTS_PATH_MAX +3);

			Zstrcpy(lpath, temp_path());
			Zstrcat(lpath, "ENAMETOOLONG");
			if(access(lpath, F_OK))
				Zmkdir(lpath, PROT_ALL);
			ptr = lpath +Zstrlen(lpath);

			ptr_end = PCTS_PATH_MAX +lpath;
			for( ; ptr +16 < ptr_end; ptr += 16 )
        		(void) strcpy(ptr, "/../ENAMETOOLONG");

			for( ; ptr < ptr_end; ptr += 2 )
        			(void) strcpy(ptr, "/.");

							/* End of path may be "/.//X" "or /./X". */
        	(void) strcpy(ptr_end, "/X");
		finis
	}

	if(test_pm = tc_path_max()) {
		testing("For path exceeds {PATH_MAX}.");
		start
			if((pid = fork()) == (pid_t)0) {
				expecting(-1);
				expecting(ENAMETOOLONG);
				exec_tst(lpath, p_arg, varg_set);

				done();
			}
			WAIT_FOR_PID(pid);
		finis
	}
	else {
		testing("For path length of {PCTS_PATH_MAX}.");
		start
			if((pid = fork()) == (pid_t)0) {
				expecting(SUCCESS);
				exec_tst(lpath, p_arg, varg_set);

				done();
			}
			WAIT_FOR_PID(pid);
		finis
	}

	cleanup("Remove buffer space allocated for long path.");
	start
		Zfree(lpath);
	finis
}
