#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<sys/Ztimes.h>
#include	<Zfcntl.h>
#include	<Zdirent.h>

#include	<sv_env.h>
#include	<pcts_timer.h>

#define	FILENAME	"3.1.1_04"
#define	DIRNAME		"3.1.1_04.d"
#define	String	"DaTa"
#define	Strlen	(sizeof(String) +1)

int	main(void)
{
	pid_t	pid;
	char	path[_POSIX_PATH_MAX +1], path2[_POSIX_PATH_MAX +1];
	int		fd;

	setup(E_FORK);
	funct(E_FORK);

	assertion(4, BASE);
	start
		FORK
			testing("Child process ID is different from parent process ID.");
			start
				if(Zgetpid() == Zgetppid())
					error("%s\n%s%ld.\n%s%ld.",
						"Process IDs are equal.",
						"Child process ID is ",
						(long) getpid(),
						"Parent process ID is ",
						(long) getppid() );
			finis
		KROF
	finis

	assertion(5, EXTENDED);
	assertion(6, EXTENDED);

	assertion(7, BASE);
	start
		pid_t	p_pid;

		prep("Get process ID of parent.");
		start
			p_pid = getpid();
		finis

		FORK
			testing("Parent process ID of child process is the process ID");
			testing("of the parent process.");
			start
				if(p_pid != Zgetppid())
					error("%s%ld.\n%s%ld.",
						"Parent process ID of child process is ",
						(long) getppid(),
						"Process ID of parent process is ",
						(long) p_pid);
			finis
		KROF
	finis

		
	assertion(8, BASE);
	start
		off_t	p_offset, c_offset;

		prep("Create test file.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, FILENAME);
			fd = Zopen(path, O_CREAT | O_TRUNC | O_RDWR | O_APPEND, PROT_ALL);
			expecting(Strlen);
			Zwrite(fd, String, Strlen);

			p_offset = Zlseek(fd, (off_t)0, SEEK_CUR);
		finis

		FORK
			testing("File offset of test file for parent process is the same");
			testing("as that of the child process.");
			start
				c_offset = Zlseek(fd, (off_t)0, SEEK_CUR);
				if(c_offset != p_offset)
					error("%s\n%s%lu.\n%s%lu.",
							"Offset incorrect.",
							"Expected an offset of ",
							(long)p_offset,
							"Actual return was ",
							(long)c_offset);
			finis

			testing("File status and file access modes are the same.");
			start
				int	ret_val;

				ret_val = Zfcntl(fd, F_GETFL, 0);

				if((ret_val & O_APPEND) != O_APPEND)
					error("%s\n%s",
						"Expected status flag for O_APPEND to be set.",
						"O_APPEND status flag was not set.");

				if((ret_val & O_NONBLOCK) == O_NONBLOCK)
					error("%s\n%s",
						"Expected status flag for O_NONBLOCK to be clear.",
						"O_NONBLOCK status flag was set.");

				if((ret_val & O_ACCMODE) != O_RDWR)
					error("%s\n%s%s.",
						"Expected access mode to be O_RDWR.",
						"Actual access mode is ",
						((ret_val & O_ACCMODE)==O_RDONLY)?"O_RDONLY":
								"O_WRONLY");
			finis
		KROF
	finis
					/*	When a call to fork() completes successfully,
						then the child process has its own copy of the
						open directory streams of the parent process.
					*/
#define	NUMBER_OF_ENTRIES	10
	assertion(9, BASE);
	start
		char	d_entries[NUMBER_OF_ENTRIES +2][3];
		DIR		*dirp;
		struct	dirent *dirent;
		int		i;


		prep("Establish entries in a directory.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, DIRNAME);
			Zmkdir(path, PROT_ALL);
			for(i=0; i<NUMBER_OF_ENTRIES; i++) {
				sprintf(path2, "%s%s/%d", temp_path(), DIRNAME, i);
				Zmkdir(path2, PROT_ALL);
			}
		finis

		prep("Parent process opens directory stream and reads the entries.");
		start
			dirp = Zopendir(path);
			for(i=0; ((dirent = readdir(dirp)) != (struct dirent *)0); i++)
				Zstrcpy(d_entries[i], dirent->d_name);
			if((i<NUMBER_OF_ENTRIES) || (i>NUMBER_OF_ENTRIES+2))
					error("Read %d entries, expected %d or %d entries.",
								i, NUMBER_OF_ENTRIES, NUMBER_OF_ENTRIES +2);
		finis

		FORK
			testing("Child process has its own copy of the parent's open");
			testing("directory stream.");
			start
				rewinddir(dirp);

				for(i=0; ((dirent = readdir(dirp)) != (struct dirent *)0); i++){
					if(strcmp(dirent->d_name, d_entries[i]))
						error("Entry %d %s\n%s\"%s\"\n%s\"%s\"",
							i,
							"does not match for child and parent.",
							"Child entry is ",
							dirent->d_name,
							"Parent entry was ",
							d_entries[i]);
				}
				if((i<NUMBER_OF_ENTRIES) || (i>NUMBER_OF_ENTRIES+2))
						error("Read %d entries, expected %d or %d entries.",
								i, NUMBER_OF_ENTRIES, NUMBER_OF_ENTRIES +2);
			finis
		KROF
	finis


	assertion(10, BASE);
	start
		struct	tms	beg, end;
		struct	tms	c_buffer, p_buffer;

		prep("Accumulate some time, then obtain parent process times.");
		start
			int	i, dum_cnt;
			clock_t	clk_tck;

			exit_on(ERROR);
			clk_tck = (clock_t) Zsysconf(_SC_CLK_TCK);

								/* Accumulate some user and system CPU time. */
			Ztimes(&beg);
			Ztimes(&end);
			do {
				for(i=0, dum_cnt=0; i<10; i++) {
					dum_cnt = (dum_cnt++%3)*2;		/* Acquire user time.	*/
				}
				Ztimes(&end);					/* Acquire system time.	*/
			} while(((end.tms_utime -beg.tms_utime) < clk_tck) ||
					((end.tms_stime -beg.tms_stime) < clk_tck) );

			Ztimes(&p_buffer);
		finis

		FORK
			testing("Child process's values of tms_utime, tms_stime,");
			testing("tms_cutime, and tms_cstime are reset.");
			start
				Ztimes(&c_buffer);

				if(c_buffer.tms_utime >= p_buffer.tms_utime)
					error("%s\n%s%lu\n%s%lu\n",
						"Value of child's \"tms_utime\" value incorrect.",
						"Expected time < ",
						(unsigned long) p_buffer.tms_utime,
						"Actual time obtained ",
						(unsigned long) c_buffer.tms_utime);

				if(c_buffer.tms_stime >= p_buffer.tms_stime)
					error("%s\n%s%lu\n%s%lu\n",
						"Value of child's \"tms_stime\" value incorrect.",
						"Expected time < ",
						(unsigned long) p_buffer.tms_stime,
						"Actual time obtained ",
						(unsigned long) c_buffer.tms_stime);

				if(c_buffer.tms_cutime)
					error("%s\n%s%lu\n",
						"Expected child's \"tms_cutime\" value to be zero.",
						"Actual value was ",
						(unsigned long) c_buffer.tms_cutime);

				if(c_buffer.tms_cstime)
					error("%s\n%s%lu\n",
						"Expected child's \"tms_cstime\" value to be zero.",
						"Actual value was ",
						(unsigned long) c_buffer.tms_cstime);
			finis
		KROF
	finis
		
	done();
}
