#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>

int	main(void)
{
	pid_t	pid;
	int		i;

	setup(E_FORK);
	funct(E_FORK);

	assertion(18, BASE);
	start
		gid_t	*g_beg_p, *g_beg_c;
		int		n_grps_p, n_grps_c;

		prep("Obtain supplementary groups for parent process.");
		start
			n_grps_p = Zgetgroups(0, (gid_t *)0);	/* Get number of groups */
			g_beg_p =(gid_t *)Zmalloc((unsigned int)(sizeof(gid_t) * n_grps_p));
			(void) Zgetgroups(n_grps_p, g_beg_p);
		finis

		FORK
			prep("Obtain supplementary groups for child process.");
			start
				n_grps_c = Zgetgroups(0, (gid_t *)0); /* Get number of groups */
				g_beg_c =
					(gid_t *)Zmalloc((unsigned int)(sizeof(gid_t) * n_grps_c));
				(void) Zgetgroups(n_grps_c, g_beg_c);
			finis

			testing("Supplementary groups are the same.");
			start
				break_on(ERROR);
				if(n_grps_c != n_grps_p)
					error("%s\n\t%s\n%s%ld\n%s%ld",
						"Expected the number of supplementary groups of",
						"child to be equal to that of parent.",
						"Child supplementary groups ",
						(long) n_grps_c,
						"Parent supplementary groups ",
						(long) n_grps_p);
				
				continue_on(ERROR);
				for(i=0; i<n_grps_c; i++)
					if(g_beg_p[i] != g_beg_c[i])
						error("%s\n%s%lu\n%s%lu",
								"Supplementary group ID not the same.",
								"Expected group ID for child of:	",
								(unsigned long) g_beg_p[i],
								"Actual group located in its place is:	",
								(unsigned long) g_beg_c[i]);
			finis
		KROF
	finis

			
	assertion(19, BASE);
	start
		pid_t	p_pgrp, c_pgrp;

		prep("Get parent process group ID.");
		start
			p_pgrp = getpgrp();
		finis

		FORK
			testing("Child process group ID is same as parent's.");
			start
				if(p_pgrp != (c_pgrp = getpgrp()) )
					error("%s\n%s%ld\n%s%ld\n",
						"Process IDs are equal.",
						"Child process ID is ",
						(long) c_pgrp,
						"Parent process ID is ",
						(long) p_pgrp );
			finis
		KROF
	finis
	

	assertion(21, BASE);
	start
		mode_t	umask_p;

		prep("Set and obtain file mode creation mask for parent.");
		start
			(void) umask((mode_t)(S_IRWXO));
			umask_p = umask((mode_t)(S_IRWXO));
		finis

		FORK
			testing("File mode creation mask of child same as parent.");
			start
				mode_t	umask_c;

				umask_c = umask((mode_t)0);
				if(umask_c != umask_p)
					error("%s\n%s\"%lu\"\n%s\"%lu\"\n",
						"File mode creation masks not the same.",
						"Parent's file mode creation mask was ",
						(unsigned long) umask_p,
						"Child's file mode reation mask is ",
						(unsigned long) umask_c);
			finis
		KROF
	finis

	assertion(22, BASE);
	start
		sigset_t	set, set_p;

		prep("Set process signal mask to block a signal.");
		start
			Zsigemptyset(&set_p);
			SIGBLOCK(set_p, SIGUSR1);
		finis

		prep("Ensure process signal mask is blocking signal.");
		start
			Zsigprocmask(SIG_BLOCK,(sigset_t *)0, &set_p);
			if( ! sigismember(&set_p, SIGUSR1) )
				error("Process signal mask not blocking SIGUSR1.");
			if(sigismember(&set_p, SIGALRM))
				error("Process signal mask blocking SIGALRM.");
			if(sigismember(&set_p, SIGUSR2))
				error("Process signal mask blocking SIGUSR2.");
		finis

		FORK
			testing("Process signal mask for child is the same as parent.");
			start
				Zsigprocmask(SIG_BLOCK, (sigset_t *)0, &set);
				for(i=0; i<NSIGS; i++) {
					if(sigismember(&set, sigs[i].sig) !=
									sigismember(&set_p, sigs[i].sig))
						error("%s%s%s\n",
							"Parent and child process signal mask differ for ",
							"signal ",
							sigs[i].sig);
				}
			finis
		KROF
	finis

	assertion(23, BASE);
	start
		pid_t	p_pgrp, c_pgrp;

		prep("Obtain parent's process group.");
		start
			p_pgrp = getpgrp();
		finis

		FORK
			testing("Child process is in same session as the parent.");
			testing("It must be in the same process group as the parent.");
			start
				c_pgrp = getpgrp();
				if(p_pgrp != c_pgrp)
					error("%s\n%s%ld\n%s%ld\n",
						"Child is not in same process group as parent.",
						"Expected parent's process group of ",
						(long) p_pgrp,
						"Actual child's process group is ",
						(long) c_pgrp);
			finis
		KROF
	finis

	assertion(24, BASE);
	start
		ENABLE_PSYNC
			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child and parent are capable of executing");
				testing("independently before either process terminates.");
				start
					SEND_PSYNC
						/* Wait for parent to complete tests */
					SET_TIMEOUT
						WAIT_PSYNCA
					CLEAR_ALARM
				finis

				done();
			}

		testing("Parent and child are capable of executing independently,");
		testing("before either process terminates.");
		start
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			SEND_PSYNCA
		finis

		WAIT_FOR_PID(pid);

		DISABLE_PSYNC
	finis

	assertion(25, BASE);
	start
		ENABLE_PSYNC
			prep("Fork a child.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("Child process was returned a value of zero.");
				start
				finis

				testing("Parent process is returned child's process ID.");
				start
					SIGNAL(SIGINT, sig_int);
					sig_rcvd = 0;

					SET_TIMEOUT
						SEND_PSYNC
						while(sig_rcvd == 0)
							dt_secs(1);
					CLEAR_ALARM
				finis

				done();
			}

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			Zkill(pid, SIGINT);
			WAIT_FOR_PID(pid);
		DISABLE_PSYNC
	finis

	done();
}
