#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>

#define		CORE		"./core"
#define		FILE_EXISTS	0

int	main(void)
{
	int		stat_val, i, j;
	struct	sigaction act1, act2;
	pid_t	pid;

	setup(E_SIGACTION);
	funct(E_SIGACTION);

	assertion(5, BASE);
	start
		FORK
			prep("Establish a reference action for a signal.");
			start
				act1.sa_handler = sig_int;
				Zsigemptyset(&act1.sa_mask);
				act1.sa_flags = 0;
				expecting(0);
				Zsigaction(SIGALRM, &act1, (struct sigaction  *) 0);
			finis

			testing("Call to sigaction(sig, NULL, oact) won't alter");
			testing("signal handling and returns a value of zero.");
			start
				expecting(0);
				Zsigaction(SIGALRM, (struct sigaction *)0, &act2);

							/* Test that action is still to catch signal. */
				sig_rcvd = 0;
				Zkill(getpid(), SIGALRM);
				SLEEP(WAIT_FOR_TIME);
				if(sig_rcvd == 0)
					error("Signal SIGALRM was not caught.");
			finis

			testing("Previous action associated with the signal is placed");
			testing("in oact.");
			start
		  		for(i=0; i<NSIGS; i++)
					if(sigismember(&act2.sa_mask, sigs[i].sig))
						error("Signal mask altered.");

				if(act2.sa_handler != act1.sa_handler)
					error("The sa_handler field has changed.");

				if((act2.sa_flags & SA_NOCLDSTOP) != 0)
					error("%s\n%s%d",
						"Expected \"sa_flags\" field value to be 0.",
						"Actual \"sa_flags\" field has value of ",
						act2.sa_flags);
			finis
		KROF
	finis

	assertion(6, BASE);		/* Also tests Assertions #8 and #9. */
	start
		ENABLE_PSYNC
		for(j=0; j<3; j++) {
		  for(i=0; i<NSIGS_JC; i++) {
			if((sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGALRM))
				continue;

			prep("Fork a child");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Set action of %s to %s.", sigs[i].signame,
							(j==0)?"SIG_DFL":(j==1)?"SIG_IGN":"catch signal");
				start
					act1.sa_handler = (j==0)?SIG_DFL:((j==1)?SIG_IGN:sig_int);
					Zsigemptyset(&act1.sa_mask);
					act1.sa_flags = 0;
					expecting(0);
					Zsigaction(sigs[i].sig, &act1, (struct sigaction  *) 0);
					SEND_PSYNC
				finis

				if(j<2) {
					SLEEP(2 * WAIT_FOR_TIME);
				}
				else {
					testing("Signal is caught when %s signal is received.",
															sigs[i].signame);
					start
						sig_rcvd = 0;
						Zkill(getpid(), sigs[i].sig);
						SLEEP(WAIT_FOR_TIME);
						if( ! sig_rcvd)
							error("Signal %s not caught.", sigs[i].sig);
					finis
				}

				done();
			}

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			if(j<2) {
				testing("%s termination occurs when %s signal is received.",
							(j==0)?"Abnormal":"Normal", sigs[i].signame);
				start
					Zkill(pid, sigs[i].sig);
					waitpid(pid, &stat_val, 0);
					if(j==0) {
						if(WTERMSIG(stat_val) != sigs[i].sig)
							error("%s %s.\n%s%d.",
								"Expected child process to be terminated by ",
								sigs[i].signame,
								"Value returned from macro WTERMSIG is ",
								WTERMSIG(stat_val) );
					}
					else
						if( ! WIFEXITED(stat_val))
							error("Child process did not terminate normally.");
				finis
			}
			else
				waitpid(pid, &stat_val, 0);
		  }
		}
		DISABLE_PSYNC

		cleanup("If SIGABRT caused a core dump file, remove it.");
		start
			if(access(CORE, F_OK) == FILE_EXISTS) {
				Zunlink(CORE);
			}
		finis
	finis


	assertion(7, BASE);		/* Also retests Assertions #8 and #9. */
	start
		ENABLE_PSYNC
		for(j=0; j<3; j++) {
		  for(i=0; i<NSIGS_JC; i++) {
			if( (sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGALRM) )
				continue;

			prep("Fork a child");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Set action of %s to %s.", sigs[i].signame,
							(j==0)?"SIG_DFL":(j==1)?"SIG_IGN":"catch signal");
				start
											/* Set action to test on oact. */
					act1.sa_handler = (j==0)?SIG_IGN:(j==1)?SIG_DFL:sig_int;
					Zsigemptyset(&act1.sa_mask);
					Zsigaddset(&act1.sa_mask, SIGALRM);
					act1.sa_flags = 0;
					Zsigaction(sigs[i].sig, &act1, &act2);

					act1.sa_handler = (j==0)?SIG_DFL:(j==1)?SIG_IGN:sig_int;
					Zsigemptyset(&act1.sa_mask);
					act1.sa_flags = 0;
					expecting(0);
					Zsigaction(sigs[i].sig, &act1, &act2);
				finis

				testing("Previous action for signal is stored in oact.");
				start
					if(act2.sa_handler != ( (j==0)?SIG_IGN: (j==1)?SIG_DFL:
																	sig_int) )
						error("Expected: %ul\nActual: %ul",
									(unsigned long) ((j==0)?SIG_IGN:SIG_DFL),
									(unsigned long) act2.sa_handler);

					if(j==2) {
						/*	sa_handler specifies signal catching function.	*/
						if( ! sigismember(&act2.sa_mask, SIGALRM))
						  error("Previous action for sa_mask field incorrect.");
					}

					if((act2.sa_flags & SA_NOCLDSTOP) != 0)
						error("%s\n%s%d.",
							"Expected sa_flags field to have value 0.",
							"Actual sa_flags field has value ", act2.sa_flags);
				finis

				SEND_PSYNC

				if(j<2) {
					SLEEP(2 * WAIT_FOR_TIME);
				}
				else {
					testing("Signal is caught when %s signal is received.",
															sigs[i].signame);
					start
						sig_rcvd = 0;
						Zkill(getpid(), sigs[i].sig);
						SLEEP(WAIT_FOR_TIME);
						if( ! sig_rcvd)
							error("Signal %s not caught.\n", sigs[i].sig);
					finis
				}

				done();
			}

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			if(j<2) {
				testing("%s termination occurs when %s signal is received.",
							(j==0)?"Abnormal":"Normal", sigs[i].signame);
				start
					Zkill(pid, sigs[i].sig);
					waitpid(pid, &stat_val, 0);
					if(j==0) {
						if(WTERMSIG(stat_val) != sigs[i].sig)
							error("%s %s.\n%s%d.",
								"Expected child process to be terminated by ",
								sigs[i].signame,
								"Value returned from macro WTERMSIG is ",
								WTERMSIG(stat_val) );
					}
					else
						if( ! WIFEXITED(stat_val))
							error("Child process did not terminate normally.");
				finis
			}
			else
				waitpid(pid, &stat_val, 0);
		  }
		}
		DISABLE_PSYNC

		cleanup("If SIGABRT caused a core dump file, remove it.");
		start
			if(access(CORE, F_OK) == FILE_EXISTS) {
				Zunlink(CORE);
			}
		finis
	finis

	assertion(8, BASE);
	start
		testing("Tested in Assertions #6 and #7.");
		start
		finis
	finis

	assertion(9, BASE);
	start
		testing("Tested in Assertions #6 and #7.");
		start
		finis
	finis

	done();
}
