#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>

static	struct		sigaction	act1;
static	pid_t		pid;
static	sigset_t	set, oset;
static	int			int_cnt;

void sig_catch(int n)
{
	prep("Interrupt routine sets process signal mask to SIGUSR1.");
	start
		Zsigemptyset(&set);
		Zsigaddset(&set, SIGUSR1);
		Zsigprocmask(SIG_SETMASK, &set, (sigset_t *)0);
	finis
}

void	ass_10(void (*)(int));
void	sig1_int(int);
void	sig2_int(int);
void	sig0_int(int);


int	main(void)
{
	int	i, j;

	setup(E_SIGACTION);
	funct(E_SIGACTION);

	assertion(10, BASE);
	start
						/* Test only for SIGINT */
		ass_10(sig0_int);

		ass_10(sig1_int);

		ass_10(sig2_int);
	finis


	assertion(11, BASE);
	start
		for(j=0; j<2; j++) {
		  for(i=0; i<NSIGS_JC; i++) {
			if( (sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGALRM) ||
						(sigs[i].sig == SIGUSR1) || (sigs[i].sig == SIGUSR2) )
				continue;

			FORK
				prep("Set process signal mask to SIGUSR2.");
				start
					Zsigemptyset(&set);
					Zsigaddset(&set, SIGUSR2);
					Zsigprocmask(SIG_SETMASK, &set, (sigset_t *)0);
				finis
				
				prep("Set action for signal %s to %s.",
								sigs[i].signame,
								(j==0)?"SIG_IGN":"signal catching function");
				prep("Interrupt this process with signal %s.",sigs[i].signame);
				start
					SIGNAL(sigs[i].sig, (j==0)?(SIG_IGN):sig_catch);

					Zkill(getpid(), sigs[i].sig);
					SLEEP(WAIT_FOR_TIME);
				finis

				testing("On return from signal being delivered,");
				testing("the original signal mask is restored.");
				start
					Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &set);
					if(sigismember(&set, sigs[i].sig))
						error("%s set in current process signal mask.",
															sigs[i].sig);
					if(sigismember(&set, SIGUSR1))
						error("SIGUSR1 set in process signal mask.");

					if( ! sigismember(&set, SIGUSR2))
						error("SIGUSR2 not set, process signal mask.");
				finis
			KROF
		  }
		}
	finis


	done();
}

void ass_10(void (*funct)(int))
{
						/* Test only for SIGINT */
	prep("Fork a child, have child set action to catch the signal.");
	start
		sv_flush();
		pid = Zfork();
	finis

	if(pid == (pid_t)0) {
		prep("Set process signal mask to SIGUSR1.");
		start
			Zsigemptyset(&set);
			Zsigaddset(&set, SIGUSR1);
			Zsigprocmask(SIG_SETMASK, &set, (sigset_t *)0);
		finis

		prep("Set action to catch the signal SIGINT.");
		start
			act1.sa_handler = funct;
			Zsigemptyset(&act1.sa_mask);
			Zsigaddset(&act1.sa_mask, SIGUSR2);
			act1.sa_flags = 0;
			Zsigaction(SIGINT, &act1, (struct sigaction  *) 0);
		finis

		prep("Send signal SIGINT to this child.");
		start
			Zkill(getpid(), SIGINT);
			if( ! int_cnt)
				error("Signal SIGINT not caught.");
		finis

		testing("On return from signal handler,");
		testing("original signal mask is restored.");
		start
			Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &set);

			if(sigismember(&set, SIGINT))
				error("SIGINT set in current process signal mask.");

			if(sigismember(&set, SIGUSR2))
				error("SIGUSR2 set in current process signal mask.");

			if( ! sigismember(&set, SIGUSR1))
				error("SIGUSR1 not set, current process signal mask.");
		finis

		done();
	}
	WAIT_FOR_PID(pid);
}

void	sig0_int(int i)
{
	testing("Signal catching function signal process mask is");
	testing("the union of current process's signal mask");
	testing("sigaction signal mask for the signal being delivered,");
	testing("and the signal being delivered.");
	start
		Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &set);

		if( ! sigismember(&set, SIGINT))
			error("SIGINT not set in current process signal mask.");

		if( ! sigismember(&set, SIGUSR2))
			error("SIGUSR2 not set in current process signal mask.");

		if( ! sigismember(&set, SIGUSR1))
			error("SIGUSR1 not set, current process signal mask.");

		int_cnt++;
	finis
}

void	sig1_int(int i)
{
	testing("Signal catching function signal process mask is");
	testing("the union of current process's signal mask");
	testing("sigaction signal mask for the signal being delivered,");
	testing("and the signal being delivered.");
	start
		Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &set);

		if( ! sigismember(&set, SIGINT))
			error("SIGINT not set in current process signal mask.");

		if( ! sigismember(&set, SIGUSR2))
			error("SIGUSR2 not set in current process signal mask.");

		if( ! sigismember(&set, SIGUSR1))
			error("SIGUSR1 not set, current process signal mask.");

		int_cnt++;
	finis

	testing("Signal mask remains in effect until sigprocmask() is called.");
	start
		Zsigemptyset(&set);
		Zsigprocmask(SIG_SETMASK, &set, &oset);

		if(sigismember(&set, SIGINT))
			error("SIGINT set in current process signal mask.");

		if(sigismember(&set, SIGUSR2))
			error("SIGUSR2 set in current process signal mask.");

		if(sigismember(&set, SIGUSR1))
			error("SIGUSR1 set, current process signal mask.");
	finis
}

void	sig2_int(int i)
{
	testing("Signal catching function signal process mask is");
	testing("the union of current process's signal mask");
	testing("sigaction signal mask for the signal being delivered,");
	testing("and the signal being delivered.");
	start
		Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &set);

		if( ! sigismember(&set, SIGINT))
			error("SIGINT not set in current process signal mask.");

		if( ! sigismember(&set, SIGUSR2))
			error("SIGUSR2 not set in current process signal mask.");

		if( ! sigismember(&set, SIGUSR1))
			error("SIGUSR1 not set, current process signal mask.");

		int_cnt++;
	finis

	testing("Signal mask remains in effect until sigsuspend() is called.");
	start
		int	stat_val;

		Zsigemptyset(&set);
		Zsigaddset(&set, SIGINT);
		pid = Zfork();
		if(pid == (pid_t)0) {
			Zsigsuspend(&set);			/* Suspend process and block SIGINT */
			done();
		}

		SLEEP(WAIT_FOR_TIME);
		Zkill(pid, SIGINT);			/* This signal should be blocked. */
		SLEEP(WAIT_FOR_TIME);
									/* Test process is still active. */
		if(Zwaitpid(pid, &stat_val, WNOHANG | WUNTRACED) != (pid_t)0)
			error("Child has exited or is stopped.");

		Zkill(pid, SIGKILL);		/* Terminate child process */
	finis

	expecting(VOID);
	WAIT_FOR_PID(pid);
}
