#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>


int	main(void)
{
	pid_t	pid;
	int		stat_val;
	struct	sigaction act1, act2, act3;
	sigset_t	set;

	setup(E_SIGACTION);
	funct(E_SIGACTION);

	assertion(13, BASE);
	start
		ENABLE_PSYNC
		FORK
			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Call sigaction() with \"sig\" of SIGCHLD and");
				prep("SA_NOCLDSTOP clear.");
				start
					Zsigemptyset(&act1.sa_mask);
					act1.sa_handler = SIG_DFL;
					act1.sa_flags = SA_NOCLDSTOP & (~SA_NOCLDSTOP);
					Zsigaction(SIGCHLD, &act1, &act2);
				finis

				prep("Delay while awaiting STOP signal from parent process.");
				start
					sv_flush();
					SEND_PSYNC
					sleep(2 * WAIT_FOR_TIME);
				finis

				done();
			}

			prep("Catch SIGCHLD signal");
			start
				sig_rcvd = 0;
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = sig_int;
				act1.sa_flags = 0;
				Zsigaction(SIGCHLD, &act1, &act2);
			finis

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			prep("Send SIGSTOP signal to child.");
			start
				Zkill(pid, SIGSTOP);
				sleep(WAIT_FOR_TIME);

				Zwaitpid(pid, &stat_val, WNOHANG | WUNTRACED);
				if( ! WIFSTOPPED(stat_val))
					error("Child process not stopped.");
			finis

			testing("SIGCHLD signal received by parent process.");
			start
				if(sig_rcvd == 0)
					error("SIGCHLD not received by parent process.");
				Zkill(pid, SIGCONT);
			finis

			expecting(VOID);	/* Child process reports it was interrupted. */
			WAIT_FOR_PID(pid);
		KROF
		DISABLE_PSYNC
	finis

	assertion(14, BASE);
	start
		ENABLE_PSYNC
		FORK
			prep("Set action for SIGCHLD to default settings.");
			start
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = SIG_DFL;
				act1.sa_flags = 0;
				Zsigaction(SIGCHLD, &act1, &act2);
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Delay while awaiting STOP signal from parent process.");
				start
					sv_flush();
					SEND_PSYNC
					sleep(2 * WAIT_FOR_TIME);
				finis

				done();
			}

			prep("Set action to catch SIGCHLD and SA_NOCLDSTOP set.");
			start
				sig_rcvd = 0;
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = sig_int;
				act1.sa_flags = SA_NOCLDSTOP;
				Zsigaction(SIGCHLD, &act1, &act2);
			finis

			WAIT_PSYNC

			prep("Send SIGSTOP signal to child.");
			start
				Zkill(pid, SIGSTOP);
				sleep(WAIT_FOR_TIME);

				Zwaitpid(pid, &stat_val, WNOHANG | WUNTRACED);
				if( ! WIFSTOPPED(stat_val))
					error("Child process not stopped.");
			finis

			testing("SIGCHLD signal not received by parent process.");
			start
				if(sig_rcvd != 0)
					error("SIGCHLD signal received by parent process.");
				Zkill(pid, SIGCONT);
			finis

			expecting(VOID);
			WAIT_FOR_PID(pid);
		KROF
		DISABLE_PSYNC
	finis

	assertion(15, BASE);
	start
		FORK
			prep("Call sigaction() to catch SIGINT.");
			start
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = sig_int;
				act1.sa_flags = 0;
				Zsigaction(SIGINT, &act1, &act2);
			finis

			testing("Once an action is installed for a specific signal,");
			testing("then it remains installed.");
			testing("SIGINTs received will be processed identically.");
			start
				Zkill(getpid(), SIGINT);
				SLEEP(WAIT_FOR_TIME);
				Zkill(getpid(), SIGINT);
				SLEEP(WAIT_FOR_TIME);

				if(sig_rcvd != 2)
					error("%s\n%s\n%s%d\n",
						"Number of SIGINT interrupts received incorrect.",
						"Expected: 2",
						"Actual: ",
						sig_rcvd);
			finis

			prep("Call sigaction() to ignore SIGINTs.");
			start
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = SIG_IGN;
				act1.sa_flags = 0;
				Zsigaction(SIGINT, &act1, &act2);
			finis

			testing("Once an action is installed for a specific signal,");
			testing("then it remains installed until another action is");
			testing("explicitly requested.");
			testing("SIGINTs received will now be ignored.");
			start
				Zkill(getpid(), SIGINT);
				SLEEP(WAIT_FOR_TIME);

				if(sig_rcvd != 2)
					error("%s\n%s\n%s%d\n",
						"Number of SIGINT interrupts received incorrect.",
						"Expected: 2",
						"Actual: ",
						sig_rcvd);

				Zsigpending(&set);
				if(sigismember(&set, SIGINT))
					error("SIGINT signal is pending.");
			finis
		KROF
	finis

	assertion(16, BASE);
	start
		FORK
			prep("Establish an action via sigaction() to ignore SIGINTs.");
			start
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = SIG_IGN;
				act1.sa_flags = 0;
				Zsigaction(SIGINT, &act1, &act2);
			finis

			prep("Establish an action via signal() to catch SIGINTs.");
			start
				signal(SIGINT, sig_int);
			finis

			testing("SIGINT signal will be caught.");
			start
				sig_rcvd = 0;
				Zkill(getpid(), SIGINT);
				SLEEP(WAIT_FOR_TIME);

				if(sig_rcvd != 1)
					error("%s\n%s\n%s%d\n",
						"Number of SIGINT interrupts received incorrect.",
						"Expected: 1",
						"Actual: ",
						sig_rcvd);
			finis

			testing("The signal handler supplied by \"oact\" can be used by");
			testing("sigaction to re-establish the original signal() invoked");
			testing("signal handling function.");
			start
										/* Set action to ignore SIGINIT. */
				Zsigaction(SIGINT, &act1, &act2);
				Zkill(getpid(), SIGINT);
				SLEEP(WAIT_FOR_TIME);

				if(sig_rcvd != 1)
					error("%s\n%s\n%s%d\n",
						"Number of SIGINT interrupts received incorrect.",
						"Expected: 1",
						"Actual: ",
						sig_rcvd);
										/* Restore action to previous action,
										   catch SIGINIT.
										*/
				Zsigaction(SIGINT, &act2, (struct sigaction *)0);
				Zkill(getpid(), SIGINT);
				SLEEP(WAIT_FOR_TIME);

				if(sig_rcvd != 2)
					error("%s\n%s\n%s%d\n",
						"Number of SIGINT interrupts received incorrect.",
						"Expected: 2",
						"Actual: ",
						sig_rcvd);
			finis
		KROF
	finis

	assertion(17, BASE);
	start
		int	ret;

		prep("Set action for SIGKILL to SIG_DFL.");
		start
			act1.sa_handler = SIG_DFL;
			Zsigemptyset(&act1.sa_mask);
			act1.sa_flags = 0;
			errno = 0;
			ret = sigaction(SIGKILL, &act1, (struct sigaction *)0);
		finis

		testing("Return is either -1 with errno set to [EINVAL] or a");
		testing("value of 0 and errno is not set.");
		start
			if(ret == -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s",
						"Value of errno incorrect.",
						"Expected: [EINVAL].",
						"Actual: ",
						err_lookup(errno) );
			}
			else {
				if(ret == 0) {
					if(errno != 0)
						error("%s\n%s\n%s%s",
							"Value of errno incorrect.",
							"Expected: errno not set.",
							"Actual: ",
							err_lookup(errno) );
				} else
					error("Return value was %d.", ret);
			}
		finis
	finis

	assertion(18, BASE);
	start
		int	ret;

		prep("Set action for SIGSTOP to SIG_DFL.");
		start
			act1.sa_handler = SIG_DFL;
			Zsigemptyset(&act1.sa_mask);
			act1.sa_flags = 0;
			errno = 0;
			ret = sigaction(SIGSTOP, &act1, (struct sigaction *)0);
		finis

		testing("Return is either -1 with errno set to [EINVAL] or a");
		testing("value of 0 and errno is not set.");
		start
			if(ret == -1) {
				if(errno != EINVAL)
					error("%s\n%s\n%s%s",
						"Value of errno incorrect.",
						"Expected: [EINVAL].",
						"Actual: ",
						err_lookup(errno) );
			}
			else {
				if(ret == 0) {
					if(errno != 0)
						error("%s\n%s\n%s%s",
							"Value of errno incorrect.",
							"Expected: errno not set.",
							"Actual: ",
							err_lookup(errno) );
				} else
					error("Return value was %d.", ret);
			}
		finis
	finis

	if(PCTS_INVALID_SIGNAL == 0) {
		assertion(19, NOT_TESTABLE);
	}
	else {
		assertion(19, BASE);
		start
			testing("Signal number argument is not a valid signal number.");
			start
				expecting(-1);
				expecting(EINVAL);
				Zsigaction(PCTS_INVALID_SIGNAL_NUMBER, &act1,
														(struct sigaction *)0);
			finis
		finis
	}

	assertion(20, BASE);
	start
		prep("Get current signal action for SIGKILL.");
		start
			Zsigaction(SIGKILL, (struct sigaction *)0, &act3);
		finis

		testing("Action specified to catch SIGKILL.");
		start
			act1.sa_handler = sig_int;
			Zsigemptyset(&act1.sa_mask);
			act1.sa_flags = 0;
			expecting(-1);
			expecting(EINVAL);
			Zsigaction(SIGKILL, &act1, (struct sigaction *)0);
		finis

		testing("No new signal action was installed.");
		start
			Zsigaction(SIGKILL, (struct sigaction *)0, &act2);
			if(act3.sa_handler != act2.sa_handler)
				error("Signal handler was changed.");
		finis

		testing("Action specified to ignore SIGKILL.");
		start
			act1.sa_handler = SIG_IGN;
			expecting(-1);
			expecting(EINVAL);
			Zsigaction(SIGKILL, &act1, (struct sigaction *)0);
		finis

		testing("No new signal action was installed.");
		start
			Zsigaction(SIGKILL, (struct sigaction *)0, &act2);
			if(act3.sa_handler != act2.sa_handler)
				error("Signal handler was changed.");
		finis
	finis

	assertion(21, BASE);
	start
		prep("Get current signal action for SIGSTOP.");
		start
			Zsigaction(SIGSTOP, (struct sigaction *)0, &act3);
		finis

		testing("Action specified to catch SIGSTOP.");
		start
			act1.sa_handler = sig_int;
			Zsigemptyset(&act1.sa_mask);
			act1.sa_flags = 0;
			expecting(-1);
			expecting(EINVAL);
			Zsigaction(SIGSTOP, &act1, (struct sigaction *)0);
		finis

		testing("No new signal action was installed.");
		start
			Zsigaction(SIGSTOP, (struct sigaction *)0, &act2);
			if(act3.sa_handler != act2.sa_handler)
				error("Signal handler was changed.");
		finis

		testing("Action specified to ignore SIGSTOP.");
		start
			act1.sa_handler = SIG_IGN;
			expecting(-1);
			expecting(EINVAL);
			Zsigaction(SIGSTOP, &act1, (struct sigaction *)0);
		finis

		testing("No new signal action was installed.");
		start
			Zsigaction(SIGSTOP, (struct sigaction *)0, &act2);
			if(act3.sa_handler != act2.sa_handler)
				error("Signal handler was changed.");
		finis
	finis

	done();
}
