#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_sigs.h>
#include	<sv_errno.h>
#include	<termios.h>
#include	<pcts_timer.h>

static	int	sigcont_rcvd;
static	void sig_cont(int i) {sigcont_rcvd++;}

int main(void)
{
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	pid_t	pid;
	struct	sigaction act;
	int		stat_val;

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	assertion(12, BASE);
	start
		ENABLE_PSYNC
		FORK
			prep("Non-controlling parent forks a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				pid_t	p_pid;

				prep("Have child delay.");
				start
					p_pid = Zgetppid();
					SEND_PSYNC
					SLEEP(WAIT_FOR_TIME);
				finis

				testing("When the non-controlling parent process is");
				testing("terminated by a signal without waiting for its");
				testing("child processes to terminate, child processes are");
				testing("not terminated and are assigned a new parent");
				testing("process ID.");
				start
					if(p_pid == Zgetppid())
						error("%s\n%s(%ld) has not changed.",
							"Expecting parent process ID to have changed.",
							"Actual parent process ID ",
							(long) p_pid);
					SEND_PSYNC
				finis

				done();
			}

			prep("Send signal to terminate parent process.");
			start
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
				SIGNAL(SIGINT, SIG_DFL);
				Zkill(getpid(), SIGINT);
			finis
		KROF
		SET_TIMEOUT
						/* Wait for child to be ready to exit. */
			WAIT_PSYNC
		CLEAR_ALARM
		SLEEP(WAIT_FOR_TIME)

		DISABLE_PSYNC
	finis


	assertion(13, BASE);
	start
		ENABLE_PSYNC
		FORK
			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				SIGNAL(SIGINT, SIG_DFL);
				SEND_PSYNC
				sleep(2 * WAIT_FOR_TIME);
				done();
			}

			prep("Have parent catch expected SIGCHLD signal.");
			start
				sig_rcvd = 0;
				SIGNAL(SIGCHLD, sig_int);
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
				SIGNAL(SIGINT, SIG_DFL);
				Zkill(pid, SIGINT);
			finis

			testing("SIGCHLD signal received by parent process when child");
			testing("is terminated by a signal.");
			start
				SET_TIMEOUT
					waitpid(pid, &stat_val, 0);
				CLEAR_ALARM
				if(sig_rcvd == 0)
					error("SIGCHLD not received by parent when child exited.");
			finis
		KROF
		DISABLE_PSYNC
	finis

	assertion(16, BASE);
	start
	  ENABLE_PSYNC
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {			/* Child process */
			prep("Establish a new process group, then fork a grandchild");
			prep("process.");
			start
				Zsetpgid((pid_t)0, (pid_t)0);
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {		/* Grandchild process */
				prep("Have grandchild process catch SIGHUP and SIGCONT.");
				start
					sig_rcvd = 0;
					SIGNAL(SIGHUP, sig_int);

					sigcont_rcvd = 0;
					Zsigemptyset(&act.sa_mask);
					act.sa_handler = sig_cont;
					act.sa_flags = 0;
					Zsigaction(SIGCONT, &act, (struct sigaction *)0);
				finis

				prep("Fork a great-grandchild process.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {	/* great-grandchild */
					prep("Have great-grandchild send SIGSTOP signal");
					prep("to itself.");
					start
						SEND_PSYNC
						sv_flush();
						Zkill(getpid(), SIGSTOP);
					finis

					testing("SIGHUP and SIGCONT received by stopped");
					testing("great-grandchild when orphaned.");
					start
						if(sig_rcvd == 0)
							error("SIGHUP not received by great-grandchild.");
						if(sigcont_rcvd == 0)
							error("SIGCONT not received by great-grandchild.");
					finis

					done();
				}

				/* Grandchild process */
				prep("Have grandchild send SIGSTOP signal to itself.");
				start
					sv_flush();
					Zkill(getpid(), SIGSTOP);
				finis

				testing("SIGHUP and SIGCONT received by grandchild.");
				start
					if(sig_rcvd == 0)
						error("SIGHUP not received by grandchild.");
					if(sigcont_rcvd == 0)
						error("SIGCONT not received by grandchild.");
				finis

				done();
			}

			prep("Have child process be terminated by a signal.");
			start
							/* Wait till grand-children are setup. */
				SET_ALARM(MAX_TIME_TO_RUN)
					WAIT_PSYNC
				CLEAR_ALARM
				SLEEP(WAIT_FOR_TIME);	/* Allow SIGSTOPs to be processed. */
				SIGNAL(SIGINT, SIG_DFL);
				Zkill(getpid(), SIGINT);
			finis

		  	done();
		}
		WAIT_FOR_PID(pid);				/* Orphaned process group created. */

		SLEEP(2 * WAIT_FOR_TIME);		/* Allow grandchild processes to exit.*/

	  DISABLE_PSYNC
	finis

	assertion(17, BASE);
	start
		ENABLE_PSYNC
			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Clear current process signal mask.");
				start
					sigset_t	mask;

					Zsigemptyset(&mask);
					Zsigprocmask(SIG_SETMASK, &mask, (sigset_t *)0);
				finis

				prep("Have child process wait for termination signal.");
				start
					sv_flush();
					SEND_PSYNC
					SLEEP(TIMEOUT);
				finis

				done();
			}

			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM

			prep("Send SIGCHLD signal to child.");
			start
				Zkill(pid, SIGCHLD);
			finis

			testing("Default action for SIGCHLD is to ignore the signal.");
			start
				Zwaitpid(pid, &stat_val, 0);
				if(WIFSIGNALED(stat_val))
					error("%s\n%s%s",
							"Expected child to exit normally.",
							"Value returned from WTERMSIG is ",
							sig_lookup(WTERMSIG(stat_val)) );
			finis
		DISABLE_PSYNC
	finis


	if(GTI_NOT_PROVIDED) {
		GTI_NOT_PROVIDED_RESULT(14, BASE)
		GTI_NOT_PROVIDED_RESULT(15, BASE)
		done();
	}

	assertion(14, BASE);
	start
		ENABLE_PSYNC
		FORK
			prep("Establish controlling terminal process.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);
			finis

			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {			/* Child process */
				pid_t	c_ppid, gc_ppid, ggc_ppid;
	
				prep("Have child processes catch SIGHUP.");
				start
					SIGNAL(SIGHUP, sig_int);
					sig_rcvd = 0;
				finis
	
				prep("Fork a grandchild process.");
				start
					sv_flush();
					pid = Zfork();
				finis
	
				if(pid == (pid_t)0) {		/* Grandchild process */
					prep("Fork a great grandchild process.");
					start
						sv_flush();
						pid = Zfork();
					finis
	
					if(pid == (pid_t)0) {	/* great grandchild */
						testing("Great grandchild received SIGHUP signal.");
						start
							ggc_ppid = getppid();
							SEND_PSYNC
							sleep(3 * WAIT_FOR_TIME);  /* Interrupt by SIGHUP */
							if(sig_rcvd == 0)
								error("SIGHUP not received.");
						finis
	
						done();
					}
	
					/* Grand child process */
					testing("Grandchild process received a SIGHUP signal.");
					start
						gc_ppid = getppid();
						sleep(3 * WAIT_FOR_TIME);	/* Interrupted by SIGHUP */
						if(sig_rcvd == 0)
							error("SIGHUP not received.");
					finis
	
					WAIT_FOR_PID(pid);
	
					done();
				}
	
				/* Child process */
				testing("Child process received a SIGHUP signal.");
				start
					c_ppid = getppid();
 								/* Allow ample time for parent to exit. */
					sleep(3 * WAIT_FOR_TIME);	/* Interrupted by SIGHUP */
					if(sig_rcvd == 0)
						error("SIGHUP not received.");
				finis
	
				testing("Child's parent process ID has changed.");
				start
					if(c_ppid == getppid()) {
								/*	Allow for SIGHUP to be processed before
									the implementation has changed the ppid
									of the children.
								*/
						SLEEP(WAIT_FOR_TIME);
						if(c_ppid == getppid())
							error("Child's ppid still is %ld.", (long)c_ppid);
					}
				finis
	
				WAIT_FOR_PID(pid);
	
				done();
			}
	
			prep("Parent process is terminated by a signal.");
			start
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
				SLEEP(WAIT_FOR_TIME);	/* All child processes now sleeping. */
				SIGNAL(SIGINT, SIG_DFL);
				Zkill(getpid(), SIGINT);
			finis
		KROF
		DISABLE_PSYNC

		SLEEP(WAIT_FOR_TIME);		/* Sleep, ensure completion of test. */
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	assertion(15, BASE);
	start
		pid_t	ct_pgrp;

		FORK
			prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
			start
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

				SIGNAL(SIGHUP, SIG_IGN);
				ct_pgrp = Ztcgetpgrp(tty_fildes);
			finis

			prep("Fork a process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				testing("When the controlling parent process is terminated");
				testing("by a signal, the controlling terminal is");
				testing("disassociated from the session.");
				start
					SLEEP(WAIT_FOR_TIME);	/* Allow time for parent to exit. */
					if(tcgetpgrp(tty_fildes) == ct_pgrp)
						error("Controlling terminal not disassociated from session.");
				finis

				done();
			}

			SIGNAL(SIGINT, SIG_DFL);
			Zkill(getpid(), SIGINT);
		KROF
		SLEEP(WAIT_FOR_TIME +1)
		SLEEP(GTI_LINE_CLOSE_DELAY)
	finis

	done();
}
