#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>


void sig_term(int i)
{	
	testing("Child process executed interrupt routine.");
	start
	finis

	done();
}
	
	
int main(void)
{
	pid_t		pid;
	int			stat_val, i;

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	assertion(32, BASE);
	start
		FORK
			prep("Fork a grandchild process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				sleep(2 * WAIT_FOR_TIME);

				done();
			}

			prep("Set action to catch SIGCHLD signal, clear SA_NOCLDSTOP");
			start
				sig_rcvd = 0;
				SIGNAL(SIGCHLD, sig_int);
			finis

			prep("Send SIGSTOP signal to child.");
			start
				Zkill(pid, SIGSTOP);
				sleep(WAIT_FOR_TIME);	/* Allow kill() to get processed. */
			finis

			testing("SIGCHLD signal received by parent when child stops.");
			start
				if( ! sig_rcvd)
					error("SIGCHLD not received when child stopped.");
			finis

			cleanup("Allow child to exit.");
			start
				SIGNAL(SIGCHLD, SIG_DFL);
				Zkill(pid, SIGCONT);
			finis

			WAIT_FOR_PID(pid);
		KROF
	finis
	
	assertion(33, BASE);
	start
		for(i=0; i<NSIGS; i++) {
			if(	(sigs[i].sig == SIGKILL) ||
				(sigs[i].sig == SIGCONT) ||
				(sigs[i].sig == SIGSTOP) ||
				(sigs[i].sig == SIGTSTP) ||
				(sigs[i].sig == SIGTTIN) ||
				(sigs[i].sig == SIGTTOU) )
					continue;
	 		prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Have child catch signal %s.",  sigs[i].signame);
				start
					SIGNAL(sigs[i].sig, sig_term);
					sig_rcvd = 0;
				finis

				prep("Stop child.");
				start
					Zkill(getpid(), SIGSTOP);
					pause();

					error("Routine sig_term() not executed.");
				finis

				done();
			}
	
			prep("Wait for child process to stop.");
			start
				SET_ALARM(TIMEOUT)
					Zwaitpid(pid, &stat_val, WUNTRACED);
				CLEAR_ALARM

				if( ! WIFSTOPPED(stat_val)) 
					error("Process not stopped.");
			finis

			prep("Generate %s signal to stopped child.", sigs[i].signame);
			start
				Zkill(pid, sigs[i].sig);
				SLEEP(WAIT_FOR_TIME);
			finis

			testing("Child process still active or stopped.");
			start
				if(Zwaitpid(pid, &stat_val, WNOHANG) != (pid_t)0)
					if(WIFEXITED(stat_val) || WIFSIGNALED(stat_val))
						error("Child has exited.");
			finis

			prep("Generate SIGCONT for child process.");
			start
				sv_flush();
				Zkill(pid, SIGCONT);
			finis

			testing("Child process terminated.");
			start
				WAIT_FOR_PID(pid);
			finis

		}
	finis
	
	assertion(34, BASE);
	start
		prep("Fork a process.");
		start
			sv_flush();
			pid = Zfork();
		finis
		
		if(pid == (pid_t)0) {
			prep("STOP child process.");
			start
				sv_flush();
				Zkill(getpid(), SIGSTOP);
				SLEEP(WAIT_FOR_TIME);
				error("Stopped process receiving SIGKILL signal continued.");
			finis

			done();
		}
		
		prep("Wait for child process to stop,");
		start
			SET_ALARM(TIMEOUT)
				Zwaitpid(pid, &stat_val, WUNTRACED);
			CLEAR_ALARM
	
			if( ! WIFSTOPPED(stat_val)) 
				error("Process not stopped.");
		finis
		
		testing("Stopped process is terminated by signal SIGKILL..");
		start
			Zkill(pid, SIGKILL);
			SET_ALARM(TIMEOUT)
				WAIT_FOR_PID(pid);
			CLEAR_ALARM
		finis
	finis
	
	done();
}
