#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<pcts_timer.h>

	/* Testing:  fcntl() is reentrant with respect to signals. */

#define	FILENAME	"3.3.1_45a"

static	struct	flock flock1, flock2;
static	int	fd;

static	int	catch_sig;
void sig_catch2(int i)
{	
	prep("Entering second signal catching routine.");
	start
		catch_sig++;
	finis
}

void sig_catch(int i)
{	
	prep("Unblock SIGINT in interrupt catcher.");
	start
		sigset_t sigint;

		sigemptyset(&sigint);
		sigaddset(&sigint, SIGINT);
		sigprocmask(SIG_UNBLOCK, &sigint, NULL);
	finis

	prep("Set second signal catcher for SIGINT.");
	start
		expecting(SUCCESS);
		SIGNAL(SIGINT , sig_catch2);
		catch_sig++;
	finis

	prep("Signal catcher calls fcntl() with F_SETLKW to wait");
	prep("to place a write lock on the file.");
	start
		errno = 0;
		fcntl(fd, F_SETLKW, &flock2);
		if(errno != EINTR)
			error("%s\n%s\n%s%s",
				"Value of errno incorrect from fcntl() in sig_catch().",
				"Expected value corresponding to [EINTR].",
				"Actual value corresponds to ",
				sig_lookup(errno) );
		errno = 0;
		catch_sig++;
	finis
}

int main(void)
{
	pid_t	pid;
	char	inbuf[50];

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	assertion(45, BASE);
	start
		prep("Create a temp file, set a shared lock on the entire file.");
		start
			flock1.l_type = F_RDLCK;
			flock1.l_whence = SEEK_SET;
			flock1.l_start = ((off_t)0);
			flock1.l_len = ((off_t)20);

			fd = temp_open(FILENAME, O_CREAT | O_TRUNC | O_RDWR, S_IRWXU);
			Zwrite(fd, inbuf, 20);
			Zfcntl(fd, F_SETLK, &flock1);
		finis

	 	prep("Fork child process.");
		   start
			sv_flush();
			pid = Zfork();
		finis
	
		if(pid == (pid_t)0) {
			prep("Set signal catcher for SIGINT.");
			start
				SIGNAL(SIGINT, sig_catch);
			finis
			
			flock2.l_type = F_WRLCK;
			flock2.l_whence = SEEK_SET;
			flock2.l_start = ((off_t)0);
			flock2.l_len = ((off_t)5);

			prep("Child calls fcntl() with F_SETLKW to wait to place");
			prep("a write lock on the file.");
			start
				errno = 0;
				fcntl(fd, F_SETLKW, &flock2);
				if(errno != EINTR)
					error("%s\n%s\n%s%s",
						"Value of errno incorrect from fcntl() in main().",
						"Expected value corresponding to [EINTR].",
						"Actual value corresponds to ",
						sig_lookup(errno) );
			finis

			testing("Function fcntl() is reentrant with respect to signals.");
			start
				if(catch_sig != 3)
					error("The fcntl() function is not reentrant.");
			finis

			done();
		}

		prep("Issue SIGINT to child.");
		start
			SLEEP(WAIT_FOR_TIME);	/* Allow time for child to call fcntl() */
			Zkill(pid, SIGINT);
		finis

		prep("Issue a second SIGINT to child to interrupt the interrupt.");
		start
			SLEEP(WAIT_FOR_TIME);	/* Allow for sig_catch() to call fcntl() */
			Zkill(pid, SIGINT);
		finis 

		WAIT_FOR_PID(pid);
	finis

	done();
}
