#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>

#include	<sv_env.h>
#include	<pcts_timer.h>

	/* Testing:  open() is reentrant with respect to signals. */

#define	FILENAME	"3.3.1_45c"

static	char path[_POSIX_PATH_MAX +1];

static int	catch_sig;
void sig_catch2(int i)
{	
	prep("Entering second signal catching routine.");
	start
		catch_sig++;
	finis
}

void sig_catch(int i)
{	
	prep("Unblock SIGINT in interrupt catcher.");
	start
		sigset_t sigint;

		sigemptyset(&sigint);
		sigaddset(&sigint, SIGINT);
		sigprocmask(SIG_UNBLOCK, &sigint, NULL);
	finis

	prep("Set second signal catcher for SIGINT.");
	start
		SIGNAL(SIGINT, sig_catch2);
		catch_sig++;
	finis

	SET_TIMEOUT
		prep("Signal catcher calls open().");
		start
			errno = 0;
			open(path, O_WRONLY);
			if(errno != EINTR)
				error("%s\n%s\n%s%s",
					"Value of errno incorrect from open() in sig_catch().",
					"Expected value corresponding to EINTR.",
					"Actual value corresponds to ",
					sig_lookup(errno) );
			errno = 0;
			catch_sig++;
		finis
	CLEAR_ALARM
}

int main(void)
{
	pid_t	pid;

	setup(E_SIGNAL_CON);
	ass_nu = 45;

	prep("Establish a path to test file and create FIFO.");
	start
		Zstrcpy(path, temp_path());
		Zstrcat(path, FILENAME);
		Zmkfifo(path, PROT_ALL);
	finis

 	prep("Fork a child process to test open() is reentrant.");
	   start
		sv_flush();
		pid = Zfork();
	finis

	if(pid == (pid_t)0) {
		prep("Set signal catcher for SIGINT.");
		start
			SIGNAL(SIGINT, sig_catch);
		finis
		
		SET_TIMEOUT
			prep("Child calls open().");
			start
				errno = 0;
				open(path, O_WRONLY);
				if(errno != EINTR)
					error("%s\n%s\n%s%s",
						"Value of errno incorrect from open() in main().",
						"Expected value corresponding to [EINTR].",
						"Actual value corresponds to ",
						sig_lookup(errno) );
			finis
		CLEAR_ALARM

		testing("The function open() is reentrant with respect to signals.");
		start
			if(catch_sig != 3)
				error("The open() function is not reentrant.");
		finis

		done();
	}

	prep("Issue SIGINT to child.");
	start
		SLEEP(WAIT_FOR_TIME);		/* Allow time for child to call open(). */
		Zkill(pid, SIGINT);
	finis

	prep("Issue a second SIGINT to child to interrupt the interrupt.");
	start
		SLEEP(WAIT_FOR_TIME);	/* Allow time for sig_catch() to call open(). */
		Zkill(pid, SIGINT);
	finis 

	WAIT_FOR_PID(pid);

	done();
}
