#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<pcts_timer.h>

	/* Testing:  waitpid() is reentrant with respect to signals. */

static	int	stat_val;
static int	catch_sig;
static	pid_t	pid;
void sig_catch2(int i)
{	
	prep("Entering second signal catching routine.");
	start
		catch_sig++;
	finis
}

void sig_catch(int i)
{	
	prep("Unblock SIGINT in interrupt catcher.");
	start
		sigset_t sigint;

		sigemptyset(&sigint);
		sigaddset(&sigint, SIGINT);
		sigprocmask(SIG_UNBLOCK, &sigint, NULL);
	finis

	prep("Set second signal catcher for SIGINT.");
	start
		SIGNAL(SIGINT, sig_catch2);
		catch_sig++;
	finis

	prep("Signal catcher calls waitpid().");
	start
		errno = 0;
		waitpid(pid, &stat_val, 0);
		if(errno != EINTR)
			error("%s\n%s\n%s%s",
				"Value of errno incorrect from waitpid() in sig_catch().",
				"Expected value corresponding to [EINTR].",
				"Actual value corresponds to ",
				sig_lookup(errno) );
		errno = 0;
		catch_sig++;
	finis
}

int main(void)
{
	setup(E_SIGNAL_CON);
	ass_nu = 45;

 	prep("Fork a child process to test waitpid() is reentrant.");
	   start
		sv_flush();
		pid = Zfork();
	finis

	if(pid == (pid_t)0) {
		prep("Set signal catcher for SIGINT.");
		start
			SIGNAL(SIGINT, sig_catch);
		finis
		
		prep("Fork a grandchild process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Grandchild goes to sleep.");
			start
				SLEEP(TIMEOUT);
			finis

			done();
		}


		prep("Child calls waitpid().");
		start
			SLEEP(WAIT_FOR_TIME);	/* Allow time for grandchild to sleep. */
			errno = 0;
			waitpid(pid, &stat_val, 0);
			if(errno != EINTR)
				error("%s\n%s\n%s%s",
					"Value of errno incorrect from waitpid() in main().",
					"Expected value corresponding to [EINTR].",
					"Actual value corresponds to ",
					sig_lookup(errno) );
		finis

		testing("The function waitpid() is reentrant with respect to signals.");
		start
			if(catch_sig != 3)
				error("The function waitpid() is not reentrant.");
		finis

		cleanup("Wait for \"grandchild\" process to successfully exit.");
		start
			expecting(SUCCESS);
			Zwaitpid(pid, &stat_val, 0);
		finis

		done();
	}

	prep("Issue SIGINT to child.");
	start
		SLEEP(2 * WAIT_FOR_TIME);  /* Allow time for child to call waitpid(). */
		Zkill(pid, SIGINT);
	finis

	prep("Issue a second SIGINT to child to interrupt the interrupt.");
	start
		SLEEP(WAIT_FOR_TIME);	/* Allow for sig_catch() to call waitpid(). */
		Zkill(pid, SIGINT);
	finis

	WAIT_FOR_PID(pid);

	done();
}
