#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<termios.h>
#include	<pcts_timer.h>

	/* Testing:  tcdrain() is reentrant with respect to signals. */

static	int		tty_fildes, loop_fildes;
static	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;

void sig_catch(int i)
{	
	testing("tcdrain() is reentrant with respect to signals.");
	start
		expecting(SUCCESS);
		Ztcflow(tty_fildes, TCOON);		/* Enable output data flow. */

		SET_TIMEOUT
			errno = 0;
			tcdrain(tty_fildes);
			if(errno != 0)
				error("%s\n%s\n%s%s",
					"Error resulted from call to tcdrain() in sig_catch().",
					"Expected no error.",
					"Actual value of errno is ",
					sig_lookup(errno) );
			errno = 0;
		CLEAR_ALARM
	finis
}

int main(void)
{
	pid_t	pid;
	size_t	nb_xmit;
	char 	outbuf[_POSIX_MAX_INPUT +1];
	int		flags;

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	assertion(47, EXTENDED);
	assertion(48, EXTENDED);

	if(GTI_NOT_PROVIDED || (PCTS_GTI_BUFFERS_OUTPUT == 0)) {
		assertion(46, NOT_TESTABLE);
		done();
	}

	assertion(46, BASE);
	start
		ENABLE_PSYNC
		prep("Fork a child process for testing tcdrain().");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Opening terminal file and set initial state.");
			start
				exit_on(ERROR);
				termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

				cfsetispeed(tty_termios_p, baud_max());
				cfsetospeed(tty_termios_p, baud_max());
				cfsetispeed(loop_termios_p, baud_max());
				cfsetospeed(loop_termios_p, baud_max());
				tty_termios_p->c_lflag &= ~ICANON;

				Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
				Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);

								/*	Ensure upcoming write does not block
									because output high water mark is exceeded.
								*/
				flags = fcntl(tty_fildes, F_GETFL);
				fcntl(tty_fildes, F_SETFL, flags | O_NONBLOCK);
			finis

			prep("Set up output buffer.");
			start
					/* Number of bytes to transmit to avoid timer expiration. */
				nb_xmit = (baud_value(baud_max())/10)*TIMEOUT/2;
				if(nb_xmit > _POSIX_MAX_INPUT)
					nb_xmit = _POSIX_MAX_INPUT;
				memset(outbuf, (int)'A', nb_xmit);
				Ztcflush(tty_fildes, TCIOFLUSH);
			finis

			prep("Write data to controlling terminal.");
			start
				Ztcflow(tty_fildes, TCOOFF);	/* Inhibit output data flow. */
									/* Since the file description for
										tty_fildes is nonblocking the actual
										amount written may be less than nb_xmit.
									*/
				expecting(SUCCESS);
				Zwrite(tty_fildes, outbuf, nb_xmit);
			finis

			prep("Have tcdrain() get interrupted.");
			start
				SIGNAL(SIGINT, sig_catch);
				SEND_PSYNC
				tcdrain(tty_fildes);
			finis

			done();
		}

		prep("Wait for child to buffer output data.");
		start
			sv_flush();
			SET_TIMEOUT
				WAIT_PSYNC
			CLEAR_ALARM
		finis

		prep("Interrupt child processing tcdrain().");
		start
			sv_flush();
			Zkill(pid, SIGINT);
		finis

		WAIT_FOR_PID(pid);
		SLEEP(GTI_LINE_CLOSE_DELAY)
		DISABLE_PSYNC
	finis

	done();
}
