#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<sv_sigs.h>
#include	<termios.h>
#include	<pcts_timer.h>


	/* Testing: - tcdrain() stopped/restarted with interruption. */

int main(void)
{
	pid_t	pid;
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	char 	*outbuf;
	int		stat_val, i;
	size_t	max_input;
	size_t	nb_xmit;

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	assertion(57, EXTENDED);

	if(GTI_NOT_PROVIDED || (PCTS_GTI_BUFFERS_OUTPUT == 0)) {
		assertion(56, NOT_TESTABLE);
		done();
	}

	assertion(56, BASE);
	start
		max_input = g_max_input(PORT_TTY);
		if(max_input > (unsigned)(PCTS_MAX_INPUT)) /* Limit buffer size */
			max_input = (unsigned)(PCTS_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
		nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
		if(nb_xmit > max_input)
			nb_xmit = max_input;

		prep("Allocate outbuf buffer.");
		start
			outbuf = (char *) Zmalloc(max_input +1);
		finis

		ENABLE_PSYNC
		for(i=NSIGS_STOP; i<NSIGS; i++) {
			prep("Fork an isolation process for testing.");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Establish PORT_TTY as the control terminal and.");
				prep("PORT_LOOP.  Clear ICANON on PORT_LOOP, set speed to");
				prep("lowest rate supported.");
				start
					exit_on(ERROR);
					termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
								loop_termios_p);

					tty_termios_p->c_lflag &= ~ICANON;
					tty_termios_p->c_cc[VMIN] = 0;
					tty_termios_p->c_cc[VTIME] = 0;
					Zcfsetospeed(loop_termios_p, baud_min());
					Zcfsetispeed(loop_termios_p, baud_min());
					Zcfsetospeed(tty_termios_p, baud_min());
					Zcfsetispeed(tty_termios_p, baud_min());
						
					Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
					Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
				finis
	
				prep("Fork a child process for testing tcdrain().");
				start
					sv_flush();
					pid = Zfork();
				finis
		
				if(pid == (pid_t)0) {
					prep("Open terminal file and set initial state.");
					prep("Set up output buffer.");
					start
						memset(outbuf, (int)'A', max_input);
						Ztcflush(tty_fildes, TCIOFLUSH);
	
						if(sigs[i].sig != SIGSTOP) {
							SIGNAL(sigs[i].sig, SIG_DFL);
						}
						SIGNAL(SIGCONT, sig_int);
					finis
		
					prep("Write data to controlling terminal.");
					start
						expecting(nb_xmit);
						Zwrite(tty_fildes, outbuf, nb_xmit);	
								/*	Baud rate should be low enough so that
									the following tcdrain() is interrupted
									while data is being transmitted.
								*/
					finis
		
					testing("Call to tcdrain() reports being interrupted.");
					start
						SEND_PSYNC
						expecting(EINTR);
						Ztcdrain(tty_fildes);
					finis

					cleanup("Flush tty queues.");
					start
						Ztcflush(loop_fildes, TCIOFLUSH);
						Ztcflush(tty_fildes, TCIOFLUSH);
					finis
		
					done();
				}
		
				prep("Wait for child to write out data and call tcdrain().");
				start
					SET_TIMEOUT
						WAIT_PSYNC
					CLEAR_ALARM
					SLEEP(WAIT_FOR_TIME);
				finis
		
						/* Establish a non-orphaned process group. */
				prep("Make child foreground process group.");
				start
					Zsetpgid(pid, pid);
					Ztcsetpgrp(tty_fildes, pid);
				finis
		
				prep("Issue a %s while tcdrain() is executing.",
														sigs[i].signame);
				start
					Zkill(pid, sigs[i].sig);
					SLEEP(WAIT_FOR_TIME)
				finis
		
				testing("Child process stopped by %s.", sigs[i].signame);
				start
					Zwaitpid(pid, &stat_val, WUNTRACED);
			
					if( ! WIFSTOPPED(stat_val)) 
						error("Process did not stop.");
					else
						if(WSTOPSIG(stat_val) != sigs[i].sig) 
							error("%s\n%s%s\n%s%s",
								"WSTOPSIG not equal to expected signal.",
								"Expected: ",
								sigs[i].signame,
								"Actual: ",
								sig_lookup(stat_val) );
				finis
		
				prep("Issue signal SIGCONT to child.");
				start
					Zkill(pid, SIGCONT);
				finis
		
				testing("Child completes successfully.");
				start
					SET_TIMEOUT
						Zwaitpid(pid, &stat_val, 0);
					CLEAR_ALARM
				finis

				done();
			}

			WAIT_FOR_PID(pid);
			SLEEP(GTI_LINE_CLOSE_DELAY)
		}
		DISABLE_PSYNC
	finis
	
	done();
}
