#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<sv_sigs.h>
#include	<termios.h>
#include	<pcts_timer.h>

	/* Testing: tcdrain() not affected by blocked signal. */


int main(void)
{
	pid_t	pid;
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	int		i;
	char	*outbuf;
	int		t_sec;
	size_t	max_input;
	size_t	nb_xmit;
	sigset_t	mask;

	setup(E_SIGNAL_CON);
	funct(E_SIGNAL_CON);

	if(GTI_NOT_PROVIDED) {
		assertion(61, NOT_TESTABLE);
		done();
	}

	assertion(61, BASE);
	start
		max_input = g_max_input(PORT_TTY);
		if(max_input > (unsigned)(PCTS_MAX_INPUT)) /* Limit buffer size */
			max_input = (unsigned)(PCTS_MAX_INPUT);

					/* Number of bytes to transmit to avoid timer expiration. */
		nb_xmit = (baud_value(baud_min())/10)*TIMEOUT/2;
		if(nb_xmit > max_input)
			nb_xmit = max_input;

		prep("Allocate outbuf buffer.");
		start
			outbuf = (char *) Zmalloc(max_input +1);
		finis

		ENABLE_PSYNC
		for(i=0; i<NSIGS_JC; i++) {
			if(sigs[i].sig == SIGKILL)
				continue;
			prep("Fork a child process to test tcdrain().");
			start
				sv_flush();
				pid = Zfork();
			finis
	
			if(pid == (pid_t)0) {
				prep("Establish PORT_TTY as the control terminal and.");
				prep("PORT_LOOP.  Clear ICANON on PORT_LOOP, set speed to");
				prep("lowest rate supported.");
				start
					exit_on(ERROR);
					termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
												loop_termios_p);

					tty_termios_p->c_lflag &= ~ICANON;
					tty_termios_p->c_cc[VMIN] = 0;
					tty_termios_p->c_cc[VTIME] = 0;
					Zcfsetospeed(loop_termios_p, baud_min());
					Zcfsetispeed(loop_termios_p, baud_min());
					Zcfsetospeed(tty_termios_p, baud_min());
					Zcfsetispeed(tty_termios_p, baud_min());

					Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
					Ztcsetattr(tty_fildes, TCSANOW, loop_termios_p);
				finis
	
				prep("Set up output buffer.");
				start
					memset(outbuf, (int)'A', sizeof(outbuf));
	
					Ztcflush(tty_fildes, TCIOFLUSH);
				finis
	
	
				prep("Write data to controlling terminal.");
				start
					expecting(nb_xmit);
					Zwrite(tty_fildes, outbuf, nb_xmit);	
				finis
	
				prep("Set process signal mask to block signal %s.",
														sigs[i].signame);
				start
					SIGBLOCK(mask, sigs[i].sig);
				finis
	
				testing("tcdrain() not affected by blocked signal.");
				start
					SEND_PSYNC
					Ztcdrain(tty_fildes);
				finis
	
				cleanup("Flush tty queues.");
				start
					Ztcflush(loop_fildes, TCIOFLUSH);
					Ztcflush(tty_fildes, TCIOFLUSH);
				finis

				done();
			}
	
			prep("Wait for child to write out data.");
			start
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM
			finis
	
						/* Ensure "testing" last journal message. */
			testing("Issue signal %s to child process.", sigs[i].signame);
			start
						/*	Compute in tenths-of-seconds time to transfer
							nb_xmit/4 bytes at transfer rate selected.
						*/
				t_sec = ((nb_xmit * 10)/4)/(baud_value(baud_min())/10);
				dt_secs(t_sec);

				Zkill(pid, sigs[i].sig);
			finis
	
			WAIT_FOR_PID(pid);
			SLEEP(GTI_LINE_CLOSE_DELAY)
		}
		DISABLE_PSYNC
	finis
		
	done();
}
