#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>


int	main(void)
{
	pid_t	pid;
	int		i, j, k;
	sigset_t	set, oset;

	setup(E_SIGPROCMASK);
	funct(E_SIGPROCMASK);


		/*	A call to sigprocmask(SIG_BLOCK, set, oset) changes the
			signals blocked for the process to the union of the current
			blocked signals and the signal set pointed to by the non-NULL
			argument set, and returns zero.

			When oset is non-NULL, the previous mask is stored in the
			location pointed to by oset.
		*/
	assertion(5, BASE);
	start
		prep("Initialize process signals blocked to the empty set.");
		start
			Zsigemptyset(&set);
			Zsigprocmask(SIG_SETMASK, &set, &oset);
		finis

		for(i=0; i<NSIGS; i++) {
			if((sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGSTOP))
				continue;

			FORK
				testing("Setting process signal mask with SIG_BLOCK");
				testing("to block signal %s returns zero.", sigs[i].signame);
				testing("Send blocked signal to this process.");
				testing("All signals have been blocked.");
				start
					for(j=0; j<=i; j++) {		/* SIGBLOCK is the union */
						if((sigs[j].sig == SIGKILL) || (sigs[j].sig == SIGSTOP))
							continue;

						SIGNAL(sigs[j].sig, sig_int);	/* Catch signal. */

						Zsigemptyset(&set);
						Zsigaddset(&set, sigs[j].sig);
						expecting(0);
						Zsigprocmask(SIG_BLOCK, &set, &oset);

						if(sigs[j].sig != SIGCHLD)
							Zkill(getpid(), sigs[j].sig);
					}

							/*	Generate SIGCHLD signal. */
					FORK
					KROF

						/*	Allow time for signals to get delivered, avoid
							possible interaction of SIGALRM with sleep(). */
					dt_secs(WAIT_FOR_TIME *10);
			
					for(j=0; j<=i; j++) {
						/* NOTE: Pending SIGCONT discarded by STOP signal. */
						if(	(sigs[j].sig == SIGKILL) ||
							(sigs[j].sig == SIGSTOP) ||
							(sigs[j].sig == SIGCONT) )
								continue;
						Zsigpending(&set);
						if( ! sigismember(&set, sigs[j].sig) )
							error("Signal %s not a pending signal.",
															sigs[j].signame);
					}
				finis

				prep("Obtain the new oset for testing.");
				start
					Zsigprocmask(SIG_BLOCK, &set, &oset);
				finis

				testing("Previous mask is stored in oset.");
				start
					for(j = 0; j <= i; j++) {
						if((sigs[j].sig == SIGKILL) || (sigs[j].sig == SIGSTOP))
							continue;

						if( ! sigismember(&oset, sigs[j].sig))
							error("%s Not member of signal set stored in oset.",
															sigs[j].signame);
					}
				finis
			KROF
		}
	finis


		/*	A call to sigprocmask(SIG_UNBLOCK, set, oset) changes the
			signals blocked for the process to the intersection of the
			current blocked signals and the complement of the signal set
			pointed to by the non-NULL argument set, and returns zero.
			
			When oset is non-NULL the previous mask is stored in the
			location pointed to by oset.
		*/
	assertion(6, BASE);
	start
		for(i=0; i<NSIGS; i++) {
			if((sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGSTOP))
				continue;

			FORK
				prep("Initialize process mask to all signals blocked.");
				start
					Zsigfillset(&set);
					Zsigprocmask(SIG_SETMASK, &set, &oset);
				finis

				testing("Set process signal mask with SIG_UNBLOCK");
				testing("to unblock some signals returns zero.");
				start
					for(j = 0; j <= i; j++) {
						if((sigs[j].sig == SIGKILL) || (sigs[j].sig == SIGSTOP))
							continue;
						Zsigaddset(&set, sigs[i].sig);
					}

					expecting(0);
					Zsigprocmask(SIG_UNBLOCK, &set, &oset);
				finis

				sig_rcvd = 0;
				for(j=0, k=0; j<=i; j++) {
					if((sigs[j].sig == SIGKILL) || (sigs[j].sig == SIGSTOP))
						continue;

					prep("Set process to catch signal %s.", sigs[j].signame);
					prep("Send signal %s to this process. ", sigs[j].signame);
					start
						SIGNAL(sigs[j].sig, sig_int);
						Zkill(getpid(), sigs[j].sig);

						k++;
					finis
				}
				dt_secs(WAIT_FOR_TIME *10);

				testing("Signals sent are received.", sigs[j].signame);
				start
					if(sig_rcvd != k)
						error("Sent %d signals and received only %d signals.",
											k, sig_rcvd);
				finis

				prep("Obtain the new oset for testing.");
				start
					Zsigprocmask(SIG_BLOCK, &set, &oset);
				finis

				testing("Previous mask stored in oset when oset is non-NULL.");
				start
					for(j = 0; j <= i; j++) {
						if((sigs[j].sig == SIGKILL) || (sigs[j].sig == SIGSTOP))
							continue;

						if(sigismember(&oset, sigs[j].sig))
							error("%s member of signal set stored in oset.",
															sigs[j].signame);
					finis
				}
			KROF
		}
	finis

	done();
}
