#include	<pcts.h>

#include	<Zunistd.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<sv_sigs.h>
#include	<pcts_timer.h>

void	sig_term(int sig_num)
{
	error("Signal %s was caught, it should have been blocked.",
														sig_lookup(sig_num));
}


int	main(void)
{
	pid_t	pid;
	int		i, j;
	sigset_t	set, oset;

	setup(E_SIGPROCMASK);
	funct(E_SIGPROCMASK);

		/*	A call to sigprocmask(SIG_SETMASK, set, oset) changes the signals
			blocked for the process to the signal set pointed to by the
			non-NULL argument set and returns zero.
			
			When oset is non-NULL, the previous mask is stored in the location
			pointed to by oset.
		*/
	assertion(7, BASE);
	start
		prep("Initialize signal mask.");
		start
			Zsigemptyset(&set);
		finis

		j = 0;				/* Previous index to signal processed. */
		for(i=0; i<NSIGS; i++) {
			if((sigs[i].sig == SIGKILL) || (sigs[i].sig == SIGSTOP))
				continue;
			FORK
				prep("Set sigaction to catch signal.");
				start
					SIGNAL(sigs[i].sig, sig_term);

					if(j>0)
						SIGNAL(sigs[j].sig, sig_int);
				finis

				testing("Setting process signal mask with SIG_SETMASK");
				testing("to block signal %s returns zero.", sigs[i].signame);
				start
					Zsigaddset(&set, sigs[i].sig);
					expecting(0);
					Zsigprocmask(SIG_SETMASK, &set, &oset);
				finis

				prep("Send the blocked %s signal", sigs[i].signame);
				prep("to this process to obtain a pending signal.");
				start
					Zkill(getpid(), sigs[i].sig);
					dt_secs(WAIT_FOR_TIME *10);
				finis

			
				testing("Signal %s is blocked.", sigs[i].signame);
				start
					Zsigpending(&set);
					if( ! sigismember(&set, sigs[i].sig) )
						error("%s\n%s%s\n%s\n",
							"Signal not reported as pending.",
							"Expected pending signal:	",
							sigs[i].signame,
							"Actual signal number was not pending.");
				finis

				if(j>0) {
					prep("Send non-blocked %s signal to this process.",
															sigs[i].signame);
					start
						Zkill(getpid(), sigs[j].sig);
						dt_secs(WAIT_FOR_TIME *10);
					finis

					testing("Signal %s is no longer blocked.", sigs[j].signame);
					start
						Zsigpending(&set);
						if(sigismember(&set, sigs[j].sig) )
							error("Signal %s reported as pending.",
															sigs[j].signame);
					finis
				}

				testing("Previous mask is stored in oset.");
				start
					Zsigprocmask(SIG_SETMASK, &set, &oset);
					if( ! sigismember(&oset, sigs[i].sig))
						error("%s Not member of signal set stored in oset.",
															sigs[i -1].signame);
				finis

				j = i;
			KROF
		}
	finis

			/*	A call to sigprocmask(how, NULL, oset) does not change the
				process's signal mask and when oset is non-NULL and stores
				the process's signal mask in oset.
			*/
	assertion(8, BASE);
	start
		FORK
			prep("Initialize process signal mask to only SIGINT set.");
			start
				Zsigemptyset(&set);
				Zsigaddset(&set, SIGINT);
				Zsigprocmask(SIG_SETMASK, &set, &oset);
			finis

			testing("Call to sigprocmask(how, NULL, oset)");
			testing("does not change the process's signal mask and");
			testing("the previous mask is stored in oset.");
			start
				Zsigprocmask(SIG_BLOCK, (sigset_t *)0, &oset);
				Zsigprocmask(SIG_UNBLOCK, (sigset_t *)0, &oset);
				Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &oset);
				Zsigprocmask(SIG_SETMASK, (sigset_t *)0, &oset);

				Zkill(getpid(), SIGINT);
				SLEEP(WAIT_FOR_TIME)
				Zsigpending(&set);
				if( ! sigismember(&set, SIGINT) )
					error("Signal SIGINT not pending.");

				for(i=0; i<NSIGS; i++) {
					if(sigs[i].sig == SIGINT) {
						if( ! sigismember(&oset, SIGINT))
							error("SIGINT not set, signal mask altered.");
					}
					else {
						if(sigismember(&oset, sigs[i].sig))
							error("Process's signal mask altered.");
					}	
				}
			finis
		KROF
	finis

	done();
}
