#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Ztimes.h>

#include	<sv_env.h>
#include	<pcts_timer.h>

#define	DELAY_TIME	5
#define	FUDGE_FACTOR	1
#define	REP_COUNT	3


int	main(void)
{
	time_t	time_beg, time_end;
	int		stat_val;
	pid_t	pid, ret;
	struct	sigaction	act1, act2;

	setup(E_WAITPID);
	funct(E_WAITPID);

	assertion(4, BASE);
	start
		clock_t	t_start, t_end, clk_tcks;
		struct	tms	tmsbuf; 

		break_on(ERROR);

		ENABLE_PSYNC

		prep("Determine delay time in clock-ticks.");
		start
			clk_tcks = (clock_t) (Zsysconf(_SC_CLK_TCK) *DELAY_TIME);
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			testing("Have child sleep a fixed time.");
			start
				sv_flush();
				SET_TIMEOUT
					WAIT_PSYNC
				CLEAR_ALARM

				t_start = Ztimes(&tmsbuf);
				SLEEP(1);	/* Ensure/allow parent is scheduled. */
							/*	Delay in seconds + or - one clock-tick.  */
				while((t_start +clk_tcks) > times(&tmsbuf))
						;
			finis

			done();
		}

		testing("Call to waitpid() suspends execution until child terminates.");
		start
			SEND_PSYNC
			t_start = times(&tmsbuf);
			Zwaitpid((pid_t)-1, &stat_val, 0);
			t_end = times(&tmsbuf);
			if((t_end -t_start +FUDGE_FACTOR) < clk_tcks)
				error("%s\n%s%lu\n%s%lu\n",
					"Child process suspended for too short a time.",
					"Expected time (clock_ticks) >= to ",
					(unsigned long) clk_tcks,
					"Actual time (clock_ticks) was: ",
					(unsigned long) (t_end -t_start));
		finis
		DISABLE_PSYNC

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Fork a grandchild process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Have grandchild send parent a signal.");
				start
					sv_flush();
					SLEEP(WAIT_FOR_TIME);
					Zkill(getppid(), SIGINT);
					SLEEP(MAX_TIME_TO_RUN);
				finis

				done();
			}

			prep("Establish an action to terminate on signal.");
			start
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = SIG_DFL;
				act1.sa_flags = 0;
				Zsigaction(SIGINT, &act1, &act2);
			finis

			Zwaitpid((pid_t)-1, &stat_val, 0);

			done();
		}

		SLEEP(WAIT_FOR_TIME);

		testing("Call to waitpid(pid, stat_loc, 0) suspends execution");
		testing("until calling process receives a signal whose action is");
		testing("to terminate.");
		start
			Ztime(&time_beg);
			Zwaitpid((pid_t)-1, &stat_val, 0);
			Ztime(&time_end);

			if((time_end -time_beg) > WAIT_FOR_TIME)
				error("%s\n%s%lu\n%s%lu\n",
					"Child process suspended for too long a time.",
					"Expected: a time <= to ",
					(unsigned long) WAIT_FOR_TIME,
					"Actual: time was: ",
					(unsigned long) (time_end -time_beg));
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Fork a grandchild process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Have grandchild send parent a signal.");
				start
					sv_flush();
					SLEEP(WAIT_FOR_TIME);
					Zkill(getppid(), SIGINT);
					SLEEP(MAX_TIME_TO_RUN);
				finis

				done();
			}

			prep("Establish an action to catch a signal.");
			start
				sv_flush();
				Zsigemptyset(&act1.sa_mask);
				act1.sa_handler = sig_int;
				act1.sa_flags = 0;
				Zsigaction(SIGINT, &act1, &act2);
			finis

			waitpid((pid_t)-1, &stat_val, 0);

			done();
		}

		SLEEP(WAIT_FOR_TIME);

		testing("Call to waitpid(pid, stat_loc, 0) suspends execution");
		testing("until calling process receives a signal whose action is");
		testing("to catch a signal.");
		start
			Ztime(&time_beg);
			Zwaitpid((pid_t)-1, &stat_val, 0);

			Ztime(&time_end);

			if((time_end -time_beg) > WAIT_FOR_TIME)
				error("%s\n%s%lu\n%s%lu\n",
					"Child process suspended for too long a time.",
					"Expected a time <= to ",
					(unsigned long) WAIT_FOR_TIME,
					"Actual time was: ",
					(unsigned long) (time_end -time_beg));
		finis
	finis

	assertion(5, BASE);
	start
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Child process exits.");
			start
			finis

			done();
		}

		SLEEP(WAIT_FOR_TIME);

		testing("When status information is available prior to a call to");
		testing("waitpid(-1, &stat_val, 0), return shall be immediate.");
		start
			Ztime(&time_beg);
			Zwaitpid(-1, &stat_val, 0);
			Ztime(&time_end);

			if((time_end -time_beg) > 1)
				error("%s\n%s%lu\n",
					"Call to waitpid() did not return immediately.",
					"Actual time was: ",
					(unsigned long) (time_end -time_beg));
		finis
	finis

	assertion(6, BASE);
	start
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Child calls exit().");
			start
				sv_flush();
			finis

			done();
		}

		testing("Call to waitpid(pid, &stat_loc, 0) returns the child");
		testing("process ID.");
		start
			pid_t ret;

			ret = waitpid(-1, &stat_val, 0);

			if(ret != pid)
				error("%s\n%s%ld\n%s%ld\n",
					"Return from waitpid(pid, &stat_loc, 0) incorrect.",
					"Expected: child process ID of ",
					(long) pid,
					"Actual: value returned was ",
					(long) ret);
		finis

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Child calls exit().");
			start
				sv_flush();
			finis

			done();
		}

		SLEEP(WAIT_FOR_TIME); /* Ensure prep() is outputted before testing(). */

		testing("Call to waitpid(pid, NULL, 0) returns the child process ID.");
		start
			pid_t ret;

			ret = waitpid(pid, (int *)0, 0);

			if(ret != pid)
				error("%s\n%s%ld\n%s%ld\n",
					"Return from waitpid(pid, NULL, 0) is incorrect.",
					"Expected: child process ID of ",
					(long) pid,
					"Actual: value returned was ",
					(long) ret);
		finis
	finis

	assertion(7, BASE);
	start
		int	i;

		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			for(i=0; i<REP_COUNT; i++) {
				prep("Have child fork() grandchild.");
				start
					sv_flush();
					pid = Zfork();
				finis

				if(pid == (pid_t)0) {
					prep("Have child exit.");
					start
					finis

					done();
				}
				Zwaitpid(pid, &stat_val, 0);
			}
			done();
		}

		ret = Zwaitpid(pid, &stat_val, 0);

		testing("waitpid(pid, ...) returns only when child terminates.");
		start
			if(ret != pid)
				error("%s\n%s%ld\n%s%ld\n",
					"Return from waitpid(pid, &stat_loc, 0) incorrect.",
					"Expected: child process ID of ",
					(long) pid,
					"Actual: value returned was ",
					(long) ret);
		finis
	finis

	assertion(8, BASE);
	start
		int	i;

		for(i=0; i<REP_COUNT; i++) {
			prep("Fork a child process.");
			start
				sv_flush();
				pid = Zfork();
			finis

			if(pid == (pid_t)0) {
				prep("Have child exit.");
				start
				finis

				done();
			}
		}
		SLEEP(WAIT_FOR_TIME);

		testing("waitpid(0, ...) returns when any child process whose process");
		testing("group ID is equal to that of the calling process terminates.");
		start
			SET_ALARM(TIMEOUT)
				for(i=0; i<REP_COUNT; i++)
					Zwaitpid(0, &stat_val, 0);		/* Wait for any child. */
			CLEAR_ALARM
		finis
	finis

	assertion(9, BASE);
	start
		break_on(ERROR);
		prep("Fork a child process.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Have child exit.");
			start
			finis

			done();
		}

		SLEEP(WAIT_FOR_TIME);

		testing("When the argument pid is less than -1, waitpid() returns");
		testing("status when any child process whose process group ID is");
		testing("equal to the absolute value of the argument pid terminates.");
		start
			SET_ALARM(TIMEOUT)
				ret = Zwaitpid((pid_t)(-1*getpgrp()), &stat_val, 0);

				if(ret != pid)
					error("%s\n%s%ld\n%s%ld\n",
						"Return from waitpid(-pid, &stat_loc, 0) incorrect.",
						"Expected: child process ID of ",
						(long) pid,
						"Actual: value returned was ",
						(long) ret);
				CLEAR_ALARM
		finis
	finis

	done();
}
