/*	pcts_ass.h - NIST-PCTS Assertion Configuration Parameters */

#ifndef	_POSIX_SOURCE
#define	_POSIX_SOURCE
#endif

#include <config.h>
#include <type_chk.h>		/* Type checking macros */
#include <element.h>


/*	Assertion macros to provide consistent coding format 	*/
extern	char	test_rc;	/* Assertion Test Result Code		*/
extern	int		ass_nu;		/* Assertion Test number			*/

/*	Assertion Classifications and some test result codes		*/
#define	UNUSED				'1'
#define	UNSUPPORTED			'2'
#define	UNTESTED			'3'
#define	BASE				'4'
#define	EXTENDED			'8'
#define	EXTENDED_TESTED		'9'
#define	TO_BE_WRITTEN		'a'		/* Test has not yet been coded.			*/
#define	NOT_TESTABLE		'b'		/* Testing Constraint is not supported. */
#define	NOT_APPLICABLE		'c'		/* Assertion is not congruent with the
									   requirements for FIPS 151-2. 		*/

/*	A set of values is bitwise distinct if the set of
	values obtained by logically oring the members of each
	subset, of the set, is a set of unique integral values."
*/
#define	BITWISE_DISTINCT(str_name, num_values) { \
	int i; \
	unsigned long d_mask; \
	for(i=0, d_mask=0; i<num_values; i++) \
		d_mask |= str_name[i].mask; \
	for(i=0; i<num_values; i++) \
		if(d_mask & str_name[i].mask)  \
			d_mask = d_mask & (~str_name[i].mask); \
		else \
			error("Element %s is not bitwise distinct.", str_name[i].name); \
}

#define	UNIQUE_VALUES(str_name, num_values) { \
	int i, j; \
	for(i=0; i<num_values -1; i++) { \
		for(j=i+1; j<num_values; j++) { \
			if(str_name[i].value == str_name[j].value) \
				error("Value of %s same as value of %s.", \
						str_name[i].name, \
						str_name[j].name); \
		} \
	} \
}

/*	Each element is assigned a unique identifier.
	The identifier function is assigned a unique value			*/
enum element_values {
	/*				General Requirements		*/
	E_GEN_TERMS,
	E_GEN_CONCEPTS,
	E_ERR_NUMBS,
	E_PSD_TYPES,
	E_ENV_DESCR,
	E_C_LANG,
	E_NUM_LMTS,
	E_SYM_CONST,
	/*				Process Primitives			*/
	E_FORK,
	E_EXECL,
	E_EXECLE,
	E_EXECV,
	E_EXECVE,
	E_EXECLP,
	E_EXECVP,
	E_WAIT,
	E_WAITPID,
	E__EXIT,
	E_SIGNAL_CON,
	E_KILL,
	E_SIGEMPTYSET,
	E_SIGFILLSET,
	E_SIGADDSET,
	E_SIGDELSET,
	E_SIGISMEMBER,
	E_SIGACTION,
	E_SIGPROCMASK,
	E_SIGPENDING,
	E_SIGSUSPEND,
	E_ALARM,
	E_PAUSE,
	E_SLEEP,
	/*			Process Environment								*/
	E_GETPID,
	E_GETPPID,
	E_GETUID,
	E_GETEUID,
	E_GETGID,
	E_GETEGID,
	E_SETUID,
	E_SETGID,
	E_GETGROUPS,
	E_GETLOGIN,
	E_GETPGRP,
	E_SETSID,
	E_SETPGID,
	E_UNAME,
	E_TIME,
	E_TIMES,
	E_GETENV,
	E_CTERMID,
	E_TTYNAME,
	E_ISATTY,
	E_SYSCONF,
	/*				Files & Directories				*/
	E_DIRENT_H,
	E_OPENDIR,
	E_READDIR,
	E_REWINDDIR,
	E_CLOSEDIR,
	E_CHDIR,
	E_GETCWD,
	E_OPEN,
	E_CREAT,
	E_UMASK,
	E_LINK,
	E_MKDIR,
	E_MKFIFO,
	E_UNLINK,
	E_RMDIR,
	E_RENAME,
	E_STAT_H,
	E_STAT,
	E_FSTAT,
	E_ACCESS,
	E_CHMOD,
	E_CHOWN,
	E_UTIME,
	E_PATHCONF,
	E_FPATHCONF,
	/*				Input & Output Primitives			*/
	E_PIPE,
	E_DUP,
	E_DUP2,
	E_CLOSE,
	E_READ,
	E_WRITE,
	E_FCNTL_H,
	E_FCNTL,
	E_LSEEK,
	/*				Device & Class Specific				*/
	E_INTERFACE,
	E_SETTABLE,
	E_CFGETOSPEED,
	E_CFSETOSPEED,
	E_CFGETISPEED,
	E_CFSETISPEED,
	E_TCGETATTR,
	E_TCSETATTR,
	E_TCSENDBREAK,
	E_TCDRAIN,
	E_TCFLUSH,
	E_TCFLOW,
	E_TCGETPGRP,
	E_TCSETPGRP,
	/*			LANGUAGE SPECIFIC SERVICES	*/
	/*			CHAPTER 8					*/
	E_ASSERT,
	E_ISALNUM,
	E_ISALPHA,
	E_ISCNTRL,
	E_ISDIGIT,
	E_ISGRAPH,
	E_ISLOWER,
	E_ISPRINT,
	E_ISPUNCT,
	E_ISSPACE,
	E_ISUPPER,
	E_ISXDIGIT,
	E_TOLOWER,
	E_TOUPPER,
	E_ACOS,
	E_ASIN,
	E_ATAN,
	E_ATAN2,
	E_COS,
	E_SIN,
	E_TAN,
	E_COSH,
	E_SINH,
	E_TANH,
	E_EXP,
	E_FREXP,
	E_LDEXP,
	E_LOG,
	E_LOG10,
	E_MODF,
	E_POW,
	E_SQRT,
	E_CEIL,
	E_FABS,
	E_FLOOR,
	E_FMOD,
	E_SETJMP,
	E_LONGJMP,
	E_CLEARERR,
	E_FCLOSE,
	E_FEOF,
	E_FERROR,
	E_FFLUSH,
	E_FGETC,
	E_FGETS,
	E_FOPEN,
	E_FPUTC,
	E_FPUTS,
	E_FREAD,
	E_FREOPEN,
	E_FSEEK,
	E_FTELL,
	E_FWRITE,
	E_GETC,
	E_GETCHAR,
	E_GETS,
	E_PERROR,
	E_PRINTF,
	E_FPRINTF,
	E_SPRINTF,
	E_VPRINTF,
	E_VFPRINTF,
	E_PUTC,
	E_PUTCHAR,
	E_PUTS,
	E_REMOVE,
	E_REWIND,
	E_SCANF,
	E_FSCANF,
	E_SSCANF,
	E_SETBUF,
	E_TMPFILE,
	E_TMPNAM,
	E_UNGETC,
	E_ABS,
	E_ATOF,
	E_ATOI,
	E_ATOL,
	E_RAND,
	E_SRAND,
	E_CALLOC,
	E_FREE,
	E_MALLOC,
	E_REALLOC,
	E_ABORT,
	E_EXIT,
	E_BSEARCH,
	E_QSORT,
	E_STRCPY,
	E_STRCAT,
	E_STRTOK,
	E_STRNCPY,
	E_STRNCAT,
	E_STRCMP,
	E_STRNCMP,
	E_STRCHR,
	E_STRCSPN,
	E_STRSPN,
	E_STRPBRK,
	E_STRRCHR,
	E_STRSTR,
	E_STRLEN,
	E_ASCTIME,
	E_CTIME,
	E_GMTIME,
	E_LOCALTIME,
	E_MKTIME,
	E_STRFTIME,
	E_SETLOCALE,
	E_FILENO,
	E_FDOPEN,
	E_SIGSETJMP,
	E_SIGLONGJMP,
	E_TZSET,
	/*				System Database				*/
	E_GETGRGID,
	E_GETGRNAM,
	E_GETPWUID,
	E_GETPWNAM,
	/*				Data Interchange Format		*/
	E_DIF,
	E_TAR,
	E_CPIO
};
extern void e_NO_TRUNC(enum element_values);
extern void e_ENAMETOOLONG(enum element_values);

