/* pcts_timer.h - test timeout macros and constants */

#include <Zunistd.h>
#include <Zsignal.h>
#include <Zfcntl.h>
#include <Ztermios.h>

/*	Signal handling common routines and defines in
	"../src/tools/sv_sys/sv_env.c" */
extern	sigjmp_buf sigenv;
extern	volatile int sigalrm_rcvd;
extern	volatile int sig_rcvd;
extern	void sigalrm(int);
extern	void sig_int(int);
extern	struct sigaction s_sigaction;

#define SIGNAL(signal,handler) \
	expecting(SUCCESS); \
	Zsigemptyset(&s_sigaction.sa_mask); \
	s_sigaction.sa_handler = handler; \
	s_sigaction.sa_flags = 0; \
	Zsigaction(signal, &s_sigaction, NULL);

#define SIGBLOCK(set,signal) \
	expecting(SUCCESS); \
	Zsigemptyset(&set); \
	Zsigaddset(&set,signal); \
	Zsigprocmask(SIG_BLOCK, &set, NULL);

#define SIGUNBLOCK(set,signal) \
	expecting(SUCCESS); \
	Zsigemptyset(&set); \
	Zsigaddset(&set,signal); \
	Zsigprocmask(SIG_UNBLOCK, &set, NULL);

#define SIGHANDLE(signal)	SIGNAL(signal,sighandler)

/********************************
* Global Signal-handler flags
* Note: sigsetjmp/siglongmp cannot be used in the same test with
*  the PCTS break_on() feature.
********************************/

#define SET_ALARM(time) { \
	int	status; \
	SIGNAL(SIGALRM, sigalrm); \
	sigalrm_rcvd = 0; \
	alarm(time); \
	status = sigsetjmp(sigenv, 1); \
	if (status == 0) {

extern	unsigned unused_time;
#define CLEAR_ALARM } \
	expecting(SUCCESS); \
	unused_time = alarm(0); \
	SIGNAL(SIGALRM, SIG_DFL); \
	if (sigalrm_rcvd != 0) \
		error("Timed out."); \
	}

#define EXPECT_ALARM } \
	expecting(SUCCESS); \
	unused_time = alarm(0); \
	SIGNAL(SIGALRM,SIG_DFL); \
	if (sigalrm_rcvd == 0) \
		error("Did not time out."); \
	}

#define IGNORE_ALARM } \
	expecting(SUCCESS); \
	unused_time = alarm(0); \
	SIGNAL(SIGALRM,SIG_DFL); \
	}

#define TIMEOUT MAX_TIME_TO_RUN
#define SET_TIMEOUT SET_ALARM(TIMEOUT)

#define ENABLE_PSYNC { \
	int		p_syn[2], p_syna[2]; \
	char	b_syn[2], b_syna[2]; \
	expecting(SUCCESS); \
	Zpipe(p_syn); \
	Zpipe(p_syna);


#define SEND_PSYNC \
	expecting((unsigned int)1); \
	Zwrite(p_syn[1], "A", (unsigned int)1);

#define WAIT_PSYNC \
	expecting((unsigned int)1); \
	Zread(p_syn[0], b_syn, (unsigned int)1);

#define SEND_PSYNCA \
	expecting((unsigned int)1); \
	Zwrite(p_syna[1], "B", (unsigned int)1);

#define WAIT_PSYNCA \
	expecting((unsigned int)1); \
	Zread(p_syna[0], b_syna, (unsigned int)1);

#define DISABLE_PSYNC \
	expecting(SUCCESS); \
	Zclose(p_syn[0]); \
	Zclose(p_syn[1]); \
	Zclose(p_syna[0]); \
	Zclose(p_syna[1]); \
}
