/*
 *  This version assumes a ufs file system type
 *  change the following define if you want a different type
 *  i.e. sfs, s5 or vxfs
 */
#define __FSTYPE "ufs"
/*#define __FSTYPE "vxfs" */

/*
	The routines in this file interface with the NIST-PCTS to
	provide the following "target system support facilities" on
	some implementations:

		mount a file system read-only,
		mount a file system for reading and writing,
		unmount a file system,
		establish a controlling terminal.
*/

#include	<pcts.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<sys/stat.h>


/* #includes for UNIX SVR4 based implementations */

#include <sys/mount.h>
#include <sys/fstyp.h>

/*
	Routine to interface the NIST-PCTS to an implementation
	provided routine to mount a file system as read-only.
*/
mtrofs(dirname, fsname)
	char	*dirname, *fsname;
{
	return(mount(fsname, dirname, MS_FSS|MS_RDONLY, 
			sysfs(GETFSIND, __FSTYPE),NULL,0));

}


/*
	Routine to interface the NIST-PCTS to an implementation
	provided routine to mount a file system for reading and writing.
*/
mtrwfs(dirname, fsname)
	char	*dirname, *fsname;
{
	return(mount(fsname, dirname, MS_FSS, 
			sysfs(GETFSIND, __FSTYPE),NULL,0));
}


/*
	Routine to interface the NIST-PCTS to an implementation
	provided routine to mount a file system for reading and writing.
*/
unmtfs(dirname, fsname)
	char	*dirname, *fsname;
{
	return(umount(fsname));

	/*	If not supported, then use "return(-1);" instead.	*/
}


/*
	Routine to interface the NIST-PCTS to an implementation
	provided routine to establish a controlling terminal.
*/
int open_ctty(path, flags, mode)
	char	*path;
	int		flags;
	mode_t	 mode;
{
	if (flags & O_CREAT) 
		return(open(path, flags & ~O_NOCTTY, mode));
	else
		return(open(path, flags & ~O_NOCTTY));

}
