#ifndef	_POSIX_SOURCE
#define	_POSIX_SOURCE
#endif

#define SV_ENV
#define SV_ELEMENT

#include <pcts.h>

#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sv_env.h>

#define	PSIZE 128

extern	char	*sv_temp_path(void);
extern	char	*sig_lookup(int);
extern	void	sv_set_exit_val(int);
extern	int		sv_get_outfd(void);
extern	void	sv_catchsig(void (*)(int));
extern	void	sv_set_outfd(int);
extern	char	*sv_getlev(char *);

static	struct	sigaction sact_alrm;
static	struct	sigaction sact_kbd;

static char sv_prefix[ PSIZE ] = "";

struct sv_env sv_deflt = {
	(struct sv_env *) 0,	/* ptr to prev frame			*/
	SV_SUCC,		/* expecting success failure or void	*/
	1,			/* id of next test blk at current frame	*/
	1,			/* id of next child at current frame	*/
	1,			/* need current level "1.2.3" printed?	*/
	{	ERROR,		/* severity for unexp failure error	*/
		ERROR,		/* severity for unexp success error	*/
		ERROR,		/* severity for unexp errno error	*/
		SEVERR,		/* severity for unexp signal error	*/
		ERROR,		/* severity for unexp ret val error	*/
		ERROR,		/* severity for undef ret val error	*/
	},
	{	CONTINUE,	/* action for normal messages		*/
		CONTINUE,	/* action for errors of sev warning	*/
		CONTINUE,	/* action for errors of sev error	*/
		ABORT,		/* action for errors of sev severr	*/
	},
};

struct sv_env *sv_env = &sv_deflt;
struct sv_env *sv_xenv = &sv_deflt;

/*
 *	generate table index for severity
 */
int sv_sid(int severity)
{
	int i;
	for( i=0; i<SV_NSEVS; i++ )
		if( sv_sev_table[i]==severity )
			return(i);
	return(0);
}

/*
 *	return severity name
 */
char *sv_snm(int severity)
{
	return( sv_snam_table[sv_sid(severity)] );
}


/*
 *	severity declaration function
 */
void sv_sevset(int conditions, int severity)
{
	int i;
	for( i = 0; i < SV_NERRS; i++ )
		if( conditions & sv_err_table[i] )
			sv_xenv->sv_sev[i] = severity;
}


/*
 *	action declaration function
 */
void sv_actset(int severities, int action)
{
	int i;
	for( i = 0; i < SV_NSEVS; i++ )
		if( severities & sv_sev_table[i] )
			if(( sv_xenv->sv_act[i] = action ) == BREAK )
				(void)sv_jcpy( sv_xenv->sv_jmpb[i], sv_xenv->sv_jmp );
}

/*
 *	reset actions and/or severities, replace break points
 */
void sv_reset(int typs)
{
	int i;
	if( !sv_xenv->sv_oenv ) return;
	for( i = 0 ; i < SV_NSEVS; i++ )
		if( typs & sv_sev_table[i] ) {
			if( sv_xenv->sv_act[i] & BREAK)
				(void)sv_jcpy( sv_xenv->sv_jmpb[i],
				  sv_xenv->sv_oenv->sv_jmpb[i] );
			sv_xenv->sv_act[i] = sv_xenv->sv_oenv->sv_act[i];
		}
	for( i = 0 ; i < SV_NERRS; i++ )
		if( typs & sv_err_table[i] )
			sv_xenv->sv_sev[i] = sv_xenv->sv_oenv->sv_sev[i];
}

void sv_cproc(void)
{
	char	buf[ PSIZE ];

	(void) sv_getlev( buf );
	(void) sprintf(buf+strlen(buf), "<p%d>.",sv_xenv->sv_fcnt);	
	(void) strcpy( sv_prefix, buf );
	sv_xenv->sv_fcnt = 1;
	sv_set_exit_val( 0 );

}

void sv_pproc(void)
{
	sv_xenv->sv_fcnt++;
}

void sv_mproc(void)
{
	int fd;
	int outfd;
	unsigned len;
	char name[ 64 ];
	char buf[ PSIZE ];

 	(void) sprintf(name,"%s/tmp/%s/pid.%ld", PCTS_PATH, SUBTMP, (long)getpid());
	fd = open(name, O_CREAT | O_WRONLY, PROT_ALL);
	if( fd < 0 )
		warning("Can't create temp file: '%s'\n",name);
	else {
		outfd = sv_get_outfd();
		if( write(fd,(char *)&outfd,sizeof(int)) != sizeof(int))
			warning("Can't write to temp file: '%s'\n",name);
		else {
			(void) sv_getlev( buf );
			len = strlen( buf );
			if( write(fd,buf,len) != len)
				warning("Can't write temp file: '%s'\n",name);
		}
		if( close(fd) )
			warning("Can't close temp file: '%s'\n",name);
	}

}
/*
 *	Since test is running with a different process group id, 
 *	it cannot be killed with signals from tty.
 *	Leave a signal catching proc around with tty grpid which
 *	can pass the signal on to the test.
 *	If desired, this can be removed for security reasons.
 */
static int child_pid = -1;
static int caught_sig = 0;
static int got_ext_sig = 0;

static void timed_out(void)
{
	severr("%s%s%s\n%s\n%s%d\n",
		"test caught signal: ",
		sig_lookup( caught_sig ),
		" from outside its process group -",
		"Test NOT terminated.",
		"pid: ",
		child_pid );
}


static void keybd_sig(int sig)
{
	sact_kbd.sa_handler =(void(*)(int)) keybd_sig;
	sigemptyset(&sact_kbd.sa_mask);
	sact_kbd.sa_flags = 0;

	sact_alrm.sa_handler =(void(*)(int)) timed_out;

	sigemptyset(&sact_alrm.sa_mask);
	sact_alrm.sa_flags = 0;

	got_ext_sig = 1;	/* if wait ret's due to this, wait again */
	caught_sig = sig;

	if( child_pid <= 0 )	/* should never happen */
		_exit((int) SEVERR );
	sigaction(sig, &sact_kbd, (struct sigaction  *) 0);
	if( child_pid > 0 )
		(void) kill( (pid_t)(-1*child_pid), sig );
	sigaction(SIGALRM, &sact_alrm, (struct sigaction  *) 0);
	(void) alarm( (unsigned) 5 );/* wait 5 sec for children to suicide */
}

static void keybd_trap(void)
{
	int wpid;
	int sts;
	int sv_errno;

	child_pid = fork();
	if( child_pid < 0 ) {
		exit_on( SEVERR );
		severr("test cannot fork");
	}
	if( child_pid ) {
		exit_on( WARNING | ERROR | SEVERR );
/*
 *		if catch sig, kill child, set 5 sec alarm, and wait
 *		if alarm times out, note that child is still alive
 *		and exit
 */
		sv_catchsig( keybd_sig );
		errno = 0;
		do {
			got_ext_sig = 0;
			errno = 0;
			wpid = wait(&sts);
			sv_errno = errno;
		} while( got_ext_sig );
		(void) alarm( (unsigned) 0); /* may have come back from alarm */
		if ( wpid != child_pid ) {
			if( sv_errno == EINTR ) {
/*
 *				alarm went off, should have called 
 *				timed_out via alarm mechanism, so
 *				something is broken, but call it anyway
 */
				severr("%s\n%s\n%s%s%s\n%s\n%s%ld\n",
						"wait() failed.",
						"Signal handling failed.",
						"Test caught signal: ",
						sig_lookup(caught_sig),
						" from outside its process group.",
						"Test NOT terminated.",
						"pid: ",
						(long) child_pid);

			}
			severr("Wait returned unexpected value: %ld", (long)wpid );
		}
		if(WIFEXITED(sts) == 0)
			if(WIFSIGNALED(sts))
				error("%s %s\n",
					"Test terminated due to unexpected signal: ",
					sig_lookup(WTERMSIG(sts)) );
		_exit(WEXITSTATUS(sts));
	}
}


void sv_nproc(void)
{
	int		fd, cnt;
	int 	outfd;
	char 	name[64];
	extern	char	*err_lookup(int);

	sv_prefix[0] = '\0';
 	(void) sprintf(name, "%s/tmp/%s/pid.%ld", PCTS_PATH,SUBTMP, (long)getpid());
 	fd = open(name, O_RDONLY);
			/* First process of a test will not have a tmp/pid.???? file. */
	if(fd<0) {
		keybd_trap();		/* Enable keyboard signal trapping. */

							/*	Establish a new process group so errant
								child processes can be easily terminated.
							*/
		if(setpgid((pid_t)0, (pid_t)0) == -1)
			error("%s\n%s%s\n",
					"Call to setpgid() failed.",
					"Errno set to  ",
					err_lookup(errno));
		return;
	}

	if(read(fd, (char *)&outfd, sizeof(int)) != sizeof(int))
		error("Cannot initialize: read failed for temp file: '%s'\n", name);
	else {
		sv_set_outfd(outfd);
		if( ! (cnt=read(fd,sv_prefix,(PSIZE-1))) )
			error("Cannot initialize: read failed for temp file: '%s'\n",name);
		else
			sv_prefix[cnt] = '\0';
	}
	if(close(fd))
		error("cannot initialize: close failed for temp file: '%s'\n", name);
}


char *sv_newlev(struct sv_env *sv_xnv, char *buf)
{
	if(sv_xnv->sv_oenv)	
		(void) sv_newlev(sv_xnv->sv_oenv, buf);
	else
		if(sv_prefix[0])
			(void) sprintf(buf,"%s",sv_prefix);
	(void) sprintf(buf+strlen(buf), "%d." ,sv_xnv->sv_cnt);
	return(buf);
}


char *sv_getlev(char *buf)
{
	buf[0] = '\0';
	if(sv_xenv->sv_oenv)
		(void) sv_newlev(sv_xenv->sv_oenv, buf);
	return(buf);
}


void sv_prlev(void)
{
	char buf[256];
	buf[0] = '\0';
	sv_printf("%s ",sv_newlev( sv_xenv , buf ));
}


void sv_einit(struct sv_env *sv_env)
{
	sv_printf("\n");
	sv_flush();
	sv_ecpy( sv_env, sv_xenv );
	sv_env->sv_oenv = sv_xenv;
	sv_env->sv_cnt=1;
	sv_env->sv_fcnt=1;
	sv_env->sv_levflg=1;
	sv_xenv=sv_env;
}



/*  Common signal handling interrupts */
sigjmp_buf sigenv;
volatile int sigalrm_rcvd;
unsigned unused_time;
void sigalrm(int signo)
{
	siglongjmp(sigenv, ++sigalrm_rcvd);
}

volatile int sig_rcvd;
void sig_int(int signo)
{
	sig_rcvd++;
}

struct sigaction s_sigaction = {0, 0, 0};



/*
** Wait for a child process to exit, then return child's process status.
**
** Arguments
**	pid			Process ID of the child we are waiting for.
**	timeout		Maximum time (in seconds) to wait.
*/

#define EXPECTING		(sv_xenv->sv_expecting)
#define EXP_SUCCESS		(EXPECTING == SV_SUCC)
#define EXP_FAILURE		(EXPECTING == SV_FAIL || EXPECTING & SV_ERRN)
#define EXP_VOID		(EXPECTING == SV_VOID)

#define EXP_FAILURE		(EXPECTING == SV_FAIL || EXPECTING & SV_ERRN)
#define EXP_VOID		(EXPECTING == SV_VOID)

int wait_for(pid_t pid, unsigned timeout)
{
	int		stat_val, s_errno = 0;
	int		exp_success = FALSE;
	pid_t	ret_val = (pid_t)-1;
	struct	sigaction oact;

	if(EXP_SUCCESS)
		exp_success = TRUE;
	sigemptyset(&s_sigaction.sa_mask);
	s_sigaction.sa_handler = sigalrm;	/* Catch SIGALRM signal if it occurs. */
	s_sigaction.sa_flags = 0;
	sigaction(SIGALRM, &s_sigaction, &oact);
	sigalrm_rcvd = 0;
	alarm(timeout);
	if(sigsetjmp(sigenv, 1) == 0) {
		ret_val = wait(&stat_val);
		s_errno = errno;
	}

	unused_time = alarm(0);
	sigemptyset(&s_sigaction.sa_mask);
	s_sigaction.sa_handler = oact.sa_handler;	/* Reset SIGALRM action. */
	s_sigaction.sa_flags = oact.sa_flags;
	sigaction(SIGALRM, &s_sigaction, (struct sigaction *)NULL);
	if(sigalrm_rcvd != 0) {
		error("wait_for(): Timed-out waiting for child to exit.");
		(void) kill(pid, SIGKILL);
		ret_val = wait(&stat_val);
		s_errno = errno;
	}

	if(exp_success == TRUE) {
		if(ret_val == (pid_t) -1)
			error("wait_for: Call to wait() returned -1\n\t\twith errno of %s.",
								err_lookup(s_errno));
		else
			if(ret_val != pid) {
				if(sigalrm_rcvd)
					;
				else
					error("%s\n%s%ld\n%s%ld.",
						"wait_for(): Unexpected child returned from wait().",
						"Expected pid: ",
						(long)pid,
						"Actual pid: ",
						(long)ret_val);
			}
	}
	return(stat_val);
}
