#include	<pcts.h>
#include	<pcts_testv.h>

#include	<Zunistd.h>
#include	<sys/Zstat.h>
#include	<Zfcntl.h>
#include	<Zdirent.h>
#include	<Zutime.h>

#include	<sv_env.h>
#include	<sv_errno.h>

void perform(int, enum element_values, char *, char *);
char	*pathm_u;
static	struct stat st_buf;

/*  g_lname
	Create a pathname containing a component of length
			{NAME_MAX} +1 bytes,   If {NAME_MAX} < {PCTS_NAME_MAX}
		Otherwise
			{PCTS_NAME_MAX} bytes.
*/

char	*g_lname(void)
{
	char	*ptr;
	char	*ptr_end;
	long	path_lng, name_lng;
	char	path[_POSIX_PATH_MAX +1];
	int		len;

	prep("Obtain implementation's value for {NAME_MAX} or {PCTS_NAME_MAX}.");
	start
		Zstrcpy(path, temp_path());

		name_lng = pathconf(path, _PC_NAME_MAX);
		if((name_lng == (long) -1) || (name_lng >= PCTS_NAME_MAX)) {
			name_lng = PCTS_NAME_MAX;
			len = FALSE;
		}
		else {
			len = TRUE;
			name_lng++;			/* {NAME_MAX} +1 */
		}
	finis

	prep("Obtain implementation's value for PATH_MAX.");
	start
		exit_on(ERROR);
		path_lng = pathconf(temp_path(), _PC_PATH_MAX);
		if(path_lng == (long) -1)
			path_lng = PCTS_PATH_MAX;
		if(path_lng <= name_lng)
			error("Path length \"%ld\" is <= name length \"%ld\".", 
										path_lng, name_lng);
	finis

	prep("Obtain buffer for implementation defined path length, %ld bytes.",
								path_lng);
	if(len == TRUE) {
		prep("Implementation supports [NAME_MAX] of %ld bytes.", name_lng);
		prep("Create pathname component of {NAME_MAX} +1 bytes.");
	}
	else
		prep("Create pathname component of [PCTS_NAME_MAX] bytes.");
	start
		exit_on(ERROR);
		pathm_u = (char *) Zmalloc((size_t)(path_lng +1) );

		Zstrcpy(pathm_u, temp_path());
		if((Zstrlen(pathm_u) + name_lng) > path_lng)
			error("Length of pathname from its components too large for path.");

		ptr = pathm_u +Zstrlen(pathm_u);
		ptr_end = ptr +name_lng;
		for ( ; ptr < ptr_end; ptr++ )
        		*ptr = 'E';
		*ptr = '\000';
	finis
	return(pathm_u);
}


/*
**  gi_path_max - Get implementations value for {PATH_MAX};
*/

long gi_path_max(void)
{
	long	path_lng, fpath_lng;
	int		fd;

	prep("Obtain implementation's value for {PATH_MAX}.");
	start
		Zstrcpy(pathm_u, temp_path());
								/* Must get values via pathconf() and
								   fpathconf() in case assertion is
								   testing one or the other.
								*/
		fd = Zopen(pathm_u, O_RDONLY);
		fpath_lng = fpathconf(fd, _PC_PATH_MAX);
		if(fpath_lng == (long) -1)
			error("%s\n%s\n",
				"fpathconf(_PC_PATH_MAX) failed.",
				"Expected value for {PATH_MAX}.");
		close(fd);
		path_lng = pathconf(pathm_u, _PC_PATH_MAX);
		if(path_lng == (long) -1)
			error("%s\n%s\n",
				"pathconf(_PC_PATH_MAX) failed.",
				"Expected value for {PATH_MAX}.");
		if(fpath_lng != path_lng)
			error("%s\n%s%ld\n%s%ld\n",
				"Different results obtained from pathconf() and fpathconf().",
				"Return from pathconf() was ",
				path_lng,
				"Return from fpathconf() was ",
				fpath_lng);
	finis

	return(path_lng);
}


/*
	g_lpath - Generate a path greater than {PATH_MAX}
	Create a pathname of length {PATH_MAX} +2("/X") bytes.
	PATH_MAX} is determined by the value returned from pathconf().

	NOTE:
		POSIX.1 does not specify whether the maximum length of
		a relative pathname, when returned by pathconf(), includes
		or excludes the terminating null.
*/

char	*g_lpath(void)
{
	char	*ptr;
	char	*ptr_end;
	long	path_lng;

	exit_on(ERROR);
	prep("Obtain buffer for minimum path length.");
	start
		pathm_u = (char *) Zmalloc(_POSIX_PATH_MAX +1);
	finis

	path_lng = gi_path_max();

	prep("Create path of length {PATH_MAX} +2 (%ld) bytes.", path_lng +2);
	start
		Zfree(pathm_u);
			/* Allocate buffer for {PATH_MAX} +1 + "/X" */
		pathm_u = (char *) Zmalloc((size_t)(path_lng +3));

		Zstrcpy(pathm_u, temp_path());
		Zstrcat(pathm_u, "ENAMETOOLONG");
		if(access(pathm_u, F_OK))
			Zmkdir(pathm_u, PROT_ALL);
		ptr = pathm_u +Zstrlen(pathm_u);

		ptr_end = path_lng +pathm_u;
		for( ; ptr +16 < ptr_end; ptr += 16 )
        		(void) strcpy(ptr, "/../ENAMETOOLONG");

		for( ; ptr < ptr_end; ptr += 2 )
        		(void) strcpy(ptr, "/.");

							/* End of path may be "/.//X" "or /./X". */
        (void) strcpy(ptr_end, "/X");
	finis
	return(pathm_u);
}

/*
	g_path_max - Generate a path of {PATH_MAX} bytes.
	PATH_MAX} is determined by the value returned from pathconf().

	NOTE:
		POSIX.1 does not specify whether the maximum length of
		a relative pathname, when returned by pathconf(), includes
		or excludes the terminating null.
*/

char	*g_path_max(void)
{
	char	*ptr;
	char	*ptr_end;
	long	path_lng;

	exit_on(ERROR);
	prep("Obtain buffer for minimum path length.");
	start
		pathm_u = (char *) Zmalloc(_POSIX_PATH_MAX +1);
	finis

	path_lng = gi_path_max();

	prep("Create path of length {PATH_MAX} (%ld) bytes.", path_lng);
	start
		Zfree(pathm_u);
			/* Allocate buffer for {PATH_MAX} +1 */
		pathm_u = (char *) Zmalloc((size_t) (path_lng +1));

		Zstrcpy(pathm_u, temp_path());
		Zstrcat(pathm_u, "ENAMETOOLONG");
		if(access(pathm_u, F_OK))
			Zmkdir(pathm_u, PROT_ALL);
		ptr = pathm_u +Zstrlen(pathm_u);

					/* The following line of code is only needed while
					   POSIX.1 allows implementations to include or
					   exclude the terminating null when the maximum
					   length of a pathname is returned by pathconf().

					   If POSIX.1 excludes the terminating null remove
					   the following line of code.
					*/
					path_lng = path_lng -sizeof(char);

		ptr_end = path_lng +pathm_u -2;
		for( ; ptr +16 < ptr_end; ptr += 16 )
        		(void) strcpy(ptr, "/../ENAMETOOLONG");

		for( ; ptr < ptr_end; ptr += 2 )
        		(void) strcpy(ptr, "/.");

							/* End of path may be "/.//X" "or /./X". */
        (void) strcpy(ptr_end, "/X");
	finis
	return(pathm_u);
}


void e_NO_TRUNC(enum element_values element)
{
	char	*lpath;
	long	res_p;
	char	path[_POSIX_PATH_MAX +1];
	char	*path_tmp;

	Zstrcpy(path, temp_path());
	res_p = pathconf(path, _PC_NO_TRUNC);
	if(res_p != (long) -1 )  {

		lpath = g_lname();

		if((element == E_LINK) || (element == E_RENAME)) {
			prep("Initialize a second pathname.");
			start
				path_tmp = (char *) Zmalloc(_POSIX_PATH_MAX +1);
				Zstrcpy(path_tmp, temp_path());
				Zstrcat(path_tmp, "path_2nd");
			finis
		}

		if(tc_name_max() == TRUE) {
			testing("Path component longer than {NAME_MAX}.");
			start
				perform(FALSE, element, lpath, path_tmp);
			finis
		}
		else {
			testing("Path component equal to {PCTS_NAME_MAX}.");
			start
				perform(TRUE, element, lpath, path_tmp);
			finis
		}

		cleanup("Remove buffer space allocated for long path.");
		start
			Zfree(lpath);
			if((element == E_LINK) || (element == E_RENAME))
				Zfree(path_tmp);
		finis
	}
	else
		error("{_POSIX_NO_TRUNC} not supported.");
}

void e_ENAMETOOLONG(enum element_values element)
{
	extern	char	*g_lpath(void);
	char	*lpath, *path_tmp;
	int		test_pm;

	if( (test_pm = tc_path_max()) ) {
		prep("Establish pathname of {PATH_MAX} +2 bytes.");
		start
			lpath = g_lpath();
		finis
	}
	else {
		prep("Establish pathname of {PCTS_PATH_MAX} +2 bytes.");
		start
			char	*ptr, *ptr_end;

					/* Allocate buffer for {PCTS_PATH_MAX} +1 + "/X" */
			lpath = (char *) Zmalloc(PCTS_PATH_MAX +3);

			Zstrcpy(lpath, temp_path());
			Zstrcat(lpath, "ENAMETOOLONG");
			if(access(lpath, F_OK))
				Zmkdir(lpath, PROT_ALL);
			ptr = lpath +Zstrlen(lpath);

			ptr_end = PCTS_PATH_MAX +lpath;
			for( ; ptr +16 < ptr_end; ptr += 16 )
        		(void) strcpy(ptr, "/../ENAMETOOLONG");

			for( ; ptr < ptr_end; ptr += 2 )
        			(void) strcpy(ptr, "/.");

							/* End of path may be "/.//X" "or /./X". */
        	(void) strcpy(ptr_end, "/X");
		finis
	}

	if((element == E_LINK) || (element == E_RENAME)) {
		prep("Initialize second pathname.");
		start
			path_tmp = (char *) Zmalloc(_POSIX_PATH_MAX +1);
			Zstrcpy(path_tmp, temp_path());
			Zstrcat(path_tmp, "path_scnd");
		finis
	}

	if( (test_pm = tc_path_max()) ) {
		testing("For path exceeds {PATH_MAX}.");
		start
			perform(FALSE, element, lpath, path_tmp);
		finis
	}
	else {
		testing("For path length of {PCTS_PATH_MAX}.");
		start
			switch(element) {
				case E_OPEN:
				case E_OPENDIR:
				case E_CHDIR:
				case E_RMDIR:
				case E_STAT:
				case E_ACCESS:
				case E_CHMOD:
				case E_UTIME:
				case E_CHOWN:			/* Needs ROOT permission. */
							Zmkdir(path_tmp, PROT_ALL);
							break;
				case E_UNLINK:
							Zclose(Zcreat(path_tmp, PROT_ALL));
							break;
				default:
							error("%s\nBad value of \"%d\" for \"element\"",
									"PCTS coding error.",
									element);
			}
			perform(TRUE, element, lpath, path_tmp);
		finis
	}

	cleanup("Remove buffer space allocated for long path.");
	start
		Zfree( lpath );
		if ((element == E_LINK) || (element == E_RENAME))
			Zfree(path_tmp);
	finis
}

void perform(int s_or_f, enum element_values element, char *lpath,
								char *path_tmp)
{
	FILE	*sfd;

	if(element == E_FREOPEN) {
		char	path_t[_POSIX_PATH_MAX +1];

		Zstrcpy(path_t, temp_path());
		Zstrcat(path_t, "tmp_err_cks");
		Zclose(Zcreat(path_t, PROT_ALL));
		sfd = Zfopen(path_t, "r");
	}

	if(s_or_f == FALSE) {
		if(	(element == E_OPENDIR) || (element == E_FOPEN) ||
			(element == E_FREOPEN)	)
			expecting(NULL);
		else
			expecting(-1);
		expecting(ENAMETOOLONG);
	}
	else
		expecting(SUCCESS);
	switch(element) {
		case E_OPENDIR:
			(void) Zopendir(lpath);
			break;
		case E_CHDIR:
			Zchdir(lpath);
			break;
		case E_OPEN:
			Zopen(lpath, O_RDONLY);
			break;
		case E_CREAT:
			Zcreat(lpath, (mode_t)S_IRWXU);
			break;
		case E_LINK:
			Zlink(lpath, path_tmp);
			break;
		case E_MKDIR:
			Zmkdir(lpath, PROT_ALL);
			break;
		case E_MKFIFO:
			Zmkfifo(lpath, PROT_ALL);
			break;
		case E_RMDIR:
			Zrmdir(lpath);
			break;
		case E_RENAME:
			Zrename(lpath, path_tmp);
			break;
		case E_UNLINK:
			Zunlink(lpath);
			break;
		case E_STAT:
			Zstat(lpath, &st_buf);
			break;
		case E_ACCESS:
			Zaccess(lpath, R_OK);
			break;
		case E_CHMOD:
			Zchmod(lpath, PROT_ALL);
			break;
		case E_CHOWN:
			Zchown(lpath, UID, GID);
			break;
		case E_UTIME:
			(void) Zutime(lpath, (struct utimbuf *) 0);
			break;
		case E_FOPEN:
			Zfopen(lpath, "r");
			break;
		case E_FREOPEN:
			Zfreopen(lpath, "r", sfd);
			break;
		default:
			error("%s\nBad value of \"%d\" for \"element\"\n",
				"PCTS coding error.",
				element);
			break;
	}
}
