#include <pcts.h>

#include <Zunistd.h>
#include <Zfcntl.h>
#include <Ztermios.h>

#include <sv_env.h>
#include <sv_errno.h>
#include <pcts_timer.h>


/*	Return the value of {OPEN_MAX} as a type int.  */
int	g_open_max()
{
	int	open_max;

	/*
	prep("Get value of {OPEN_MAX}.");
	start
	*/
		open_max = get_open_max();
	/*
	finis
	*/

	return open_max;
}

int	get_open_max()
{
	int	open_max;

	errno = 0;				/* If value not defined, errno unchanged. */

							/*	Get working value of {OPEN_MAX} for IUT.
								(i.e., {OPEN_MAX} from sysconf() can be
								greater than that from <limits.h>.
							*/
	open_max = (int) sysconf(_SC_OPEN_MAX);
	if(open_max == -1) {
		if(errno == 0) {
#ifdef	OPEN_MAX
			open_max = OPEN_MAX;
#else
			open_max = INT_MAX;		/* Allow for no limit */
#endif
		}
		else
			error("%s[%s]\n",
					"A call to sysconf(_SC_OPEN_MAX) returns -1 with errno of ",
					err_lookup(errno));
	}
	return open_max;
}

/*
		g_pipe_buf()	->	{PIPE_BUF}  Atomic write size
		Provide:
			value of {pipe_BUF} for input specified FIFO,
			value of {pipe_BUF} for temporary generated pipe,
		Return:
			maximum value of these two values.
*/

long	g_pipe_buf(char *path, size_t *pipe_buf_sz, size_t *fifo_buf_sz)
{
	long	p_ret, f_ret;

	prep("Determine {PIPE_BUF} for FIFO.");
	start
		errno = 0;				/* If value not defined, errno unchanged. */
		f_ret = pathconf(path, _PC_PIPE_BUF);
		if(f_ret == -1) {
			if(errno == 0)
				f_ret = LONG_MAX;			/* Allow for no limit. */
			else
				error("%s\n%s%s\n",
					"Call to pathconf(path, _PC_PIPE_BUF) failed.",
					"The errno was ",
					err_lookup(errno));
		}

		if(f_ret < _POSIX_PIPE_BUF)
			error("%s\n%s%d\n%s%d\n",
				"pathconf() - value returned is too low.",
				"Expected >= ",
				_POSIX_PIPE_BUF,
				"Received: ",
				f_ret);
		
		*fifo_buf_sz = f_ret;

		prep("{PIPE_BUF} for FIFO is %ld.\n", f_ret);
	finis

	prep("Determine {PIPE_BUF} for pipe.");
	start
		int	pipfd[2];

		Zpipe(pipfd);
		errno = 0;				/* If value not defined, errno unchanged. */
		p_ret = fpathconf(pipfd[0], _PC_PIPE_BUF);
		if(p_ret == -1) {
			if(errno == 0)
				p_ret = LONG_MAX;			/* Allow for no limit. */
			else
				error("%s\n%s%s\n",
					"Call to fpathconf(fildes, _PC_PIPE_BUF) failed.",
					"The errno was ",
					err_lookup(errno));
		}


		if(p_ret < _POSIX_PIPE_BUF)
			error("%s\n%s%d\n%s%d\n",
				"fpathconf() - value returned is too low.",
				"Expected >= ",
				_POSIX_PIPE_BUF,
				"Received: ",
				p_ret);

		*pipe_buf_sz = p_ret;

		prep("{PIPE_BUF} for pipe is %ld.\n", p_ret);

		Zclose(pipfd[0]);
		Zclose(pipfd[1]);
	finis

	return (long)((*pipe_buf_sz > *fifo_buf_sz) ? *pipe_buf_sz : *fifo_buf_sz);
}


/*
	Get the value of {MAX_INPUT}.
*/

size_t g_max_input(char *path)
{
	size_t	max_input;

	prep("Get value of {MAX_INPUT}.");
	start
		exit_on(ERROR);
		max_input = (size_t) Zpathconf(path, _PC_MAX_INPUT);
		if(max_input == -1) {
			if(errno == 0)
				max_input = (size_t)((cc_t)-1);		/* Allow for no limit */
			else
				error("%s[%s]\n",
					"pathconf(path, _PC_MAX_INPUT) returns -1 with errno of ",
					err_lookup(errno));
		}
	finis
		return max_input;
}


/*
		g_max_pipe()	->	
		Provides:
			Atomic write size {PIPE_BUF} for pipe.
			Atomic write size {PIPE_BUF} for FIFO.
			Maximum number of bytes that can be placed in a pipe.
			Maximum number of bytes that can be placed in a fifo.
		Returns:
			The maximum atomic write size {PIPE_BUF} for a pipe or FIFO.
*/

#define	FIFO_NBYTE	"det_nbytes"
long	g_max_pipe(size_t *pipe_buf_sz, size_t *fifo_buf_sz, size_t *pipe_max,
						size_t *fifo_max)
{
	char	fifop[_POSIX_PATH_MAX +1];
	int		pipfd[2];
	char	*buf_out;
	int 	fifowr;							/* FIFO write fildes	*/
	int 	fiford;							/* FIFO read fildes		*/
	int		flags;
	long	nbytes, nb_wr, nb_to_write;
	size_t		max_buf_sz;

	exit_on(ERROR);
	prep("Establish path to a FIFO file and create it");
	start
		Zstrcpy(fifop, temp_path());
		Zstrcat(fifop, FIFO_NBYTE);
		Zmkfifo(fifop, PROT_ALL);
	finis

	max_buf_sz = g_pipe_buf(fifop, pipe_buf_sz, fifo_buf_sz);

	prep("Allocate and initialize output buffer.");
	start
		buf_out = (char *) Zmalloc(max_buf_sz);
		memset(buf_out, 'D', max_buf_sz);
	finis

	prep("Create a pipe, set O_NONBLOCK on write channel.");
	start
		Zpipe(pipfd);
		flags = Zfcntl(pipfd[1], F_GETFL);
		Zfcntl(pipfd[1], F_SETFL, flags | O_NONBLOCK);
	finis

	prep(" Determine the maximum nuber of bytes that can be");
	prep("written to a pipe.");
	start
		SET_ALARM(TIMEOUT)
			expecting(*pipe_buf_sz);
			nbytes = Zwrite(pipfd[1], buf_out, *pipe_buf_sz);
		CLEAR_ALARM
						/*	Must eventually write one byte at a time to ensure
							that if data can be placed in pipe it will be
							transferred to the pipe for all POSIX.1
							implementations.
						*/
		SET_ALARM(5 * TIMEOUT)
			for(nb_to_write=*pipe_buf_sz; nb_to_write; nb_to_write /= 2)
				while((nb_wr = write(pipfd[1], buf_out, nb_to_write)) != -1)
					nbytes += nb_wr;
		CLEAR_ALARM

		*pipe_max = nbytes;
	finis

	prep("Opening FIFO with O_NONBLOCK flag on write channel.");
	start
		fiford = Zopen(fifop, O_RDONLY | O_NONBLOCK);
		fifowr = Zopen(fifop, O_WRONLY | O_NONBLOCK);
	finis

	prep("Determine the maximum nuber of bytes that can be");
	prep("written to a FIFO.");
	start
		SET_ALARM(TIMEOUT)
			expecting(*fifo_buf_sz);
			nbytes = Zwrite(fifowr, buf_out, *fifo_buf_sz);
		CLEAR_ALARM
						/*	Must eventually write one byte at a time to ensure
							that if data can be placed in FIFO it will be
							transferred to the FIFO for all POSIX.1
							implementations.
						*/
		SET_ALARM(TIMEOUT)
			for(nb_to_write=*fifo_buf_sz; nb_to_write; nb_to_write /= 2)
				while((nb_wr = write(fifowr, buf_out, nb_to_write)) != -1)
					nbytes += nb_wr;
		CLEAR_ALARM

		*fifo_max = nbytes;

		prep("Maximum number of bytes that can be written to FIFO is %ld.",
				(long)*fifo_max);
		prep("Maximum number of bytes that can be written to pipe is %ld.",
				(long)*pipe_max);
		prep("Remove pipe and FIFO.");

		Zclose(pipfd[0]);
		Zclose(pipfd[1]);
		Zclose(fifowr);
		Zclose(fiford);
		Zunlink(fifop);
		Zfree(buf_out);
	finis

	return (long)(max_buf_sz);
}


/*	Return a supplementary group ID from the list supplied by the
	<config> parameter SUPPLEMENTARY_GROUPS.
*/
gid_t	gsup_gid(void)
{
	gid_t	s_gid;
	int		n_grps, i;
	char	*p_sg;

	n_grps = Zsysconf(_SC_NGROUPS_MAX);
	if(n_grps >= MIN_SGA_NGROUPS_MAX) {
		for(i=0; i<NGROUPS_MAX/2; i++) {
			if(i==0)
				p_sg = strtok(SUPPLEMENTARY_GROUPS, ",");
			else
				p_sg = strtok(NULL, ",");
			if(p_sg == NULL) {						/* test for NULL */
				error("Config parameter SUPPLEMENTARY_GROUPS in error.");
				error("Return value of current group ID.");
				return(getgid());
			}
		   	s_gid = (gid_t) atoi(p_sg);
		}
		return(s_gid);
	}
	error("{NGROUPS_MAX} is %d, FIPS 151-2 requires 8 minimum.",n_grps);
	error("Return value of current group ID.");
	return(getgid());
}

/*	Change or add new value to an environment.
		Mimick putenv().
*/
extern	char **environ;

void put_env(char *var)
{
	char **new_environ, *ptr;
	int offset, length, elength;

	offset = 0;
	if( (ptr=(char *)strchr(var,'=')) == (char *) 0)
		length = strlen(var);
	else
		length = ptr -var;

	while(environ[offset] != (char *)0) {
		if( (ptr=(char *)strchr(environ[offset],'=')) == (char *)0)
			elength = strlen(environ[offset]);
		else
			elength = ptr -environ[offset];

		if((length == elength) && !strncmp(environ[offset], var, length)) {
			environ[offset] = var; 	/* Change existant variable. */
			return;
		}
		else
			offset++;
	}

	offset += 2;					/* Offset currently is last element. */
	if(environ[0] == (char *)0)		/* Environment being reset.	*/
		new_environ = (char **)malloc(offset*sizeof(char *));
	else
		new_environ = (char **)realloc(environ, offset*sizeof(char *));

	if(new_environ==NULL)
		severr("Not enough memory to satisfy putenv() request.");

	new_environ[offset -2]= var;	/* Add new entry to end. */
	new_environ[offset -1]= (char *)0;
	environ = (char **)new_environ;
	return;
}
